{ Telesis_Bits.pas                                SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Some generic bit manipulation functions.
<br>
@desc Primarily used with the BitCheckBox and BitCheckList of the
IBO-aware controls.
<br>
<br>
<strong>History</strong><br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_Bits;

{$INCLUDE Telesis.inc}

interface

uses SysUtils;

type

  {: @summary Available bit positions for use as parameters to the functions
  defined in <See Unit=Telesis_Bits>.
  <br>
  @desc Requires a 32bit integer. Avoids inclusion of the sign bit.}
  TBitPosition = 0..30;
  {: @summary Bitwise comparison instructions.
  <br>
  @desc See the <See Unit=Telesis_Bits Routine=GetBitsWhereSQL> routine.<br>
  <br>
  @enum bcAtLeast 
  At least the bits set in the test value are also set in the source value.  
  There may be other bits also set in the source.
  <br>
  @enum bcExact 
  Exactly the same bits are set in the test value as the source
  value.  Works out the same as a direct comparison.
  <br>
  @enum bcAnyOf 
  At least one of the bits provided in the test value is also
  set in the source value.  The source value may also have other bits set.
  <br>
  @enum bcNoneOf 
  None of the bits in the test value are set in the source
  value.  The source value may also have other bits set.
  <br>
  }
  TBitComparison = (bcAtLeast, bcExact, bcAnyOf, bcNoneOf);

function BitIsSet(const Flags: longint; const BitPos: TBitPosition): boolean;
procedure BitTurnOn(var Flags: longint; const BitPos: TBitPosition);
procedure BitTurnOff(var Flags: longint; const BitPos: TBitPosition);
procedure BitToggle(var Flags: longint; const BitPos: TBitPosition);
function GetBitsWhereSQL(const FieldName: string; const SelectVal: longint;
  Comparison: TBitComparison): string;


implementation

{: Return true if the bit position indicated is set (1) in the Flags
parameter.}
function BitIsSet(const Flags: longint; const BitPos: TBitPosition): boolean;
begin
  Result := (Flags and (1 shl BitPos)) <> 0;
end;

{: Ensure that the bit position indicated is set (1) in the var Flags
parameter.}
procedure BitTurnOn(var Flags: longint; const BitPos: TBitPosition);
begin
  Flags := Flags or (1 shl BitPos);
end;

{: Ensure that the bit position indicated is not set (0) in the var Flags
parameter.}
procedure BitTurnOff(var Flags: longint; const BitPos: TBitPosition);
begin
  Flags := Flags and ((1 shl BitPos) xor $FFFFFFFF);
end;

{: Toggle the status of the bit position indicated.  If set (1) then clear (0),
if clear (0) then set (1).}
procedure BitToggle(var Flags: longint; const BitPos: TBitPosition);
begin
  Flags := Flags xor (1 shl BitPos);
end;

{: @summary Create Bit comparison SQL Where clause as required.
<br>
@desc Assumes that the server has a BIN_AND function available for processing
bit information.<br>
<br>
<br>
If Comparison=bcAtLeast and SelectVal=4 then:<br>
<Code>Result := 'BIN_AND(MyField,4)=4';
</Code><br>
<br>
If Comparison=bcExact and SelectVal=6 then:<br>
<Code>Result := 'MyField=6';
</Code><br>
<br>
If Comparison=bcAnyOf and SelectVal=7 then:<br>
<Code>Result := 'BIN_AND(MyField,7)<>0';
</Code><br>
<br>
If Comparison=bcNoneOf and SelectVal=15 then:<br>
<Code>Result := 'BIN_AND(MyField,15)=0';
</Code><br>
<br>
}
function GetBitsWhereSQL(const FieldName: string; const SelectVal: longint;
  Comparison: TBitComparison): string;
begin
  Result := '';
  case Comparison of
    bcAtLeast:
      Result := 'BIN_AND(' + FieldName + ',' +
        IntToStr(SelectVal) + ')=' + IntToStr(SelectVal);
    bcExact:
      Result := FieldName + '=' + IntToStr(SelectVal);
    bcAnyOf:
      Result := 'BIN_AND(' + FieldName + ',' +
        IntToStr(SelectVal) + ')<>0';
    bcNoneOf:
      Result := 'BIN_AND(' + FieldName + ',' +
       IntToStr(SelectVal) + ')=0';
  end;
end;

end.

