{ Telesis_CheckBoxEnh.pas                         SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary CheckBox derived from TCustomComboEnh.
<br>
<br>
<strong>History</strong><br>
<br>
06-Aug-2009 - Not detecting change of value when SetChecked(false).<br>
<br>
16-Jan-2008 - Adjust AllowGrayed - see Telesis_IB_CheckBoxEnh.<br>
<br>
21-Jun-2005 - CheckBoxList fix from Cao Ancoinc - thank you.<br>
<br>
22-Jan-2002 - Some additional documentation.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_CheckBoxEnh;

{$INCLUDE Telesis.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  ExtCtrls,
  Telesis_EditEnh,
  Telesis_ComboEnh,
  Telesis_Utils;


type

{: @summary Dropbox used by TCheckBoxEnh.
<br>
@desc Set to display a simple list of boolean related values.}
  TCustomCheckBoxListEnh = class(TCustomComboControlEnh)
  private
    FList: TListBox;

  protected
    procedure SizeControl; override;
    procedure SyncToParentValue; override;
    procedure SyncParentToValue; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  end;

{: @summary Base class for TCheckBoxEnh
<br>
@desc Set to display a simple list of boolean strings.<br>
<br>
Note that this control is derived from TCustomComboEnh (which derives from
TCustomEditEnh).  It is NOT based on the Windows API CheckBox!  This may have
various implications on how you use the control.<br>
<br>
In its default configuration the control displays only the "glyph" (see the
various Glyph properties on TCustomEditEnh), showing a check-mark when true,
blank when false and grayed otherwise.  So in its default configuration you
will need to use the OnGlyphClick event instead of the OnClick event to
respond to mouse clicks.<br>
<br>
You may be interested in experiementing with alternative configurations for
the control.  For example if you set the following properties you get a combo
styled control supporting true/false...<br>
  Button.Style = eebDownArrow<br>
  GlyphAlign = alLeft<br>
  GlyphSize = 17<br>
  Width = 80<br>
<br>
Please read the details for the base class if you intend to use the OnDropDown
and BeforeDropDown event handlers: <see Class=TCustomComboEnh> }
  TCustomCheckBoxEnh = class(TCustomComboEnh)
  private
    FAllowGrayed: boolean;
    FCheckedString: string;
    FUncheckedString: string;
    FGrayedString: string;
    FListWidth: integer;

    function GetChecked: boolean;
    procedure SetChecked(AVal: boolean);
    function GetState: TCheckBoxState;
    procedure SetState(AVal: TCheckBoxState);
    procedure SetCheckedString(AVal: string);
    procedure SetUncheckedString(AVal: string);
    procedure SetGrayedString(AVal: string);
    function CheckedStringStored: boolean;
    function UncheckedStringStored: boolean;
    function GrayedStringStored: boolean;

  protected
    procedure SetAllowGrayed(AVal: boolean); virtual;
    function GetButtonStyleDefault: TEditEnhButtonStyle; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: char); override;
    procedure DoDrawGlyph(var DrawInfo: TDrawInfoEnh); override;
    procedure DoGlyphClick; override;

    function IncrementalSearch(TxtVal: string; ChkLen: integer): integer;
      override;

    function CreateComboControl: TCustomComboControlEnh; override;
    procedure DestroyComboControl(CurrentControl: TCustomComboControlEnh);
      override;

    procedure Toggle; virtual;

{: @summary Allow control to be set "grayed" - unknown/undefined.
<br>
@desc While boolean normally indicates only true or false, it is sometimes
convenient to be able to indicate an indeterminate state.  Set this property
to true to allow the third unknown/undefined state.}
    property AllowGrayed: boolean read FAllowGrayed write SetAllowGrayed default false;
{: @summary True if CheckBox contains True, returns false otherwise.}
    property Checked: boolean read GetChecked write SetChecked default false;
{: @summary String representation of "True" state.}
    property CheckedString: string read FCheckedString write SetCheckedString
      stored CheckedStringStored;
{: @summary String representation of "False" state.}
    property UncheckedString: string read FUncheckedString
      write SetUncheckedString stored UncheckedStringStored;
{: @summary String representation of grayed/unknown/undefined state.}
    property GrayedString: string read FGrayedString write SetGrayedString
      stored GrayedStringStored;

{: @summary Controls the width of the dropbox.
<br>
@desc If this property is set to 0 (the default) then the dropbox will be the
same width as the combobox edit control.<br>
<br>
Normally the dropbox will not display, but if the button is made visible the
dropbox can be used to select between true/false values in the list.}
    property ListWidth: integer read FListWidth write FListWidth default 0;

{: @summary Control display/alignment of the Glyph indicate the checked
state of the control.
<br>
@desc Normally set to alClient so the it appears similar to a normal checkbox,
however it can be set to alLeft (or whatever) so that the text value
representing the checked state will be displayed.}
    property GlyphAlign default alClient;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

{: @summary Current state of the checkbox.
<br>
@desc Where the Checked property can only indicate whether or not the control
is in the Checked state, this State property can detect all three possible
states.}
    property State: TCheckBoxState read GetState write SetState
      default cbUnchecked;
{: @summary Reads the string value of the current control state.}
    property Text stored false;
  end;

{: @summary Published variation of TCustomCheckBoxEnh.
<br>
@desc Refer to the base class <see Class=TCustomCheckBoxEnh> for
more details.}
  TCheckBoxEnh = class(TCustomCheckBoxEnh)
    {$DEFINE PUBLISH_CustomCheckBoxEnh}
    {$INCLUDE Telesis_Publish.INC}
  end;

var
{: @summary The shared dropbox used by TCustomCheckBoxEnh.
<br>
@desc DO NOT ALTER THIS VARIABLE UNLESS WRITING A NEW DERIVATION.}
  CheckBoxEnh_ListBox: TCustomCheckBoxListEnh;
{: @summary The number of TCustomCheckBoxEnh components accessing the shared
dropbox (CheckBoxEnh_ListBox).
<br>
@desc DO NOT ALTER THIS VARIABLE UNLESS WRITING A NEW DERIVATION.}
  CheckBoxEnh_ListBoxRefCount: integer;
{: @summary Default CheckedString property value.
<br>
@desc Assigned when the unit is initialised from the Telesis_Locale.pas
definitions.  However you can change this at runtime if you wish.}
  CheckBoxEnh_DefaultCheckedString: string;
{: @summary Default UncheckedString property value.
<br>
@desc Assigned when the unit is initialised from the Telesis_Locale.pas
definitions.  However you can change this at runtime if you wish.}
  CheckBoxEnh_DefaultUncheckedString: string;
{: @summary Default GrayedString property value.
<br>
@desc Assigned when the unit is initialised from the Telesis_Locale.pas
definitions.  However you can change this at runtime if you wish.}
  CheckBoxEnh_DefaultGrayedString: string;


implementation

uses Telesis_Locale;

type
  // A slightly specialised ListBox for the combo...
  TCheckBoxListBoxEnh = class(TListBox)
  protected
    //procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;

  public
  end;

procedure TCheckBoxListBoxEnh.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  i: integer;
  Pt: TPoint;
begin
  inherited;
  Pt.X := X;
  Pt.Y := Y;
  i := ItemAtPos(Pt, false);
  if i >= 0 then
  begin
    ItemIndex := i;
  end
  else
    ItemIndex := -1;
end;

procedure TCheckBoxListBoxEnh.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  i: integer;
  Pt: TPoint;
begin
  inherited;
  Pt.X := X;
  Pt.Y := Y;
  i := ItemAtPos(Pt, false);
  if i >= 0 then
  begin
    ItemIndex := i;
  end
  else
    ItemIndex := -1;
  with TCustomCheckBoxListEnh(Parent) do
  begin
    if Controller.EditBeginModify then
    begin
      Change;
      CloseUp;
    end;
  end;
end;


constructor TCustomCheckBoxListEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ActiveWhenInvisible := false; // We use a shared listbox so cannot be active
  // until visible
  FList := TCheckBoxListBoxEnh.Create(Self);
  // It is replicatable because there is only one!
  FList.ControlStyle := FList.ControlStyle + [csReplicatable];
  FList.Parent := Self;
  FList.Align := alClient;
  FList.BorderStyle := bsNone;
  FList.Visible := true;
  FList.ExtendedSelect := false;
  FList.MultiSelect := false;
end;

destructor TCustomCheckBoxListEnh.Destroy;
begin
  if FList <> nil then
  begin
    FList.Free;
    FList := nil;
  end;
  inherited Destroy;
end;

procedure TCustomCheckBoxListEnh.SizeControl;
var
  i, j, k: integer;
begin
  if TCustomCheckBoxEnh(Controller).ListWidth <= 0 then
    Width := Controller.Width
  else
    Width := TCustomCheckBoxEnh(Controller).ListWidth;
  if TCustomCheckBoxEnh(Controller).AllowGrayed then
    i := 3
  else
    i := 2;
  j := SendMessage(FList.Handle, LB_GETITEMHEIGHT, 0, 0);
  k := Height - FList.Height;           // Get the border size adjustment
  Height := k + (i * j);
end;

procedure TCustomCheckBoxListEnh.SyncToParentValue;
var
  i: integer;
begin
  if FInternalChange = 0 then
  begin
    // We are using a shared listbox, so clear and re-initialise the listbox
    // contents first...
    with Controller as TCustomCheckBoxEnh do
    begin
      FList.Clear;
      FList.Items.Add(FCheckedString);
      FList.Items.Add(FUncheckedString);
      if AllowGrayed then
        FList.Items.Add(FGrayedString);
      case State of
        cbChecked: i := 0;
        cbUnchecked: i := 1;
      else
        if AllowGrayed then
          i := 2
        else
          i := 1;
      end;
      FList.ItemIndex := i;
    end;
  end
  else
  begin
    with Controller as TCustomCheckBoxEnh do
    begin
      case State of
        cbChecked: i := 0;
        cbUnchecked: i := 1;
      else
        if AllowGrayed then
          i := 2
        else
          i := 1;
      end;
      if FList.ItemIndex <> i then
      begin
        FList.ItemIndex := i;
      end;
    end;
  end;
end;

procedure TCustomCheckBoxListEnh.SyncParentToValue;
begin
  if FList.ItemIndex = 0 then
  begin
    TCustomCheckBoxEnh(Controller).State := cbChecked;
  end
  else
    if FList.ItemIndex = 1 then
    begin
      TCustomCheckBoxEnh(Controller).State := cbUnchecked;
    end
    else
      TCustomCheckBoxEnh(Controller).State := cbGrayed;
end;



constructor TCustomCheckBoxEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAllowGrayed := false;
  FCheckedString := CheckBoxEnh_DefaultCheckedString;
  FUncheckedString := CheckBoxEnh_DefaultUncheckedString;
  FGrayedString := CheckBoxEnh_DefaultGrayedString;
  FListWidth := 0;
  Text := CheckBoxEnh_DefaultUncheckedString;
  GlyphSize := 0;
  GlyphAlign := alClient;
  Width := 20;
end;

destructor TCustomCheckBoxEnh.Destroy;
begin
  inherited Destroy;
end;

function TCustomCheckBoxEnh.GetChecked: boolean;
begin
  Result := (State = cbChecked);
end;

procedure TCustomCheckBoxEnh.SetChecked(AVal: boolean);
begin
  if AVal then
    State := cbChecked
  else
    State := cbUnchecked;
end;

function TCustomCheckBoxEnh.GetState: TCheckBoxState;
begin
  if Text = FCheckedString then
    Result := cbChecked
  else
    if Text = FUncheckedString then
      Result := cbUnchecked
    else
      // if specifically set to FGrayedString then we assume the
      // value was set using SetState and so it is truly an indication
      // of a grayed/indeterminate checkbox - EVEN IF allowgrayed was
      // set to true.
      if Text = FGrayedString then
        Result := cbGrayed
      else
      begin
        // any other value returns grayed if AllowGrayed, otherwise
        // we assume any non true value is really a false value.
        if AllowGrayed then
          Result := cbGrayed
        else
          Result := cbUnchecked;
      end;
end;

procedure TCustomCheckBoxEnh.SetState(AVal: TCheckBoxState);
begin
  if (AVal <> State) and EditCanModify then
  begin
    EditBeginModify;
    case AVal of
      cbChecked: Text := FCheckedString;
      cbUnchecked: Text := FUncheckedString;
    else
      // If the State is being set directly, we ignore the
      // AllowGrayed flag.  Leave it upto the user interface
      // elements to control whether AllowGrayed is used.
      Text := FGrayedString;
      //if AllowGrayed then
      //  Text := FGrayedString
      //else
      //  Text := FUncheckedString;
    end;
    if (ComboControl <> nil) then
      ComboControl.DoParentValueChanged;
    Modified := true;
    if HandleAllocated then
      InvalidateGlyph;
    Change; // We stop the text change indicating change so this is to
    // generate a change event when the selection has changed.
  end;
end;

procedure TCustomCheckBoxEnh.Toggle;
begin
  case State of
    cbUnchecked:
      if AllowGrayed then
        State := cbGrayed
      else
        State := cbChecked;
    cbChecked: State := cbUnchecked;
    cbGrayed: State := cbChecked;
  end;
end;

procedure TCustomCheckBoxEnh.SetCheckedString(AVal: string);
var
  tmpState: TCheckBoxState;
begin
  if AVal <> FCheckedString then
  begin
    tmpState := GetState;
    FCheckedString := AVal;
    State := tmpState;
  end;
end;

procedure TCustomCheckBoxEnh.SetUncheckedString(AVal: string);
var
  tmpState: TCheckBoxState;
begin
  if AVal <> FUncheckedString then
  begin
    tmpState := GetState;
    FUncheckedString := AVal;
    State := tmpState;
  end;
end;

procedure TCustomCheckBoxEnh.SetGrayedString(AVal: string);
var
  tmpState: TCheckBoxState;
begin
  if AVal <> FGrayedString then
  begin
    tmpState := GetState;
    FGrayedString := AVal;
    State := tmpState;
  end;
end;

function TCustomCheckBoxEnh.CheckedStringStored: boolean;
begin
  Result := (FCheckedString <> CheckBoxEnh_DefaultCheckedString);
end;

function TCustomCheckBoxEnh.UncheckedStringStored: boolean;
begin
  Result := (FUncheckedString <> CheckBoxEnh_DefaultUncheckedString);
end;

function TCustomCheckBoxEnh.GrayedStringStored: boolean;
begin
  Result := (FGrayedString <> CheckBoxEnh_DefaultGrayedString);
end;

procedure TCustomCheckBoxEnh.SetAllowGrayed(AVal: boolean);
begin
  FAllowGrayed := AVal;
end;

function TCustomCheckBoxEnh.GetButtonStyleDefault: TEditEnhButtonStyle;
begin
  Result := eebNone;
end;

procedure TCustomCheckBoxEnh.KeyDown(var Key: Word; Shift: TShiftState);
begin
  //if (ComboControl <> nil) and (not ComboControl.Active) and
  //    (not ComboControl.ActiveWhenInvisible) then begin
  if EditCanModify then
  begin
    if Shift = [] then
    begin
      case Key of
        VK_DOWN:
          begin
            Toggle;
            Key := 0;
          end;
        VK_UP:
          begin
            case State of
              cbUnchecked: State := cbChecked;
              cbChecked:
                if AllowGrayed then
                  State := cbGrayed
                else
                  State := cbUnchecked;
              cbGrayed: State := cbUnchecked;
            end;
            Key := 0;
          end;
      end;
    end;
  end;
  inherited;
end;

procedure TCustomCheckBoxEnh.KeyPress(var Key: char);
begin
  if Key = ' ' then
  begin
    Toggle;
  end
  else
    if (Length(FCheckedString) > 0) and
    (AnsiUpperCase(Key) = AnsiUpperCase(FCheckedString[1])) then
    begin
      State := cbChecked;
    end
    else
      if (Length(FUncheckedString) > 0) and
      (AnsiUpperCase(Key) = AnsiUpperCase(FUncheckedString[1])) then
      begin
        State := cbUnchecked;
      end
      else
        if (Length(FGrayedString) > 0) and
        (AnsiUpperCase(Key) = AnsiUpperCase(FGrayedString[1])) then
        begin
          State := cbGrayed;
        end;
  Key := #0;
  inherited;
end;

procedure TCustomCheckBoxEnh.DoDrawGlyph(var DrawInfo: TDrawInfoEnh);
begin
  if assigned(OnDrawGlyph) then
    OnDrawGlyph(Self, DrawInfo)
  else
  begin
    with DrawInfo, DrawInfo.Canvas do
    begin
      // Initialise brush and pen settings
      Brush.Style := bsSolid;
      Brush.Color := Self.Color;
      FillRect(Rect);
      VectorBrush := bsSolid;
      if Self.Focused then
      begin
        Brush.Color := clHighlight;
        Pen.Color := clHighlightText;
        BkColor := clHighlight;
        VectorColor := clHighlightText;
        InflateRect(Rect, -1, -1);
        DrawFocusRect(Rect);
        InflateRect(Rect, -1, -1);
        FillRect(Rect);
        InflateRect(Rect, 2, 2);
      end
      else
      begin
        BkColor := Brush.Color;
        VectorColor := Self.Font.Color;
      end;
      case State of
        cbChecked:
          DrawCheckMark(DrawInfo);
        // cbUnchecked: nothing to draw
        cbGrayed:
          DrawQuery(DrawInfo);
      end;
    end;
  end;
end;

procedure TCustomCheckBoxEnh.DoGlyphClick;
begin
  Toggle;
  inherited;
end;

function TCustomCheckBoxEnh.IncrementalSearch(TxtVal: string; ChkLen: integer):
  integer;
begin
  Result := -1;
end;

function TCustomCheckBoxEnh.CreateComboControl: TCustomComboControlEnh;
begin
  if CheckBoxEnh_ListBox = nil then
  begin
    CheckBoxEnh_ListBox := TCustomCheckBoxListEnh.Create(nil);
  end;
  Inc(CheckBoxEnh_ListBoxRefCount);
  Result := CheckBoxEnh_ListBox;
end;

procedure TCustomCheckBoxEnh.DestroyComboControl(CurrentControl:
  TCustomComboControlEnh);
begin
  if CheckBoxEnh_ListBox.Parent = Self then
  begin
    CheckBoxEnh_ListBox.Parent := nil;
  end;
  if CurrentControl = CheckBoxEnh_ListBox then
  begin
    Dec(CheckBoxEnh_ListBoxRefCount);
    if CheckBoxEnh_ListBoxRefCount = 0 then
    begin
      CheckBoxEnh_ListBox.Free;
      CheckBoxEnh_ListBox := nil;
    end;
  end;
end;

initialization;
  CheckBoxEnh_DefaultCheckedString := ENHSTR_Bool_True;
  CheckBoxEnh_DefaultUncheckedString := ENHSTR_Bool_False;
  CheckBoxEnh_DefaultGrayedString := ENHSTR_Bool_Unknown;
  CheckBoxEnh_ListBox := nil;
  CheckBoxEnh_ListBoxRefCount := 0;

finalization;
  CheckBoxEnh_ListBox := nil;
  CheckBoxEnh_ListBoxRefCount := 0;

end.
