{ Telesis_ComboEnh.pas                            SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Generic Combo-Styled control and specific TComboBoxEnh variation.
<br>
@desc Sets up a base combobox styled class that is NOT based on the Windows
API combobox.  This allows greater flexibility in what dropbox can be display
and allows such things as increment searching of a list.
<br>
<strong>History</strong><br>
<br>
15-Nov-2008 - MouseWheel support had altered keyboard behaviour in combobox,
making alt+down inconsistent.  Now fixed so that scrolling only occurs with
empty keyboard shift state, a "slide" movement works with ssCtrl (relates to
the mousewheel).<br>
<br>
12-Feb-2008 - MouseWheel support.<br>
<br>
02-Jun-2007 - Fixed problem with ComboBox IncrementalSearch (repeated 
characters cause search to skip to end of repeats).  This is an old bug I
cant believe has lasted this long.  Thanks to Roberto for pointing it out.<br>
<br>
06-Aug-2006 - Adjustments to support multiple monitors (use Screen.DesktopWidth
etc).<br>
<br>
11-May-2006 - Memory leak; FDropDownFont was not being released.<br>
<br>
22-Sep-2002 - Introduced ListOwnerDraw and OnListDrawItem properties for
CustomComboBox which gives access to the TList.OnDrawItem capability.  This is
intentially restricted to the OwnerDrawFixed capability as the variable option
would get complicated to synchronise with the rest of the control.<br>
<br>
29-May-2002 - Introduced DropDownFont and DropDownFontAuto properties.<br>
<br>
29-Apr-2002 - Altered the GetComboBoxEnhDisplay* global functions to take
constant TStrings parameters - was getting an exception in some instances.<br>
<br>
22-Jan-2002 - Altered the base DropDown procedure so that it does not
disable dropbox TabStops until after the dropbox has been displayed.
Previously done before display, which did not take proper effect for
the datetime derivative.<br>
<br>
28-Dec-2001 - Added support for string localisation.<br>
<br>
17-Dec-2001 - GetItemValues was returning the display value and not the
item value.  Added AutoDropDown capability to base class.<br>
<br>
01-Nov-2001 - Extra check to avoid problems with SelectAll, IncrementalSearch
and OnChange event.  (Was preventing incremental search in some circumstances
relating to code in OnChange event handlers.)<br>
<br>
30-Sep-2001 - Prevent change event when clicking on already selected
entry in dropdown list.<br>
<br>
12-Sep-2001 - Better support for Style=csDropDown.  Previously only
really supported csDropDownList.<br>
<br>
23-Aug-2001 - Support to synchronise colors on drop box.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_ComboEnh;

{$INCLUDE Telesis.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  Telesis_MaskEnh,
  Telesis_Utils,
  Telesis_EditEnh;

type
{: @summary Combobox dropdown list style.
<br>
@desc I do not (yet) support all options of the normal ComboBox style.}
  TComboBoxEnhStyle = (
    csDropDown,
    csDropDownList
    );

  TCustomComboEnh = class;

{: @summary Base class for dropboxes.
<br>
@desc Serves as a base class for boxes which are to be dropped "dropped-down"
from a TCustomCombEnh component.  These are made to "stay-on-top" while
visible but automatically hide when the the master control loses focus.<br>
<br>
The control itself should never gain focus but can/should respond
appropriately to mouse input.  This class defines a set of common
interaction functions which should be overridden by derived controls as
required.}
  TCustomComboControlEnh = class(TCustomPanel) //(TCustomControl)
  private
    FController: TCustomComboEnh;
    FActiveWhenInvisible: boolean;
    function GetActive: boolean;

  protected
    FInternalChange: integer;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;

    procedure DoChange; virtual;
    procedure DoCloseUp; virtual;
    procedure DoBeforeDropDown; virtual;
    procedure DoDropDown; virtual;

    function CanChange: boolean; virtual;
    procedure Change; virtual;

    procedure ParentKeyDown(var Key: Word; Shift: TShiftState); virtual;
    procedure PositionControl; virtual;
    procedure SizeControl; virtual;
    procedure SyncToParentValue; virtual;
    procedure SyncParentToValue; virtual;
    procedure SyncColors; virtual;

    property ActiveWhenInvisible: boolean read FActiveWhenInvisible write FActiveWhenInvisible;
    property ParentFont default true;
    property ParentColor default true;
    property ParentCtl3D default true;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure CloseUp; virtual;
    procedure DropDown; virtual;

    // These two procedures made available for access by derivatives
    // of the edit control component, not really intended for use by
    // the application.
    procedure DoParentKeyDown(var Key: Word; Shift: TShiftState);
    procedure DoParentValueChanged;

    property Active: boolean read GetActive;
    property Controller: TCustomComboEnh read FController;

    // Made these public to give runtime access to display attributes.
    // NB. I did not try to make these part of the edit control designtime
    // interface because most of the derivatives use a single global instance
    // of a list and it could get confusing trying to maintain these.
    // Instead just use runtime access to the specific global instances.
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property Font;
    property ShowHint;

  end;

{: @summary Base class for all combobox styled controls derived from
TCustomEditEnh.
<br>
@desc Defines the default interaction between the edit control and the
generic drop box (TCustomComboControlEnh).<br>
<br>
One difference between this (and all the derived controls) is that the
OnDropDown event occurs after the dropdown list has already been initialised,
just before it becomes visible.  If the list box is shared - as all the
derivations in this collection are - then any changes you make to the list
attributes in the OnDropDown event will be ignored until you next dropdown.<br>
<br>
This was done to ensure that a shared list is properly defined for the current
control when OnDropDown occurs.   If you want to make changes to the list
attributes then use the BeforeDropDown event - which is generated after the
control claims the list for its own use, but before the list is initialised
with the current control settings.  (Side note.  I am not a big fan of
changing combo items in response to dropdown events because usually the set
of values available should exist - and be consistent - for use even without
the user intervention of dropping the list.  See my notes about the
DisplayField property below.}
  TCustomComboEnh = class(TCustomEditEnh)
  private
    FComboControl: TCustomComboControlEnh;
    FActiveWhenReadOnly: boolean;
    FSearching: integer;
    FAutoDropDown: boolean;
    FBeforeDropDown: TNotifyEvent;
    FOnDropDown: TNotifyEvent;
    FOnCloseUp: TNotifyEvent;
    FStyle: TComboBoxEnhStyle;
    FDropDownColor: TColor;
    FDropDownFont: TFont;
    FDropDownFontAuto: boolean;

    function GetDroppedDown: boolean;
    procedure DoDropDownFontChange( Sender: TObject );
    procedure SetDropDownFont( AVal: TFont );
    function IsDropDownFontStored: boolean;
    procedure SetDropDownFontAuto( AVal: boolean );

    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;

  protected
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    function GetButtonStyleDefault: TEditEnhButtonStyle; override;
    function GetButton1KeyDefault: TShortCut; override;
    function AllowButtonClickOnFocus: boolean; override;

    procedure KeyDown(var Key: Word; Shift: TShiftState); override;

    function GetButtonEnabled: boolean; override;
    function DoTextChange: boolean; override;

    procedure DoButton1; override;

    procedure IBG_SetColor( AColor: TColor ); override;

    function IncrementalSearch(TxtVal: string; ChkLen: integer): integer;
      virtual;
    procedure NoSearchChange; virtual;

    function CreateComboControl: TCustomComboControlEnh; virtual;
    procedure DestroyComboControl(CurrentControl: TCustomComboControlEnh);
      virtual;

    property Searching: integer read FSearching;

{: @summary Controls whether the dropbox can be displayed when the
control is set to readonly.}
    property ActiveWhenReadOnly: boolean read FActiveWhenReadOnly write
      FActiveWhenReadOnly default false;

{: @summary When true the dropdown blox will be displayed automatically on
enter and during any select activity.}
    property AutoDropDown: boolean read FAutoDropDown write FAutoDropDown
      default false;
{: @summary Readonly access to the dropbox component.
<br>
@desc Beware!  Many of the controls derived from this base use a shared
dropbox control to minimise resources.  As such this property is not
guaranteed to refer to anything until the particular control takes
possession of the shared dropbox.}
    property ComboControl: TCustomComboControlEnh read FComboControl;
{: @summary Font for drop down component.
<br>
@desc When DropDownFontAuto is true this property is ignored (and not stored),
instead the font of the control itself is used.  When DropDownFontAuto is false
then this font is used by the dropdownbox.}
    property DropDownFont: TFont read FDropDownFont write SetDropDownFont
      stored IsDropDownFontStored;
{: @summary Determines if DropDownFont property is used to control the font
of the dropdown box.}
    property DropDownFontAuto: boolean read FDropDownFontAuto
      write SetDropDownFontAuto default true;

{: @summary Combobox style.
<br>
@desc This property may need to be used with care, depending on the
particular derivation being used.}
    property Style: TComboBoxEnhStyle read FStyle write FStyle default
      csDropDownList;
{: @summary Handler fired when dropbox is closed.}
    property OnCloseUp: TNotifyEvent read FOnCloseUp write FOnCloseUp;
{: @summary Handler fired BEFORE dropbox is displayed.}
    property BeforeDropDown: TNotifyEvent read FBeforeDropDown
      write FBeforeDropDown;
{: @summary Handler fired when dropbox is displayed.}
    property OnDropDown: TNotifyEvent read FOnDropDown write FOnDropDown;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function EditBeginModify: boolean; override;
{: @summary Shows DropDown box if permitted.
<br>
@desc Returns false if DropDown box could not be shown.}
    function ShowDropDown: boolean;
{: @summary Hides DropDown box.}
    procedure HideDropDown;
{: @summary Returns true when the dropbox is visible.}
    function HasActiveChild: boolean; override;
{: @summary Call to have dropbox color synchronised - if displayed.}
    procedure SyncDropBoxColors; virtual;
{: @summary True when dropbox is visible.}
    property DroppedDown: boolean read GetDroppedDown;
{: @summary Base window color for dropbox.
<br>
@desc Determin the window background color on the dropbox - defaults to
clWindow.  Set this to clNone to have the background color automatically
follow the control color.}
    property DropDownColor: TColor read FDropDownColor write FDropDownColor
      default clWindow;

  end;                                  // TCustomComboEnh

{: @summary Dropbox used by TComboBoxEnh.
<br>
@desc Set to display a simple list of values.}
  TCustomComboListBoxEnh = class(TCustomComboControlEnh)
  private
    FList: TListBox;

  protected
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    procedure ParentKeyDown(var Key: Word; Shift: TShiftState); override;
    procedure SizeControl; override;
    procedure SyncToParentValue; override;
    procedure SyncParentToValue; override;
    procedure SyncColors; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  end;

  TCustomComboBoxEnh = class;
  TComboBoxEnhDrawItemEvent = procedure(Sender: TCustomComboBoxEnh;
    Control: TListBox; Index: Integer; Rect: TRect; State: TOwnerDrawState)
    of object;

{: @summary Base class for TComboBoxEnh
<br>
@desc Setup to allow selection from a list.<br>
<br>
Please read the details for the base class if you intend to use the OnDropDown
and BeforeDropDown event handlers: <see Class=TCustomComboEnh> }
  TCustomComboBoxEnh = class(TCustomComboEnh)
  private
    FListCols: integer;
    FListRows: integer;
    FDropDownRows: integer;
    FListWidth: integer;
    FItems: TStringList;
    FItemIndex: integer;
    FTabWidth: integer;
    FListOwnerDraw: boolean;
    FOnListDrawItem: TComboBoxEnhDrawItemEvent;

    procedure SetItems(AValue: TStringList);
    procedure SetItemIndex(AValue: integer);
    function GetItemValue: string;
    procedure SetItemValue(AValue: string);
    function GetItemValues(Index: integer): string;
    procedure SetItemValues(Index: integer; AValue: string);
    function GetSorted: boolean;
    procedure SetSorted(AValue: boolean);
    // OnDrawItem event procedure for TList
    procedure DoListDrawItem(Control: TWinControl;
      Index: Integer; Rect: TRect; State: TOwnerDrawState);

  protected
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;

    procedure LoadItemList; virtual;
    function IsSortedStored: boolean;
    function AreItemsStored: boolean; virtual;

    function IncrementalSearch(TxtVal: string; ChkLen: integer): integer;
      override;
    procedure NoSearchChange; override;

    function CreateComboControl: TCustomComboControlEnh; override;
    procedure DestroyComboControl(CurrentControl: TCustomComboControlEnh);
      override;

{: @summary Controls how many columns are used to display the list in the
dropbox.}
    property ListColumns: integer read FListCols write FListCols default 0;
{: @summary Controls the maximum number of rows displayed in the dropbox.
<br>
@desc If there are less than this number of items available then the control
will attempt to size the dropbox accordingly.}
    property ListRows: integer read FListRows write FListRows default 8;
{: @summary Controls the width of the dropbox.
<br>
@desc If this property is set to 0 (the default) then the dropbox will be the
same width as the combobox edit control.}
    property ListWidth: integer read FListWidth write FListWidth default 0;
{: @summary List of items that are selectable from the dropbox.
<br>
@desc This control allows the Items stringlist to be defined as a set of:<br>
<code>
    display text=value
</code><br>
strings.   Where "display text" is what appears in the edit-box and drop-down
list, but "value" can be used to set/retrieve the intended value of control.
Having the list defined this way (as a single list rather than as two separate
lists) means that you can sort the list of items without impacting upon the
value matching.}
    property Items: TStringList read FItems write SetItems
      stored AreItemsStored;
{: @summary Index of the item currently selected.
<br>
@desc If no item has been selected this property is set to -1.}
    property ItemIndex: integer read FItemIndex write SetItemIndex default -1;
{: @summary Value of the item currently selected.
<br>
@desc Allows you to get/set the defined value of the control without regard
to the actual displayed text.  (If there is no "=value" specified then the
displayed text is returned.)<br>
<br>
If is no item selected this property will return a blank string.}
    property ItemValue: string read GetItemValue write SetItemValue;
{: @summary Return an Item value by index into the Items list.
<br>
@desc Returns the "value" part of the item definition.  See the Items
property for details.}
    property ItemValues[Index: Integer]: string read GetItemValues
      write SetItemValues;
{: @summary determines if the Items list should be sorted.}
    property Sorted: boolean read GetSorted write SetSorted stored IsSortedStored;
{: @summary Assigned to the TListBox used to display the Items.
<br>
@desc See TListBox.TabWidth for details.}
    property TabWidth: integer read FTabWidth write FTabWidth default 0;
{: @summary Turn on owner draw access to the dropdown TList (see the
OnListDrawItem event).}
    property ListOwnerDraw: boolean read FListOwnerDraw write FListOwnerDraw
        default false;
{: @summary Access to the dropdown TList OnDrawItem event (see the
ListOwnerDraw property).}
    property OnListDrawItem: TComboBoxEnhDrawItemEvent read FOnListDrawItem
        write FOnListDrawItem;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

{: @summary Number of rows actually displayed on the dropbox.
<br>
@desc The value of this property is only valid if the dropbox is currently
visible (DroppedDown = true). }
    property DropDownRows: integer read FDropDownRows;

  end;

{: @summary Published variation to TCustomComboBoxEnh.}
  TComboBoxEnh = class(TCustomComboBoxEnh)
    {$DEFINE PUBLISH_CustomComboBoxEnh}
    {$DEFINE PUBLISH_MaskEnh}
    {$DEFINE PUBLISH_Text}
    {$INCLUDE Telesis_Publish.INC}
  end;

{: @summary Return Display value from ComboBox Items by providing the Value
string for the required item.
<br>
@desc Use this to obtain the display string from a ComboBoxEnh string list
given a particular value.  Simply pass the "Items" (or similar StringList)
value from the required combo-box and the "Value" what you want the
display string for.}
function GetComboBoxEnhDisplayString(const StrList: TStrings; AValue: string):
  string;
{: @summary Return Display value from ComboBox Items by providing the Index
of the required item.
<br>
@desc Use this to obtain the display string from a ComboBoxEnh string list
given a particular item index.  Simply pass the "Items" (or similar StringList)
value from the required combo-box and the index of the item you want the
display string for.}
function GetComboBoxEnhDisplayIndex(const StrList: TStrings; AValue: string):
  integer;

var
{: @summary The shared dropbox used by TCustomComboBoxEnh.
<br>
@desc DO NOT ALTER THIS VARIABLE UNLESS WRITING A NEW DERIVATION.}
  ComboBoxEnh_ListBox: TCustomComboListBoxEnh;
{: @summary The number of TCustomCombBoxEnh components accessing the shared
dropbox (ComboBoxEnh_ListBox).
<br>
@desc DO NOT ALTER THIS VARIABLE UNLESS WRITING A NEW DERIVATION.}
  ComboBoxEnh_ListBoxRefCount: integer;



implementation

uses Telesis_Locale;

type
  // A slightly specialised ListBox for the combo...
  TListBoxEnh = class(TListBox)
  protected
    //procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;
  end;

constructor TCustomComboControlEnh.Create(AOwner: TComponent);
begin
  FController := nil;
  if AOwner <> nil then
  begin
    if not (AOwner is TCustomComboEnh) then
      raise
        Exception.Create(ENHERR_COMBOENH_INVALID_PARENT);
    FController := TCustomComboEnh(AOwner);
  end;
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csNoDesignVisible, csReplicatable];
  FInternalChange := 0;
  Visible := false;
  BevelInner := bvNone;                 //bvLowered;
  BevelOuter := bvRaised;
  BevelWidth := 1;
  BorderWidth := 1;
end;

destructor TCustomComboControlEnh.Destroy;
begin
  inherited Destroy;
end;

function TCustomComboControlEnh.GetActive: boolean;
begin
  if HandleAllocated then
  begin
    Result := IsWindowVisible(Handle);
  end
  else
  begin
    Result := false;
  end;
end;

procedure TCustomComboControlEnh.DoParentKeyDown(var Key: Word; Shift:
  TShiftState);
begin
  if Active or ActiveWhenInvisible then
    ParentKeyDown(Key, Shift);
end;

procedure TCustomComboControlEnh.DoParentValueChanged;
begin
  if Active or ActiveWhenInvisible then
  begin
    Inc(FInternalChange);
    try
      SyncToParentValue;
    finally
      Dec(FInternalChange);
    end;
  end;
end;

procedure TCustomComboControlEnh.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    // Note that we are *keeping* the "WS_CHILD" flag which does cause some
    // complications because the Parent is not actually defined!  However this
    // is necessary to have the window operate the way we want it to.
    // The problem is that the VCL will check that a Parent handle exists when
    // this style is defined so we are forced to take some unusual precautions
    // in DropDown.
    WindowClass.Style := WindowClass.Style or CS_SAVEBITS;
    ExStyle := WS_EX_TOOLWINDOW or WS_EX_TOPMOST;
  end;
end;

procedure TCustomComboControlEnh.MouseDown(Button: TMouseButton; Shift:
  TShiftState; X, Y: Integer);
var
  Pt: TPoint;
begin
  Pt.X := X;
  Pt.Y := Y;
  if not PtInRect(ClientRect, Pt) then
    CloseUp;
  inherited;
end;

procedure TCustomComboControlEnh.DoChange;
// When the controller is in "list" mode the change events are generated
// from here...
begin
  if FInternalChange = 0 then
  begin
    if //(FController.Style = csDropDownList) and
      assigned(FController.OnChange) and
      FController.IsChangeEventEnabled then
      FController.OnChange(FController);
  end;
end;

procedure TCustomComboControlEnh.DoCloseUp;
begin
  if assigned(FController.FOnCloseUp) then
    FController.FOnCloseUp(FController);
end;

procedure TCustomComboControlEnh.DoBeforeDropDown;
begin
  if assigned(FController.FBeforeDropDown) then
    FController.FBeforeDropDown(FController);
end;

procedure TCustomComboControlEnh.DoDropDown;
begin
  if assigned(FController.FOnDropDown) then
    FController.FOnDropDown(FController);
end;

function TCustomComboControlEnh.CanChange: boolean;
begin
  if Controller.EditCanModify then
    Result := true
  else
    Result := false;
end;

procedure TCustomComboControlEnh.Change;
begin
  SyncParentToValue;
  //??? DoChange;
end;

procedure TCustomComboControlEnh.ParentKeyDown(var Key: Word; Shift:
  TShiftState);
begin
  // Just provide the default exits...
  if Active then
  begin
    if Key = VK_ESCAPE then
    begin
      CloseUp;
      Key := 0;
    end
    else
      if Key = VK_TAB then
      begin
        CloseUp;                        // Closeup but dont clear the key!
      end
      else
        if Key = VK_RETURN then
        begin
          if Controller.EditBeginModify then
          begin
            Change;
            CloseUp;
          end
          else
            CloseUp;
          Key := 0;
        end;
  end;
end;

procedure TCustomComboControlEnh.PositionControl;
var
  Pt: TPoint;
  //Rect: TRect;
begin
  // Initially try the origin at immediately below the left-bottom point of
  // the parent control...
  Pt.X := Controller.Left;
  Pt.Y := Controller.Top + Controller.Height;
  Pt := Controller.Parent.ClientToScreen(Pt);
  if (Pt.Y + Height) > Screen.DesktopHeight then
  begin                                 // If too high to fit
    Pt.Y := Pt.Y - (Controller.Height + Height); //   display above the parent
  end;
  if (Pt.X + Width) > Screen.DesktopWidth then
  begin                                 // If too wide to fit
    Pt.X := Pt.X - ((Pt.X + Width) - Screen.DesktopWidth); //   slide left just enough
  end;
  Left := Pt.X;
  Top := Pt.Y;
end;

procedure TCustomComboControlEnh.SizeControl;
// This will nearly always be overridden...
begin
  Height := 21;                         // just a height to get started with
  Width := Controller.Width;
end;

procedure TCustomComboControlEnh.SyncToParentValue;
begin
  // Nothing to do here
end;

procedure TCustomComboControlEnh.SyncParentToValue;
begin
  // Nothing to do here
end;

procedure TCustomComboControlEnh.SyncColors;
begin
  // abstract
end;

procedure TCustomComboControlEnh.CloseUp;
begin
  if Active then
  begin
    DoCloseUp;
    Visible := false; //Does NOT always work (if parent form is hidden) so...
    ShowWindow(Handle, SW_HIDE);
  end;
end;

procedure AllTabStopsOff( AWinControl: TWinControl );
var ii: integer;
begin
  for ii := 0 to AWinControl.ControlCount - 1 do
  begin
    if AWinControl.Controls[ii] is TWinControl then
    begin
      AllTabStopsOff( TWinControl(AWinControl.Controls[ii]) );
      TWinControl(AWinControl.Controls[ii]).TabStop := false;
    end;
  end;
end;

procedure TCustomComboControlEnh.DropDown;
begin
  // if not already visible...
  if not Active then
  begin
    if (not HandleAllocated) or (Parent <> Controller) then
    begin
      // Because of the WS_CHILD style we will need a parent while the
      // handle is created - after which we no longer want a parent.
      Parent := Controller;
    end;
    // Someone has been resetting our parent handle, so ensure it is clear
    Windows.SetParent(Handle, 0);
    if Controller.DropDownFontAuto then
      Font.Assign( Controller.Font )
    else
      Font.Assign( Controller.DropDownFont );
    DoBeforeDropDown; // allow developer to initialise before we drop
    if not ActiveWhenInvisible then     // if we haven't been kept upto date
      SyncToParentValue;                //   then synchronise now
    SizeControl;
    PositionControl;
    DoDropDown;
    Visible := true;
    AllTabStopsOff( Self );
  end
    // if already visible just reposition...
  else
  begin
    SizeControl;
    PositionControl;
  end;
end;



constructor TCustomComboEnh.Create(AOwner: TComponent);
begin
  FComboControl := nil;
  inherited Create(AOwner);
  FActiveWhenReadOnly := false;
  FSearching := 0;
  FAutoDropDown := false;
  FStyle := csDropDownList;
  FDropDownColor := clWindow;
  FDropDownFont := TFont.Create;
  FDropDownFont.OnChange := DoDropDownFontChange;
  FDropDownFontAuto := true;
  FComboControl := CreateComboControl;
end;

destructor TCustomComboEnh.Destroy;
begin
  if FDropDownFont <> nil then
  begin
    FDropDownFont.Free;
    FDropDownFont := nil;
  end;
  if FComboControl <> nil then
  begin
    DestroyComboControl(FComboControl);
    FComboControl := nil;
  end;
  inherited Destroy;
end;

procedure TCustomComboEnh.SyncDropBoxColors;
begin
  if (FComboControl <> nil) and FComboControl.Active then
    FComboControl.SyncColors;
end;

function TCustomComboEnh.GetDroppedDown: boolean;
begin
  if (FComboControl <> nil) and FComboControl.Active then
    Result := true
  else
    Result := false;
end;

procedure TCustomComboEnh.DoDropDownFontChange( Sender: TObject );
begin
  if (FComboControl <> nil) and FComboControl.Active then
  begin
    if FDropDownFontAuto then
      FComboControl.Font.Assign( Font )
    else
      FComboControl.Font.Assign( FDropDownFont );
  end;
end;

procedure TCustomComboEnh.SetDropDownFont( AVal: TFont );
begin
  FDropDownFont.Assign( AVal );
  FDropDownFontAuto := false;
end;

function TCustomComboEnh.IsDropDownFontStored: boolean;
begin
  Result := not FDropDownFontAuto;
end;

procedure TCustomComboEnh.SetDropDownFontAuto( AVal: boolean );
begin
  if AVal <> FDropDownFontAuto then
  begin
    FDropDownFontAuto := AVal;
    if FDropDownFontAuto then
      FDropDownFont.Assign( Font );
  end;
end;

function TCustomComboEnh.EditBeginModify: boolean;
begin
  Result := inherited EditBeginModify;
  if Result and AutoDropDown then
    ShowDropDown;
end;

function TCustomComboEnh.ShowDropDown: boolean;
begin
  if (FComboControl <> nil) and (not DroppedDown) and Focused and
    (FComboControl.FController = Self) and
    (EditCanModify or ActiveWhenReadOnly) then
    FComboControl.DropDown;
  Result := DroppedDown;
end;

procedure TCustomComboEnh.HideDropDown;
begin
  if DroppedDown then
    FComboControl.CloseUp;
end;

function TCustomComboEnh.HasActiveChild: boolean;
begin
  Result := GetDroppedDown;
end;

procedure TCustomComboEnh.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if (FComboControl <> nil) and (FComboControl.Active and not AutoDropDown) then
    Message.Result := Message.Result or DLGC_WANTTAB;
end;

procedure TCustomComboEnh.CMCancelMode(var Message: TCMCancelMode);
// Note: VCL does not handle my setup very well, if an item is clicked on
// the dropbox then the VCL thinks it is the ActiveControl even though
// it has not actually got the focus.  So we have to do a bit of extra
// work to make sure the VCL knows who is really the active control.
var
  tmpForm: TCustomForm;
begin
  with Message do
  begin
    // This is what I had, and it worked as long as no normally focusable
    // item was clicked on in the dropbox...
    //if (FComboControl <> nil) and
    //    (Sender <> FComboControl) and
    //    (Sender <> Self) and
    //    (not FComboControl.ContainsControl( Sender ) ) then begin
    //    FComboControl.CloseUp;
    //    end;
    // Should only get a cancel message from ourselves when a child
    // has previously sent a cancel message.  We use this to indicate
    // that the Form.ActiveControl needs updating...
    if Sender = Self then
    begin
      tmpForm := GetParentForm(Self);
      if CanFocus then
        tmpForm.ActiveControl := Self;
    end
    else
      if FComboControl.ContainsControl(Sender) then
      begin
        // Post a message back to ourselves to notify of the need
        // to update the form ActiveControl
        PostMessage(Handle, CM_CANCELMODE, 0, longint(Self));
      end
      else
      begin
        FComboControl.CloseUp;
        inherited;
      end;
  end;
end;

procedure TCustomComboEnh.WMSetFocus(var Message: TWMSetFocus);
begin
  if (FComboControl <> nil) then
  begin
    if FComboControl.Controller <> Self then
      FComboControl.FController := Self;
  end;
  inherited;
  if AutoDropDown then
    ShowDropDown;
end;

procedure TCustomComboEnh.WMKillFocus(var Message: TWMKillFocus);
begin
  HideDropDown;
  inherited;
end;

function TCustomComboEnh.DoMouseWheelDown(Shift: TShiftState;
  MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);
  if (not Result) and (FComboControl <> nil) and
    (FComboControl.Active or FComboControl.ActiveWhenInvisible) and
    (not (eeoNoDefaultMouseWheel in Options)) then
    Result := FComboControl.DoMouseWheelDown(Shift, MousePos);
end;

function TCustomComboEnh.DoMouseWheelUp(Shift: TShiftState;
  MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);
  if (not Result) and (FComboControl <> nil) and
    (FComboControl.Active or FComboControl.ActiveWhenInvisible) and
    (not (eeoNoDefaultMouseWheel in Options)) then
    Result := FComboControl.DoMouseWheelUp(Shift, MousePos);
end;

function TCustomComboEnh.DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
  MousePos: TPoint): Boolean;
begin
  Result := inherited DoMouseWheel(Shift, WheelDelta, MousePos);
  if (not Result) and (FComboControl <> nil) and
    (FComboControl.Active or FComboControl.ActiveWhenInvisible) and
    (not (eeoNoDefaultMouseWheel in Options)) then
    Result := FComboControl.DoMouseWheel(Shift, WheelDelta, MousePos);
end;

function TCustomComboEnh.GetButtonStyleDefault: TEditEnhButtonStyle;
begin
  Result := eebDownArrow;
end;

function TCustomComboEnh.GetButton1KeyDefault: TShortCut;
begin
  Result := TextToShortCut('Alt+Down');
end;

function TCustomComboEnh.AllowButtonClickOnFocus: boolean;
begin
  if AutoDropDown then
    Result := false
  else
    Result := inherited AllowButtonClickOnFocus;
end;

procedure TCustomComboEnh.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if FComboControl <> nil then
  begin
    FComboControl.DoParentKeyDown(Key, Shift);
  end;
  inherited;
end;

function TCustomComboEnh.GetButtonEnabled: boolean;
begin
  Result := inherited GetButtonEnabled;
  if Result then
  begin
    if ActiveWhenReadOnly or EditCanModify then
    begin
      Result := true;
    end
    else
      Result := false;
  end;
end;

function TCustomComboEnh.DoTextChange: boolean;
// When the control is in "list" mode (csDropDownList) we process the
// input as incremental searches of the list - since all input must
// correspond to a list entry.  In this case the function always returns
// "false" to prevent TCustomEditEnh from generating change events, we will
// generate the change events from the list itself - as list selection actually
// changes.
//
// At this stage, if mask processing is enabled while in list mode the
// results are somewhat unpredictable - due to the re-selection of text
// performed by this procedure, which may disrupt mask processing.
//
// When the control is not in list mode then input is interpretted only as
// input to the text box like a normal edit control.  The listbox becomes
// just a method of looking up values to assign to the text box.
// IMPORTANT:  When not in list mode the "Display=Value" capability of the
// items string list is redundant because there is no guarantee that the
// input to the edit box corresponds to a particular entry in the list.
var
  SelStart, SelStop: integer;
  WasSelectAll: boolean;
  tmpText: string;
  FndPos: integer;
begin
  if Style = csDropDownList then
  begin
    if (FSearching = 0) and not (csDesigning in ComponentState) and
      HandleAllocated then
    begin
      Inc(FSearching);
      try
        tmpText := Text;
        SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
        if (SelStart = 0) and (SelStop > 0) and (SelStop = Length(tmpText)) then
          WasSelectAll := true
        else
          WasSelectAll := false;
        if WasSelectAll then
          FndPos := IncrementalSearch(tmpText, -1)
        else
          FndPos := IncrementalSearch(tmpText, SelStart);
        if FndPos < 0 then
        begin
          if WasSelectAll then
            SelectAll
          else
          begin
            SelStart := 0;
            SendMessage(Handle, EM_SETSEL, SelStart, SelStart);
            SyncPositionToCaret;
          end;
        end
        else
        begin
          SelStart := FndPos;
          SendMessage(Handle, EM_SETSEL, SelStart, SelStart);
          SyncPositionToCaret;
        end;
      finally
        Dec(FSearching);
      end;
    end;
    Result := false;                    // we generate our own change events
  end
  else
  begin
    if (FSearching = 0) and not (csDesigning in ComponentState) and
      HandleAllocated then
      NoSearchChange;
    Result := inherited DoTextChange;
  end;
end;

procedure TCustomComboEnh.DoButton1;
begin
  if assigned(OnButton1) then
  begin
    inherited DoButton1;
  end
  else
    if FComboControl <> nil then
    begin
      if FComboControl.Active then
        FComboControl.CloseUp
      else
        FComboControl.DropDown;
    end;
end;

procedure TCustomComboEnh.IBG_SetColor( AColor: TColor );
begin
  inherited IBG_SetColor( AColor );
  SyncDropBoxColors;
end;

function TCustomComboEnh.IncrementalSearch(TxtVal: string; ChkLen: integer):
  integer;
begin
  Result := -1;
end;

procedure TCustomComboEnh.NoSearchChange;
begin
  // abstract - called when no incremental search (Style <> csDropDownList)
end;

function TCustomComboEnh.CreateComboControl: TCustomComboControlEnh;
// Will need to override this in all instances to give a useful combo control.
// See notes below as to whether you actually want to create an instance of
// the combo control for every edit control, or whether you can get away with
// a single instance for every editor.
begin
  Result := TCustomComboControlEnh.Create(Self);
  //Result.Parent := Self;
end;

procedure TCustomComboEnh.DestroyComboControl(CurrentControl:
  TCustomComboControlEnh);
// In many instances a single combo-control can be shared amongst all instances
// of an edit control, in which case a derived type may not wish to actually
// destroy the combo-control.  I will leave this upto particular
// implementations.
begin
  CurrentControl.Free;
end;


procedure TListBoxEnh.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  i: integer;
  Pt: TPoint;
begin
  inherited;
  Pt.X := X;
  Pt.Y := Y;
  i := ItemAtPos(Pt, false);
  ItemIndex := i;
end;

procedure TListBoxEnh.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y:
  Integer);
var
  i: integer;
  Pt: TPoint;
begin
  inherited;
  Pt.X := X;
  Pt.Y := Y;
  i := ItemAtPos(Pt, false);
  ItemIndex := i;
  with TCustomComboListBoxEnh(Parent) do
  begin
    if TCustomComboBoxEnh(Controller).ItemIndex <> i then
    begin
      if Controller.EditBeginModify then
        Change;
    end;
    CloseUp;
  end;
end;

constructor TCustomComboListBoxEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ActiveWhenInvisible := false; // We use a shared listbox so cannot be active
  // until visible
  FList := TListBoxEnh.Create(Self);
  // It is replicatable because there is only one!
  FList.ControlStyle := FList.ControlStyle + [csReplicatable];
  FList.Parent := Self;
  FList.Align := alClient;
  FList.BorderStyle := bsNone;
  FList.Visible := true;
  FList.ExtendedSelect := false;
  FList.MultiSelect := false;
end;

destructor TCustomComboListBoxEnh.Destroy;
begin
  if FList <> nil then
  begin
    FList.Free;
    FList := nil;
  end;
  inherited Destroy;
end;

// I chose to use [ssCtrl] shift calls to ParentKeyDown in order to get the
// slide effect rather than the scroll effect.  This just seems to work better
// with the way the mouse interacts with listbox (see also TListBoxEnh above).
function TCustomComboListBoxEnh.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean;
var
  k: WORD;
begin
  k := VK_DOWN;
  ParentKeyDown(k, [ssCtrl]);  // [ssCtrl] uses slide not scroll
  Result := (k = 0); // Was processed if k was reset to 0
  //Result := FList.DoMouseWheelDown(Shift, MousePos);
end;

function TCustomComboListBoxEnh.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean;
var
  k: WORD;
begin
  k := VK_UP;
  ParentKeyDown(k, [ssCtrl]);  // [ssCtrl] uses slide not scroll
  Result := (k = 0); // Was processed if k was reset to 0
  //Result := FList.DoMouseWheelUp(Shift, MousePos);
end;

function TCustomComboListBoxEnh.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
begin
  //Result := FList.DoMouseWheel(Shift, WheelDelta, MousePos);
  Result := false;
end;

procedure TCustomComboListBoxEnh.ParentKeyDown(var Key: Word; Shift: TShiftState);

  function PageSize: integer;
  var
    ih: integer;
  begin
    ih := SendMessage(FList.Handle, LB_GETITEMHEIGHT, 0, 0);
    Result := FList.Height div ih; // number of full items visible (page size)
  end;

  procedure ScrollUp(AMove: integer);
  var
    i: integer;
  begin
    i := FList.ItemIndex - AMove;     // We allow scroll up to scroll to none
    if i < -1 then
      i := -1;
    FList.ItemIndex := i;
    Key := 0; // Mark as used
  end;

  procedure ScrollDown(AMove: integer);
  var
    i: integer;
  begin
    i := FList.ItemIndex + AMove;
    if i >= FList.Items.Count then
      i := FList.Items.Count - 1;
    FList.ItemIndex := i;
    Key := 0; // Mark as used
  end;

  procedure SlideUp(AMove: integer);
  var
    i: integer;
  begin
    i := FList.TopIndex - AMove;
    if i < 0 then
      i := 0;
    if i >= FList.Items.Count then
      i := FList.Items.Count - 1;
    FList.TopIndex := i;
    Key := 0; // Mark as used
  end;

  procedure SlideDown(AMove: integer);
  var
    i: integer;
  begin
    i := FList.TopIndex + AMove;
    if i > (FList.Items.Count - PageSize) then
    begin
      i := FList.Items.Count - PageSize;
      if i < 0 then
        i := 0;
    end;
    if i >= FList.Items.Count then
      i := FList.Items.Count - 1;
    FList.TopIndex := i;
    Key := 0; // Mark as used
  end;

begin
  inherited ParentKeyDown(Key, Shift);
  if (Key <> 0) and Active then
  begin
    if Shift = [ssCtrl] then
    begin
      if Key = VK_PRIOR then       SlideUp(PageSize)
      else if Key = VK_UP then     SlideUp(1)
      else if Key = VK_DOWN then   SlideDown(1)
      else if Key = VK_NEXT then   SlideDown(PageSize);
    end
    else if Shift = [] then
    begin
      if Key = VK_PRIOR then       ScrollUp(PageSize)
      else if Key = VK_UP then     ScrollUp(1)
      else if Key = VK_DOWN then   ScrollDown(1)
      else if Key = VK_NEXT then   ScrollDown(PageSize);
    end;
  end;
end;

procedure TCustomComboListBoxEnh.SizeControl;
var
  i, j, k: integer;
begin
  if TCustomComboBoxEnh(Controller).ListWidth <= 0 then
    Width := Controller.Width
  else
    Width := TCustomComboBoxEnh(Controller).ListWidth;
  i := TCustomComboBoxEnh(Controller).DropDownRows;
  if i < 0 then
    i := 1;                             // reset to minimal height on invalid
  j := SendMessage(FList.Handle, LB_GETITEMHEIGHT, 0, 0);
  k := Height - FList.Height;           // Get the border size adjustment
  Height := k + (i * j);
end;

procedure TCustomComboListBoxEnh.SyncToParentValue;
var
  i: integer;
  tmpStr: string;
begin
  if FInternalChange = 0 then
  begin
    // We are using a shared listbox, so clear and re-initialise the listbox
    // contents first...
    with Controller as TCustomComboBoxEnh do
    begin
      FList.Clear;
      FList.Sorted := Sorted;
      FList.Columns := ListColumns;
      FList.TabWidth := TabWidth;
      if ListOwnerDraw then
        FList.Style := lbOwnerDrawFixed
      else
        FList.Style := lbStandard;
      FList.OnDrawItem := DoListDrawItem;
      Self.SyncColors;
      for i := 0 to (Items.Count - 1) do
      begin
        tmpStr := Items.Names[i];
        if tmpStr = '' then
          FList.Items.Add(Items.Strings[i])
        else
          FList.Items.Add(tmpStr);
      end;
      if Items.Count < ListRows then
      begin
        FDropDownRows := Items.Count;
      end
      else
        FDropDownRows := ListRows;
      FList.ItemIndex := ItemIndex;
    end;
  end
  else
  begin
    with Controller as TCustomComboBoxEnh do
    begin
      if FList.ItemIndex <> ItemIndex then
      begin
        // If gone back to no selection force the list to
        // back to the top...
        if (ItemIndex < 0) and (FList.Items.Count > 0) then
          FList.ItemIndex := 0;
        FList.ItemIndex := ItemIndex;
      end;
    end;
  end;
end;

procedure TCustomComboListBoxEnh.SyncParentToValue;
begin
  TCustomComboBoxEnh(Controller).ItemIndex := FList.ItemIndex;
end;

procedure TCustomComboListBoxEnh.SyncColors;
begin
  with Controller as TCustomComboEnh do
  begin
    if DropDownColor = clNone then
      Self.FList.Color := Color
    else
      Self.FList.Color := DropDownColor;
  end;
end;


constructor TCustomComboBoxEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FListCols := 0;
  FListRows := 8;
  FDropDownRows := 0;
  FListWidth := 0;
  FItems := TStringList.Create;
  FItemIndex := -1;
  FTabWidth := 0;
end;

destructor TCustomComboBoxEnh.Destroy;
begin
  if FItems <> nil then
  begin
    FItems.Free;
    FItems := nil;
  end;
  inherited Destroy;
end;

procedure TCustomComboBoxEnh.Loaded;
begin
  inherited;
  LoadItemList;
end;

procedure TCustomComboBoxEnh.KeyDown(var Key: Word; Shift: TShiftState);
begin
  //if (ComboControl <> nil) and (not ComboControl.Active) and
  //    (not ComboControl.ActiveWhenInvisible) then begin
  if EditCanModify then
  begin
    if Shift = [] then
    begin
      case Key of
        VK_DOWN:
          begin
            if (ItemIndex < (Items.Count - 1)) and EditBeginModify then
              ItemIndex := ItemIndex + 1;
            Key := 0;
          end;
        VK_UP:
          begin
            // Dont allow null select with up key, so removed
            // the ItemIndex >= 0 and made ItemIndex > 0
            if (ItemIndex > 0) and EditBeginModify then
              ItemIndex := ItemIndex - 1;
            Key := 0;
          end;
        VK_NEXT:
          begin
            if (ItemIndex < (Items.Count - 1)) and EditBeginModify then
            begin
              if (ItemIndex + FListRows) >= Items.Count then
                ItemIndex := Items.Count - 1
              else
                ItemIndex := ItemIndex + FListRows;
            end;
            Key := 0;
          end;
        VK_PRIOR:
          begin
            if (ItemIndex >= 0) and EditBeginModify then
            begin
              if (ItemIndex - FListRows) < -1 then
                ItemIndex := -1
              else
                ItemIndex := ItemIndex - FListRows;
            end;
            Key := 0;
          end;
      end;
    end;
    if Shift = [ssCtrl] then
    begin
      case Key of
        VK_HOME:
          begin
            if EditBeginModify then
              ItemIndex := 0;
            Key := 0;
          end;
        VK_END:
          begin
            if EditBeginModify then
              ItemIndex := Items.Count - 1;
            Key := 0;
          end;
      end;
    end;
  end;
  inherited;
end;

procedure TCustomComboBoxEnh.LoadItemList;
begin
  // abstract - provided to allow the automatic loading of the item
  // list from some other source if required.  Called by Loaded after
  // all standard resource definitions have been loaded.
end;

function TCustomComboBoxEnh.IsSortedStored: boolean;
begin
  Result := (Sorted and AreItemsStored);
end;

function TCustomComboBoxEnh.AreItemsStored: boolean;
begin
  Result := true;
  // abstract - provided to allow automation of items without wasting
  // resource space with storing automated values.
end;

function TCustomComboBoxEnh.IncrementalSearch(TxtVal: string; ChkLen: integer):
  integer;
var
  i, j, Lim, FoundAt, FoundPos, RC: integer;
  SrchVal, tmpStr: string;
begin
  if ChkLen > 0 then
    SrchVal := Copy(TxtVal, 1, ChkLen)
  else
    SrchVal := TxtVal;
  // We search through our string list for the closest (>=) match in the
  // list of "names"...
  // I could optimise this for sorted strings but I wont bother for the moment.
  FoundAt := -1;
  FoundPos := -1;
  for i := 0 to (Items.Count - 1) do
  begin
    tmpStr := Items.Names[i];
    if tmpStr = '' then
      tmpStr := Items.Strings[i];
    if Length(tmpStr) > Length(SrchVal) then
      Lim := Length(SrchVal)
    else
      Lim := Length(tmpStr);
    for j := 1 to Lim do
    begin
      RC := CompareText(tmpStr[j], SrchVal[j]);
      if RC <> 0 then
        break
      else
      begin
        if j > FoundPos then
        begin
          FoundAt := i;
          FoundPos := j;
        end;
      end;
    end;
  end;
  if FoundAt >= 0 then
    ItemIndex := FoundAt
  else
    ItemIndex := -1;
  Result := FoundPos;
end;

procedure TCustomComboBoxEnh.NoSearchChange;
begin
  ItemIndex := -1;
end;

function TCustomComboBoxEnh.CreateComboControl: TCustomComboControlEnh;
begin
  if ComboBoxEnh_ListBox = nil then
  begin
    ComboBoxEnh_ListBox := TCustomComboListBoxEnh.Create(nil);
  end;
  Inc(ComboBoxEnh_ListBoxRefCount);
  Result := ComboBoxEnh_ListBox;
end;

procedure TCustomComboBoxEnh.DestroyComboControl(CurrentControl:
  TCustomComboControlEnh);
begin
  if ComboBoxEnh_ListBox.Parent = Self then
  begin
    ComboBoxEnh_ListBox.Parent := nil;
  end;
  if CurrentControl = ComboBoxEnh_ListBox then
  begin
    Dec(ComboBoxEnh_ListBoxRefCount);
    if ComboBoxEnh_ListBoxRefCount = 0 then
    begin
      ComboBoxEnh_ListBox.Free;
      ComboBoxEnh_ListBox := nil;
    end;
  end;
end;
//---------------------------------------------------------------------------

procedure TCustomComboBoxEnh.SetItems(AValue: TStringList);
begin
  FItems.Clear;
  FItems.AddStrings(AValue);
  SetItemIndex(FItemIndex);
end;

procedure TCustomComboBoxEnh.SetItemIndex(AValue: integer);
var
  OldIdx: integer;
  HoldMod: boolean;
  tmpStr: string;
begin
  Inc(FSearching);
  try
    OldIdx := FItemIndex;
    if (AValue < 0) or (AValue >= Items.Count) then
      FItemIndex := -1
    else
      FItemIndex := AValue;
    // Set the text value without changing the modified status
    HoldMod := Modified;
    if (FItemIndex < 0) then
    begin
      // dont clear the text if not a dropdown list
      if Style = csDropDownList then
        Text := '';
    end
    else
    begin
      tmpStr := Items.Names[FItemIndex];
      if tmpStr = '' then
        tmpStr := Items.Strings[FItemIndex];
      Text := tmpStr;
    end;
    if HoldMod <> Modified then
      Modified := HoldMod;
    // Now update the modified status according to a change in index
    if OldIdx <> FItemIndex then
    begin
      if (ComboControl <> nil) then
        ComboControl.DoParentValueChanged;
      Modified := true;
      Change; // We stop the text change indicating change so this is to
    end; // generate a change event when the selection has changed.
  finally
    Dec(FSearching);
  end;
end;

function TCustomComboBoxEnh.GetItemValue: string;
var
  tmpStr: string;
begin
  if (ItemIndex < 0) or (ItemIndex >= Items.Count) then
  begin
    if Style = csDropDownList then
      Result := ''
    else
      Result := Text;
  end
  else
  begin
    tmpStr := Items.Names[ItemIndex];
    if tmpStr = '' then
      Result := Items.Strings[ItemIndex]
    else
      Result := Items.Values[tmpStr];
  end;
end;

procedure TCustomComboBoxEnh.SetItemValue(AValue: string);
var
  tmpIdx: integer;
  OldVal: string;
begin
  if Style = csDropDownList then
    ItemIndex := GetComboBoxEnhDisplayIndex(Items, AValue)
  else
  begin
    OldVal := GetItemValue;
    tmpIdx := ItemIndex;
    ItemIndex := GetComboBoxEnhDisplayIndex(Items, AValue);
    // If change not initiated via ItemIndex change then
    // specifically call change if the ItemValue has changed.
    if (tmpIdx < 0) and (ItemIndex < 0) and
      (OldVal <> AValue) then
    begin
      Text := AValue;
      Change;
    end;
  end;
end;

function TCustomComboBoxEnh.GetItemValues(Index: integer): string;
var
  tmpStr: string;
begin
  if (Index < 0) or (Index >= Items.Count) then
    Result := ''
  else
  begin
    tmpStr := Items.Names[Index];
    if tmpStr = '' then
      Result := Items.Strings[Index]
    else
      Result := Items.Values[tmpStr];
  end;
end;

procedure TCustomComboBoxEnh.SetItemValues(Index: integer; AValue: string);
var
  tmpStr: string;
begin
  if (Index >= 0) and (Index < Items.Count) then
  begin
    tmpStr := Items.Names[Index];
    if tmpStr = '' then
      Items.Strings[Index] := AValue
    else
      Items.Strings[Index] := tmpStr + '=' + AValue;
  end;
end;

function TCustomComboBoxEnh.GetSorted: boolean;
begin
  Result := TStringList(FItems).Sorted;
end;

procedure TCustomComboBoxEnh.SetSorted(AValue: boolean);
begin
  TStringList(FItems).Sorted := AValue;
end;

procedure TCustomComboBoxEnh.DoListDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  if assigned(FOnListDrawItem) then
    FOnListDrawItem(Self, Control as TListBox, Index, Rect, State);
end;

function GetComboBoxEnhDisplayString(const StrList: TStrings; AValue: string):
  string;
var
  i: integer;
  tmpStr, tmpName: string;
begin
  Result := '';
  if AValue <> '' then
  begin
    for i := 0 to (StrList.Count - 1) do
    begin
      tmpName := StrList.Names[i];
      if tmpName = '' then
      begin
        tmpName := StrList.Strings[i];
        tmpStr := tmpName;
      end
      else
      begin
        tmpStr := StrList.Values[tmpName];
      end;
      if tmpStr = AValue then
      begin
        Result := tmpName;
        break;
      end;
    end;
  end;
end;

function GetComboBoxEnhDisplayIndex(const StrList: TStrings; AValue: string):
  integer;
var
  i: integer;
  tmpStr, tmpName: string;
begin
  Result := -1;
  if AValue <> '' then
  begin
    for i := 0 to (StrList.Count - 1) do
    begin
      tmpName := StrList.Names[i];
      if tmpName = '' then
      begin
        tmpName := StrList.Strings[i];
        tmpStr := tmpName;
      end
      else
      begin
        tmpStr := StrList.Values[tmpName];
      end;
      if tmpStr = AValue then
      begin
        Result := i;
        break;
      end;
    end;
  end;
end;


initialization;
  ComboBoxEnh_ListBox := nil;
  ComboBoxEnh_ListBoxRefCount := 0;


finalization;
  ComboBoxEnh_ListBox := nil;
  ComboBoxEnh_ListBoxRefCount := 0;

end.
