{ Telesis_DateTimeEnh.pas                         SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary DateTime panel and combo-styled control.
<br>
<br>
<strong>History</strong><br>
<br>
12-Feb-2008 - MouseWheel support.<br>
<br>
19-Dec-2007 - Direct assigning DateTime property now always results in
full assignment of text.<br>
<br>
24-Apr-2007 - The internal stringgrid was still stealing focus at odd times,
making the datetimeenh derivatives very difficult to use on TIB_Grid.  I think
I have solved the issue now - by using a derivative of stringgrid, that
overrides CanFocus to return false, when used on combo controls.<br>
<br>
23-May-2006 - Attempt to ensure that edit mode always has access to full
date details - revert to shortdateformat if DateFormat value is missing any
of day, month or year attributes.  Also take account of possible editmask
property (use DisplayText to avoid problems when not editing).  Also correct
and annoying focus issue - having clicked on the grid during dropdown the
next tab would carry you back to the first control on the form (because, I
believe, the stringgrid was stealing focus.<br>
<br>
21-Apr-2006 - Adjust internal update to allow data-aware version to reset
the datetime property without necessarily adjusting the text.<br>
<br>
19-Jun-2005 - Extended the default display formatting to check whether
seconds are specified (in which case use long time format).<br>
<br>
22-Sep-2002 - Tried to improve the autosizing of the DateTimeEnh dropbox,
not sure if the change will cause excessive flicker but we will see.<br>
<br>
28-Dec-2001 - Inspired by a request from Marcin Bury the DateTime control
has now significantly changed its processing.  It now carries DateTime,
DateFormat and DateInputLayout properties.  DateTime can be used to set and
read the actual datetime value of the control.  Indeed the DateTime property
is now the actual value of the control, with the Text property simply being
a display of the current DateTime value.  DateInputLayout is used to assist
with determining input formatting when input can differ to display.<br>
<br>
23-Aug-2001 - Support to synchronise colors on drop box.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_DateTimeEnh;

{$INCLUDE Telesis.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  ExtCtrls,
  Mask,
  Buttons,
  Grids,
  Telesis_MaskEnh,
  Telesis_EditEnh,
  Telesis_ComboEnh;

type
{: @summary Display options for DateTimeEnh controls.}
  TDateTimeEnhOptions = set of (
    dtoShowButtons,
    dtoShowTime,
    dto24HourTime,
    dtoShowSeconds
    );

{: @summary Control which day of the week is considered the first day
of the week.}
  TDateTimeEnhStartDay = (
    dtsSunday,
    dtsMonday,
    dtsTuesday,
    dtsWednesday,
    dtsThursday,
    dtsFriday,
    dtsSaturday
    );

{: @summary Base class for simply datetime selection panel.}
  TCustomDateTimePnlEnh = class(TCustomPanel)
  private
    FHeaderPnl: TPanel;
    FClientPnl: TPanel;
    FTimePnl: TPanel;
    FButtonPnl: TPanel;
    FDateGrid: TStringGrid;
    FMonthEdit: TComboBoxEnh;
    FYearEdit: TEditEnh;
    FHourEdit: TEditEnh;
    FMinuteEdit: TEditEnh;
    FSecondEdit: TEditEnh;
    FAmPmEdit: TComboBoxEnh;
    FTodayButton: TSpeedButton;
    FTodayLabel: string;
    FNowLabel: string;
    FClearLabel: string;
    FClearButton: TSpeedButton;
    FAdjusting: integer;
    FOptions: TDateTimeEnhOptions;
    FStartDay: TDateTimeEnhStartDay;
    FTitleColor: TColor;
    FDateCellColor: TColor;
    FYear: integer;
    FMonth: integer;
    FDay: integer;
    FHour: integer;
    FMinute: integer;
    FSecond: integer;
    FAmPm: integer;
    FOnChange: TNotifyEvent;
    FOnPopupClose: TNotifyEvent;

    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure SetTitleColor(AValue: TColor);
    procedure SetDateCellColor(AValue: TColor);
    procedure SetOptions(AValue: TDateTimeEnhOptions);
    procedure SetStartDay(AValue: TDateTimeEnhStartDay);
    function GetTimeStamp: TDateTime;
    procedure SetTimeStamp(AValue: TDateTime);
    procedure TodayClick(Sender: TObject);
    procedure ClearClick(Sender: TObject);

  protected
    procedure CreateWnd; override;
    procedure Change; virtual;

    function CreatePanelControl(CtrlParent: TWinControl): TPanel;
    function CreateEditControl(CtrlParent: TPanel): TEditEnh;
    function CreateComboControl(CtrlParent: TPanel): TComboBoxEnh;
    function CreateButtonControl(CtrlParent: TPanel): TSpeedButton;
    procedure DrawDateCell(Sender: TObject; Col, Row: longint; Rect: TRect; State: TGridDrawState);
    procedure SelectDateCell(Sender: TObject; Col, Row: longint; var CanSelect: boolean);
    procedure EditChange(Sender: TObject);
    procedure Incrementer(Sender: TObject);
    procedure Decrementer(Sender: TObject);

    function GetCellTitle(Col, DrawWidth: integer): string;
    function GetCellDateValue(Col, Row: integer): integer;
    procedure UpdateGridSelect;
    procedure UpdateDisplay;

    property TitleColor: TColor read FTitleColor write SetTitleColor default clInfoBk;
    property DateCellColor: TColor read FDateCellColor write SetDateCellColor default clWindow;
    property Options: TDateTimeEnhOptions read FOptions write SetOptions default [dtoShowButtons, dtoShowTime, dtoShowSeconds];
    property StartDay: TDateTimeEnhStartDay read FStartDay write SetStartDay default dtsMonday;
    property TimeStamp: TDateTime read GetTimeStamp write SetTimeStamp;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnPopupCloseEvent: TNotifyEvent read FOnPopupClose write FOnPopupClose;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure AdjustLayout;
    procedure Today;
    procedure Clear;
    procedure IncDay;
    procedure IncWeek;
    procedure IncMonth;
    procedure IncYear;
    procedure DecDay;
    procedure DecWeek;
    procedure DecMonth;
    procedure DecYear;

  published

  end;                                  // TCustomDateTimePnlEnh

{: @summary Simple DateTime selection panel.}
  TDateTimePnlEnh = class(TCustomDateTimePnlEnh)
    {$DEFINE PUBLISH_CustomDateTimePnlEnh}
    {$INCLUDE Telesis_Publish.INC}
  end;


{: @summary Dropbox used by TDateTimeEnh.}
  TCustomDateTimeBoxEnh = class(TCustomComboControlEnh)
  private
    FDateTimePnl: TDateTimePnlEnh;
    procedure DateTimeCloseEvent(Sender: TObject);

  protected
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    procedure ParentKeyDown(var Key: Word; Shift: TShiftState); override;
    procedure SizeControl; override;
    procedure SyncToParentValue; override;
    procedure SyncParentToValue; override;
    procedure SyncColors; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure DropDown; override;

  end;

{: @summary Base class for TDateTimeEnh
<br>
@desc Setup to allow input of datetime or just date values.<br>
<br>
This control has been setup to support relatively freeform input of date and
time text, it attempts to extract the relevent information using the
ExtractDateTime function - refer to that function for full details.<br>
<br>
Please read the details for the base class if you intend to use the OnDropDown
and BeforeDropDown event handlers: <see Class=TCustomComboEnh> }
  TCustomDateTimeEnh = class(TCustomComboEnh)
  private
    FDateOptions: TDateTimeEnhOptions;
    FDateStartDay: TDateTimeEnhStartDay;
    FDateTitleColor: TColor;
    FDateDateCellColor: TColor;
    FDateTime: TDateTime;
    FDateFormat: string;
    FDateInputLayout: string;
    FDTUpdate: integer; // track direct/internal updates

    procedure SetDateTime( AVal: TDateTime );
    procedure SetDateFormat( AVal: string );
    procedure SetDateInputLayout( AVal: string );
    function GetDateTimeDisplayText( ADateTime: TDateTime ): string;

  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;

    function IncrementalSearch(TxtVal: string; ChkLen: integer): integer;
      override;

    function CreateComboControl: TCustomComboControlEnh; override;
    procedure DestroyComboControl(CurrentControl: TCustomComboControlEnh);
      override;

    function DoTextChange: boolean; override;
    procedure DoValidateEdit(var ErrCode: integer); override;
    function TextToDateTime( AStr: string; var ADateTime: TDateTime ): integer;

{: @summary DateTime display options.}
    property DateOptions: TDateTimeEnhOptions read FDateOptions write
      FDateOptions default [dtoShowButtons, dtoShowTime, dtoShowSeconds];
{: @summary Choose which day is considered first day of week.}
    property DateStartDay: TDateTimeEnhStartDay read FDateStartDay write
      FDateStartDay default dtsMonday;
{: @summary Choose color for title area of dropbox.}
    property DateTitleColor: TColor read FDateTitleColor write FDateTitleColor
      default clInfoBk;
{: @summary Choose color for date grid cell color of dropbox.}
    property DateDateCellColor: TColor read FDateDateCellColor write
      FDateDateCellColor default clWindow;
{: @summary Get/Set the DateTime value of the control directly.
<br>
@desc The control attempts to synchronise the text of the control with this
TDateTime property.  The synchronisation occurs when the text is validated, so
the value of this DateTime property is only predictable when Modified = false.
To get a reliable value you need to do something similar to...<br>
<br>
  if DateTimeEnh1.Modified then<br>
    DateTimeEnh1.ValidateEdit;<br>
  MyVar := DateTimeEnh1.DateTime;<br> }
    property DateTime: TDateTime read FDateTime write SetDateTime;
{: @summary Override date display formatting.
<br>
@desc If left blank then the operating system default locale is used
to determine the display format.  If an editmask is defined it must be
compatible with the DateFormat value or input errors may occur.<br>
<br>
In the IBO aware variation of this control the DateFormat is set from
the field DisplayFormat (as determined by the connection or query
FieldsDisplayFormat setting).}
    property DateFormat: string read FDateFormat write SetDateFormat;
{: @summary Support an alternative input format.
<br>
@desc If you have a special display format, but still want to allow users to
input the date in a different form, then you can provide the layout of the
alternative input in this property.<br>
<br>
Refer to the ExtractDateTime function for a full description of values
supported in this string property.<br>
<br>
An example of using this property would be as follows...<br>
<br>
If you have a DateFormat similar to: 'yyyy mm dd'<br>
<br>
then in Australia you may want to support input of dates as 'dd/mm/yyyy'
(the date without the day name prefix).<br>
<br>
to do this simply provide a DateInputLayout of 'dmy'.<br>
<br>
Note that I do not recommend mixing American and European date ordering of
dates since this may introduce inconsistent results and confuse the user.}
    property DateInputLayout: string read FDateInputLayout write SetDateInputLayout;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published

  end;                                  // TCustomDateTimeEnh

{: @summary Published variation of TCustomDateTimeEnh.}
  TDateTimeEnh = class(TCustomDateTimeEnh)
    {$DEFINE PUBLISH_CustomDateTimeEnh}
    {$DEFINE PUBLISH_MaskEnh}
    {.$.DEFINE PUBLISH_Text - use DateTime instead}
    {$INCLUDE Telesis_Publish.INC}
  end;                                  // TDateTimeEnh


var
{: @summary The shared dropbox used by TDateTimeEnh.
<br>
@desc DO NOT ALTER THIS VARIABLE UNLESS WRITING A NEW DERIVATION.}
  DateTimeEnh_Panel: TCustomDateTimeBoxEnh;
{: @summary The number of TDateTimeEnh components accessing the shared
dropbox (DateTimeEnh_Panel).
<br>
@desc DO NOT ALTER THIS VARIABLE UNLESS WRITING A NEW DERIVATION.}
  DateTimeEnh_PanelRefCount: integer;

implementation


uses Telesis_Utils,
  Telesis_Constants,
  Telesis_Locale;


type
  TDateEnhStringGrid = class(TStringGrid)
  protected
    function CanFocus: boolean; override;
  end;


function TDateEnhStringGrid.CanFocus: boolean;
begin
  Result := false;
end;



function LastDayInMonth(MonthVal, YearVal: Word): Word;
begin
  Result := MonthDays[IsLeapYear(YearVal)][MonthVal]
end;


constructor TCustomDateTimePnlEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height := 220;
  Width := 222;
  BorderStyle := bsNone;
  BevelInner := bvNone;
  BevelOuter := bvRaised;
  BorderWidth := 4;
  FTodayLabel := ENHSTR_Today;
  FNowLabel := ENHSTR_Now;
  FClearLabel := ENHSTR_Clear;
  FAdjusting := 0;
  FHeaderPnl := nil;
  FClientPnl := nil;
  FTimePnl := nil;
  FButtonPnl := nil;
  FDateGrid := nil;
  FMonthEdit := nil;
  FYearEdit := nil;
  FHourEdit := nil;
  FMinuteEdit := nil;
  FSecondEdit := nil;
  FAmPmEdit := nil;
  FTodayButton := nil;
  FClearButton := nil;
  FOptions := [dtoShowButtons, dtoShowTime, dtoShowSeconds];
  FStartDay := dtsMonday;
  FTitleColor := clInfoBk;
  FDateCellColor := clWindow;
  Clear;
end;

destructor TCustomDateTimePnlEnh.Destroy;
begin
  inherited Destroy;
end;

procedure TCustomDateTimePnlEnh.WMSize(var Message: TWMSize);
begin
  inherited;
  AdjustLayout;
end;

procedure TCustomDateTimePnlEnh.Change;
begin
  if assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TCustomDateTimePnlEnh.SetTitleColor(AValue: TColor);
begin
  if AValue <> FTitleColor then
  begin
    FTitleColor := AValue;
    if HandleAllocated then
      Invalidate;
  end;
end;

procedure TCustomDateTimePnlEnh.SetDateCellColor(AValue: TColor);
begin
  if AValue <> FDateCellColor then
  begin
    FDateCellColor := AValue;
    if HandleAllocated then
      Invalidate;
  end;
end;

procedure TCustomDateTimePnlEnh.SetOptions(AValue: TDateTimeEnhOptions);
begin
  if AValue <> FOptions then
  begin
    FOptions := AValue;
    if HandleAllocated then
    begin
      AdjustLayout;
    end;
  end;
end;

procedure TCustomDateTimePnlEnh.SetStartDay(AValue: TDateTimeEnhStartDay);
begin
  if AValue <> FStartDay then
  begin
    FStartDay := AValue;
    if HandleAllocated then
    begin
      Invalidate;
      UpdateGridSelect;
    end;
  end;
end;

function TCustomDateTimePnlEnh.GetTimeStamp: TDateTime;
begin
  try
    if FYear = 0 then
    begin
      Result := 0;
    end
    else
    begin
      Result := EncodeDate(FYear, FMonth, FDay);
      if dto24HourTime in FOptions then
      begin
        Result := Result + EncodeTime(FHour, FMinute, FSecond, 0);
      end
      else
      begin
        if FAmPm = 1 then
        begin
          if FHour = 12 then
            Result := Result + EncodeTime(0, FMinute, FSecond, 0)
          else
            Result := Result + EncodeTime(FHour, FMinute, FSecond, 0);
        end
        else
          if FAmPm = 2 then
          begin
            if FHour = 12 then
              Result := Result + EncodeTime(12, FMinute, FSecond, 0)
            else
              Result := Result + EncodeTime(FHour + 12, FMinute, FSecond, 0);
          end;
      end;
    end;
  except
    Result := 0;
  end;
end;

procedure TCustomDateTimePnlEnh.SetTimeStamp(AValue: TDateTime);
var
  A, B, C, D: Word;
begin
  DecodeDate(AValue, A, B, C);
  FYear := A;
  FMonth := B;
  FDay := C;
  DecodeTime(AValue, A, B, C, D);
  if dto24HourTime in FOptions then
  begin
    FHour := A;
    FMinute := B;
    FSecond := C;
    FAmPm := 0;
  end
  else
  begin
    FMinute := B;
    FSecond := C;
    if A = 0 then
    begin
      FHour := 12;
      FAmPm := 1;
    end
    else
      if A <= 11 then
      begin
        FHour := A;
        FAmPm := 1;
      end
      else
        if A = 12 then
        begin
          FHour := 12;
          FAmPm := 2;
        end
        else
        begin
          FHour := A - 12;
          FAmPm := 2;
        end;
  end;
  UpdateDisplay;
  Change;
end;

procedure TCustomDateTimePnlEnh.Today;
var
  DT: TDateTime;
begin
  DT := SysUtils.Date;
  DT := DT + SysUtils.Time;
  TimeStamp := DT;
end;

procedure TCustomDateTimePnlEnh.TodayClick(Sender: TObject);
begin
  Today;
  if assigned(FOnPopupClose) then
    FOnPopupClose(Self);
end;

procedure TCustomDateTimePnlEnh.Clear;
begin
  FYear := 0;
  FMonth := 1;
  FDay := 1;
  FMinute := 0;
  FSecond := 0;
  if dto24HourTime in FOptions then
  begin
    FHour := 0;
    FAmPm := 0;
  end
  else
  begin
    FHour := 12;
    FAmPm := 1;
  end;
end;

procedure TCustomDateTimePnlEnh.ClearClick(Sender: TObject);
begin
  Clear;
  if assigned(FOnPopupClose) then
    FOnPopupClose(Self);
end;

procedure TCustomDateTimePnlEnh.CreateWnd;
var
  i: integer;
begin
  inherited;
  // Only create this stuff of first pass
  if FHeaderPnl <> nil then
    Exit;
  FHeaderPnl := CreatePanelControl(Self);
  FHeaderPnl.Align := alTop;
  FHeaderPnl.Font.Style := FHeaderPnl.Font.Style + [fsBold];
  FClientPnl := CreatePanelControl(Self);
  FClientPnl.Align := alClient;
  //FClientPnl.BevelInner := bvLowered;
  // If creating for combo then use special stringgrid that wont focus
  if (Owner <> nil) and (Owner is TCustomDateTimeBoxEnh) then
    FDateGrid := TDateEnhStringGrid.Create(FClientPnl)
  // else use normal string grid that can focus
  else
    FDateGrid := TStringGrid.Create(FClientPnl);
  // It is replicatable because there is only one!
  FDateGrid.ControlStyle := FDateGrid.ControlStyle + [csReplicatable];
  FDateGrid.Parent := FClientPnl;
  FDateGrid.ParentColor := true;
  FDateGrid.OnDrawCell := DrawDateCell;
  FDateGrid.OnSelectCell := SelectDateCell;
  FDateGrid.ScrollBars := ssNone;
  FDateGrid.BorderStyle := bsNone;
  FDateGrid.Align := alClient;
  FDateGrid.DefaultColWidth := 20;
  FDateGrid.DefaultRowHeight := 17;
  FDateGrid.ColCount := 7;
  FDateGrid.RowCount := 7;
  FDateGrid.FixedRows := 0;
  FDateGrid.FixedCols := 0;
  FDateGrid.Options := FDateGrid.Options + [goDrawFocusSelected];
  FDateGrid.Options := FDateGrid.Options - [goRangeSelect, goVertLine,
    goHorzLine];
  FButtonPnl := CreatePanelControl(Self);
  FButtonPnl.Align := alBottom;
  FTimePnl := CreatePanelControl(Self);
  FTimePnl.Align := alBottom;
  FMonthEdit := CreateComboControl(FHeaderPnl);
  for i := 1 to 12 do
  begin
    FMonthEdit.Items.Add(LongMonthNames[i] + '=' + IntToStr(i));
  end;
  FYearEdit := CreateEditControl(FHeaderPnl);
  FYearEdit.EditMaskMode := eemMaskEnh;
  FYearEdit.EditMask := '!0000;0;0;0';
  FHourEdit := CreateEditControl(FTimePnl);
  FHourEdit.EditMaskMode := eemNumeric;
  FHourEdit.EditMask := '!90;0;0;0';
  FMinuteEdit := CreateEditControl(FTimePnl);
  FMinuteEdit.EditMaskMode := eemNumeric;
  FMinuteEdit.EditMask := '!90;0;0;0';
  FSecondEdit := CreateEditControl(FTimePnl);
  FSecondEdit.EditMaskMode := eemNumeric;
  FSecondEdit.EditMask := '!90;0;0;0';
  FAmPmEdit := CreateComboControl(FTimePnl);
  FAmPmEdit.Button.Align := alRight;
  if (TimeAMString = '') then
    FAmPmEdit.Items.Add('AM=1')
  else
    FAmPmEdit.Items.Add(TimeAMString + '=1');
  if (TimePMString = '') then
    FAmPmEdit.Items.Add('PM=2')
  else
    FAmPmEdit.Items.Add(TimePMString + '=2');
  FTodayButton := CreateButtonControl(FButtonPnl);
  FTodayButton.OnClick := TodayClick;
  FClearButton := CreateButtonControl(FButtonPnl);
  FClearButton.OnClick := ClearClick;
  AdjustLayout;
  Today;
end;

function TCustomDateTimePnlEnh.CreatePanelControl(CtrlParent: TWinControl):
  TPanel;
begin
  Result := TPanel.Create(CtrlParent);
  Result.Parent := CtrlParent;
  Result.BorderStyle := bsNone;
  Result.Height := 30;
  Result.BevelInner := bvNone;
  Result.BevelOuter := bvNone;
  Result.BorderWidth := 4;
  Result.Caption := '';
  Result.ParentColor := true;
end;

function TCustomDateTimePnlEnh.CreateEditControl(CtrlParent: TPanel): TEditEnh;
begin
  Result := TEditEnh.Create(CtrlParent);
  Result.Parent := CtrlParent;
  Result.AutoSize := true;
  Result.BorderStyle := bsSingle;
  Result.Ctl3D := true;
  Result.Ctl3DShallow := CTL3DSHALLOW_DEFAULT;
  Result.ParentColor := true;
  Result.ParentFont := true;
  Result.Text := '';
  Result.Button.Style := eebUpDownArrow;
  Result.Button.Size := 12;
  Result.Button.Align := alRight;
  Result.Button.AutoRepeat := true;
  Result.Alignment := taRightJustify;
  Result.OnButton1 := Incrementer;
  Result.OnButton2 := Decrementer;
end;

function TCustomDateTimePnlEnh.CreateComboControl(CtrlParent: TPanel):
  TComboBoxEnh;
begin
  Result := TComboBoxEnh.Create(CtrlParent);
  Result.Parent := CtrlParent;
  Result.AutoSize := true;
  Result.BorderStyle := bsSingle;
  Result.Ctl3D := true;
  Result.Ctl3DShallow := CTL3DSHALLOW_DEFAULT;
  Result.ParentColor := true;
  Result.ParentFont := true;
  Result.Text := '';
  Result.Button.Style := eebUpDownArrow;
  Result.Button.Size := 12;
  Result.Button.Align := alLeft;
  Result.Button.AutoRepeat := true;
  Result.Alignment := taLeftJustify;
  Result.OnButton1 := Incrementer;
  Result.OnButton2 := Decrementer;
end;

function TCustomDateTimePnlEnh.CreateButtonControl(CtrlParent: TPanel):
  TSpeedButton;
begin
  Result := TSpeedButton.Create(CtrlParent);
  // It is replicatable because there is only one!
  Result.ControlStyle := Result.ControlStyle + [csReplicatable];
  Result.Parent := CtrlParent;
  Result.Flat := true;
end;

procedure TCustomDateTimePnlEnh.AdjustLayout;
var
  TxtWidth, i, j: integer;
begin
  if (FAdjusting = 0) and (FHeaderPnl <> nil) then
  try
    Inc(FAdjusting);
    //--------------------------------
    // First Adjust the Header...
    //--------------------------------
    TxtWidth := Canvas.TextWidth('XXXXXXXXX'); // width required for month
    FMonthEdit.Width := TxtWidth + FMonthEdit.Button.Size + 20;
    TxtWidth := Canvas.TextWidth('0000'); // width required for year
    FYearEdit.Width := TxtWidth + FYearEdit.Button.Size + 20;
    i := (FHeaderPnl.Width - (FMonthEdit.Width + FYearEdit.Width));
    if i < 0 then
    begin
      Width := Width - i;
      i := (FHeaderPnl.Width - (FMonthEdit.Width + FYearEdit.Width));
    end;
    i := i div 2;
    FMonthEdit.Left := i;
    FYearEdit.Left := i + FMonthEdit.Width;
    FHeaderPnl.Height := FMonthEdit.Height + (2 * FHeaderPnl.BorderWidth);
    i := FHeaderPnl.Height - FMonthEdit.Height;
    i := i div 2;
    FMonthEdit.Top := i;
    FYearEdit.Top := i;
    //--------------------------------
    // Then the footer - button panel
    //--------------------------------
    TxtWidth := Canvas.TextWidth(FTodayLabel);
    FTodayButton.Width := TxtWidth + 20;
    TxtWidth := Canvas.TextWidth(FClearLabel);
    FClearButton.Width := TxtWidth + 20;
    TxtWidth := Canvas.TextHeight(FTodayLabel);
    FTodayButton.Height := TxtWidth + 6; // Only small border top/bottom
    FClearButton.Height := TxtWidth + 6;
    // I am not too worried about the width of the panel for the buttons
    // they can just take what they get.
    i := FButtonPnl.Width - (FTodayButton.Width + FClearButton.Width);
    i := i div 3;
    FTodayButton.Left := i;
    FClearButton.Left := i + FTodayButton.Width + i;
    FButtonPnl.Height := FTodayButton.Height + (2 * FButtonPnl.BorderWidth);
    i := FButtonPnl.Height - FTodayButton.Height;
    i := i div 2;
    FTodayButton.Top := i;
    FTodayButton.Caption := FTodayLabel;
    FClearButton.Top := i;
    FClearButton.Caption := FClearLabel;
    if dtoShowButtons in FOptions then
    begin
      if not FButtonPnl.Visible then
      begin
        Height := Height + FButtonPnl.Height;
        FButtonPnl.Visible := true;
      end;
    end
    else
    begin
      if FButtonPnl.Visible then
      begin
        Height := Height - FButtonPnl.Height;
        FButtonPnl.Visible := false;
      end;
    end;
    //--------------------------------
    // Then the time
    //--------------------------------
    if dto24HourTime in FOptions then
    begin
      if dtoShowSeconds in FOptions then
        j := 3
      else
        j := 2
    end
    else
    begin
      if dtoShowSeconds in FOptions then
        j := 4
      else
        j := 3;
    end;
    TxtWidth := Canvas.TextWidth('XX'); // width required for hr/min/sec/pm
    FHourEdit.Width := TxtWidth + FHourEdit.Button.Size + 12;
    FMinuteEdit.Width := FHourEdit.Width;
    FSecondEdit.Width := FHourEdit.Width;
    FAmPmEdit.Width := FHourEdit.Width;
    i := (FTimePnl.Width - (j * FHourEdit.Width));
    if i < 0 then
    begin
      Width := Width - i;
      i := (FTimePnl.Width - (j * FHourEdit.Width));
    end;
    i := i div 2;
    FHourEdit.Left := i;
    FMinuteEdit.Left := i + FHourEdit.Width;
    if dtoShowSeconds in FOptions then
    begin
      FSecondEdit.Left := i + (2 * FHourEdit.Width);
      FSecondEdit.Visible := true;
    end
    else
    begin
      FSecondEdit.Left := FMinuteEdit.Left;
      FSecondEdit.Visible := false;
    end;
    if dto24HourTime in FOptions then
    begin
      FAmPmEdit.Left := i;
      FAmPmEdit.Visible := false;
    end
    else
    begin
      FAmPmEdit.Left := FSecondEdit.Left + FSecondEdit.Width;
      FAmPmEdit.Visible := true;
    end;
    FTimePnl.Height := FHourEdit.Height + (2 * FTimePnl.BorderWidth);
    i := FTimePnl.Height - FHourEdit.Height;
    i := i div 2;
    FHourEdit.Top := i;
    FMinuteEdit.Top := i;
    FSecondEdit.Top := i;
    FAmPmEdit.Top := i;
    if dtoShowTime in FOptions then
    begin
      FTodayButton.Caption := FNowLabel;
      if not FTimePnl.Visible then
      begin
        Height := Height + FTimePnl.Height;
        FTimePnl.Visible := true;
      end;
    end
    else
    begin
      if FTimePnl.Visible then
      begin
        Height := Height - FTimePnl.Height;
        FTimePnl.Visible := false;
      end;
    end;
    //--------------------------------
    // and finally the date grid
    //--------------------------------
    TxtWidth := Canvas.TextWidth('WXX') + 4; // width required for date entry
    i := FClientPnl.ClientWidth - (2 * BorderWidth);
    if (7 * TxtWidth) > i then
    begin
      TxtWidth := Canvas.TextWidth('WX') + 4; // width required for date entry
      if (7 * TxtWidth) > i then
      begin
        Width := (Width - FClientPnl.ClientWidth) +
          (2 * FClientPnl.BorderWidth) + (7 * TxtWidth);
      end;
    end;
    i := Canvas.TextHeight('W') + 6;
    j := (Height - FClientPnl.ClientHeight) +
      (2 * FClientPnl.BorderWidth) + (7 * i);
    if j > Height then
      Height := j;
    i := FClientPnl.ClientWidth - (2 * BorderWidth);
    FDateGrid.DefaultColWidth := (i div 7) - 1;
    i := FClientPnl.ClientHeight - (2 * BorderWidth);
    FDateGrid.DefaultRowHeight := (i div 7) - 1;
  finally
    Dec(FAdjusting);
  end;
end;

procedure TCustomDateTimePnlEnh.DrawDateCell(Sender: TObject; Col, Row: longint;
  Rect: TRect; State: TGridDrawState);
var
  i, j: integer;
  CellTxt: string;
begin
  with FDateGrid.Canvas do
  begin
    Brush.Style := bsSolid;
    if Row = 0 then
    begin
      Brush.Color := FTitleColor;
      Font.Color := TextColorFromColor(FTitleColor);
    end
    else
    begin
      if gdSelected in State then
      begin
        Brush.Color := clHighlight;
        Font.Color := clHighlightText;
      end
      else
      begin
        Brush.Color := FDateCellColor;
        Font.Color := TextColorFromColor(FDateCellColor);
      end;
    end;
    FillRect(Rect);
    if Row = 0 then
    begin
      CellTxt := GetCellTitle(Col, (Rect.Right - Rect.Left));
    end
    else
    begin
      i := GetCellDateValue(Col, Row);
      if i > 0 then
        CellTxt := IntToStr(i)
      else
        CellTxt := '';
    end;
    i := (((Rect.Right - 1) - Rect.Left) - TextWidth(CellTxt)) div 2;
    j := (((Rect.Bottom - 1) - Rect.Top) - TextHeight(CellTxt)) div 2;
    TextRect(Rect, Rect.Left + i, Rect.Top + j, CellTxt);
    Pen.Width := 1;
    Pen.Style := psSolid;
    // Last thing we do is draw the cell lines...
    Pen.Color := clBtnShadow;
    if (Col = 0) then
    begin
      MoveTo(Rect.Left, Rect.Top);
      LineTo(Rect.Left, Rect.Bottom);
    end;
    if (Row = 0) then
    begin
      MoveTo(Rect.Left, Rect.Top);
      LineTo(Rect.Right, Rect.Top);
    end;
    MoveTo(Rect.Right - 1, Rect.Top);
    LineTo(Rect.Right - 1, Rect.Bottom - 1);
    LineTo(Rect.Left - 1, Rect.Bottom - 1);
  end;
end;

procedure TCustomDateTimePnlEnh.SelectDateCell(Sender: TObject; Col, Row:
  longint;
  var CanSelect: boolean);
var
  i, Hold: integer;
begin
  if Row = 0 then
  begin
    CanSelect := false;
  end
  else
  begin
    i := GetCellDateValue(Col, Row);
    if i > 0 then
    begin
      if FAdjusting = 0 then
      begin
        Inc(FAdjusting);
        try
          Hold := FDay;
          FDay := i;
          FDateGrid.Invalidate;
          if Hold <> FDay then
          begin
            Change;
          end;
        finally
          Dec(FAdjusting);
        end;
      end;
      if assigned(FOnPopupClose) then
        FOnPopupClose(Self);
    end
    else
      CanSelect := false;
  end;
end;

procedure TCustomDateTimePnlEnh.EditChange(Sender: TObject);
begin
  if FAdjusting = 0 then
  try
    Inc(FAdjusting);
    if Sender = FYearEdit then
    begin
      FYear := StrToInt(FYearEdit.Text);
    end
    else
      if Sender = FMonthEdit then
      begin
        FMonth := StrToInt(FMonthEdit.ItemValue);
        if FMonth < 1 then
        begin
          FMonth := 1;
          FMonthEdit.ItemValue := '1';
        end
        else
          if FMonth > 12 then
          begin
            FMonth := 12;
            FMonthEdit.ItemValue := '12';
          end;
      end
      else
        if Sender = FHourEdit then
        begin
          FHour := StrToInt(FHourEdit.Text);
          if dto24HourTime in FOptions then
          begin
            if FHour > 23 then
            begin
              FHour := 23;
              FHourEdit.Text := '23';
            end;
          end
          else
          begin
            if FHour < 1 then
            begin
              FHour := 1;
              FHourEdit.Text := '1';
            end
            else
              if FHour > 12 then
              begin
                FHour := 12;
                FHourEdit.Text := '12';
              end;
          end;
        end
        else
          if Sender = FMinuteEdit then
          begin
            FMinute := StrToInt(FMinuteEdit.Text);
            if FMinute > 59 then
            begin
              FMinute := 59;
              FMinuteEdit.Text := '59';
            end;
          end
          else
            if Sender = FSecondEdit then
            begin
              FSecond := StrToInt(FSecondEdit.Text);
              if FSecond > 59 then
              begin
                FSecond := 59;
                FSecondEdit.Text := '59';
              end;
            end
            else
              if Sender = FAmPmEdit then
              begin
                FAmPm := StrToInt(FAmPmEdit.ItemValue);
                if dto24HourTime in FOptions then
                begin
                  if FAmPm <> 0 then
                  begin
                    FAmPm := 0;
                    FAmPmEdit.ItemValue := '0';
                  end;
                end
                else
                begin
                  if FAmPm < 1 then
                  begin
                    FAmPm := 1;
                    FAmPmEdit.ItemValue := '1';
                  end
                  else
                    if FAmPm > 2 then
                    begin
                      FAmPm := 2;
                      FAmPmEdit.ItemValue := '2';
                    end;
                end;
              end;
    UpdateGridSelect;
    Change;
  finally
    Dec(FAdjusting);
  end;
end;

procedure TCustomDateTimePnlEnh.Incrementer(Sender: TObject);
begin
  try
    Inc(FAdjusting);
    if Sender = FYearEdit then
    begin
      if FYear < 9999 then
      begin
        Inc(FYear);
        FYearEdit.Text := IntToStr(FYear);
        if FDay > LastDayInMonth(FMonth, FYear) then
          FDay := LastDayInMonth(FMonth, FYear);
        UpdateGridSelect;
      end;
    end
    else
      if Sender = FMonthEdit then
      begin
        if FYear < 9999 then
        begin
          Inc(FMonth);
          if FMonth > 12 then
          begin
            FMonth := 1;
            Incrementer(FYearEdit);
          end;
          FMonthEdit.ItemValue := IntToStr(FMonth);
          if FDay > LastDayInMonth(FMonth, FYear) then
            FDay := LastDayInMonth(FMonth, FYear);
          UpdateGridSelect;
        end;
      end
      else
        if Sender = FDateGrid then
        begin
          if FYear < 9999 then
          begin
            Inc(FDay);
            if FDay > LastDayInMonth(FMonth, FYear) then
            begin
              FDay := 1;
              Incrementer(FMonthEdit);
            end;
            UpdateGridSelect;
          end;
        end
        else
          if Sender = FHourEdit then
          begin
            if FYear < 9999 then
            begin
              Inc(FHour);
              if dto24HourTime in FOptions then
              begin
                if FHour > 23 then
                begin
                  FHour := 0;
                  Incrementer(FDateGrid);
                end;
              end
              else
              begin
                if FHour > 12 then
                begin
                  FHour := 1;
                end
                else
                  if FHour > 11 then
                  begin
                    FHour := 12;
                    Incrementer(FAmPmEdit);
                  end;
              end;
              FHourEdit.Text := IntToStr(FHour);
            end;
          end
          else
            if Sender = FMinuteEdit then
            begin
              if FYear < 9999 then
              begin
                Inc(FMinute);
                if FMinute > 59 then
                begin
                  FMinute := 1;
                  Incrementer(FHourEdit);
                end;
                FMinuteEdit.Text := IntToStr(FMinute);
              end;
            end
            else
              if Sender = FSecondEdit then
              begin
                if FYear < 9999 then
                begin
                  Inc(FSecond);
                  if FSecond > 59 then
                  begin
                    FSecond := 1;
                    Incrementer(FMinuteEdit);
                  end;
                  FSecondEdit.Text := IntToStr(FSecond);
                end;
              end
              else
                if Sender = FAmPmEdit then
                begin
                  if FYear < 9999 then
                  begin
                    Inc(FAmPm);
                    if FAmPm > 2 then
                    begin
                      FAmPm := 1;
                      Incrementer(FDateGrid);
                    end;
                    FAmPmEdit.ItemValue := IntToStr(FAmPm);
                  end;
                end;
    Change;
  finally
    Dec(FAdjusting);
  end;
end;

procedure TCustomDateTimePnlEnh.Decrementer(Sender: TObject);
begin
  try
    Inc(FAdjusting);
    if Sender = FYearEdit then
    begin
      if FYear > 0 then
      begin
        Dec(FYear);
        FYearEdit.Text := IntToStr(FYear);
        UpdateGridSelect;
        if FDay > LastDayInMonth(FMonth, FYear) then
          FDay := LastDayInMonth(FMonth, FYear);
      end;
    end
    else
      if Sender = FMonthEdit then
      begin
        if FYear > 0 then
        begin
          Dec(FMonth);
          if FMonth < 1 then
          begin
            FMonth := 12;
            Decrementer(FYearEdit);
          end;
          FMonthEdit.ItemValue := IntToStr(FMonth);
          if FDay > LastDayInMonth(FMonth, FYear) then
            FDay := LastDayInMonth(FMonth, FYear);
          UpdateGridSelect;
        end;
      end
      else
        if Sender = FDateGrid then
        begin
          if FYear > 0 then
          begin
            Dec(FDay);
            if FDay < 1 then
            begin
              Decrementer(FMonthEdit);
              FDay := LastDayInMonth(FMonth, FYear);
            end;
            UpdateGridSelect;
          end;
        end
        else
          if Sender = FHourEdit then
          begin
            if FYear > 0 then
            begin
              Dec(FHour);
              if dto24HourTime in FOptions then
              begin
                if FHour < 0 then
                begin
                  FHour := 23;
                  Decrementer(FDateGrid);
                end;
              end
              else
              begin
                if FHour < 1 then
                begin
                  FHour := 12;
                end
                else
                  if FHour = 11 then
                  begin
                    //FHour := 12;
                    Decrementer(FAmPmEdit);
                  end;
              end;
              FHourEdit.Text := IntToStr(FHour);
            end;
          end
          else
            if Sender = FMinuteEdit then
            begin
              if FYear > 0 then
              begin
                Dec(FMinute);
                if FMinute < 0 then
                begin
                  FMinute := 59;
                  Decrementer(FHourEdit);
                end;
                FMinuteEdit.Text := IntToStr(FMinute);
              end;
            end
            else
              if Sender = FSecondEdit then
              begin
                if FYear > 0 then
                begin
                  Dec(FSecond);
                  if FSecond < 0 then
                  begin
                    FSecond := 59;
                    Decrementer(FMinuteEdit);
                  end;
                  FSecondEdit.Text := IntToStr(FSecond);
                end;
              end
              else
                if Sender = FAmPmEdit then
                begin
                  if FYear > 0 then
                  begin
                    Dec(FAmPm);
                    if FAmPm < 1 then
                    begin
                      FAmPm := 2;
                      Decrementer(FDateGrid);
                    end;
                    FAmPmEdit.ItemValue := IntToStr(FAmPm);
                  end;
                end;
    Change;
  finally
    Dec(FAdjusting);
  end;
end;

function TCustomDateTimePnlEnh.GetCellTitle(Col, DrawWidth: integer): string;
var
  i: integer;
begin
  // Start day plus column offset gives offset of desired title
  i := Ord(FStartDay) + Col + 1;
  if i > 7 then
    i := i - 7;
  Result := ShortDayNames[i];
  if Canvas.TextWidth(Result) > (DrawWidth - 4) then
    Result := Copy(Result, 1, 2);
end;

function TCustomDateTimePnlEnh.GetCellDateValue(Col, Row: integer): integer;
var
  FirstDate: TDateTime;
  i, j: integer;
begin
  // This gives the offset into the grid (1 as the first returnable cell)
  i := ((Row - 1) * 7) + Col + 1;
  // But we dont want to start until the first day of the month
  FirstDate := EncodeDate(FYear, FMonth, 1); // date of the first
  j := DayOfWeek(FirstDate) - 1;
  if Ord(FStartDay) > j then
    j := j + 7;
  j := j - Ord(FStartDay);
  Result := i - j;
  if Result > LastDayInMonth(FMonth, FYear) then
    Result := 0;
end;

procedure TCustomDateTimePnlEnh.UpdateGridSelect;
var
  FirstDate: TDateTime;
  i, j: integer;
  GRect: TGridRect;
begin
  Inc(FAdjusting);
  try
    i := FDay;
    FirstDate := EncodeDate(FYear, FMonth, 1); // date of the first
    j := DayOfWeek(FirstDate) - 1;
    if Ord(FStartDay) > j then
      j := j + 7;
    j := j - Ord(FStartDay);
    i := i + j - 1;
    GRect.Left := i mod 7;
    GRect.Top := (i div 7) + 1;
    GRect.Right := GRect.Left;
    GRect.Bottom := GRect.Top;
    FDateGrid.Selection := GRect;
  finally
    Dec(FAdjusting);
  end;
  FDateGrid.Invalidate;
end;

procedure TCustomDateTimePnlEnh.UpdateDisplay;
begin
  FYearEdit.Text := IntToStr(FYear);
  FMonthEdit.ItemValue := IntToStr(FMonth);
  UpdateGridSelect;
  FHourEdit.Text := IntToStr(FHour);
  FMinuteEdit.Text := IntToStr(FMinute);
  FSecondEdit.Text := IntToStr(FSecond);
  FAmPmEdit.ItemValue := IntToStr(FAmPm);
end;

procedure TCustomDateTimePnlEnh.IncDay;
begin
  Incrementer(FDateGrid);
end;

procedure TCustomDateTimePnlEnh.IncWeek;
begin
  // There are better ways but Im in a hurry
  Incrementer(FDateGrid);
  Incrementer(FDateGrid);
  Incrementer(FDateGrid);
  Incrementer(FDateGrid);
  Incrementer(FDateGrid);
  Incrementer(FDateGrid);
  Incrementer(FDateGrid);
end;

procedure TCustomDateTimePnlEnh.IncMonth;
begin
  Incrementer(FMonthEdit);
end;

procedure TCustomDateTimePnlEnh.IncYear;
begin
  Incrementer(FYearEdit);
end;

procedure TCustomDateTimePnlEnh.DecDay;
begin
  Decrementer(FDateGrid);
end;

procedure TCustomDateTimePnlEnh.DecWeek;
begin
  Decrementer(FDateGrid);
  Decrementer(FDateGrid);
  Decrementer(FDateGrid);
  Decrementer(FDateGrid);
  Decrementer(FDateGrid);
  Decrementer(FDateGrid);
  Decrementer(FDateGrid);
end;

procedure TCustomDateTimePnlEnh.DecMonth;
begin
  Decrementer(FMonthEdit);
end;

procedure TCustomDateTimePnlEnh.DecYear;
begin
  Decrementer(FYearEdit);
end;




constructor TCustomDateTimeBoxEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ActiveWhenInvisible := false; // We use a shared datetimepnl so cannot be
  // active until visible
  FDateTimePnl := TDateTimePnlEnh.Create(Self);
  FDateTimePnl.Parent := Self;
  FDateTimePnl.Top := BorderWidth;
  FDateTimePnl.Left := BorderWidth;
  FDateTimePnl.BorderStyle := bsNone;
  FDateTimePnl.BevelOuter := bvNone;
  FDateTimePnl.BorderWidth := 2;
  FDateTimePnl.Width := 172;
  FDateTimePnl.Height := 236;
  FDateTimePnl.Visible := true;
  FDateTimePnl.OnPopupCloseEvent := DateTimeCloseEvent;
end;

destructor TCustomDateTimeBoxEnh.Destroy;
begin
  if FDateTimePnl <> nil then
  begin
    FDateTimePnl.Free;
    FDateTimePnl := nil;
  end;
  inherited Destroy;
end;

procedure TCustomDateTimeBoxEnh.DateTimeCloseEvent(Sender: TObject);
begin
  if Controller.EditBeginModify then
  begin
    Change;
    CloseUp;
  end;
end;

function TCustomDateTimeBoxEnh.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean;
var
  k: WORD;
begin
  k := VK_DOWN;
  ParentKeyDown(k, Shift);
  Result := (k = 0); // Was processed if k was reset to 0
end;

function TCustomDateTimeBoxEnh.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean;
var
  k: WORD;
begin
  k := VK_UP;
  ParentKeyDown(k, Shift);
  Result := (k = 0); // Was processed if k was reset to 0
end;

function TCustomDateTimeBoxEnh.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
begin
  Result := false;
end;

procedure TCustomDateTimeBoxEnh.ParentKeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if (Key <> 0) and Controller.EditCanModify then
  begin
    if Shift = [] then
    begin
      case Key of
        VK_RIGHT:
          begin
            FDateTimePnl.IncDay;
            Key := 0;
          end;
        VK_LEFT:
          begin
            FDateTimePnl.DecDay;
            Key := 0;
          end;
        VK_DOWN:
          begin
            FDateTimePnl.IncWeek;
            Key := 0;
          end;
        VK_UP:
          begin
            FDateTimePnl.DecWeek;
            Key := 0;
          end;
        VK_NEXT:
          begin
            FDateTimePnl.IncMonth;
            Key := 0;
          end;
        VK_PRIOR:
          begin
            FDateTimePnl.DecMonth;
            Key := 0;
          end;
        VK_HOME:
          begin
            FDateTimePnl.IncYear;
            Key := 0;
          end;
        VK_END:
          begin
            FDateTimePnl.DecYear;
            Key := 0;
          end;
        VK_SPACE:
          begin
            FDateTimePnl.Today;
            Key := 0;
          end;
        VK_DELETE:
          begin
            FDateTimePnl.Clear;
            Change;
            CloseUp;
            Key := 0;
          end;
      end;
    end
    else if Shift = [ssShift] then
    begin
      case Key of
        VK_DOWN:
          begin
            FDateTimePnl.IncMonth;
            Key := 0;
          end;
        VK_UP:
          begin
            FDateTimePnl.DecMonth;
            Key := 0;
          end;
      end;
    end
    else if Shift = [ssCtrl, ssShift] then
    begin
      case Key of
        VK_DOWN:
          begin
            FDateTimePnl.IncYear;
            Key := 0;
          end;
        VK_UP:
          begin
            FDateTimePnl.DecYear;
            Key := 0;
          end;
      end;
    end;
  end;
end;

procedure TCustomDateTimeBoxEnh.SizeControl;
begin
  // try to set control to minimum size
  FDateTimePnl.SetBounds(FDateTimePnl.Left, FDateTimePnl.Top, 172, 236);
  // Now adjust outer panel size
  ClientHeight := FDateTimePnl.Height + (2 * BorderWidth);
  ClientWidth := FDateTimePnl.Width + (2 * BorderWidth);
end;

procedure TCustomDateTimeBoxEnh.SyncToParentValue;
var
  DT: TDateTime;
begin
  if FInternalChange = 0 then
  begin
    try
      FDateTimePnl.HandleNeeded;        // Force all the bits to be created;
      with Controller as TCustomDateTimeEnh do
      begin
        FDateTimePnl.Options := FDateOptions;
        FDateTimePnl.StartDay := FDateStartDay;
        FDateTimePnl.TitleColor := FDateTitleColor;
        FDateTimePnl.DateCellColor := FDateDateCellColor;
      end;
      SyncColors;
      with Controller as TCustomDateTimeEnh do
      begin
        if Modified then
          ValidateEdit;
        DT := DateTime;
      end;
      if DT = 0 then
        FDateTimePnl.Today
      else
        FDateTimePnl.TimeStamp := DT;
    except
      FDateTimePnl.Today;
    end;
  end;
end;

procedure TCustomDateTimeBoxEnh.SyncParentToValue;
begin
  if Controller.EditBeginModify then
    TCustomDateTimeEnh(Controller).DateTime := FDateTimePnl.TimeStamp;
end;

procedure TCustomDateTimeBoxEnh.SyncColors;
begin
  with Controller as TCustomComboEnh do
  begin
    if DropDownColor = clNone then
      Self.FDateTimePnl.DateCellColor := Color
    else
      Self.FDateTimePnl.DateCellColor := DropDownColor;
  end;
end;

procedure TCustomDateTimeBoxEnh.DropDown;
begin
  inherited;
  // This is messy and horrible, and likely to cause flicker when the dropbox
  // needs to be resized.  I just cant find any other way of making this thing
  // work.  (It should all work from inside the SizeControl method, thats what
  // it is for, but the changes just arent getting through in time.)
  // Hopefully the impact will be minimal - only showing on those apps that use
  // multiple font sizes for their datetime controls.
  if Visible then
  begin
      Application.ProcessMessages;
      FDateTimePnl.SetBounds(FDateTimePnl.Left, FDateTimePnl.Top, 172, 236);
      FDateTimePnl.AdjustLayout;
      ClientHeight := FDateTimePnl.Height + (2 * BorderWidth);
      ClientWidth := FDateTimePnl.Width + (2 * BorderWidth);
  end;
end;


constructor TCustomDateTimeEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csSetCaption];
  Caption := '';
  Style := csDropDown;                  // turn off list mode!
  FDateOptions := [dtoShowButtons, dtoShowTime, dtoShowSeconds];
  FDateStartDay := dtsMonday;
  FDateTitleColor := clInfoBk;
  FDateDateCellColor := clWindow;
  FDateTime := 0;
  FDateFormat := '';
  FDateInputLayout := '';
  FDTUpdate := 0;
end;

destructor TCustomDateTimeEnh.Destroy;
begin
  inherited Destroy;
end;

procedure TCustomDateTimeEnh.SetDateTime( AVal: TDateTime );
begin
  if FDTUpdate = 0 then
  try
    Inc(FDTUpdate);
    // We rely on running this code whetehr AVal is an actual
    // change in value or not - it is not ideal but works.
    FDateTime := AVal;
    Text := GetDateTimeDisplayText( FDateTime );
    if (ComboControl <> nil) then
      ComboControl.DoParentValueChanged;
  finally
    Dec(FDTUpdate);
  end;
end;

procedure TCustomDateTimeEnh.SetDateFormat( AVal: string );
begin
  if (AVal <> FDateFormat) and (FDTUpdate = 0) then
  try
    Inc(FDTUpdate);
    FDateFormat := AVal;
    //Text := GetDateTimeDisplayText( FDateTime );
    SetDateTime( FDateTime );
  finally
    Dec(FDTUpdate);
  end;
end;

procedure TCustomDateTimeEnh.SetDateInputLayout( AVal: string );
begin
  if AVal <> FDateInputLayout then
  begin
    FDateInputLayout := AVal;
  end;
end;

function TCustomDateTimeEnh.GetDateTimeDisplayText( ADateTime: TDateTime ): string;
var
  tmpFormat: string;
begin
  if FDateTime = 0 then
    Result := ''
  else
  begin
    if FDateFormat = '' then
    begin
      if dtoShowTime in DateOptions then
      begin
        if dtoShowSeconds in DateOptions then
          tmpFormat := 'ddddd tt' // shortdateformat + longtimeformat
        else
          tmpFormat := 'ddddd t'; // shortdateformat + shorttimeformat
      end
      else
        tmpFormat := 'ddddd' // shortdateformat
    end
    else
    begin
      tmpFormat := FDateFormat;
      // If we are focused we want to be certain that all available details
      // are displayed for editing.  That is; if the DateFormat has been
      // restricted to not show the year (or whatever) then we need to ensure
      // such details are visible during editing (while focused).  We assume
      // that the system defined short and long date formats will provide
      // all necessary information.  I ignore the read-only state in order
      // that it is possible to see the full details when focused even if not
      // actually editing.
      if Focused and (Pos('ddddd', tmpFormat) <= 0) then
      begin
        // if any of day, month or year are missing then revert to default
        // (system) formats while editing.
        if (Pos('d', tmpFormat) <= 0) or
          (Pos('m', tmpFormat) <= 0) or
          (Pos('y', tmpFormat) <= 0) then
        begin
          if dtoShowTime in DateOptions then
          begin
            if dtoShowSeconds in DateOptions then
              tmpFormat := 'ddddd tt' // shortdateformat + longtimeformat
            else
              tmpFormat := 'ddddd t'; // shortdateformat + shorttimeformat
          end
          else
            tmpFormat := 'ddddd' // shortdateformat
        end;
      end;
    end;
    Result := FormatDateTime( tmpFormat, ADateTime );
  end;
end;

procedure TCustomDateTimeEnh.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
end;

function TCustomDateTimeEnh.IncrementalSearch(TxtVal: string; ChkLen: integer):
  integer;
begin
  // No incremental search with date time
  Result := ChkLen;
end;

function TCustomDateTimeEnh.CreateComboControl: TCustomComboControlEnh;
begin
  if DateTimeEnh_Panel = nil then
  begin
    DateTimeEnh_Panel := TCustomDateTimeBoxEnh.Create(nil);
  end;
  Inc(DateTimeEnh_PanelRefCount);
  Result := DateTimeEnh_Panel;
end;

procedure TCustomDateTimeEnh.DestroyComboControl(CurrentControl:
  TCustomComboControlEnh);
begin
  if DateTimeEnh_Panel.Parent = Self then
  begin
    DateTimeEnh_Panel.Parent := nil;
  end;
  if CurrentControl = DateTimeEnh_Panel then
  begin
    Dec(DateTimeEnh_PanelRefCount);
    if DateTimeEnh_PanelRefCount = 0 then
    begin
      DateTimeEnh_Panel.Free;
      DateTimeEnh_Panel := nil;
    end;
  end;
end;

function TCustomDateTimeEnh.DoTextChange: boolean;
var
  tmpDT: TDateTime;
begin
  // detect direct text changes and reset internal datetime directly
  // according to the current text value - ignoring errors so that DateTime
  // gets reset to 0 whenever the input value is not valid.
  if FDTUpdate = 0 then
  begin
    if IsEmpty then
      FDateTime := 0
    else
    begin
      tmpDT := 0;
      TextToDateTime( Text, tmpDT );
      FDateTime := tmpDT;
    end;
  end;
  Result := inherited DoTextChange;
end;

procedure TCustomDateTimeEnh.DoValidateEdit(var ErrCode: integer);
var
  tmpDT: TDateTime;
begin
  if FInternalUpd = 0 then
  begin
    inherited DoValidateEdit(ErrCode);
    if ErrCode = E_EDITENH_NOERROR then
    begin
      // if we got through default and (optional) user processing
      // then we have to extract the datetime and assign to the internal
      // variable to keep ourselves synchronised...
      if IsEmpty then
        DateTime := 0
      else
      begin
        tmpDT := 0;
        ErrCode := TextToDateTime( Text, tmpDT );
        if ErrCode = E_EDITENH_NOERROR then
          DateTime := tmpDT;
      end;
    end;
  end;
end;

function TCustomDateTimeEnh.TextToDateTime( AStr: string; var ADateTime: TDateTime ): integer;
begin
  // First we attempt to interpret the date according to the
  // given display format - in case the user has simply changed
  // a value in the displayed text...
  Result := ExtractDateTime(
    DateTimeFormatToLayout( DateFormat ),
    AStr, ADateTime );
  // if failed to get a date via the display format then use the
  // DateInputLayout and see if we can understand the date that way.
  if Result <> E_EDITENH_NOERROR then
    Result := ExtractDateTime(
      DateInputLayout,
      AStr, ADateTime );
end;

initialization;
  DateTimeEnh_Panel := nil;
  DateTimeEnh_PanelRefCount := 0;

finalization;
  DateTimeEnh_Panel := nil;
  DateTimeEnh_PanelRefCount := 0;

end.

