{ Telesis_EditEnh.pas                             SearchTag: ~~EnhComponents~~

 Copyright 1999-2007 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Enhanced edit control.
<br>
@desc Defines TCustomEditEnh and TEditEnh - the published variation.
Base class (TCustomEditEnh) is used for many other controls in the Enh
collection.<br>
<br>
Used to be used by IBO directly, but we have since managed to setup
interfaces support so that this unit can be compiled with or without IBO.<br>
<br>
If you intend to try and compile this unit on a system without IBObjects
installed you need to edit the Telesis.inc unit and undefine (or comment out)
the __COMPILE_WITH_IBO__ declaration.<br>
<br>
If you are compiling this unit with IBObjects installed, IBObjects must be
in the search path.<br>
<br>
<br>
<strong>History</strong><br>
<br>
22-Jan-2008 - Adjustments to suit MemoEnh.UseTextEditor property.  Some
clean up of uses clause.  Disable KeyPress responses while ButtonState shows
there is a button down - allowing it to ignore events generated by any
short-cuts for the buttons (such as Ctrl+Enter).<br>
<br>
19-Dec-2007 - Adjust clipboard behaviour; remove redundant open/close calls
and avoid the internal updating of the clipboard contents during paste
(which was there to temporarily remove unwanted CR/LF characters).  Alter
masked display so that an empty mask always shows as an empty string.<br>
<br>
24-Sep-2007 - Override Repaint to make it update masked text properly.<br>
<br>
15-May-2006 - Made PasswordChar public in TCustomEditEnh.<br>
<br>
19-Jun-2005 - Installed Validating property to indicate when validation
processing is taking place.  Useful in data-aware derivatives to prevent
recursion.<br>
<br>
17-Feb-2004 - Fix to button processing, so second click in double click is
not missed (causes a second button-down event).  Thanks Ales Kahanek.<br>
<br>
09-Oct-2003 - Support for XP Themes.<br>
<br>
11-Apr-2003 - Made the DoAutoExit procedure protected so that it can be used
by derivatives.<br>
<br>
22-Sep-2002 - Improved handling of range deletions.<br>
<br>
22-Sep-2002 - Improved handling of mouse clicks to reposition in mask.<br>
<br>
29-Apr-2002 - Added AfterValidateEdit event.<br>
<br>
22-Jan-2002 - Implement support for IBO4.2.Fp - GetAutoLabel method.<br>
<br>
22-Jan-2002 - Created virtual methods AllowButtonClickOnFocus and
AllowGlyphClickOnFocus (both of which simply use the the eeoNoClickOnFocus
option in this base class).  Provided mainly to allow the combo derivation
to ignore clicks when focusing and AutoDropDown is true.<br>
<br>
04-Jan-2002 - Altered CMWantSpecialKey to better support combo styled
derivations.  (Need Escape and Enter when HasActiveChild is true.)<br>
<br>
28-Dec-2001 - Prepare for possible change to IBO interface definition
(addition of GetAutoLabel method).  Addition of Locale strings.  Addition of
the eeoIgnoreKeyPress option.  Moved a private property (FInternalUpd) to
protected for use by DateTime control.<br>
<br>
08-Sep-2001 - Added a ResetButton function to public, to allow manual
clearing of the button down states.  Discovered that PopupMenu (when triggered
by a button click via the shortcut) steals the keyup message without stealing
the focus, thus preventing the button from resetting as it should.
<br>
23-Aug-2001 - Added an InvalidateButton function that can be called
by derivations to cause button to be redrawn.  Also using this function
internally to be sure the button is redrawn on focus change.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_EditEnh;

{$INCLUDE Telesis.inc}

interface

uses Classes,
  Controls,
  ImgList,
  StdCtrls,
  Graphics,
  Types,
  Messages,
  Forms,
  ExtCtrls,
  {$IFDEF __COMPILE_WITH_IBO__}
  IB_ControlInterface,
  {$ENDIF}
  Telesis_Constants,
  Telesis_MaskEnh,
  Telesis_Utils;

const

{: @summary Simple constant representing the Key value of Ctrl+Enter
in KeyPress routines.}
  EE_VK_CTRLENTER = 10;

{: @summary Constant controlling the default value of Ctl3DShallow properties.
<br>
@desc Unless overridden at the package compile time (by defining
USECTL3DSHALLOW as a compiler directive) then the default is false for
compatibility with standard VCL components.}
  {$IFDEF USECTL3DSHALLOW}
  CTL3DSHALLOW_DEFAULT = true;
  {$ELSE}
  CTL3DSHALLOW_DEFAULT = false;
  {$ENDIF}


type

{: @summary Defines input character conversions performed.
<br>
@desc Note that this is merely an extension to the standard VCL character
case type to provide support for "ProperCase" character conversions. }
TCharCaseEnh = (
    ecNormal,          // no special character case processing
    ecUpperCase,       // all characters to upper case
    ecLowerCase,       // all characters to lower case
    ecProperCase       // words are "Proper Cased"
    );

{: @summary Set of options to control the behaviour of TCustomEditEnh and its
derivatives.
<br>
@desc Allows individual instances to control some of the more obscure default
behaviour aspects.<br>
<br>
@enum eeoAllowEmptyExit
If a mask is defined then this setting controls whether a user is permitted
to exit the control while it is empty.  The default VCL mask edit will not
permit exiting a masked control if anything has been input previously - which
can be inappropriate.  By default I permit the user to leave the control if
all input has been cleared.<br>

@enum eeoAutoExit
If masked and the user has entered data into all available positions the
cursor will automatically be jumped to the next control in the tab-order.<br>

@enum eeoBeepOnBadChar
If the global variable EditEnh_BeepOverride is false then this option
controls whether the control issues a beep when the user inputs an
invalid character.<br>

@enum eeoBeepOnValidateErr
If the global variable EditEnh_BeepOverride is false then this option
controls whether the control issues a beep when it finds an error during
validation.<br>

@enum eeoBeepOnTruncateErr
Defaults to true, so a beep will be issued whenever an input string is
truncated - when it is assigned directly to the Text property.  Generally you
will only want to set this to false when your control is receiving input
from floating point values - in such cases the number of decimal places
actually received is uncertain due to the approximations inherant to these
data types.  (Ultimately the beep is only issued according to the setting
of eeoBeepOnBadChar, this is provided as a separate setting in order to
allow the exclusion of beeps only when assigning text directly to the
control and you want truncated characters to be ignored.)<br>

@enum eeoBtnGlyphStretch
If a button is defined using a custom glyph this controls whether the glyph
will be stretched/shrunk to fit the size of the button.<br>

@enum eeoBtnProportional
If a button is defined this controls whether the glyph (or arrow drawings)
are constrained to maintain their proportions.<br>

@enum eeoExceptions
Controls whether the control will raise an exception when it recieves a
validation error.<br>

@enum eeoWordWrap
Controls whether text is allowed to word wrap.  The control could be resized
to display multiple lines of text.  I haven't tried out this option with
masking enabled.<br>

@enum eeoNoClickOnFocus
Controls when click events are generated for the button or glyph parts of
the control.  Normally (when this option is false) a mouse down event on
the button or glyph will focus the control and generate a click event on
the button or glyph.  If you set this option to true, then clicking on the
button or glyph of a control that does not have focus will cause the
control to gain focus, but will not generate a click event.  A second click
on the now focused control will be required to generate a click event.<br>

@enum eeoInputDecimalTrans
When EditMaskMode is eemNumeric and this option is true the keyboard input
of a '.' character will be translated to the current locale decimal point
character.  This is to make it easier to input numbers from the numeric
keypad which often only contains the '.' character and not the character
specific to the current locale.  Set this option to false to prevent this
translation from occuring.<br>

@enum eeoIgnoreKeyPress
When this property is set then keyboard based input will be ignored (the
KeyPress event will ignore input characters).  This is provided primarily to
support combo styled derivations - where you want to insist that the user
selects input from the drop box.  Setting this property should not prevent
copying of text to the clipboard - but will prevent paste.  It should not
prevent keyboard based navigation.

@enum eeoNoDefaultMouseWheel
Normally the control detects an enabled eebUpDownArrow button and generates
appropriate button1 and button2 calls on mousewheel events.  Setting this
option will turn off the default behaviour.
}
  TEditEnhOptions = set of (
    eeoAllowEmptyExit,
    eeoAutoExit,
    eeoBeepOnBadChar,
    eeoBeepOnValidateErr,
    eeoBeepOnTruncateErr,
    eeoBtnGlyphStretch,
    eeoBtnProportional,
    eeoExceptions,
    eeoWordWrap,
    eeoNoClickOnFocus,
    eeoInputDecimalTrans,
    eeoIgnoreKeyPress,
    eeoNoDefaultMouseWheel
    );


{: @summary EditEnh event used to allow developer override of button drawing. }
  TEditEnhDrawEvent = procedure(Sender: TObject; var DrawInfo: TDrawInfoEnh) of
    object;

{: @summary EditEnh event used to allow developer override of mask error
processing. }
  TEditEnhErrorEvent = procedure(Sender: TObject; ErrCode: integer;
    var RaiseExc: boolean) of object;

{: @summary EditEnh event see OnValidateEdit property. }
  TEditEnhValidateEditEvent = procedure(Sender: TObject; var ErrCode: integer) of
    object;


  TCustomEditEnh = class;

{: @summary EditEnh Button definition
<br>
@desc This class is provided as an interface to manage the properties
relating the button(s) of the TCustomEditEnh control and its derivatives.}
  TEditEnhButtonDef = class(TPersistent)
  private
    FOwner: TCustomEditEnh;
    function GetAlign: TAlign;
    procedure SetAlign(AValue: TAlign);
    function GetAutoRepeat: boolean;
    procedure SetAutoRepeat(AValue: boolean);
    function GetEnabled: boolean;
    procedure SetEnabled(AValue: boolean);
    function GetImageList: TImageList;
    procedure SetImageList(AValue: TImageList);
    function GetImageIndex: TImageIndex;
    procedure SetImageIndex(AValue: TImageIndex);
    function GetImagePressed: TImageIndex;
    procedure SetImagePressed(AValue: TImageIndex);
    function GetImageDisabled: TImageIndex;
    procedure SetImageDisabled(AValue: TImageIndex);
    function GetKey1: string;
    procedure SetKey1(AValue: string);
    function GetKey2: string;
    procedure SetKey2(AValue: string);
    function GetKey3: string;
    procedure SetKey3(AValue: string);
    function GetSize: integer;
    procedure SetSize(AValue: integer);
    function GetStyle: TEditEnhButtonStyle;
    procedure SetStyle(AValue: TEditEnhButtonStyle);

    function IsAlignStored: boolean;
    function IsAutoRepeatStored: boolean;
    function IsEnabledStored: boolean;
    function IsImageListStored: boolean;
    function IsImageIndexStored: boolean;
    function IsImagePressedStored: boolean;
    function IsImageDisabledStored: boolean;
    function IsKey1Stored: boolean;
    function IsKey2Stored: boolean;
    function IsKey3Stored: boolean;
    function IsSizeStored: boolean;
    function IsStyleStored: boolean;

  protected

  public
    constructor Create(AOwner: TCustomEditEnh);
    destructor Destroy; override;

  published

{: @summary Specify the alignment of the button.
<br>
@desc alNone and alClient options are ingored.}
    property Align: TAlign read GetAlign write SetAlign stored IsAlignStored;
{: @summary Set true to have button down event automatically repeat.
<br>
@desc Specify whether you want the button action to automatically repeat if
the button is held own (by the mouse or keyboard action).
<br><br>
Global variables impacting this property...
<br>  EditEnh_RepeatInit := 500;   The initial pause before repeating starts.
<br>  EditEnh_RepeatPause := 100;  The delay between calls.}
    property AutoRepeat: boolean read GetAutoRepeat write SetAutoRepeat
      stored IsAutoRepeatStored;
{: @summary Allows you to disable the button independantly of the control.
<br>
@desc The button is always disabled if the control is disabled.}
    property Enabled: boolean read GetEnabled write SetEnabled
      stored IsEnabledStored;
{: @summary Allows you specify an image list from which button glyphs can
be specified.}
    property ImageList: TImageList read GetImageList write SetImageList
      stored IsImageListStored;
{: @summary The index of the image to be displayed on the button.
<br>
@desc If two buttons are displayed then ImageIndex represents the image for
the first button, ImageIndex + 1 represents the image for the second button. }
    property ImageIndex: TImageIndex read GetImageIndex write SetImageIndex
      stored IsImageIndexStored;
{: @summary The index of the image to be displayed when the button is pressed.
<br>
@desc If value is < 0 then ImageIndex will be used instead. If two buttons are
displayed then ImagePressed represents the image for the first button,
ImagePressed + 1 represents the image for the second button. }
    property ImagePressed: TImageIndex read GetImagePressed write SetImagePressed
      stored IsImagePressedStored;
{: @summary The index of the image to be displayed when the button is disabled.
<br>
@desc If value is < 0 then ImageIndex will be used instead. If two buttons are
displayed then ImageDisabled represents the image for the first button,
ImageDisabled + 1 represents the image for the second button. }
    property ImageDisabled: TImageIndex read GetImageDisabled write
      SetImageDisabled stored IsImageDisabledStored;
{: @summary Shortcut key (TShortCut) which will act as a "click" of Button1.}
    property Key1: string read GetKey1 write SetKey1 stored IsKey1Stored;
{: @summary Shortcut key (TShortCut) which will act as a "click" of Button2.}
    property Key2: string read GetKey2 write SetKey2 stored IsKey2Stored;
{: @summary Shortcut key (TShortCut) which will act as a "click of Button3.
<br>
@desc Can be used as a simple way of defining a hotkey in the control.  This
hotkey will function regardless of whether a button is defined or not, since
it does not actually relate to a visible button anyway.}
    property Key3: string read GetKey3 write SetKey3 stored IsKey3Stored;
{: @summary Specifies the size of the button.
<br>
@desc If the button is left or right aligned this is the width, if the
button is top or bottom aligned then this is the height.}
    property Size: integer read GetSize write SetSize stored IsSizeStored;
{: @summary Specifies what sort of button should be displayed.
<br>
@desc Use eebNone to not have a button displayed.}
    property Style: TEditEnhButtonStyle read GetStyle write SetStyle
      stored IsStyleStored;

  end;                                  // TEditEnhButtonDef


{: @summary The base class for various Enhanced editor controls.
<br>
@desc These controls support enhanced mask processing (see TMaskEnh) and
embedded buttons amongst other things.
<br><br>
<strong>IMPORTANT:</strong> To have the control operate with PasswordChar
you MUST specify a mask.  The simplest option is to choose a mask-mode of
eemEditEnh and a mask value of "cccccccccc;0;_" (where you specify as
many "c" character positions as required/permitted for the password.}
  TCustomEditEnh = class(TCustomEdit
    {$IFDEF __COMPILE_WITH_IBO__}
    , IIB_GridWinControl
    {$ENDIF}
    )
  private
    FAlignment: TAlignment;
    FButton: TEditEnhButtonDef;
    FButtonAlign: TAlign;
    FButtonEnabled: boolean;
    FButtonSize: integer;
    FButtonStyle: TEditEnhButtonStyle;
    FButtonRepeat: boolean;
    FButtonState: TEditEnhButtonState;
    FButtonHover: TEditEnhButtonState;
    FButton1Key: TShortCut;
    FButton2Key: TShortCut;
    FButton3Key: TShortCut;
    FButtonImageList: TImageList;
    FButtonImageIndex: TImageIndex;
    FButtonImagePressed: TImageIndex;
    FButtonImageDisabled: TImageIndex;
    FGlyphAlign: TAlign;
    FGlyphSize: integer;

    FCharCaseEnh: TCharCaseEnh;
    FCtl3DShallow: boolean;

    FMask: TMaskEnh;
    FMaskChanging: integer;
    FSavedText: string;
    FDisableChangeEvents: integer;
    FNulledReturn: boolean; // Flag to indicate if CustomEdit nulled return key

    FModifyFlag: boolean;
    FValidating: boolean;
    FTextMouseDown: boolean;   // mouse is down in text area
    FOptions: TEditEnhOptions;
    FCanvas: TCanvas;

    FOnButton1: TNotifyEvent;
    FOnButton2: TNotifyEvent;
    FOnButton3: TNotifyEvent;
    FOnDrawButton: TEditEnhDrawEvent;
    FOnValidateEdit: TEditEnhValidateEditEvent;
    FAfterValidateEdit: TEditEnhValidateEditEvent;
    FOnValidateError: TEditEnhErrorEvent;
    FOnGlyphClick: TNotifyEvent;
    FOnDrawGlyph: TEditEnhDrawEvent;

    function GetButtonRect(var ARect: TRect; BdrAdjust: boolean): boolean;
    function GetGlyphRect(var ARect: TRect; BdrAdjust: boolean): boolean;
    procedure GetTextRect(var ARect: TRect); virtual;

    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMCaptureChanged(var Message: TMessage); message WM_CAPTURECHANGED;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SetFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMCopy(var Message: TMessage); message WM_COPY;
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMWantSpecialKey(var Message: TCMWantSpecialKey); message CM_WANTSPECIALKEY;
    procedure CNKeyUp(var Message: TWMKeyUp); message CN_KEYUP;
    procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;

    procedure SetAlignment(AValue: TAlignment);
    procedure SetButton(AValue: TEditEnhButtonDef);
    procedure SetCharCaseEnh(AValue: TCharCaseEnh);
    procedure SetCtl3DShallow(AValue: boolean);
    procedure MakeMask;
    function GetMask: TMaskEnh;
    function GetMaskChanging: boolean;
    function GetEditMask: string;
    procedure SetEditMask(AValue: string);
    function GetIsMasked: boolean;
    function GetIsEmpty: boolean;
    function GetMaskMode: TMaskEnhMode;
    procedure SetMaskMode(AValue: TMaskEnhMode);
    procedure SetOptions(AValue: TEditEnhOptions);
    procedure SetGlyphAlign(AValue: TAlign);
    procedure SetGlyphSize(AValue: integer);
    function GetText: string;
    procedure SetText(AValue: string);
    procedure RefreshDisplayedText; // Refresh displayed text without change event or change to modify
    procedure UpdateDisplayedText; // Refresh displayed text with change event, modify reset to false (by Windows)
    procedure ModifyText; // Refresh displayed text with change event and modify set to true
    function GetDisplayText: string;
    procedure SetDisplayText(AValue: string);

    procedure MovePosition(Key: Word; Shift: TShiftState);
    procedure DeleteChar(Key: Word; WithUpd: boolean);
    procedure InputChar(var Key: char; IncLiterals: boolean);
    procedure SyncSelection(Key: Word; ShiftIsDown: boolean);

    procedure SetButtonState(NewState: TEditEnhButtonState);
    procedure SetButtonHover(NewHover: TEditEnhButtonState);
    procedure RepeatButtonDown(Sender: TObject);
    procedure DoButtonDown;

    procedure DoMaskEnhError(Sender: TMaskEnh; ErrCode: integer);

  protected
    FInternalUpd: integer; // Internal Enh collection use only
    FDummyString: TTelesisEnh_Version;  // dummy value - not used.
    FCRLFProtection: boolean; // defaults to true, set to false for memo controls

    // Set of default value virtuals to permit easy override of
    // button attribute storage...
    function GetButtonAlignDefault: TAlign; virtual;
    function GetButtonEnabledDefault: boolean; virtual;
    function GetButtonSizeDefault: integer; virtual;
    function GetButtonStyleDefault: TEditEnhButtonStyle; virtual;
    function GetButtonAutoRepeatDefault: boolean; virtual;
    function GetButton1KeyDefault: TShortCut; virtual;
    function GetButton2KeyDefault: TShortCut; virtual;
    function GetButton3KeyDefault: TShortCut; virtual;

    // We allow a derivate to define their own Button3Key, this base defaults
    // to the user defined value (FButton3Key) but a derivative (eg: MemoEnh)
    // may choose to offer a value even when no user defined value exists.
    function GetButton3KeyEx: TShortCut; virtual;             

    function GetButtonImageListDefault: TImageList; virtual;
    function GetButtonImageIndexDefault: TImageIndex; virtual;
    function GetButtonImagePressedDefault: TImageIndex; virtual;
    function GetButtonImageDisabledDefault: TImageIndex; virtual;
    function AllowButtonClickOnFocus: boolean; virtual;
    function AllowGlyphClickOnFocus: boolean; virtual;

    procedure DoAutoExit;
    procedure SyncPositionToCaret;
    procedure SyncCaretToPosition;

    procedure LoadMaskProperty(Reader: TReader);
    procedure StoreMaskProperty(Writer: TWriter);
    procedure DefineProperties(Filer: TFiler); override;
    procedure Loaded; override;
    function IsMaskStored: boolean; virtual;

    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: char); override;
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;

    procedure Paint; virtual;
    procedure InvalidateButton;
    procedure InvalidateGlyph;
    procedure DoDrawGlyph(var DrawInfo: TDrawInfoEnh); virtual;
    procedure DoGlyphClick; virtual;
    function DoTextChange: boolean; virtual;

    function GetDBDisplayField: TObject; virtual;
    function GetButtonEnabled: boolean; virtual;
  //function GetOnButton1: TNotifyEvent; virtual;
  //procedure SetOnButton1( AValue: TNotifyEvent ); virtual;
  //function GetOnButton2: TNotifyEvent; virtual;
  //procedure SetOnButton2( AValue: TNotifyEvent ); virtual;
  //function GetOnButton3: TNotifyEvent; virtual;
  //procedure SetOnButton3( AValue: TNotifyEvent ); virtual;

    procedure DoButton1; virtual;
    procedure DoButton2; virtual;
    procedure DoButton3; virtual;
    procedure DoValidateEdit(var ErrCode: integer); virtual;
    procedure DoAfterValidateEdit(var ErrCode: integer); virtual;
    procedure DoValidateError(ErrCode: integer); virtual;

    {$IFDEF __COMPILE_WITH_IBO__}
    // These can be simply redirected
    function IIB_GridWinControl.GetText = GetDisplayText;
    function IIB_GridWinControl.DisplayColumn = GetDBDisplayField;
    // These are already defined
    //function HasActiveChild: boolean;
    //procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
    //procedure BoundsChanged;
    //function CanFocus: boolean;
    //procedure SetFocus;
    //procedure Invalidate;
    // These functions need to be defined - which I will do via
    // functions named IBG_* and redirect...
    function IBG_TranslateText(AVal: string): string; virtual;
    procedure IBG_SetBorderStyle( AVal: TBorderStyle );
    function IBG_GetFont: TFont;
    procedure IBG_SetFont( const AFont: TFont );
    procedure IBG_SetColor( AColor: TColor ); virtual;
    procedure IBG_SetVisible( AVal: boolean );
    function IBG_GetAutoLabel: TObject; virtual;
    function IIB_GridWinControl.TranslateText = IBG_TranslateText;
    procedure IIB_GridWinControl.SetBorderStyle = IBG_SetBorderStyle;
    function IIB_GridWinControl.GetFont = IBG_GetFont;
    procedure IIB_GridWinControl.SetFont = IBG_SetFont;
    procedure IIB_GridWinControl.SetColor = IBG_SetColor;
    procedure IIB_GridWinControl.SetVisible = IBG_SetVisible;
    // Support for IBO 4.2.Fp (comment out to use with previous versions)
    function IIB_GridWinControl.GetAutoLabel = IBG_GetAutoLabel;
    {$ENDIF}

    property Canvas: TCanvas read FCanvas;
{: Property is primarily used by memo derivatives to determine if the base
control has "nulled" a VK_RETURN character input. }
    property NulledReturn: boolean read FNulledReturn write FNulledReturn;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  // Moved to public for IB_Grid access
  // - we have to FORCE a BoundsChanged calculation due to some background
  // painting done by windows when embedded in the grid which bypasses the
  // normal WM_ERASEBKGND / WM_PAINT process.
  // (Has to do with WM_CTL* messages)
    procedure BoundsChanged;
    procedure Repaint; override;

{: @summary DisableChangeEvent can be called to prevent OnChange from firing.
<br>
@desc This is used internally to prevent change events while loading mask
values that may result in text change events.  However you may wish to use
this feature at other times to prevent OnChange while making multiple
changes to the contents of the edit control.<br>
<br>
DisableChangeEvent can be called multiple times, as long as each call
is accompanied by a call to EnableChangeEvent.  So use
DisableChangeEvent/EnableChangeEvent events inside try/finally blocks.}
    procedure DisableChangeEvent;
{: @summary EnableChangeEvent should be called after DisableChangeEvent to
reenable OnChange processing.
<br>
@desc Normally you will call EnableChangeEvent with the FireEvent parameter
set to true.  This will cause OnChange to be called so that your application
can resynchronise with any changes made while OnChange was disabled.<br>
<br>
Note that it is possible to call EnableChangeEvent(true) without having
called DisableChangeEvent - which will result in the OnChange event being
executed whether changes were made or not.  This may be useful sometimes
(or may not :-).<br>
<br>
DisableChangeEvent can be called multiple times, as long as each call
is accompanied by a call to EnableChangeEvent.  So use
DisableChangeEvent/EnableChangeEvent events inside try/finally blocks.}
    procedure EnableChangeEvent( FireEvent: boolean );

{: @summary Returns true if OnChange is currently enabled.
<br>
@desc See DisableChangeEvent and EnableChangeEvent.}
    function IsChangeEventEnabled: boolean;

{: @summary Set the character case by ordinal (integer) value.
<br>
@desc Allows TEditCharCase and compatible types to set the character case
without case statement translation. }
    procedure SetCharCaseByOrdinal(AValue: integer);
{: @summary Returns the full edit mask including the mode prefix.
<br>
@desc See TMaskEnhString for details.}
    function GetEditMaskEx: string;
{: @summary Called to set both the mask mode and the editmask from a single
string parameter.  See TMaskEnhString for details.}
    procedure SetEditMaskEx(AValue: string);
{: @summary Call to clear the contents of the control. }
    procedure Clear;{$IFDEF __VCL30_OR_LATER__} override;{$ENDIF}
{: @summary Returns whether the control contents can be modified.
<br>
@desc Call to find out if editing input is permitted to modify the contents
of the control.  In the base class this returns true/false depending upon
the ReadOnly status of the control.  In derived classes this is overridden
to allow editing to be permitted/denied by other factors.}
    function EditCanModify: boolean; virtual;
{: @summary Automatically called internally when the contents are about to
be modified.
<br>
@desc Call to begin modification of control contents.  Generally used by
derived classes to perform processing which needs to occur when the
controls contents are about to change due to user input.}
    function EditBeginModify: boolean; virtual;
{: @summary Called to see if the control has some form of active child control.
<br>
@desc Always returns false at this level but Combo derivations may override
this function to say when their dropboxes are active. }
    function HasActiveChild: boolean; virtual;
{: @summary Call to validate the contents of the control.
<br>
@desc If there is no mask defined it will always return true. }
    function ValidateEdit: boolean;
{: @summary Reset control back to the contents it had before it was
last modified. }
    procedure Reset; virtual;

{: @summary ResetButton provided to allow manual reset of the button state.
<br>
@desc I discovered that, if you show a PopupMenu as a result of a button
event via the keyboard shortcut, the menu steals the keyup message without
stealing the focus, preventing button from resetting.  So this function
is provided to allow manual reset of the button state. }
    procedure ResetButton;

  // make these public for easy access
  // ie. IB_Grid needs them to alter display properties dynamically
    property BorderStyle;
    property Color;
    property Font;
    property Visible;
    property ParentColor;
    property ParentFont;

    // Make these public for easy reference
    property Button1Key: TShortCut read FButton1Key;
    property Button2Key: TShortCut read FButton2Key;
    property Button3Key: TShortCut read FButton3Key;

{: @summary Control supports alignment of text.
<br>
@desc Internally this control is a multi-line edit control, but its processing
restricts input to a single line.  This attribute allows the control to
fully support the various alignment capabilities that normally only work
with memo controls.}
    property Alignment: TAlignment read FAlignment
    write SetAlignment
      default taLeftJustify;
{: @summary Override AutoSize to default to false. }
    property AutoSize default false;
{: @summary Defines the attributes (and existence) of any button to
be displayed.}
    property Button: TEditEnhButtonDef read FButton write SetButton;
{: @summary Indicates the current state of the controls button. }
    property ButtonState: TEditEnhButtonState read FButtonState;
{: @summary Indicates the current hover state of the controls button. }
    property ButtonHover: TEditEnhButtonState read FButtonHover;
{: @summary The extended CharCase property. }
    property CharCase: TCharCaseEnh read FCharCaseEnh
    write SetCharCaseEnh
      default ecNormal;
{: @summary Change the appears of the controls border so that the control has
a "shallow" 3D effect.}
    property Ctl3DShallow: boolean read FCtl3DShallow
    write SetCtl3DShallow
      default CTL3DSHALLOW_DEFAULT;
{: @summary Returns a TObject reference which is nil in the base class.
<br>
@desc This is often overridden in data-aware classes to return the
field/column that is used to provide the display value of the control.}
    property DBDisplayField: TObject read GetDBDisplayField;
{: @summary Used to return the value of the text which is actually
being displayed.
<br>
@desc If there is no mask this will be the same as the Text property.  If there
is a mask defined then the text property may return value that filters certain
characters (characters that are not to be saved).  So this property was
provided to give access to the actually displayed value.}
    property DisplayText: string read GetDisplayText write SetDisplayText;
{: @summary Set/Clear the edit mask associated with the control.
<br>
@desc You should set the EditMaskMode property BEFORE entering a mask
value here.}
    property EditMask: TMaskEnhString read GetEditMask write SetEditMask stored
      false;
{: @summary Set the desired mask interpretation mode.
<br>
@desc This should be set prior to entering a mask string in the
EditMask property.}
    property EditMaskMode: TMaskEnhMode read GetMaskMode
    write SetMaskMode
      stored false
      default eemNone;
{: @summary Alignment of glyph.
<br>
@desc Displays if GlyphSize > 0 or alClient is specified).}
    property GlyphAlign: TAlign read FGlyphAlign write SetGlyphAlign default
      alLeft;
{: @summary Size of Glyph area.
<br>
@desc Only painted if OnDrawGlyph event defined.  Glyph will only show
if GlyphSize > 0 or GlyphAlign = alClient. }
    property GlyphSize: integer read FGlyphSize write SetGlyphSize default 0;
{: @summary Indicates whether the control has received input.
<br>
@desc Mostly of use when the control IsMasked, this is the easiest way to
find out whether the control represents an "empty mask", that is, a control
which has not received any input characters, or which has been cleared - all
contents deleted. }
    property IsEmpty: boolean read GetIsEmpty;
{: @summary Indicates whether mask processing active or not. }
    property IsMasked: boolean read GetIsMasked;
{: @summary Reference to the TMaskEnh class of this control.
<br>
@desc Note that this is the mask processing object.  See the EditMask and
EditMode properties to access the actual mask string.}
    property Mask: TMaskEnh read GetMask;
{: @summary MaskChanging property is for use in the Change event to
detect changes that were due to changes in the mask. }
    property MaskChanging: boolean read GetMaskChanging;
{: @summary Operational properties of this control. }
    property Options: TEditEnhOptions read FOptions write SetOptions default
      [eeoAllowEmptyExit, eeoBeepOnTruncateErr, eeoBtnGlyphStretch,
      eeoBtnProportional, eeoExceptions, eeoInputDecimalTrans];
{: @summary Whether the control should be read-only, preventing user input. }
    property ReadOnly;
{: @summary Access the text value of the control.
<br>
@desc Note that the value of this property will pass through the mask processor
(if mask processing is active).  This means that the value returned may not be
identical to the display value.  For example; a mask that does not save its
literal character positions will filter those characters before being returned
via this property.}
    property Text: string read GetText write SetText;
{: @summary Indicates if currently in validation processing.}
    property Validating: boolean read FValidating;
{: @summary Indicates if mouse is down in the text area.}
    property TextMouseDown: boolean read FTextMouseDown;
{: @summary Event processing to be done when Button1 is clicked.
<br>
@desc Button1 is the Top, Left or Only button.}
    property OnButton1: TNotifyEvent read FOnButton1 write FOnButton1;
{: @summary Event processing to be done when Button2 is clicked.
<br>
@desc Button2 is the Bottom, or Right button.}
    property OnButton2: TNotifyEvent read FOnButton2 write FOnButton2;
{: @summary Event processing to be done when Button3 is clicked.
<br>
@desc Button3 does not actually exist as a visible button on the control.  It
can only be accessed by defining a valid shortcut to the Key3 property of the
button definition.  When the use enters that shortcut this event is fired.<br>
<br>
Setup this way to provide an easy way of adding a short cut to the control
that is not dependant on the visibility of the buttons.}
    property OnButton3: TNotifyEvent read FOnButton3 write FOnButton3;
    property OnChange;
{: @summary When button is set to eebCustom use this event to paint the button.}
    property OnDrawButton: TEditEnhDrawEvent read FOnDrawButton
    write FOnDrawButton;
{: @summary Define action here to draw the glyph.
<br>
@desc If you define the glyph properties to make room for a glyph you must
provide a handler here to draw something, or only a blank space will
appear.<br>
<br>
I recommend you use the various Draw* functions available from the
Telesis_Utils.pas unit to make life a bit easier.  Since those functions
are already setup to use the TDrawInfoEnh structure passed with this
event handler.}
    property OnDrawGlyph: TEditEnhDrawEvent read FOnDrawGlyph
    write FOnDrawGlyph;
{: @summary Event fired when click occurs on Glyph image.}
    property OnGlyphClick: TNotifyEvent read FOnGlyphClick
    write FOnGlyphClick;
{: @summary Event occurs just prior to the default validation processing
(on exit of control, or press of return key).
<br>
@desc Either process the contents of the control to make it valid, ready
for default validation, or set the ErrCode parameter to non-zero to indicate
that the value is not valid (skipping the default validation and raising
the ValidateError event).<br>
<br>
You could choose to raise an exception directly in this handler, to prevent
further processing.  However if you set ErrCode to an appropriate non-zero
value you can centralise your exception processing in the OnValidateError
handler.  (Perhaps using a single OnValidateError handler for all edit
controls, while creating specific OnValidateEdit handlers as required.)<br>
<br>
Note the predefined values that exist for ErrCode.  See E_MASKENH_*
entries in the online help or MaskEnh.pas unit.<br>
<br>
To create your own error codes for use in this handler, create values based
from E_EDITENH_USER_ERROR.  For example:<br>
<br>
<code>
const
  E_MY_OWN_ERROR = E_EDITENH_USER_ERROR + 1;
</code><br>
<br>
You can then setup a case statement in OnValidateError that will raise
exception messages appropriate to the given error codes.}
    property OnValidateEdit: TEditEnhValidateEditEvent read FOnValidateEdit
    write FOnValidateEdit;
{: @summary Event occurs just after normal validation processing, and only
occurs if validation was successful.
<br>
@desc You can choose to perform further validation and alter the ErrCode
or raise an exception. See also the notes for OnValidateEdit.<br>
<br>
The anticipated use for this event is to take some action after the contents
of the control have been validated as OK (which will occur if the control was
modified and then the Enter key pressed or the control is exited.)}
    property AfterValidateEdit: TEditEnhValidateEditEvent
      read FAfterValidateEdit write FAfterValidateEdit;
{: @summary Processing to be done when a mask validation error occurs.
<br>
@desc You can use the ErrCode parameter to determine what type of error has
occurred.  It may be one of the predefined E_MASKENH_* errors (see online
help or MaskEnh.pas for definitions) or it may be a value defined
in the OnValidateEdit handler.<br>
<br>
Only set the value the RaiseExc parameter if you want to override the default
processing of the ValidateError processing.  For example: You may decide that
you do not want a visual exception raised in some circumstances.  The
processing will still prevent the focus from moving until the control contains
a valid value.<br>
<br>
Because CustomEditEnh does not have a separate language module, you may want
to override the message displayed.  To do this simply raise your own
exceptions in this handler.  If you raise your own exception message inside
this handler any value assigned to RaiseExc is redundant and ignored.<br>
<br>
Note:  The RaiseExc parameter is initialised (before this handler is called)
according to whether CustomEditEnh thinks an error should be raised for a
particular type of error.  For example; It will be false if the error is
simply the result of a single character insertion to prevent the control
from being too annoying during dataentry.  It will also be false if
eeoExceptions is not defined in the Options property.  So the recommended
way to override exception message generation is something like...<br>
<br>
<code>
  if RaiseExc then case ErrCode of
    E_MASKENH_INVALID:        raise Exception.Create( 'my text' );
    E_EDITENH_USER_ERROR:     raise Exception.Create( 'my text' );
    ...etc...
  end;
</code><br>
}
    property OnValidateError: TEditEnhErrorEvent read FOnValidateError
    write FOnValidateError;

    property PasswordChar;

  published
    property Version_Enh: TTelesisEnh_Version read FDummyString
      write FDummyString stored false;

  end;                                  // TCustomEditEnh



{: @summary The published variation of TCustomEditEnh.
<br>
@desc Provides support for enhanced mask processing (see TMaskEnh) and
embedded buttons amongst other things.
<br><br>
<strong>IMPORTANT:</strong> To have the control operate with PasswordChar
you MUST specify a mask.  The simplest option is to choose a mask-mode of
eemEditEnh and a mask value of "cccccccccc;0;_" (where you specify as many
"c" character positions as required/permitted for the password.}
  TEditEnh = class(TCustomEditEnh)
    {$DEFINE PUBLISH_CustomEditEnh}
    {$DEFINE PUBLISH_MaskEnh}
    {$DEFINE PUBLISH_Text}
    {$INCLUDE Telesis_Publish.INC}
  end;                                  // TEditEnh




var
{: @summary Have EditEnh beep controlled by global variables.
<br>
@desc Defaults to true so that all beep functions are controlled by the
global variables.  Set to false to allow individual CustomEditEnh controls
to manage the beep of error functionality.}
  EditEnh_BeepOverride: boolean;
{: @summary EditEnh will beep when bad input character occurs.
<br>
@desc Defaults to true so CustomEditEnh controls will beep (with sound
controlled by EditEnh_BeepOnBadCharSound) when user inputs an invalid
character.  Only used if EditEnh_BeepOverride = true. }
  EditEnh_BeepOnBadChar: boolean;
{: @summary Sound made when beeping on bad character input.
<br>
@desc Controls the sound issued when an invalid character is input into a
CustomEditEnh control.  default is MB_OK, use $FFFFFFFF for computer beep.
See Windows API MessageBeep for other options }
  EditEnh_BeepOnBadCharSound: longint;
{: @summary EditEnh will beep when validation fails.
<br>
@desc Defaults to true so CustomEditEnh controls will beep (with sound
controlled by EditEnh_BeepOnValidateErrSound) when the contents of a
CustomEditEnh control fails validation.  Only used if
EditEnh_BeepOverride = true. }
  EditEnh_BeepOnValidateErr: boolean;
{: @summary Sound made when beeping on validation failure.
<br>
@desc Controls the sound issued when a CustomEditEnh control fails validation.
default is MB_ICONASTERISK, use $FFFFFFFF for computer beep.
See Windows API MessageBeep for other options }
  EditEnh_BeepOnValidateErrSound: longint;
{: @summary Control EditEnh button repeat capability.
<br>
@desc This is a global timer used by all instances of CustomEditEnh controls
that have "AutoRepeat := true" for their button properties.  The timer
controls the repeat processing while the button is held down.}
  EditEnh_RepeatTimer: TTimer;
{: @summary Control EditEnh button repeat lock.
<br>
@desc This global variable controls concurrent access to the
EditEnh_RepeatTimer global timer.}
  EditEnh_RepeatLock: integer;
{: @summary EditEnh button repeat initial delay.
<br>
@desc This is a global defines (in milliseconds) the initial delay before an
AutoRepeat button begins to repeat.}
  EditEnh_RepeatInit: integer;
{: @summary EditEnh button repeat pause delay.
<br>
@desc This is a global defines (in milliseconds) the subsequent delays (after
the initial delay defined by EditEnh_RepeatInit) for AutoRepeat button
repeats.}
  EditEnh_RepeatPause: integer;

implementation

uses Clipbrd,
  Menus,
  Windows,
  SysUtils,
  Mask,
  Telesis_Locale;
{  SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  ExtCtrls,
  Mask,
  TypInfo,
  ImgList,}

const
  MaskPropertyName = 'EditMaskEx';

constructor TEditEnhButtonDef.Create(AOwner: TCustomEditEnh);
begin
  FOwner := AOwner;
  inherited Create;
end;

destructor TEditEnhButtonDef.Destroy;
begin
  inherited Destroy;
end;

function TeditEnhButtonDef.GetAlign: TAlign;
begin
  Result := FOwner.FButtonAlign;
end;

procedure TEditEnhButtonDef.SetAlign(AValue: TAlign);
begin
  if AValue <> FOwner.FButtonAlign then
  begin
    // Only allow certain values...
    if (AValue = alTop) or (AValue = alLeft) or (AValue = alRight) or
      (AValue = alBottom) then
    begin
      FOwner.FButtonAlign := AValue;
      FOwner.BoundsChanged;
    end;
  end;
end;

function TEditEnhButtonDef.GetAutoRepeat: boolean;
begin
  Result := FOwner.FButtonRepeat;
end;

procedure TEditEnhButtonDef.SetAutoRepeat(AValue: boolean);
begin
  if AValue <> FOwner.FButtonRepeat then
  begin
    FOwner.FButtonRepeat := AValue;
  end;
end;

function TEditEnhButtonDef.GetEnabled: boolean;
begin
  Result := FOwner.FButtonEnabled;
end;

procedure TEditEnhButtonDef.SetEnabled(AValue: boolean);
begin
  if AValue <> FOwner.FButtonEnabled then
  begin
    FOwner.FButtonEnabled := AValue;
    if (not AValue) and (FOwner.FButtonHover <> eesNoButtonDown) then
      FOwner.SetButtonHover(eesNoButtonDown);
    if (not AValue) and (FOwner.FButtonState <> eesNoButtonDown) then
      FOwner.SetButtonState(eesNoButtonDown)
    else
      FOwner.Invalidate;
  end;
end;

function TEditEnhButtonDef.GetImageList: TImageList;
begin
  Result := FOwner.FButtonImageList;
end;

procedure TEditEnhButtonDef.SetImageList(AValue: TImageList);
begin
  if AValue <> FOwner.FButtonImageList then
  begin
    FOwner.FButtonImageList := AValue;
    FOwner.Invalidate;
  end;
end;

function TEditEnhButtonDef.GetImageIndex: TImageIndex;
begin
  Result := FOwner.FButtonImageIndex;
end;

procedure TEditEnhButtonDef.SetImageIndex(AValue: TImageIndex);
begin
  if AValue <> FOwner.FButtonImageIndex then
  begin
    FOwner.FButtonImageIndex := AValue;
    FOwner.Invalidate;
  end;
end;

function TEditEnhButtonDef.GetImagePressed: TImageIndex;
begin
  Result := FOwner.FButtonImagePressed;
end;

procedure TEditEnhButtonDef.SetImagePressed(AValue: TImageIndex);
begin
  if AValue <> FOwner.FButtonImagePressed then
  begin
    FOwner.FButtonImagePressed := AValue;
    FOwner.Invalidate;
  end;
end;

function TEditEnhButtonDef.GetImageDisabled: TImageIndex;
begin
  Result := FOwner.FButtonImageDisabled;
end;

procedure TEditEnhButtonDef.SetImageDisabled(AValue: TImageIndex);
begin
  if AValue <> FOwner.FButtonImageDisabled then
  begin
    FOwner.FButtonImageDisabled := AValue;
    FOwner.Invalidate;
  end;
end;

function TEditEnhButtonDef.GetKey1: string;
begin
  Result := ShortCutToText(FOwner.FButton1Key);
end;

procedure TEditEnhButtonDef.SetKey1(AValue: string);
begin
  FOwner.FButton1Key := TextToShortCut(AValue);
end;

function TEditEnhButtonDef.GetKey2: string;
begin
  Result := ShortCutToText(FOwner.FButton2Key);
end;

procedure TEditEnhButtonDef.SetKey2(AValue: string);
begin
  FOwner.FButton2Key := TextToShortCut(AValue);
end;

function TEditEnhButtonDef.GetKey3: string;
begin
  Result := ShortCutToText(FOwner.FButton3Key);
end;

procedure TEditEnhButtonDef.SetKey3(AValue: string);
begin
  FOwner.FButton3Key := TextToShortCut(AValue);
end;

function TEditEnhButtonDef.GetSize: integer;
begin
  Result := FOwner.FButtonSize;
end;

procedure TEditEnhButtonDef.SetSize(AValue: integer);
begin
  if AValue <> FOwner.FButtonSize then
  begin
    FOwner.FButtonSize := AValue;
    FOwner.BoundsChanged;
  end;
end;

function TeditEnhButtonDef.GetStyle: TEditEnhButtonStyle;
begin
  Result := FOwner.FButtonStyle;
end;

procedure TEditEnhButtonDef.SetStyle(AValue: TEditEnhButtonStyle);
begin
  if AValue <> FOwner.FButtonStyle then
  begin
    FOwner.FButtonStyle := AValue;
    FOwner.BoundsChanged;
  end;
end;

function TEditEnhButtonDef.IsAlignStored: boolean;
begin
  Result := (FOwner.FButtonAlign <> FOwner.GetButtonAlignDefault);
end;

function TEditEnhButtonDef.IsAutoRepeatStored: boolean;
begin
  Result := (FOwner.FButtonRepeat <> FOwner.GetButtonAutoRepeatDefault);
end;

function TEditEnhButtonDef.IsEnabledStored: boolean;
begin
  Result := (FOwner.FButtonEnabled <> FOwner.GetButtonEnabledDefault);
end;

function TEditEnhButtonDef.IsImageListStored: boolean;
begin
  Result := (FOwner.FButtonImageList <> FOwner.GetButtonImageListDefault);
end;

function TEditEnhButtonDef.IsImageIndexStored: boolean;
begin
  Result := (FOwner.FButtonImageIndex <> FOwner.GetButtonImageIndexDefault);
end;

function TEditEnhButtonDef.IsImagePressedStored: boolean;
begin
  Result := (FOwner.FButtonImagePressed <> FOwner.GetButtonImagePressedDefault);
end;

function TEditEnhButtonDef.IsImageDisabledStored: boolean;
begin
  Result := (FOwner.FButtonImageDisabled <> FOwner.GetButtonImageDisabledDefault);
end;

function TEditEnhButtonDef.IsKey1Stored: boolean;
begin
  Result := (FOwner.FButton1Key <> FOwner.GetButton1KeyDefault);
end;

function TEditEnhButtonDef.IsKey2Stored: boolean;
begin
  Result := (FOwner.FButton2Key <> FOwner.GetButton2KeyDefault);
end;

function TEditEnhButtonDef.IsKey3Stored: boolean;
begin
  Result := (FOwner.FButton3Key <> FOwner.GetButton3KeyDefault);
end;

function TEditEnhButtonDef.IsSizeStored: boolean;
begin
  Result := (FOwner.FButtonSize <> FOwner.GetButtonSizeDefault);
end;

function TEditEnhButtonDef.IsStyleStored: boolean;
begin
  Result := (FOwner.FButtonStyle <> FOwner.GetButtonStyleDefault);
end;





constructor TCustomEditEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
// Change the default height and width
  Height := 21;
  Width := 121;
  FCRLFProtection := true;
  ControlStyle := ControlStyle + [csReplicatable];
  FAlignment := taLeftJustify;
  AutoSize := false;
  FButtonAlign := GetButtonAlignDefault;
  FButtonSize := GetButtonSizeDefault;
  FButtonStyle := GetButtonStyleDefault;
  FButtonEnabled := GetButtonEnabledDefault;
  FButtonImageList := GetButtonImageListDefault;
  FButtonImageIndex := GetButtonImageIndexDefault;
  FButtonImagePressed := GetButtonImagePressedDefault;
  FButtonImageDisabled := GetButtonImageDisabledDefault;
  FButtonRepeat := GetButtonAutoRepeatDefault;
  FButton1Key := GetButton1KeyDefault;
  FButton2Key := GetButton2KeyDefault;
  FButton3Key := GetButton3KeyDefault;
  FButtonState := eesNoButtonDown;
  FButtonHover := eesNoButtonDown;
  FGlyphAlign := alLeft;
  FGlyphSize := 0;
  FCharCaseEnh := ecNormal;
  FCanvas := TControlCanvas.Create;
  TControlCanvas(FCanvas).Control := Self;
  FCtl3DShallow := CTL3DSHALLOW_DEFAULT;
  FMask := nil;
  FMaskChanging := 0;
  FSavedText := '';
  FOptions := [eeoAllowEmptyExit, eeoBeepOnTruncateErr, eeoBtnGlyphStretch,
    eeoBtnProportional, eeoExceptions, eeoInputDecimalTrans];
  FInternalUpd := 0;
  FDisableChangeEvents := 0;
  FNulledReturn := false;
  FModifyFlag := false;
  FValidating := false;
  FTextMouseDown := false;
  FOnValidateError := nil;
  FOnValidateEdit := nil;
  FOnButton1 := nil;
  FOnButton2 := nil;
  FOnButton3 := nil;
  FOnDrawButton := nil;
  FOnGlyphClick := nil;
  FOnDrawGlyph := nil;
  FButton := TEditEnhButtonDef.Create(Self);
end;

destructor TCustomEditEnh.Destroy;
begin
  if FMask <> nil then
  begin
    FMask.Free;
    FMask := nil;
  end;
  FButton.Free;
  FButton := nil;
  FCanvas.Free;
  FCanvas := nil;
  inherited Destroy;
end;

procedure TCustomEditEnh.Loaded;
begin
  inherited Loaded;
  Modified := false;
end;

procedure TCustomEditEnh.LoadMaskProperty(Reader: TReader);
var
  tmpStr, tmpText: string;
begin
  DisableChangeEvent;
  try
    tmpStr := Reader.ReadString;
    if Length(tmpStr) >= 0 then
    begin
      tmpText := Text;
      SetEditMaskEx(tmpStr);
      Text := tmpText;
    end;
  finally
    // do NOT fire the event as a result of loading the value
    EnableChangeEvent( false );
  end;
end;

procedure TCustomEditEnh.StoreMaskProperty(Writer: TWriter);
begin
  Writer.WriteString(GetEditMaskEx);
end;

procedure TCustomEditEnh.DefineProperties(Filer: TFiler);

  function DoWrite: boolean;
  begin
    if Filer.Ancestor <> nil then
    begin
      if TCustomEditEnh(Filer.Ancestor).FMask = nil then
        Result := FMask <> nil
      else
        if (FMask = nil) or
        (TCustomEditEnh(Filer.Ancestor).FMask.MaskEx <> FMask.MaskEx) then
          Result := True
        else
          Result := False;
    end
    else
      Result := FMask <> nil;
    Result := (Result and IsMaskStored);
  end;
begin
  inherited; // allow base classes to define properties
  Filer.DefineProperty(MaskPropertyName, LoadMaskProperty, StoreMaskProperty,
    DoWrite);
end;

function TCustomEditEnh.IsMaskStored: boolean;
begin
  Result := true;
end;

procedure TCustomEditEnh.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
  if eeoWordWrap in Options then
    Params.Style := Params.Style and not (ES_AUTOHSCROLL)
  else
    Params.Style := Params.Style or ES_AUTOHSCROLL;
  if FAlignment = taRightJustify then
    Params.Style := Params.Style or ES_RIGHT
  else
    if FAlignment = taCenter then
      Params.Style := Params.Style or ES_CENTER
    else
      Params.Style := Params.Style or ES_LEFT;
end;

procedure TCustomEditEnh.CreateWnd;
begin
  inherited CreateWnd;
  BoundsChanged;
end;

procedure TCustomEditEnh.WMSize(var Message: TWMSize);
begin
  inherited;
  BoundsChanged;
end;

procedure TCustomEditEnh.BoundsChanged;
var
  Rect: TRect;
  Rgn: HRGN;
begin
// If there is no window actually created yet then dont bother doing anything
// (otherwise we may try to create the window before we are ready).
  if WindowHandle = 0 then
    Exit;
  GetTextRect(Rect);
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Rect));
  Invalidate;                           // Invalidate the client area
// Now get the REAL window dimensions...
  GetWindowRect(Handle, Rect);
  Rect.BottomRight := ScreenToClient(Rect.BottomRight);
// Use the Windows "region" features to force a repaint of
// the non-client area (button and border).
  Rgn := CreateRectRgn(0, 0, Rect.Right + 2, Rect.Bottom + 2); // Make region
  GetWindowRgn(Handle, Rgn);            // Fill from Window
  SetWindowRgn(Handle, Rgn, true);      // Give ownership & redraw
  SetWindowRgn(Handle, 0, false);       // Clear Region
end;

procedure TCustomEditEnh.Repaint;
begin
  if IsMasked then
    RefreshDisplayedText;
  inherited Repaint;
end;

procedure TCustomEditEnh.KeyDown(var Key: Word; Shift: TShiftState);
var
  TestSC: TShortCut;
  B3Key: TShortCut;
begin
  TestSC := ShortCut(Key, Shift);
// !!! Changed to allow key-strokes to operate button events whether button
// is visible or not.  Can be a convenient way of providing default processing
// for certain keystrokes.  Can still disable the processing by disabling
// the buttons.
// !!!Button1Key and Button2Key are only available if the Button is visible
// !!!(both are available even if only one button is visible)
// !!!if (FButtonStyle <> eebNone) then begin
  if (FButton1Key <> 0) and (FButton1Key = TestSC) then
  begin
    if GetButtonEnabled then
      SetButtonState(eesButton1Down);
    Key := 0;
  end
  else
    if (FButton2Key <> 0) and (FButton2Key = TestSC) then
    begin
      if GetButtonEnabled then
        SetButtonState(eesButton2Down);
      Key := 0;
    end;
// !!!    end;
// !!!Button3Key is always available
  B3Key := GetButton3KeyEx;
  if (B3Key <> 0) and (B3Key = TestSC) then
  begin
    if GetButtonEnabled then
      SetButtonState(eesButton3Down);
    Key := 0;
  end;
  inherited KeyDown(Key, Shift);
  if (Key <> 0) and IsMasked and (not (ssAlt in Shift)) then
  begin
    if (Key = VK_LEFT) or (Key = VK_RIGHT) or
      (Key = VK_HOME) or (Key = VK_END) then
    begin
      if not (ssShift in Shift) then
      begin
        MovePosition(Key, Shift);
        Key := 0;
      end;
    end
    else
      if (KEY = VK_UP) or (KEY = VK_DOWN) then
      begin
        Key := 0;
      end
      else
        if (Key = VK_DELETE) then
        begin
          if EditBeginModify then
          begin
            DeleteChar(VK_DELETE, true);
          end;
          Key := 0;
        end;
  end
  else
    if (Key = VK_DELETE) then
    begin
      if not EditCanModify then
        Key := 0
      else
        if not EditBeginModify then
          Key := 0;
    end;
end;

procedure TCustomEditEnh.KeyUp(var Key: Word; Shift: TShiftState);
var
  TestSC: TShortCut;
  B3Key: TShortCut;
begin
  TestSC := ShortCut(Key, Shift);
// !!! See Notes for KeyDown
// !!! if (FButtonStyle <> eebNone) then begin
  if (FButton1Key <> 0) and (FButton1Key = TestSC) then
  begin
    SetButtonState(eesNoButtonDown);
    Key := 0;
  end
  else
    if (FButton2Key <> 0) and (FButton2Key = TestSC) then
    begin
      SetButtonState(eesNoButtonDown);
      Key := 0;
    end;
// !!!    end;
// Button3Key is always available
  B3Key := GetButton3KeyEx;
  if (B3Key <> 0) and (B3Key = TestSC) then
  begin
    SetButtonState(eesNoButtonDown);
    Key := 0;
  end;
  inherited KeyUp(Key, Shift);
  if (Key <> 0) and IsMasked and (not (ssAlt in Shift)) then
  begin
    if Key = VK_SHIFT then
    begin
      SyncSelection(Key, false);
    end
    else
      if (Key = VK_LEFT) or (Key = VK_RIGHT) or
      (Key = VK_HOME) or (Key = VK_END) then
      begin
        if (ssShift in Shift) then
        begin
          SyncSelection(Key, true);
        end;
      end
  end;
end;

procedure TCustomEditEnh.KeyPress(var Key: char);
begin
  // If any button is down then ignore the key-press, as this
  // ensures that any hot-key associated with button-down is
  // not also detected as a key press.
  if ButtonState <> eesNoButtonDown then
    Key := #0;
  NulledReturn := false;
  inherited KeyPress(Key);
// Ctrl+C (CopyToClipboard) excluded from further
// processing since it is not an edit function.
  if (Char(Key) in [Char(0), ^C]) then
    Exit;
// Dont process other keys if not allowed to edit
  if not EditCanModify then
  begin
    Key := Char(0);
    Exit;
  end;
// I force the processing of the clipboard keys because Windows can
// inconsistent with when it bothers to send the messages (I think
// it checks its internal read-only status before sending cut/paste).
// This becomes an issue with the IBO aware derivations which set
// the ReadOnly state, but whose actual read-only status is dependant
// upon the EditCanModify function.
  if not(eeoIgnoreKeyPress in Options) then
  begin
    if Char(Key) = ^V then
    begin
      PasteFromClipboard;
      Key := Char(0);
      Exit;
    end
    else
      if Char(Key) = ^X then
      begin
        CutToClipboard;
        Key := Char(0);
        Exit;
      end;
  end;
  if IsMasked then
  begin
    if Key = char(VK_ESCAPE) then
    begin
      if Modified then
        Reset;
    end
    else
      if (Key = char(VK_BACK)) and not(eeoIgnoreKeyPress in Options) then
      begin
        if EditBeginModify then
          DeleteChar(VK_BACK, true);
      end
      else
        if (Key = char(VK_RETURN)) or (Key = char(EE_VK_CTRLENTER)) then
        begin
        // VK_RETURN will have been validated during WantSpecialKey processing
          if (Key = char(EE_VK_CTRLENTER)) and Modified then
            ValidateEdit;
          if AutoSelect then
          begin
            SelectAll;
            SyncSelection(VK_END, false);
          end;
        end
        else
        begin
          if not(eeoIgnoreKeyPress in Options) and EditBeginModify then
          begin
            // in numeric mode we normally translate the '.' character to the
            // locale specific decimal separate to make input from the numeric
            // keypad easier.  This can be disabled by setting the
            // eeoInputDecimalTrans Option to false.
            if (EditMaskMode = eemNumeric) and
              (eeoInputDecimalTrans in Options) and (Key = '.') then
              Key := DecimalSeparator;
            InputChar(Key, false);
          end;
        end;
    Key := char(0);
  end
  else
  begin
    if (Key = char(VK_RETURN)) or (Key = char(EE_VK_CTRLENTER)) then
    begin
        // VK_RETURN will have been validated during WantSpecialKey processing
      if (Key = char(EE_VK_CTRLENTER)) and Modified then
        ValidateEdit;
      NulledReturn := true; // Tell derivatives that I nulled return key
      Key := char(0); // we are not really a multi-line control!
    end
    else
      if Key = char(VK_TAB) then
      begin
        NulledReturn := true; // Tell derivatives that I nulled the tab key
        Key := char(0);
      end
      else
        if Key = char(VK_ESCAPE) then
        begin
          if Modified then
            Reset;
          Key := char(0);
        end
        else
        begin
          if eeoIgnoreKeyPress in Options then
            Key := char(0)
          else
          begin
            if FCharCaseEnh = ecProperCase then
            begin
              if SelStart > 0 then
                Key := ConvCharCase(Text[SelStart], Key, Ord(ecProperCase))
              else
                Key := ConvCharCase(eeMskEmptyChar, Key, Ord(ecProperCase));
            end;
            if not EditBeginModify then
              Key := char(0);
          end;
        end;
  end;
end;


function TCustomEditEnh.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);
  if (not Result) and (FButtonStyle = eebUpDownArrow) and GetButtonEnabled and
    (not (eeoNoDefaultMouseWheel in Options)) then
  begin
    DoButton2;
    Result := true;
  end;
end;


function TCustomEditEnh.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);
  if (not Result) and (FButtonStyle = eebUpDownArrow) and GetButtonEnabled and
    (not (eeoNoDefaultMouseWheel in Options)) then
  begin
    DoButton1;
    Result := true;
  end;
end;


function TCustomEditEnh.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheel(Shift, WheelDelta, MousePos);
end;


function TCustomEditEnh.GetButtonRect(var ARect: TRect; BdrAdjust: boolean):
  boolean;
begin
  Result := false;
  if FButtonStyle <> eebNone then
  begin
    ARect.Top := 0;
    ARect.Left := 0;
    ARect.Right := ClientWidth;
    ARect.Bottom := ClientHeight;
    if BdrAdjust and (BorderStyle <> bsNone) then
    begin
      if Ctl3D then
      begin
        SetRect(ARect, 2, 2, ARect.Right + 2, ARect.Bottom + 2);
      end
      else
      begin
        SetRect(ARect, 1, 1, ARect.Right - 1, ARect.Bottom - 1);
      end;
    end;
    case FButtonAlign of
      alLeft:
        begin
          ARect.Right := ARect.Left + FButtonSize;
          Result := true;
        end;
      alRight:
        begin
          ARect.Left := ARect.Right - FButtonSize;
          Result := true;
        end;
      alBottom:
        begin
          ARect.Top := ARect.Bottom - FButtonSize;
          Result := true;
        end;
      alTop:
        begin
          ARect.Bottom := ARect.Top + FButtonSize;
          Result := true;
        end;
    end;
  end;
end;


function TCustomEditEnh.GetGlyphRect(var ARect: TRect; BdrAdjust: boolean):
  boolean;
begin
  Result := false;
  if (FGlyphSize > 0) or (FGlyphAlign = alClient) then
  begin
    ARect.Top := 0;
    ARect.Left := 0;
    ARect.Right := ClientWidth;
    ARect.Bottom := ClientHeight;
    if BdrAdjust and (BorderStyle <> bsNone) then
    begin
      if Ctl3D then
      begin
        SetRect(ARect, 2, 2, ARect.Right + 2, ARect.Bottom + 2);
      end
      else
      begin
        SetRect(ARect, 1, 1, ARect.Right - 1, ARect.Bottom - 1);
      end;
    end;
    if FButtonStyle <> eebNone then
    begin
        // adjust available rect is button is used.
      case FButtonAlign of
        alLeft: ARect.Left := ARect.Left + FButtonSize;
        alRight: ARect.Right := ARect.Right - FButtonSize;
        alBottom: ARect.Bottom := ARect.Bottom - FButtonSize;
        alTop: ARect.Top := ARect.Top + FButtonSize;
      end;
    end;
    case FGlyphAlign of
      alLeft:
        begin
          ARect.Right := ARect.Left + FGlyphSize;
          Result := true;
        end;
      alRight:
        begin
          ARect.Left := ARect.Right - FGlyphSize;
          Result := true;
        end;
      alBottom:
        begin
          ARect.Top := ARect.Bottom - FGlyphSize;
          Result := true;
        end;
      alTop:
        begin
          ARect.Bottom := ARect.Top + FGlyphSize;
          Result := true;
        end;
      alClient:
        begin
            // fill entire area
          Result := true;
        end;
    end;
  end;
end;


// Return the rectangle that we want to use for drawing the text.
// Again I use the Windows API to get the window rectangle - because I need
// this rectangle specified in non-client coordinates...

procedure TCustomEditEnh.GetTextRect(var ARect: TRect);
begin
  ARect := ClientRect;
  if FButtonStyle <> eebNone then
  begin
    case FButtonAlign of
      alLeft:
        begin
          ARect.Left := ARect.Left + FButtonSize;
        end;
      alRight:
        begin
          ARect.Right := ARect.Right - FButtonSize;
        end;
      alBottom:
        begin
          ARect.Bottom := ARect.Bottom - FButtonSize;
        end;
      alTop:
        begin
          ARect.Top := ARect.Top + FButtonSize;
        end;
    end;
  end;
  if (FGlyphSize > 0) or (FGlyphAlign = alClient) then
  begin
    case FGlyphAlign of
      alLeft:
        begin
          ARect.Left := ARect.Left + FGlyphSize;
        end;
      alRight:
        begin
          ARect.Right := ARect.Right - FGlyphSize;
        end;
      alBottom:
        begin
          ARect.Bottom := ARect.Bottom - FGlyphSize;
        end;
      alTop:
        begin
          ARect.Top := ARect.Top + FGlyphSize;
        end;
      alClient:
        begin
          ARect.Top := -Height;
          ARect.Left := -Width;
          ARect.Right := -1;
          ARect.Bottom := -1;
            //ARect.Right := ARect.Left; // Zero width text
            //ARect.Bottom := ARect.Top; // Zero height text
        end;
    end;
  end;
end;

procedure TCustomEditEnh.WMPaint(var Message: TWMPaint);
begin
  inherited;
  {$IFDEF __VCL30_OR_LATER__}
  Canvas.Lock;
  try
    {$ENDIF}
    Canvas.Handle := GetWindowDC(Handle);
    try
      Paint;
    finally
      ReleaseDC(Handle, Canvas.Handle);
      Canvas.Handle := 0;
    end;
    {$IFDEF __VCL30_OR_LATER__}
  finally
    Canvas.Unlock;
  end;
  {$ENDIF}
end;

procedure TCustomEditEnh.Paint;
var
  DrawInfo: TDrawInfoEnh;
  HasXPTheme: boolean;
begin
  HasXPTheme := Telesis_Utils.UseXPThemes;
  with Canvas do
  begin
    if Ctl3DShallow and (BorderStyle = bsSingle) and Ctl3D and not HasXPTheme
    then begin
      with DrawInfo do
      begin // Just using DrawInfo for its Rect property
        GetWindowRect(Self.Handle, Rect);
        Rect.TopLeft := ScreenToClient(Rect.TopLeft);
        Rect.BottomRight := ScreenToClient(Rect.BottomRight);
        Pen.Color := clBtnFace;
        Pen.Width := 1;
        MoveTo(1, Rect.Bottom - 1);
        LineTo(1, 1);
        LineTo(Rect.Right, 1);
      end;
    end;
    if GetButtonRect(DrawInfo.Rect, true) then
    begin
      if HasXPTheme then
      begin
        Brush.Color := Self.Color;
        FillRect(DrawInfo.Rect);
      end
      else
      begin
        // Fill background area
        Brush.Color := clBtnFace;
        FillRect(DrawInfo.Rect);
        // Pull in according to the alignment so highlighting does not
        // conflict with the text area.  (Only necessary with Shallow)
        if Ctl3DShallow then
        begin
          case FButtonAlign of
            alRight:
              begin
                DrawInfo.Rect.Left := DrawInfo.Rect.Left + 1;
              end;
            alBottom:
              begin
                DrawInfo.Rect.Top := DrawInfo.Rect.Top + 1;
              end;
          end;
        end;
      end;
      with DrawInfo do
      begin
        VectorGlyph := vgNone;
        Canvas := Self.Canvas;
        VectorColor := clBtnText;
        BkColor := clBtnFace;
        VectorBrush := bsSolid;
        // Rect := already done
        if HasXPTheme then
          Border := bbFlat
        else
        begin
          if Ctl3DShallow then
            Border := bbShallow
          else
            Border := bbCtl3D;
        end;
        Proportional := (eeoBtnProportional in FOptions);
        Enabled := GetButtonEnabled;
        Stretch := (eeoBtnGlyphStretch in FOptions);
        Pressed := FButtonState;
        Hover := FButtonHover;
        TestState := eesButton1Down;
        ImageList := FButtonImageList;
        ImageIndex := FButtonImageIndex;
        ImageDisabled := FButtonImageDisabled;
        ImagePressed := FButtonImagePressed;
      end;
      if FButtonStyle = eebCustom then
      begin
        if assigned(FOnDrawButton) then
          FOnDrawButton(Self, DrawInfo);
      end
      else
      begin
        DrawEditEnhButton(DrawInfo, FButtonStyle);
      end;
    end;
    // Again simply using DrawInfo for its Rect property
    if GetGlyphRect(DrawInfo.Rect, true) then
    begin
      with DrawInfo do
      begin
        VectorGlyph := vgNone;
        Canvas := Self.Canvas;
        VectorColor := clBtnText;
        BkColor := clBtnFace;
        VectorBrush := bsSolid;
        // Rect := already done
        Border := bbFlat;
        Proportional := (eeoBtnProportional in FOptions);
        Enabled := Self.Enabled;
        Stretch := (eeoBtnGlyphStretch in FOptions);
        Pressed := eesNoButtonDown;
        Hover := eesNoButtonDown;
        TestState := eesButton1Down;
        ImageList := nil;
        ImageIndex := -1;
        ImageDisabled := -1;
        ImagePressed := -1;
        // Pull in according to the alignment so highlighting does not
        // conflict with the text area.  (Only necessary with Shallow)
        if Ctl3DShallow and not HasXPTheme then
        begin
          case FGlyphAlign of
            alRight:
              begin
                Rect.Left := Rect.Left + 1;
              end;
            alBottom:
              begin
                Rect.Top := Rect.Top + 1;
              end;
          end;
        end;
      end;
      DoDrawGlyph(DrawInfo);
    end;
  end;
end;


procedure TCustomEditEnh.InvalidateButton;
var
  ARect: TRect;
begin
  if HandleAllocated and GetButtonRect(ARect, true) then
    InvalidateRect(Handle, @ARect, false);
end;


procedure TCustomEditEnh.InvalidateGlyph;
var
  ARect: TRect;
begin
  if HandleAllocated and GetGlyphRect(ARect, true) then
    InvalidateRect(Handle, @ARect, false);
end;

procedure TCustomEditEnh.DoDrawGlyph(var DrawInfo: TDrawInfoEnh);
begin
  if assigned(FOnDrawGlyph) then
    FOnDrawGlyph(Self, DrawInfo)
  else
  begin
    with DrawInfo.Canvas do
    begin
        // Initialise brush and pen settings
      Brush.Style := bsSolid;
      if Self.Focused then
      begin
        Brush.Color := clHighlight;
        Pen.Color := clHighlightText;
      end
      else
      begin
        Brush.Color := Self.Color;
        Pen.Color := Self.Font.Color;
      end;
      FillRect(DrawInfo.Rect);
    end;
  end;
end;


procedure TCustomEditEnh.DoGlyphClick;
begin
  if assigned(FOnGlyphClick) then
    FOnGlyphClick(Self);
end;


procedure TCustomEditEnh.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
  Rect: TRect;
begin
//if FButtonStyle = eebNone then Rect := ClientRect
// else
  GetTextRect(Rect);
  FillRect(Message.DC, Rect, Brush.Handle);
  Message.Result := 1;
end;

procedure TCustomEditEnh.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
// Allow the glyph to change with focus
  InvalidateGlyph;
  InvalidateButton;
end;

procedure TCustomEditEnh.WMKillFocus(var Message: TWMKillFocus);
begin
  inherited;
// If a button is/was down and we've lost focus then we'll never get the
// mouse-up/key-up events to reset the status, so I reset here...
  SetButtonState(eesNoButtonDown);
  InvalidateGlyph;
  InvalidateButton;
end;

procedure TCustomEditEnh.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  // We dont want Returns (or Escape keys???)
  Message.Result := Message.Result and not DLGC_WANTALLKEYS;
end;

procedure TCustomEditEnh.WMCopy(var Message: TMessage);
var
  ClipStr: string;
  SelStart, SelStop: integer;
begin
  if IsMasked and (PasswordChar = #0) then
  begin
    SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
    if (SelStart <> SelStop) and (SelStart >= 0) then
    begin
      ClipStr := Copy(Mask.GetMaskedText(true, true, false), SelStart + 1,
        SelStop - SelStart);
      if Length(ClipStr) > 0 then
        Clipboard.AsText := ClipStr;
    end;
  end
  else
    inherited;
end;

procedure TCustomEditEnh.WMCut(var Message: TMessage);
var
  ClipStr: string;
  SelStart, SelStop: integer;
begin
  SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
  if SelStart = SelStop then
    Exit;
  if (not EditBeginModify) then
    Exit;
  if IsMasked and (PasswordChar = #0) then
  begin
    if (SelStart <> SelStop) and (SelStart >= 0) then
    begin
      ClipStr := Copy(Mask.GetMaskedText(true, true, false), SelStart + 1,
        SelStop - SelStart);
      if Length(ClipStr) > 0 then
        Clipboard.AsText := ClipStr;
      DeleteChar(VK_DELETE, true);
    end;
  end
  else
    inherited;
end;

procedure TCustomEditEnh.WMPaste(var Message: TMessage);
var
  ClipStr, NewStr: string;
  i: integer;
begin
  if (not EditBeginModify) then
    Exit;
  ClipStr := Clipboard.AsText;
  if Length(ClipStr) > 0 then
  begin
    if IsMasked then
    begin
      if (Mask.Mode > eemMaskList) and (SelLength > 0) then
        DeleteChar(VK_DELETE, true);    // Delete selected text
      SyncCaretToPosition;
        //Inc(FInternalUpd);
      try
        for i := 1 to Length(ClipStr) do
        begin
          InputChar(ClipStr[i], true);
        end;
      finally
        //Dec(FInternalUpd);
        //ModifyText;
        SyncCaretToPosition;
      end;
    end
    else
    begin
        // For non-memo style controls (the default)...
      if FCRLFProtection then
      begin
        // We want to prevent cr/lf chars getting through even if we
        // are not masked, convert each new line to one space.
        if LastDelimiter(#13#10, ClipStr) > 0 then
        begin
          NewStr := '';
          for i := 1 to Length(ClipStr) do
          begin
            if ClipStr[i] = #13 then
              NewStr := NewStr + ' '
            else if ClipStr[i] = #10 then
            begin
              if not ((i > 1) and (ClipStr[i-1] = #13)) then
                NewStr := NewStr + ' ';
            end
            else
              NewStr := NewStr + ClipStr[i];
          end;
          ClipStr := NewStr;
        end;
        SelText := ClipStr;
      end
      else
      begin
        inherited;
      end;
    end;
  end;
end;

procedure TCustomEditEnh.CMEnter(var Message: TCMEnter);
begin
  if IsMasked then
  begin
    RefreshDisplayedText;
    if HandleAllocated and
      not (csDesigning in ComponentState) then
    begin
      if FInternalUpd = 0 then
      begin
        Mask.SaveMaskTextTo(FSavedText);
        // If mouse is down then position will be synchronised on mouse up
        if not TextMouseDown then
        begin
          Mask.InitMaskInput(false);
          SyncCaretToPosition;
        end;
        // I did not used to use mouse down test, but I want the controls
        // to be have consistently with other Windows controls (no selection
        // on mouse down) so have added in in here.
        if AutoSelect and not TextMouseDown then
        begin
          SelectAll;
          SyncSelection(VK_END, false);
        end;
      end
      else
      begin
        SyncCaretToPosition;
      end;
    end;
  end
  else
  begin
    FSavedText := Text;
    // nb. Using CRLFProtection as indicator that this is a single line control
    // cannot use the window style because created as multi-line capable.
    if AutoSelect and not TextMouseDown and
      FCRLFProtection then
      SelectAll;
  end;
  inherited;
end;

procedure TCustomEditEnh.CMExit(var Message: TCMExit);
begin
  if Modified and
    not (csDesigning in ComponentState) then
    ValidateEdit;
  if IsMasked then
  begin
    RefreshDisplayedText;
  end;
  inherited;
end;

procedure TCustomEditEnh.CMTextChanged(var Message: TMessage);
begin
  if IsMasked then
  begin
    if FInternalUpd = 0 then
    begin
      if FModifyFlag then
      begin                             // If a genuine modification to the text
        Modified := true;               // tell everyone it is modified
        SyncCaretToPosition;
      end
      else
      begin // otherwise pretend we've only just entered...
        Mask.SaveMaskTextTo(FSavedText);
        if not TextMouseDown then
        begin
          Mask.InitMaskInput(false);
          SyncCaretToPosition;
        end;
        if AutoSelect and (not TextMouseDown) and HandleAllocated then
        begin
          SelectAll;
          SyncSelection(VK_END, false);
        end;
      end;
      if DoTextChange then
        inherited;
    end;
  end
  else
  begin
    if DoTextChange then
      inherited;
    FSavedText := Text;
  end;
end;

// CMWantSpecialKey processing is required to support validation of
// edit controls on modal dialogs.  Without this the enter key goes
// directly to the default button and the contents of the control
// never get validated.

procedure TCustomEditEnh.CMWantSpecialKey(var Message: TCMWantSpecialKey);
begin
  inherited;
  if (Message.CharCode = VK_ESCAPE) then
  begin
    // If a derivation has an active child then we will want the Enter key.
    if HasActiveChild then
      Message.Result := 1
    else
      // if modified then escape key needed for reset processing
      if Modified then
        Message.Result := 1;
  end
  else
    if (Message.CharCode = VK_RETURN) then
    begin
      // If a derivation has an active child then we will want the Enter key.
      if HasActiveChild then
        Message.Result := 1
      else
        // if modified enter key need to perform validate
        if Modified then
        begin
          if not ValidateEdit then
            Message.Result := 1;
        end;
    end;
end;

// Its very frustrating, but WantSpecialKey processing may throw an
// exception (from ValidateEdit), in which case the user will get a dialog
// from which they may press Enter.  The up-stroke from the Enter key is
// not swallowed by the dialog and ends up in the edit control causing
// yet another call to WantSpecialKey - and so on while ever the user
// keeps pressing the enter key.  The only way I can work out how to
// avoid this is to ignore the CNKeyUp message for VK_RETURN.  To minimise
// and potential impacts this may cause, I only ignore it when the control
// is modified - which is the only time that ValidateEdit is called.

procedure TCustomEditEnh.CNKeyUp(var Message: TWMKeyUp);
begin
  if (Message.CharCode <> VK_RETURN) or (not Modified) then
    inherited;
end;

procedure TCustomEditEnh.CNCommand(var Message: TWMCommand);
begin
  if (Message.NotifyCode = EN_CHANGE) then
  begin
    if DoTextChange then
      inherited;                        // calls change event
  end
  else
    inherited;
end;

procedure TCustomEditEnh.WMLButtonDown(var Message: TWMLButtonDown);
var
  Rect: TRect;
  DoInherited, WasFocused: boolean;
begin
  DoInherited := true;
  WasFocused := Focused;
  if GetButtonRect(Rect, false) then
  begin
    if (Message.XPos >= Rect.Left) and (Message.XPos < Rect.Right) and
      (Message.YPos >= Rect.Top) and (Message.YPos < Rect.Bottom) then
    begin
      SendCancelMode(Self);
      DoInherited := false;
      if CanFocus and GetButtonEnabled then
      begin
        SetFocus;
        // Make sure the button state has been reset...
        if FButtonState <> eesNoButtonDown then
          SetButtonState(eesNoButtonDown);
        // before setting a new button state...
        MouseCapture := true;
        if WasFocused or AllowButtonClickOnFocus then
          case FButtonStyle of
            eebUpDownBlank, eebUpDownGlyph, eebUpDownArrow:
              begin
                if Message.YPos < (Rect.Top + ((Rect.Bottom - Rect.Top) div 2))
                  then
                  SetButtonState(eesButton1Down)
                else
                  SetButtonState(eesButton2Down);
              end;
            eebLeftRightBlank, eebLeftRightGlyph, eebLeftRightArrow,
              eebLeftRightEllipsis:
              begin
                if Message.XPos < (Rect.Left + ((Rect.Right - Rect.Left) div 2))
                  then
                  SetButtonState(eesButton1Down)
                else
                  SetButtonState(eesButton2Down);
              end;
          // Default - assume its a single button configuration
          else
            SetButtonState(eesButton1Down);
          end;                          // case
      end;
    end;
  end;
  if DoInherited then                   // if not found on button check glyph
    if GetGlyphRect(Rect, false) then
    begin
      if (Message.XPos >= Rect.Left) and (Message.XPos < Rect.Right) and
        (Message.YPos >= Rect.Top) and (Message.YPos < Rect.Bottom) then
      begin
        SendCancelMode(Self);
        DoInherited := false;
        if CanFocus and GetButtonEnabled then
        begin
          SetFocus;
          MouseCapture := true;
          if WasFocused or AllowGlyphClickOnFocus then
            DoGlyphClick;
        end;
      end;
    end;
  if DoInherited then
  begin
    FTextMouseDown := true;
    inherited;
  end;
end;

procedure TCustomEditEnh.WMLButtonUp(var Message: TWMLButtonUp);
begin
  FTextMouseDown := false;
  if (FButtonState <> eesNoButtonDown) then
  begin
    if MouseCapture then
      SendCancelMode(Self);
    SetButtonState(eesNoButtonDown);
    MouseCapture := false;
  end
  else
  begin
    SendCancelMode(Self);
    MouseCapture := false;
    SyncPositionToCaret;
  end;
  inherited;
end;

procedure TCustomEditEnh.WMMouseMove(var Message: TWMMouseMove);
var
  Rect: TRect;
begin
  Cursor := crDefault;
  inherited;
  if GetButtonRect(Rect, false) then
  begin
    if (Message.XPos >= Rect.Left) and (Message.XPos < Rect.Right) and
      (Message.YPos >= Rect.Top) and (Message.YPos < Rect.Bottom) then
    begin
      Cursor := crArrow;
      case FButtonStyle of
        eebUpDownBlank, eebUpDownGlyph, eebUpDownArrow:
        begin
          if Message.YPos < (Rect.Top + ((Rect.Bottom - Rect.Top) div 2))
          then
            SetButtonHover(eesButton1Down)
          else
            SetButtonHover(eesButton2Down);
        end;
        eebLeftRightBlank, eebLeftRightGlyph, eebLeftRightArrow,
          eebLeftRightEllipsis:
        begin
          if Message.XPos < (Rect.Left + ((Rect.Right - Rect.Left) div 2))
          then
            SetButtonHover(eesButton1Down)
          else
            SetButtonHover(eesButton2Down);
        end;
      // Default - assume its a single button configuration
      else
        SetButtonHover(eesButton1Down);
      end;                          // case
      MouseCapture := true;
    end
    else
    begin
      if (ButtonHover <> eesNoButtonDown) then
      begin
        MouseCapture := false;
        SetButtonHover(eesNoButtonDown);
      end;
    end;
  end;
  if Cursor <> crArrow then
    if GetGlyphRect(Rect, false) then
    begin
      if (Message.XPos >= Rect.Left) and (Message.XPos < Rect.Right) and
        (Message.YPos >= Rect.Top) and (Message.YPos < Rect.Bottom) then
      begin
        Cursor := crArrow;
      end;
    end;
end;

procedure TCustomEditEnh.WMCaptureChanged(var Message: TMessage);
begin
  if (FButtonState <> eesNoButtonDown) then
  begin
    SetButtonState(eesNoButtonDown);
  end;
  inherited;
end;

// Called with every text change - return true to have default Change
// processing take place with every text change.

function TCustomEditEnh.DoTextChange: boolean;
begin
  if FDisableChangeEvents = 0 then
    Result := true
  else
    Result := false;
end;

// Designed for use in derived types - particularly database derivations where
// it may be nice to have all database derivatives to return their column or
// datalink specification.  For now we just return nil.

function TCustomEditEnh.GetDBDisplayField: TObject;
begin
  Result := nil;
end;

function TCustomEditEnh.GetButtonEnabled: boolean;
var
  Control: TWinControl;
begin
  Result := (Enabled and FButtonEnabled);
  if Result then
  begin
    Control := Self.Parent;
    while Control <> nil do
    begin
      if not Control.Enabled then
      begin
        Result := false;
        break;
      end;
      Control := Control.Parent;
    end;
  end;
end;

//function TCustomEditEnh.GetOnButton1: TNotifyEvent;
//begin
//Result := FOnButton1;
//end;
//
//procedure TCustomEditEnh.SetOnButton1( AValue: TNotifyEvent );
//begin
//FOnButton1 := AValue;
//end;
//
//function TCustomEditEnh.GetOnButton2: TNotifyEvent;
//begin
//Result := FOnButton2;
//end;
//
//procedure TCustomEditEnh.SetOnButton2( AValue: TNotifyEvent );
//begin
//FOnButton2 := AValue;
//end;
//
//function TCustomEditEnh.GetOnButton3: TNotifyEvent;
//begin
//  Result := FOnButton3;
//end;
//
//procedure TCustomEditEnh.SetOnButton3( AValue: TNotifyEvent );
//begin
//  FOnButton3 := AValue;
//end;

procedure TCustomEditEnh.SetAlignment(AValue: TAlignment);
begin
  if AValue <> FAlignment then
  begin
    FAlignment := AValue;
    RecreateWnd;
  end;
end;

procedure TCustomEditEnh.SetButton(AValue: TEditEnhButtonDef);
begin
// Do nothing
end;

procedure TCustomEditEnh.SetCharCaseEnh(AValue: TCharCaseEnh);
begin
  if AValue <> FCharCaseEnh then
  begin
    FCharCaseEnh := AValue;
    case FCharCaseEnh of
      ecNormal: inherited CharCase := StdCtrls.ecNormal;
      ecLowerCase: inherited CharCase := StdCtrls.ecLowerCase;
      ecUpperCase: inherited CharCase := StdCtrls.ecUpperCase;
      ecProperCase: inherited CharCase := StdCtrls.ecNormal;
    end;
  end;
end;

procedure TCustomEditEnh.SetCtl3DShallow(AValue: boolean);
begin
  if AValue <> FCtl3DShallow then
  begin
    FCtl3DShallow := AValue;
    BoundsChanged;
  end;
end;

function TCustomEditEnh.GetMask: TMaskEnh;
begin
  Result := FMask;
end;

procedure TCustomEditEnh.MakeMask;
begin
  FMask := TMaskEnh.Create;
  FMask.OnError := DoMaskEnhError;
end;

function TCustomEditEnh.GetMaskChanging: boolean;
begin
  Result := FMaskChanging <> 0;
end;

function TCustomEditEnh.GetEditMask: string;
begin
  if Mask <> nil then
    Result := Mask.Mask
  else
    Result := '';
end;

procedure TCustomEditEnh.SetEditMask(AValue: string);
var
  SaveMaskTxt: string;
  SavePlainTxt: string;
  WasMasked: boolean;
begin
  Inc(FMaskChanging);
  try
    SaveMaskTxt := '';
    SavePlainTxt := '';
    WasMasked := IsMasked;
    if WasMasked then
    begin
      Mask.SaveMaskTextTo(SaveMaskTxt);
      SavePlainTxt := Mask.Text;
    end
    else
      if PasswordChar <> #0 then
        SavePlainTxt := DisplayText;
    if Length(AValue) <= 0 then
    begin
      if Mask <> nil then
      begin
        if Mask.Mode = eemNone then
        begin
          FMask.Free;
          FMask := nil;
        end
        else
        begin
          Mask.Mask := AValue;
        end;
      end;
    end
    else
    begin
      if Mask = nil then
      begin
        MakeMask;
      end;
      Mask.Mask := AValue;
    end;
    if WasMasked then
    begin
      if IsMasked then
        Mask.RestoreMaskTextFrom(SaveMaskTxt, false)
      else
        Text := SavePlainTxt;
    end
    else
      Text := SavePlainTxt;
    RefreshDisplayedText;
  finally
    Dec(FMaskChanging);
  end;
end;

function TCustomEditEnh.GetIsMasked: boolean;
begin
  if Mask = nil then
    Result := false
  else
  begin
    Result := ((Mask.Mode <> eemNone) and (Mask.MaskLen > 0));
  end;
end;

function TCustomEditEnh.GetIsEmpty: boolean;
begin
  if IsMasked then
  begin
    Result := Mask.IsEmpty;
  end
  else
    Result := (Text = '');
end;

function TCustomEditEnh.GetMaskMode: TMaskEnhMode;
begin
  if Mask = nil then
    Result := eemNone
  else
    Result := Mask.Mode;
end;

procedure TCustomEditEnh.SetMaskMode(AValue: TMaskEnhMode);
var
  ClearMsk: boolean;
begin
  if Mask = nil then
  begin
    if AValue <> eemNone then
    begin
      MakeMask;
      Mask.Mode := AValue;
      Inc(FMaskChanging);
      try
        RefreshDisplayedText;
      finally
        Dec(FMaskChanging);
      end;
    end;
  end
  else
  begin
    if Mask.Mode <> AValue then
    begin
      if (AValue = eemNone) and (Mask.MaskLen <= 0) then
        ClearMsk := true
      else
        ClearMsk := false;
      Mask.Mode := AValue;
      if ClearMsk then
        SetEditMask(''); // will delete the mask to release resources
      Inc(FMaskChanging);
      try
        RefreshDisplayedText;
      finally
        Dec(FMaskChanging);
      end;
    end;
  end;
end;

procedure TCustomEditEnh.SetOptions(AValue: TEditEnhOptions);
var
  OldOpt: TEditEnhOptions;
begin
  if AValue <> FOptions then
  begin
    OldOpt := FOptions;
    FOptions := AValue;
    // These options mean we need to recreate the window...
    if (eeoWordWrap in OldOpt) and not (eeoWordWrap in FOptions) then
    begin
      if HandleAllocated then
        RecreateWnd;
    end
    // Otherwise we just need to refresh the display
    else
    begin
      if HandleAllocated then
        Invalidate;
    end;
  end;
end;

procedure TCustomEditEnh.SetGlyphAlign(AValue: TAlign);
begin
  if AValue <> FGlyphAlign then
  begin
    // Only allow certain values...
    if (AValue = alTop) or (AValue = alLeft) or (AValue = alRight) or
      (AValue = alBottom) or (AValue = alClient) then
    begin
      FGlyphAlign := AValue;
      BoundsChanged;
    end;
  end;
end;

procedure TCustomEditEnh.SetGlyphSize(AValue: integer);
begin
  if AValue <> FGlyphSize then
  begin
    // Only allow certain values...
    if AValue >= 0 then
    begin
      FGlyphSize := AValue;
      BoundsChanged;
    end;
  end;
end;

function TCustomEditEnh.GetText: string;
begin
  if IsMasked then
  begin
    Result := Mask.Text; //Mask.GetMaskedText( true, true, (Focused and EditCanModify) );
  end
  else
    Result := DisplayText;
end;

procedure TCustomEditEnh.SetText(AValue: string);
var
  tmpStr: string;
  i: integer;
begin
  if IsMasked then
  begin
    try
      Mask.Text := AValue;
      UpdateDisplayedText;
    except
      RefreshDisplayedText;
      Mask.InitMaskInput(false);
      SyncCaretToPosition;
      raise;
    end;
    if not TextMouseDown then
    begin
      Mask.InitMaskInput(false);
      SyncCaretToPosition;
    end;
  end
  else
  begin
    if (Mask <> nil) and (Mask.MaskLen > 0) then
      Mask.Clear;
    // For non-memo style controls (the default)...
    if FCRLFProtection then
    begin
      tmpStr := Copy(AValue, 1, Length(AValue));
      for i := 1 to Length(tmpStr) do
      begin
        if (tmpStr[i] = #13) or (tmpStr[i] = #10) then
          tmpStr[i] := ' ';
      end;
      DisplayText := tmpStr;
    end
    else
      DisplayText := AValue;
  end;
end;

procedure TCustomEditEnh.RefreshDisplayedText;
var
  SaveMod: boolean;
begin
  if IsMasked then
  begin
    Inc(FInternalUpd);
    try
      SaveMod := Modified;
      if Mask.IsEmpty then
        DisplayText := ''
      else
        DisplayText := Mask.GetMaskedText(true, true, (Focused and EditCanModify));
      if SaveMod <> Modified then
        Modified := SaveMod;
    finally
      Dec(FInternalUpd);
    end;
  end;
end;

// Note that Modified will be reset to false by Windows
// Presumes that IsMasked has already been checked...

procedure TCustomEditEnh.UpdateDisplayedText;
begin
  if Mask.IsEmpty then
    DisplayText := ''
  else
    DisplayText := Mask.GetMaskedText(true, true, (Focused and EditCanModify));
end;

// Presumes that IsMasked has already been checked...

procedure TCustomEditEnh.ModifyText;
begin
  FModifyFlag := true; // Tell the text change event we want to be Modified!
  try
    DisplayText := Mask.GetMaskedText(true, true, (Focused and EditCanModify));
    // If setting display text did not actually cause a CMTextChanged event
    // (and so update the Modified flag) then we must do it explicitly.
    if not Modified then
      Modified := true;
  finally
    FModifyFlag := false;
  end;
end;

function TCustomEditEnh.GetDisplayText: string;
begin
  Result := inherited Text;
end;

procedure TCustomEditEnh.SetDisplayText(AValue: string);
begin
// Password char processing only available with mask (where the actual
// text is stored separately).
  if (PasswordChar = #0) or (not IsMasked) then
    inherited Text := AValue
  else
  begin
    FillChar(AValue[1], Length(AValue), PasswordChar);
    inherited Text := AValue;
  end;
end;

procedure TCustomEditEnh.MovePosition(Key: Word; Shift: TShiftState);
begin
  Mask.MovePosition(Key, Shift);
  if (Key = VK_RIGHT) and (Mask.Position = Mask.MaskLen) and
    (eeoAutoExit in Options) then
    DoAutoExit;
  SyncCaretToPosition;                  // Synchronise the caret
end;

procedure TCustomEditEnh.DeleteChar(Key: Word; WithUpd: boolean);
var
  SelStart, SelStop: integer;

  procedure StringDelete;
  begin
    if (SelStart = 0) and (SelStop = Length(DisplayText)) then
    begin
      Mask.Clear;
      Mask.AdjustMaskPos(false);
      ModifyText;
    end
    else
    begin
      if Mask.DeleteRange(WithUpd, SelStop - SelStart) then
        ModifyText;
    end;
  end;
begin
  SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
  if (Key = VK_BACK) then
  begin
    if (SelStart = (SelStop - 1)) and (Mask.Mode <= eemMaskList) then
    begin
      if Mask.DeleteBackSpace(WithUpd) then
        ModifyText;
    end
    else
      if SelStart = SelStop then
      begin
        Mask.AdjustMaskPos(false);
        if Mask.DeleteBackSpace(WithUpd) then
          ModifyText;
      end
      else
        StringDelete;
  end
  else
    if (Key = VK_DELETE) then
    begin
      if (SelStart = (SelStop - 1)) and (Mask.Mode <= eemMaskList) then
      begin
        if Mask.DeleteForward(WithUpd) then
          ModifyText;
      end
      else
        if SelStart = SelStop then
        begin
          if Mask.DeleteForward(WithUpd) then
            ModifyText;
        end
        else
          StringDelete;
    end;
  SyncCaretToPosition;
end;

procedure TCustomEditEnh.InputChar(var Key: char; IncLiterals: boolean);
var
  SavedTxt: string;
  Restore: boolean;
begin
// First check for special case of numeric sign - since there is no point
// doing deletions and so forth if we are just going to switch a literal...
  if (Mask.Mode = eemNumeric) and
    ((Key = eeMskNumSign) or (Key = eeMskNumSignOpt)) then
  begin
    if Mask.SetNumericSign(Key) then
    begin
      ModifyText;
      Mask.AdjustMaskPos(false);
      SyncCaretToPosition;
    end
  end
  else
  begin
    Restore := false;
    if (SelLength > 1) or ((SelLength = 1) and (Mask.Position >= Mask.MaskLen))
      then
    begin
      Mask.SaveMaskTextTo(SavedTxt);
      Restore := true;
      DeleteChar(VK_DELETE, false);
    end;
    if SelLength > 0 then
    begin
      Mask.AdjustMaskPos(false);
      if Mask.ReplaceMaskChar(Key, IncLiterals, true) then
      begin
        ModifyText;                     // this is the easiest approach
        SyncCaretToPosition;
        Restore := false;
        if (Mask.Position = Mask.MaskLen) and (eeoAutoExit in Options) then
          DoAutoExit;
      end
    end
    else
    begin
      Mask.AdjustMaskPos(false);
      if Mask.InsertMaskChar(Key, IncLiterals) then
      begin
        ModifyText;
        SyncCaretToPosition;
        Restore := false;
        if (Mask.Position = Mask.MaskLen) and (eeoAutoExit in Options) then
          DoAutoExit;
      end;
    end;
    if Restore then
    begin
      Mask.RestoreMaskTextFrom(SavedTxt, true);
      RefreshDisplayedText;             // make sure display matches input
      SyncCaretToPosition;
    end;
  end;
end;

function TCustomEditEnh.ValidateEdit: boolean;
var
  TmpCode: integer;
begin
  TmpCode := E_EDITENH_NOERROR;
  Result := true;
  if not FValidating then // prevent re-entry
  try
    FValidating := true;
    try
      DoValidateEdit( TmpCode );
    except
      // If an exception was generated then make sure we keep focus in the control
      Inc(FInternalUpd);
      try
        if (not Focused) and CanFocus then
          SetFocus;
      finally
        Dec(FInternalUpd);
      end;
      raise;                              // then re-raise the exception
    end;
    if TmpCode = E_EDITENH_NOERROR then
    begin
      if IsMasked then
      begin
        if (eeoAllowEmptyExit in Options) and IsEmpty then
        begin
          DoAfterValidateEdit( TmpCode );
          if TmpCode <> E_EDITENH_NOERROR then
            Result := false;
          Exit;
        end;
        // MaskEnh calls its own validate error handling - usually assigned
        // back to the controls DoMaskEnhError handler, with appropriate code.
        Result := Mask.ValidateEdit;
      end;
      if Result then
      begin
        DoAfterValidateEdit( TmpCode );
        if TmpCode <> E_EDITENH_NOERROR then
          Result := false;
      end;
    end
    else
    begin
      // ValidateError may not raise an exception to report false so caller
      // knows if the validation succeeded.
      Result := false;
      DoValidateError(TmpCode);
    end;
  finally
    FValidating := false;
  end;
end;

procedure TCustomEditEnh.DoMaskEnhError(Sender: TMaskEnh; ErrCode: integer);
begin
  DoValidateError(ErrCode);
end;

function TCustomEditEnh.GetButtonAlignDefault: TAlign;
begin
  Result := alRight;
end;

function TCustomEditEnh.GetButtonEnabledDefault: boolean;
begin
  Result := true;
end;

function TCustomEditEnh.GetButtonSizeDefault: integer;
begin
  Result := 17;
end;

function TCustomEditEnh.GetButtonStyleDefault: TEditEnhButtonStyle;
begin
  Result := eebNone;
end;

function TCustomEditEnh.GetButtonAutoRepeatDefault: boolean;
begin
  Result := false;
end;

function TCustomEditEnh.GetButton1KeyDefault: TShortCut;
begin
  Result := 0;
end;

function TCustomEditEnh.GetButton2KeyDefault: TShortCut;
begin
  Result := 0;
end;

function TCustomEditEnh.GetButton3KeyDefault: TShortCut;
begin
  Result := 0;
end;

function TCustomEditEnh.GetButton3KeyEx: TShortCut;
begin
  Result := FButton3Key;
end;

function TCustomEditEnh.GetButtonImageListDefault: TImageList;
begin
  Result := nil;
end;

function TCustomEditEnh.GetButtonImageIndexDefault: TImageIndex;
begin
  Result := 0;
end;

function TCustomEditEnh.GetButtonImagePressedDefault: TImageIndex;
begin
  Result := -1;
end;

function TCustomEditEnh.GetButtonImageDisabledDefault: TImageIndex;
begin
  Result := -1;
end;

function TCustomEditEnh.AllowButtonClickOnFocus: boolean;
begin
  Result := (not (eeoNoClickOnFocus in FOptions));
end;

function TCustomEditEnh.AllowGlyphClickOnFocus: boolean;
begin
  Result := (not (eeoNoClickOnFocus in FOptions));
end;

procedure TCustomEditEnh.Clear;
begin
  if IsMasked then
  begin
    Mask.Clear;
    RefreshDisplayedText;
  end
  else
    inherited;
end;

function TCustomEditEnh.EditCanModify: boolean;
begin
  Result := (not ReadOnly);
end;

function TCustomEditEnh.EditBeginModify: boolean;
begin
  Result := (not ReadOnly);
end;

function TCustomEditEnh.HasActiveChild: boolean;
begin
  Result := false;
end;

procedure TCustomEditEnh.DoValidateEdit(var ErrCode: integer);
begin
  if assigned(FOnValidateEdit) then
    FOnValidateEdit(Self, ErrCode);
end;

procedure TCustomEditEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  if assigned(FAfterValidateEdit) then
    FAfterValidateEdit(Self, ErrCode);
end;

procedure TCustomEditEnh.DoValidateError(ErrCode: integer);
var
  RaiseExc: boolean;
begin
  if (csDesigning in ComponentState) or (csLoading in ComponentState) then
    Exit;
  try
    // Dont raise exceptions (by default) for character insert/replace errors
    // only for entire string validation errors...
    if (eeoExceptions in Options) and (ErrCode <> E_MASKENH_CHAR_INVALID) and
      (ErrCode <> E_MASKENH_CHAR_TRUNCATED) then
      RaiseExc := true
    else
      RaiseExc := false;
    if assigned(FOnValidateError) then
      FOnValidateError(Self, ErrCode, RaiseExc);
    if ErrCode = E_MASKENH_CHAR_INVALID then
    begin
      if (EditEnh_BeepOverride and EditEnh_BeepOnBadChar) or
        ((not EditEnh_BeepOverride) and (eeoBeepOnBadChar in Options)) then
        MessageBeep(EditEnh_BeepOnBadCharSound);
    end
    else
      if ErrCode = E_MASKENH_CHAR_TRUNCATED then
      begin
        if (eeoBeepOnTruncateErr in Options) or
          not ((Mask <> nil) and Mask.AssigningText) then
        begin
          if (EditEnh_BeepOverride and EditEnh_BeepOnBadChar) or
            ((not EditEnh_BeepOverride) and (eeoBeepOnBadChar in Options)) then
            MessageBeep(EditEnh_BeepOnBadCharSound);
        end;
      end
      else
        if (ErrCode = E_MASKENH_INVALID) or
        (ErrCode >= E_EDITENH_USER_ERROR) then
        begin
          if (EditEnh_BeepOverride and EditEnh_BeepOnValidateErr) or
            ((not EditEnh_BeepOverride) and (eeoBeepOnValidateErr in Options))
              then
            MessageBeep(EditEnh_BeepOnValidateErrSound);
          if not RaiseExc then
          begin
            // If no exception, make sure focus remains with control anyway.
            Inc(FInternalUpd);
            try
              if (not Focused) and CanFocus then
                SetFocus;
            finally
              Dec(FInternalUpd);
            end;
          end;
        end
        else
        begin
          if (EditEnh_BeepOverride and EditEnh_BeepOnValidateErr) or
            ((not EditEnh_BeepOverride) and (eeoBeepOnValidateErr in Options))
              then
            MessageBeep(EditEnh_BeepOnValidateErrSound);
        end;
    if RaiseExc then
      raise EDBEditError.Create(GetEditEnhErrorMsg(ErrCode));
  except
    // If an exception was generated then make sure we keep focus in the control
    Inc(FInternalUpd);
    try
      if (not Focused) and CanFocus then
        SetFocus;
    finally
      Dec(FInternalUpd);
    end;
    raise;                              // reraise the exception
  end;
end;


{$IFDEF __COMPILE_WITH_IBO__}
function TCustomEditEnh.IBG_TranslateText(AVal: string): string;
begin
  Result := AVal;
end;

procedure TCustomEditEnh.IBG_SetBorderStyle( AVal: TBorderStyle );
begin
  BorderStyle := AVal;
end;

function TCustomEditEnh.IBG_GetFont: TFont;
begin
  Result := Font;
end;

procedure TCustomEditEnh.IBG_SetFont( const AFont: TFont );
begin
  Font.Assign( AFont );
end;

procedure TCustomEditEnh.IBG_SetColor( AColor: TColor );
begin
  Color := AColor;
end;

procedure TCustomEditEnh.IBG_SetVisible( AVal: boolean );
begin
  Visible := AVal;
end;

function TCustomEditEnh.IBG_GetAutoLabel: TObject;
begin
  Result := nil;
end;
{$ENDIF}


procedure TCustomEditEnh.Reset;
begin
  if IsMasked then
  begin
    if Modified then
    begin
      Mask.RestoreMaskTextFrom(FSavedText, true);
      UpdateDisplayedText;
      Mask.InitMaskInput(false);
      if AutoSelect then
      begin
        SelectAll;
        SyncSelection(VK_END, false);
      end;
      Modified := false;
    end;
  end
  else
  begin
    if Modified then
    begin
      Text := FSavedText;
      Modified := false;
    end;
  end;
end;

procedure TCustomEditEnh.ResetButton;
begin
  SetButtonState( eesNoButtonDown );
end;

procedure TCustomEditEnh.SyncPositionToCaret;
var
  SelStart, SelStop: Integer;
begin
  if not HandleAllocated then Exit;
  if IsMasked and (FInternalUpd = 0) then
  begin
    Inc(FInternalUpd);
    try
      SelStart := -1;
      SelStop := -1;
      SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
      if SelStart < 0 then
        SelStart := 0;
      if SelStart > Mask.MaskLen then
        SelStart := Mask.MaskLen;
      if Mask.MoveToPosition(SelStart, SelStop = SelStart) then
      begin
        SelStop := SelStart + 1;
        if SelStop > Mask.MaskLen then
          SelStop := Mask.MaskLen;
        SendMessage(Handle, EM_SETSEL, SelStop, SelStart);
      end;
    finally
      Dec(FInternalUpd);
    end;
  end;                                  // IsMasked and InternalUpd = 0
end;

procedure TCustomEditEnh.SyncCaretToPosition;
var
  SelStart, SelStop: Integer;
begin
  if not HandleAllocated then Exit;
  if IsMasked and (FInternalUpd = 0) then
  begin
    Inc(FInternalUpd);
    try
      if Mask.GetDisplayPosition(SelStart, SelStop) then
      begin
        SendMessage(Handle, EM_SETSEL, SelStart, SelStart);
      end
      else
      begin
        SendMessage(Handle, EM_SETSEL, SelStop, SelStart);
      end;
    finally
      Dec(FInternalUpd);
    end;
  end;                                  // IsMasked and InternalUpd = 0
end;

procedure TCustomEditEnh.SyncSelection(Key: Word; ShiftIsDown: boolean);
var
  SelStart, SelStop: Integer;
begin
  SelStart := -1;
  SelStop := -1;
  SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
  if (SelStart = SelStop) and (not ShiftIsDown) then
  begin
    SyncPositionToCaret;
  end
  else
  begin
    if ShiftIsDown then
    begin
      if ((SelStart = SelStop) or
        ((SelStart = 1) and (SelStop = Mask.MaskLen) and (Key = VK_END))
        ) and
        (SelStart > 0) and (SelStart <= Mask.MaskLen) and
        (Mask.Mode <= eemMaskList) and
        Mask.GetIsLeftJust(SelStart) then
      begin
        Dec(SelStart);
        if SelStop < Mask.MaskLen then
          Inc(SelStop);
      end;
      if (Key = VK_LEFT) or (KEY = VK_HOME) then
      begin
        SendMessage(Handle, EM_SETSEL, SelStart, SelStart);
        SyncPositionToCaret;
        SendMessage(Handle, EM_SETSEL, SelStop, SelStart);
      end
      else
      begin                             // (Key = VK_RIGHT) or (Key = VK_END)
        SendMessage(Handle, EM_SETSEL, SelStart, SelStart);
        SyncPositionToCaret;
        SendMessage(Handle, EM_SETSEL, SelStart, SelStop);
      end;
    end;
  end;
end;

// I dont know why, but SelectNext and FindNextControl are *protected*
// WinControl functions which means I cannot just call either of these
// functions to jump to the next control.  So what I have done is create
// my own "FindNextControl" equivalent function which uses is essentially just
// a copy of the code from TWinControl.  I have separated this from the edit
// component and declared the function so that I can reuse it later if I want.

procedure TCustomEditEnh.DoAutoExit;
var
  NxtControl: TWinControl;
begin
  if (Parent <> nil) and Parent.HandleAllocated then
  begin
    NxtControl := GetNextControl(Parent, Self, true, true, false);
    if (NxtControl <> nil) and NxtControl.HandleAllocated then
      NxtControl.SetFocus;
  end;
end;

procedure TCustomEditEnh.DisableChangeEvent;
begin
  Inc(FDisableChangeEvents);
end;

procedure TCustomEditEnh.EnableChangeEvent( FireEvent: boolean );
begin
  if (FDisableChangeEvents > 0) then
    Dec(FDisableChangeEvents);
  if (FDisableChangeEvents = 0) and FireEvent then
    Change;
end;

function TCustomEditEnh.IsChangeEventEnabled: boolean;
begin
  Result := (FDisableChangeEvents = 0);
end;

procedure TCustomEditEnh.SetCharCaseByOrdinal(AValue: integer);
begin
  if (AValue >= 0) and (AValue <= Ord(ecProperCase)) then
  begin
    CharCase := TCharCaseEnh(AValue);
  end;
end;

{: Called to retrieve a string that can be used to set both the mode and
the mask at one time.  See TMaskEnhString for details.}
function TCustomEditEnh.GetEditMaskEx: string;
begin
  if Mask = nil then
    Result := ''
  else
    Result := Mask.MaskEx;
end;

procedure TCustomEditEnh.SetEditMaskEx(AValue: string);
begin
  if (Length(AValue) <= 0) or (AnsiUpperCase(Trim(AValue)) = '::NO::') then
  begin
    //EditMask := '';
    EditMaskMode := eemNone;
  end
  else
  begin
    if Mask = nil then
      MakeMask;
    Mask.MaskEx := AValue;
    //RefreshDisplayedText;
  end;
end;

procedure TCustomEditEnh.SetButtonState(NewState: TEditEnhButtonState);
begin
  if NewState <> FButtonState then
  begin
    // Turn off any repeating that may have been active
    if (FButtonState <> eesNoButtonDown) and
      EditEnh_RepeatTimer.Enabled then
    begin
      EditEnh_RepeatTimer.Enabled := false;
    end;
    FButtonState := NewState;
    Repaint;
    if (FButtonState <> eesNoButtonDown) then
    begin
      DoButtonDown;
      if FButtonRepeat then
      begin
        EditEnh_RepeatTimer.Interval := EditEnh_RepeatInit;
        EditEnh_RepeatTimer.OnTimer := RepeatButtonDown;
        EditEnh_RepeatTimer.Enabled := true;
      end;
    end;
  end
  else
  begin
    if (FButtonState <> eesNoButtonDown) then
      DoButtonDown;
  end;
end;

procedure TCustomEditEnh.SetButtonHover(NewHover: TEditEnhButtonState);
begin
  if NewHover <> FButtonHover then
  begin
    FButtonHover := NewHover;
    Repaint;
  end;
end;

procedure TCustomEditEnh.RepeatButtonDown(Sender: TObject);
begin
// If the button is no longer down then turn off the timer...
  if FButtonState = eesNoButtonDown then
  begin
    EditEnh_RepeatTimer.Enabled := false;
  end
  else
  begin
    DoButtonDown;
    if FButtonRepeat then
    begin
      if (integer(EditEnh_RepeatTimer.Interval) <> EditEnh_RepeatPause) or
        (not EditEnh_RepeatTimer.Enabled) then
      begin
        EditEnh_RepeatTimer.Interval := EditEnh_RepeatPause;
        EditEnh_RepeatTimer.Enabled := true;
      end;
    end
    else
      EditEnh_RepeatTimer.Enabled := false;
  end;
end;

procedure TCustomEditEnh.DoButtonDown;
begin
  case FButtonState of
    eesButton1Down:
      begin
        DoButton1;
      end;
    eesButton2Down:
      begin
        DoButton2;
      end;
    eesButton3Down:
      begin
        DoButton3;
      end;
  end;
end;

procedure TCustomEditEnh.DoButton1;
begin
  if Assigned(FOnButton1) then
  begin
    FOnButton1(Self);
  end;
end;

procedure TCustomEditEnh.DoButton2;
begin
  if Assigned(FOnButton2) then
  begin
    FOnButton2(Self);
  end;
end;

procedure TCustomEditEnh.DoButton3;
begin
  if Assigned(FOnButton3) then
  begin
    FOnButton3(Self);
  end;
end;


initialization;
  EditEnh_BeepOverride := true;         // Always use these settings
  EditEnh_BeepOnBadChar := true;
  // MB_OK = default, use $FFFFFFFF for computer beep
  // See Windows API MessageBeep for other options
  EditEnh_BeepOnBadCharSound := MB_OK;
  EditEnh_BeepOnValidateErr := true;
  EditEnh_BeepOnValidateErrSound := MB_ICONASTERISK;
  EditEnh_RepeatTimer := TTimer.Create(nil);
  EditEnh_RepeatTimer.Enabled := false;
  EditEnh_RepeatLock := 0;
  EditEnh_RepeatInit := 500;
  EditEnh_RepeatPause := 100;

finalization;
  EditEnh_RepeatTimer.Free;
  EditEnh_RepeatTimer := nil;
  EditEnh_RepeatLock := 0;

end.
