{ Telesis_MaskEnh.pas                             SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Defines the TMaskEnh class and supporting datatypes.
<br>
@desc Used to be used by IBO directly, but we have since managed to setup
interfaces support so that this unit can be compiled with or without IBO.<br>
<br>
If you intend to try and compile this unit on a system without IBObjects
installed you need to edit the Telesis.inc unit and undefine (or comment out)
the __COMPILE_WITH_IBO__ declaration.<br>
<br>
If you are compiling this unit with IBObjects installed, IBObjects must be
in the search path.
<br>
<br>
<strong>History</strong><br>
<br>
22-Jan-2007 - Use Delphi6 MatchesMask (instead of IB_Utils.IsMatch) if
available.<br>
<br>
22-Sep-2002 - Improved handling of range deletions.<br>
<br>
09-Jan-2002 - Fixed problem when a locale character was used after
the literal indicator in numeric mode.<br>
<br>
28-Dec-2001 - Added support for string localisation.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_MaskEnh;


{$INCLUDE Telesis.inc}


interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  {$IFDEF __COMPILE_WITH_IBO__}
  IB_MaskInterface,
  {$ENDIF}
  Mask;



const
// Character Attributes
// ---------------------
// Used in the mask control definition to define various aspects of possible
// mask expression characters.
//
//   0  1  2  3  4  5  6  7
//   1  2  4  8  16 32 64 128
//
// Literals = 1
//     IntlLiteral = 4
//     Prefix        16
//     Postfix       32
//
// Mask    = 2
//     Mandatory   = 4
//     UpperCase   = 16
//     LowerCase   = 32
//     ProperCase  = 64
//
// Common...
//     RightJust   = 8


// Mask Character Attributes
  eecaMask = 2;
  eecaMandatory = 4;
  eecaUpperCase = 16;
  eecaLowerCase = 32;
  eecaProperCase = 64;

// Literal character attributes
  eecaLiteral = 1;
  eecaIntlLiteral = 4;
  eecaPrefix = 16;
  eecaPostfix = 32;
  eecaNumericSign = 64;

// Common character attributes
  eecaRightJust = 8;


// Mask Expression Constants
// --------------------------
// Constants which define the various characters and so forth used to
// express mask requirements.
//
// Some of these are copied (but renamed) from Mask.PAS supplied with the VCL
// in order to provide as much compatibility with the standard mask processing
// as possible.

  eeMskReverse = '!'; // removes leading blanks if true, else trailing blanks
  eeMskDefaultBlank = '_';
  eeMskRealBlank = ' ';
  eeMskFieldSeparator = ';';
  eeMskNoSave = '0';
  eeMskSave = '1';
  eeMskUpperCase = '>';                 // all chars that follow to upper case
  eeMskLowerCase = '<';                 // all chars that follow to lower case
                            // '<>' means remove casing directive
  eeMskLiteral = '\'; // char that immediately follows is a literal
  eeMskPrefixLiteral = '[';
  eeMskPostfixLiteral = ']';
  eeMskProperCase = '~';
  eeMskAlpha = 'L';                     // in US = A-Z,a-z
  eeMskAlphaOpt = 'l';
  eeMskAlphaNum = 'A';                  // in US = A-Z,a-z,0-9
  eeMskAlphaNumOpt = 'a';
  eeMskAscii = 'C';                     // any character
  eeMskAsciiOpt = 'c';
  eeMskNumeric = '0';                   // 0-9, no plus or minus
  eeMskNumericOpt = '9';
  eeMskNumSymOpt = '#';                 // 0-9, plus and minus
  eeMskTimeSeparator = ':';
  eeMskDateSeparator = '/';
  eeMskDecimal = '.';
  eeMskThousands = ',';
  eeMskNumSign = '+';
  eeMskNumSignOpt = '-';
  eeMskEmptyChar = #0;                  //char(0);

  eeMskStrStdChars = '_;09!><\LlAaCc#:/';

  eeMskStrEnhChars = '_;09!><\LlAaCc#:/,.[]~';

  eeMskStrNumChars = '_;09!><\LlAaCc#:/,.[]~-+';

// SPECIAL NOTE:
// The eeMskReverse indicator ("!") is interpreted as indicating a switch
// between that right and left justified data-entry.  Essentially this is
// usually used for numeric or date/time data-entry.  (Note that the control
// does not have to be right-justified to support right-justified input, such
// input is simply done from the right-hand side of the mask or mask section.)



type


  TMaskEnh = class;

{: @summary The mask processor TMaskEnh error event.
<br>
@desc Define code here to override the default error processing performed by
the mask processor.  The ErrCode variable returns one of the following
constants...
<br>
<br>E_MASKENH_CHAR_INVALID  Insert/Replace character invalid for position
<br>E_MASKENH_TRUNCATED     Input string was too long for mask
<br>E_MASKENH_INVALID       Input does not fit mask requirements
<br>E_MASKENH_INVALIDTRUNC  Input string was truncated due to mask mismatch
<br>E_MASKENH_CHAR_TRUNCATED Insert/Replace attempted beyond end of mask
but only optional characters remain to be filled.
<br><br>
Note that E_MASKENH_CHAR_* errors relate to character-by-character validation,
while the other error codes relate to errors detected when validating an
input string as a whole.   The char-by-char errors may occur multiple times
while processing an input string, whereas the other errors only occur once,
when validating after the input string has been assigned.
}
  TMaskEnhErrorEvent = procedure(Sender: TMaskEnh; ErrCode: integer) of object;


{: @summary Type used to define the mask processing mode.
<br>
@desc Note that eemNumeric mode offers special processing of the '0' mask value
(it is accepted as a non-mandatory input position).  Also this mode parses any
directly assigned string value for an exponent (contains 'E' or 'e') and
attempts to convert the value back to string without an exponent.  See the
help information for ExponentFilter for details.
<br>
@enum eemNone
Standard edit control, no mask processing
<br>
@enum eemMask
Mask processing as per standard TMaskEdit control
<br>
@enum eemMaskList
Mask processing from list of "standard" masks
<br>
@enum eemMaskEnh
Enhanced Mask processing
<br>
@enum eemMaskEnhList
Mask processing from list of Enhanced Masks
<br>
@enum eemNumeric
Special Mask processing for numeric values
<br>
}
  TMaskEnhMode = (
    eemNone,
    eemMask,
    eemMaskList,
    eemMaskEnh,
    eemMaskEnhList,
    eemNumeric
    );


{: @summary A string defining an enhanced (or standard) edit mask.
<br>
@desc Note that this type is simply a "string" type, renamed to make it
easier to document and search for within the documentation.  Essentially this
is what you need to know when defining edit masks.
<br><br>
Sample masks are given at the bottom of these notes, along with some
specific notes relating to the IBObjects TIB_EditEnh deriviative.
<br><br>
<b>IMPORTANT:</b> No provision has been made for multi-byte character sets
in this enhanced mask processing, not even when processing standard
edit masks.
<br><br>
Parts of this are copied/summarised from the VCL online help to put all the
details in one place.  Differences have are specified in this description.
<br><br>
A mask consists of five fields with semicolons separating the fields (as
opposed to the three of the standard VCL mask). The first part of the mask is
the mask itself. The second part is the character that determines whether the
literal characters of a mask are saved as part of the data. The third part of
the mask is the character used to represent unentered characters in the mask
while displayed in a control.  The forth part of the mask represents the
character to be used to represent unentered characters in the returned string
(if not defined defaults to a space character).  The fifth part, if provided,
permits the definition of user defined mask characters.
<br><br>
NOTE:  In eemNumeric mode, if the forth part of the mask definition is the
zero character ("0") then mandatory entry mask positions (masked with a "0")
are not enforced, since the returned value will contain a zero in that
position anyway.
<br><br><br>

These are the special characters used in the first field of the mask
<br><br>
<strong> ! </strong>  Changed from the TMaskEdit  An edit mask starts in
left-justified data-entry mode (as you enter each character the cursor moves
to the right). This character can be used to switch between left-justified and
right-justified data-entry mode.  In right-justified data entry the cursor
stays where it is, and the characters move to the left - when the
right-justified section is full the cursor automatically moves to the next
section of the mask.  There MUST be a literal character between a change from
one justification to another and this literal will ALWAYS be returned with the
resulting text regardless of the "save-literals" setting.  (If the entire mask
is right or left justified this does not matter.)  If this literal character is
input then the cursor will automatically jump to the next section avoiding the
need to fill the section.  (See the sample date entry mask.)
<br><br>
<strong> > </strong>  All characters after a > character are converted to
uppercase.
<br><br>
<strong> < </strong>  All characters after a < character are converted to
lowercase.
<br><br>
<strong> <> </strong>  Disables case conversions.
<br><br>
<strong> \ </strong>  The character that follows a \ character is a literal
character.
<br><br>
<strong> L </strong>  The L character requires an alphabetic character only in
this position. For the US, this is A-Z, a-z.
<br><br>
<strong> l </strong>  The l character permits only an alphabetic character in
this position, but doesn't require it.
<br><br>
<strong> A </strong>  The A character requires an alphanumeric character only
in this position. For the US, this is A-Z, a-z, 0-9.
<br><br>
<strong> a </strong>  The a character permits an alphanumeric character in this
position, but doesn't require it.
<br><br>
<strong> C </strong>  The C character requires an arbitrary character in this
position.
<br><br>
<strong> c </strong>  The c character permits an arbitrary character in this
position, but doesn't require it.
<br><br>
<strong> 0 </strong>  The 0 character requires a numeric character only in this
position.
<br><br>
<strong> 9 </strong>  The 9 character permits a numeric character in this
position, but doesn't require it.
<br><br>
<strong> # </strong>  The # character permits a numeric character or a plus or
minus sign in this position, but doesn't require it.
<br><br>
<strong> : </strong>  The : character represents the regional time separator
value.
<br><br>
<strong> / </strong>  The / character represents the regional date separator
value.
<br><br>
<strong> ; </strong>  The ; character is used to separate the five fields of
the mask.
<br><br>
<strong> _ </strong>  The _ character automatically inserts spaces into the
text.
<br><br>
<strong> ~ </strong>  New and ONLY in eemMaskEnh and higher modes.  All
characters after a ~ character are converted to propercase.  (This only
applies to keyboard input and NOT to assigned strings.)
<br><br>
<strong> ] </strong>  New and ONLY in eemMaskEnh and higher modes.  I call
this the "Prefix Literal".  The literal character following this is ONLY
displayed if there are masked characters (user input characters) preceding
this value.
<br><br>
<strong> [ </strong>  New and ONLY in eemMaskEnh and higher modes.  I call
this the "Postfix Literal".  The literal character following this is ONLY
displayed if there are masked characters (user input characters) following
this value.
<br><br>
<strong> , </strong>  New and ONLY in eemMaskEnh and higher modes.  The ,
character represents the regional thousands separator value.
<br><br>
<strong> . </strong>  New and ONLY in eemMaskEnh and higher modes.  The .
character represents the regional decimal separator value.
<br><br>
<strong> - </strong>  New and ONLY in eemNumeric mode.  The - character
represents the position of the "negative" sign if it is required.  No value
is displayed if the represented value is positive.  Note that input the the
sign character can occur at any point in the data-entry and the control will
automatically assign the sign value to its required position.  (It should
only occur once in the mask.)
<br><br>
<strong> + </strong>  New and ONLY in eemNumeric mode.  The + character
represents the position of the sign character which is always displayed,
either as a negative or positive according to the represented value.    Note
that input the the sign character can occur at any point in the data-entry
and the control will automatically assign the sign value to its required
position. (It should only occur once in the mask.)
<br><br><br>
A new feature of the enhanced mask processing permits the definition of
additional user defined mask characters as a fifth section/field in the
editmask.  This allows you specify that the character "X" (for example) is a
special mask character representing some specific set of characters.
<br><br>
eg. '<strong>X[0-9A-F]</strong>' says that an 'X' in the editmask is a
mandatory position in the mask that can be filled with a character out of
0..9 or A..F.  If the mask contains the letter 'x' then filling the position
is optional.
<br><br>
If required you can define multiple user defined mask characters by repeating
the defining sequence.  eg. '<strong>X[0-9A-F]Y[X-Z]Z[abc]</strong>' defines
three new mask characters 'X', 'Y' and 'Z' ('x', 'y' and 'z' for
non-mandatory positions).
<br><br>
User defined mask characters can be any letter (to allow upper and lower case
as indication of mandatory or optional positions) except for the letters that
are already used in standard edit masks (A, C, L).
<br><br>
Because this fifth section of the mask is part of the main editmask string, it
can be used with the mask list modes, but the user defined mask characters
must be defined with every mask in the list - or at least every mask to which
the user defined characters need to be applied.
<br><br>
Currently the set of characters defined inside the square brackets cannot
include the ']' (closing square bracket) character as the processing does
not provide any escape or control character sequence to support special
characters.  This may be improved in the future.
<br><br><br>
Examples using eemMask mode:<br>
<code>
    Telephone:         !\(999\)000-0000;1;_
    Social Security:   000\-00\-0000;1;_
    Long Zip Code:     00000\-9999;1;_
</code><br><br><br>

Examples using eemMaskEnh mode:<br>
<code>
    Formatted text:    ["ccc[-ccc[-ccc]";0;_
    Password input:    cccccccccccc;0;_
</code><br><br><br>

Examples using eemNumeric mode:<br>
<code>
    Currency:          !\$-999],999],990!.00;0;0;0
    Floating Point:    !+999],999],999],990!.099[,999[,999;0;0;0
    Date-Time Input:   !90!/!90!/!0000!\ !90!:!00;0;_;0
</code><br><br><br>

Example using eemMaskList mode (list of standard edit masks):<br>
<br>
Setup a string list like this...<br>
<code>
        (C0) 0000 0000
        00*=(C000) cccccccccccccccccccccccc
        01*=(C00) 000 000
        04*=(C000) 000 000
        13*=C0 0000
        18*=(C000) 000 000
        19*=(C000) cccccccccccccccccccccccc
        #*=Cccccccccccccccccccccccccccccc
</code><br><br>
then assign this to the Mask.MaskList property of the control at run-time.
This will setup mask processing which dynamically changes according to the
input.  The first value in the list is the default (the mask used when no
other match is found), subsequent lines should appear as test=mask,
where the test value is checked against the input string using the IsMatch
function from IB_Parse (an IBObjects unit).  This comparison supports the use
of wildcard characters in the comparison - hence the use of "*" at the end
of test values in the example shown above.
<br>
<br>Obviously all masks in the list must have compatible test values in
order to function correctly.  Checking for a match is done in the order that
the list is defined and will stop at the first match found.  If no match is
found then the default mask is used.
<br><br><br>

Example using eemMaskEnhList mode (list of enhanced edit masks):<br>
<br>
Setup a string list like this...<br>
<code>
        00000[-99999;1;_
        ??????*=00000-00000;1;_
</code><br><br>
then assign this to the Mask.MaskList property of the control at run-time.
This is like the previous example except that the mask list supports enhanced
mask options.
<br><br><br>
Example using User Defined editmask (standard mode):<br>
<code>
    Hex Value Entry: >XXXX\-XXXX;0;_;;X[0-9A-F]
</code><br><br><br>
NOTE:  TMaskEnh has a function called "SetMaskEx" and TCustomEditEnh and its
derivatives have a function called "SetEditMaskEx" which take a single string
parameter.  These functions can be used to set both the mask and the mode in
one hit from a single string input parameter.  Different modes are identified
using prefix values like "::SM::".
<br><br>
Examples:
<br><br>
Disable mask processing (eemNone mode)<br>
<code>
        EditEnh.SetEditMaskEx( '::NO::' );
</code>
<br><br>
Install a standard mask (eemMask mode)<br>
<code>
        EditEnh.SetEditMaskEx( '::SM::000\-00\-0000;1;_' );
</code><br><br>
Install an enhanced mask (eemMaskEnh mode)<br>
<code>
        EditEnh.SetEditMaskEx( '::EM::["ccc[-ccc[-ccc]";0;_' );
</code>
<br><br>
Install an standard mask list (eemMaskList mode)<br>
<code>
        EditEnh.SetEditMaskEx( '::SL::"CC\-CC;1;_","AB*=CC\$CC;1;_"' );
</code>
<br>
(ie. The supplied string is read into a string list as "CommaText", see
TStrings for details.)
<br><br>
Install an numeric mask (eemNumeric mode)<br>
<code>
        EditEnh.SetEditMaskEx( '::NM::!\$-999],999],990!.00;0;0;0' );
</code>
<br><br><br>
TIB_EditEnh and TIB_CustomEditEnh - The IBObjects derivatives.
<br><br>
The only way of setting an edit mask for an IBO control is to set the mask
value in the EditMask list of IB_Connection or IB_Dataset. The EditMask list
properties should be entered in the form...<br>
<br>
<code>
    FieldName1=EditMask1
    FieldName2=EditMask2
    etc
</code>
<br><br>
The enhanced masks should be entered with the appropriate mode prefix string,
for example...<br>
<br>
<code>
    MyNumericField=::NM::!\$-999],999],990!.00;0;0;0
</code>
<br><br>
If no mode prefix is found then the mask will be interpretted as a standard
mask (making the implementation compatible with existing installations).
<br><br>
Notice that these strings are the same form as passed to the SetEditMaskEx
function described above - because that is the function used to assign these
masks to the control by IBO.
}
  TMaskEnhString = string;


{: @summary TMaskEnh mask processor class.
<br>
@desc TMaskEnh class can be considered the "mask processor", into which you set
the required mask attributes and extract the resulting values.  This class
is used by TCustomEditEnh and derived controls.<br>
<br>
The mask processor is able to manipulate "standard" masks (like those
defined for the VCL TMaskEdit control) as well mask strings which use extended
character definitions for "enhanced" mask processing.<br>
<br>
NOTE: The processing of VCL standard masks is NOT 100% compatible, particularly
when the "!" control character is used.  You should test such mask prior
to implementation.<br>
<br>
Note: The StringList property MaskList is defined so that all instances
of TMaskEnh which do not actually use a specifically defined MaskList, all
use the same instance of an empty string list.  So there should be no
overhead other than the pointer/reference to this common list.}
  TMaskEnh = class(
    {$IFDEF __COMPILE_WITH_IBO__}
    TInterfacedObject, IIB_MaskProcessor
    {$ELSE}
    TObject
    {$ENDIF}
    )
  private
    FCtrlStr: string;
    FMaskStr: string;
    FTextStr: string;
    FLength: integer;
    FEditBlank: char;
    FTextBlank: char;
    FSave: boolean;
    FMode: TMaskEnhMode;
    FPos: integer;
    FMaskList: TStrings;
    FUserMask: TStrings;
    FMaskListIdx: integer;
    FChoosingMask: boolean;
    FAssigningText: boolean;            // flag to indicate SetText is active
    FValidateOnSetText: boolean;
    FMaxIntDigits: integer;
    FOnError: TMaskEnhErrorEvent;

    function GetMask: string;
    procedure SetMask(AValue: string);
    function GetMaskEx: string;
    procedure SetMaskEx(MaskStr: string);
    function GetUserMask: string;
    procedure SetFromUserMask(ValidMaskChars, MaskSrc: string);
    function GetMaskList: TStrings;
    procedure SetMaskList(AValue: TStrings);
    function GetMode: TMaskEnhMode;
    procedure SetMode(AValue: TMaskEnhMode);
    function GetIsEmpty: boolean;

    function ValidateCharInPos(var AValue: char; MskPos: integer; ChkOpt:
      boolean): boolean;
    function GetEnhMaskPos(Limit: integer; DispPos: boolean): integer;
    procedure ChooseMask;

    function GetText: string;
    procedure SetText(AValue: string);
    procedure DoValidateError(ErrCode: integer);

  {$IFDEF __COMPILE_WITH_IBO__}
  public
    // Simply redirect these interface elements
    function IIB_MaskProcessor.GetMask = GetMaskEx;
    procedure IIB_MaskProcessor.SetMask = SetMaskEx;
    // These ones we will define specifically
    function IsMasked: boolean;
    function EmptyText: string;
    function FormatText(AVal: string): string;
  {$ENDIF}

  public
    constructor Create;
    destructor Destroy; override;

    // This stuff is public specifically for CustomEditEnh and is not
    // recommended for general consumption...
{: @desc Used to save the internal representation of the input string for
restoration later if required.  The saved value should not be used directly,
but only as a parameter for RestoreMaskTextFrom.}
    procedure SaveMaskTextTo(var ToStr: string);
{: @desc Used to restore the internal representation of the input string.  The
value given as FromStr should have come from a call to SaveMaskTextTo.
The ListChk variable indicates whether the currently selected mask (from
a defined list of masks) should be re-evaluated after the restore.}
    procedure RestoreMaskTextFrom(var FromStr: string; ListChk: boolean);
{: @desc Prepares the mask processor for new input.  The WithLiterals parameter
is used to indicate whether input is expected to contain mask literals. }
    procedure InitMaskInput(WithLiterals: boolean);
{: @desc Attempts to calculate the correct mask position for the next input
character.  The WithLiterals parameter is used to indicate whether input is
expected to contain mask literals. }
    procedure AdjustMaskPos(WithLiterals: boolean);
{: @desc Used in eemNumeric mode to directly set the sign of the value being
represented.  Returns true if the sign was changed.  The input value should
be a valid sign character for the current regional settings.}
    function SetNumericSign(var AValue: char): boolean;
{: @desc Insert a character at the current position.  WithLiterals indicates
whether input is expected to contain mask literals.  Returns true if the
character was successfully inserted.  If it was not successful then
DoValidateError will have been called. }
    function InsertMaskChar(var AValue: char; WithLiterals: boolean): boolean;
{: @desc Replace the character after the current position.  WithLiterals
indicates whether input is expected to contain mask literals.  Returns true
if the character was successfully replaced.  If it was not successful then
DoValidateError will have been called. }
    function ReplaceMaskChar(var AValue: char; WithLiterals, IncPos: boolean):
      boolean;
{: @desc Called with appropriate Key board style parameters (VK_DOWN etc) this
function will cause the process to move its current position appropriately. }
    procedure MovePosition(Key: Word; Shift: TShiftState);
{: @desc Called to delete the previous input character.  If the delete
succeeded AND WithUpd was given as true then function returns true. }
    function DeleteBackSpace(WithUpd: boolean): boolean;
{: @desc Called to delete the next input character.  If the delete succeeded
AND WithUpd was given as true then function returns true.}
    function DeleteForward(WithUpd: boolean): boolean;
{: @desc Called to delete Len characters from the current position.  If the
delete succeeded AND WithUpd was given as true then function returns true.}
    function DeleteRange(WithUpd: boolean; Len: integer): boolean;
{: @desc Attempts to move the current mask position directly to the position
specified.  If standard masks are being used and Reselect was set to true
and the move was successful then the function will return true, which
will indicate to the control that it should reselect at the new position. }
    function MoveToPosition(NewPos: integer; Reselect: boolean): boolean;
{: @desc Calculates what character position in the actually displayed text
corresponds to the current mask position. }
    function GetDisplayPosition(var StartPos: integer; var StopPos: integer):
      boolean;
{: @desc Calculates if the given position should be left-justified. }
    function GetIsLeftJust(TestPos: integer): boolean;
{: @desc Calculates if the given position should be right-justified.}
    function GetIsRightJust(TestPos: integer): boolean;

{: @desc Clears the masks storage of input text.}
    procedure Clear;
{: @desc Called to validate the input against the mask.}
    function ValidateEdit: boolean;
{: @summary Returns a copy of the input text passed through the mask processing.
<br>
@desc Set IncLiterals to true to have the returned text include the literals
from the mask, false to exclude them.
<br><br>
Set IncNulls to true to include charater representations of input characters
which have not yet been entered.  If UseMaskBlank is true this will be the
character which is displayed in a control as the place-holder for un-entered
characters, otherwise un-entered characters will be represented by the character
given as the forth part of the mask (defaults to a space character).}
    function GetMaskedText(IncLiterals, IncNulls, UseMaskBlank: boolean):
      string;
{: @summary Directly set the entire input string.
<br>
@desc Use WithLiterals to indicate if the input string contains mask
literal characters.}
    procedure SetMaskedText(AValue: string; WithLiterals: boolean);

{: @desc Property set to true when text is being assigned directly - rather
than the char-by-char input that occurs during user-entry.}
    property AssigningText: boolean read FAssigningText;
{: @summary Call this to see if any input has been made to the mask.}
    property IsEmpty: boolean read GetIsEmpty;
{: @summary The number of character positions represented by the mask.}
    property MaskLen: integer read FLength;
{: @desc ReadOnly property representing the current processor position in the
mask.}
    property Position: integer read FPos;
{: @desc Determines whether the processor should validate the input when set
directly use the Text property or SetMaskedText function.  Defaults to false
for compatibility with VCL style controls and to minimise error reporting
in database situations. }
    property ValidateOnSetText: boolean read FValidateOnSetText write
      FValidateOnSetText default false;
{: @summary Get/Set the Mask definiton string.
<br>
@desc Plesae refer to TCustomEditEnh for details. }
    property Mask: TMaskEnhString read GetMask write SetMask;
{: @summary Dynamic mask list.
<br>
@desc Get/Set a list of masks which will be dynamically selected according to
how they match the current input value. }
    property MaskList: TStrings read GetMaskList write SetMaskList;
{: @summary Get/Set the current mode of the mask processor. }
    property Mode: TMaskEnhMode read GetMode write SetMode default eemNone;
{: @summary Get/Set the current text value. }
    property Text: string read GetText write SetText;

{: @summary Provide code here to override default error processing. }
    property OnError: TMaskEnhErrorEvent read FOnError write FOnError;

{: @summary Refer to TCustomEditEnh.SetEditMaskEx for details. }
    property MaskEx: string read GetMaskEx write SetMaskEx;

  end;                                  // TMaskEdit


{: @summary Convert a TMaskEnhMode type value into its string representation.
<br>
@desc Such as '::SM::' for a standard mask. }
function MaskEnhModeToString(MMode: TMaskEnhMode): string;

{: @summary Interprets a string value to see if it represents a TMaskEnhMode
value.
<br>
@desc Returns eemNone if the give string does match a known TMaskEnhMode
value. }
function StringToMaskEnhMode(PrefStr: string; var MMode: TMaskEnhMode): boolean;

{: @summary Utility function to split the given string into its appropriate
sections.}
procedure SplitMaskEnhString(Source: string;
  var MaskStr: string; var SaveLit: boolean;
  var EditBlank: char; var TextBlank: char; var UserMask: string);
{: @summary Utility function to join mask attributes into a full an EditMask
string (without any ::xx:: prefix).}
function MakeMaskEnhString(MaskStr: string; SaveLit: boolean;
  EditBlank, TextBlank: char; UserMask: string): string;





implementation

uses
    Telesis_Constants,
    Telesis_Utils,
    Masks;

var
  MaskEnh_EmptyList: TStringList;



function MaskEnhModeToString(MMode: TMaskEnhMode): string;
begin
  case MMode of
    eemNone:
      Result := '::NO::';
    eemMask:
      Result := '::SM::';
    eemMaskList:
      Result := '::SL::';
    eemMaskEnh:
      Result := '::EM::';
    eemMaskEnhList:
      Result := '::EL::';
    eemNumeric:
      Result := '::NM::';
  else
    Result := '';
  end;
end;


function StringToMaskEnhMode(PrefStr: string; var MMode: TMaskEnhMode): boolean;
begin
  Result := true;
  if PrefStr = '::NO::' then
  begin
    MMode := eemNone;
  end
  else
    if PrefStr = '::SM::' then
    begin
      MMode := eemMask;
    end
    else
      if PrefStr = '::SL::' then
      begin
        MMode := eemMaskList
      end
      else
        if PrefStr = '::EM::' then
        begin
          MMode := eemMaskEnh;
        end
        else
          if PrefStr = '::EL::' then
          begin
            MMode := eemMaskEnhList;
          end
          else
            if PrefStr = '::NM::' then
            begin
              MMode := eemNumeric;
            end
            else
            begin
              Result := false;
            end;
end;


procedure SplitMaskEnhString(Source: string;
  var MaskStr: string; var SaveLit: boolean;
  var EditBlank: char; var TextBlank: char; var UserMask: string);
var
  i, Lim: integer;
  tmpStr: string;
begin
  MaskStr := '';
  SaveLit := true;                      // default
  EditBlank := eeMskDefaultBlank;
  TextBlank := eeMskRealBlank;
  UserMask := '';

  i := 1;
  Lim := Length(Source);

  // Section 1 - the mask itself
  // parsing needs to check for literal char indicators
  while i <= Lim do
  begin
    if Source[i] = eeMskFieldSeparator then
    begin
      Inc(i);                           // skip past the separator
      break;
    end;
    if (Source[i] = eeMskLiteral) and (i < Lim) then
    begin
      MaskStr := MaskStr + Source[i];
      Inc(i);
    end;
    MaskStr := MaskStr + Source[i];
    Inc(i);
  end;

  // Section 2 - save literal indicator
  // parsing has no special parsing
  tmpStr := '';
  while i <= Lim do
  begin
    if Source[i] = eeMskFieldSeparator then
    begin
      Inc(i);                           // skip past the separator
      break;
    end;
    tmpStr := tmpStr + Source[i];
    Inc(i);
  end;
  if Length(tmpStr) > 0 then
  begin
    if tmpStr[1] = eeMskNoSave then
      SaveLit := false
    else
      SaveLit := true;
  end;

  // Section 3 - edit/display blank character
  // parsing has no special parsing
  tmpStr := '';
  while i <= Lim do
  begin
    if Source[i] = eeMskFieldSeparator then
    begin
      Inc(i);                           // skip past the separator
      break;
    end;
    tmpStr := tmpStr + Source[i];
    Inc(i);
  end;
  if Length(tmpStr) > 0 then
    EditBlank := tmpStr[1];

  // Section 4 - returned text blank character
  // parsing has no special parsing
  tmpStr := '';
  while i <= Lim do
  begin
    if Source[i] = eeMskFieldSeparator then
    begin
      Inc(i);                           // skip past the separator
      break;
    end;
    tmpStr := tmpStr + Source[i];
    Inc(i);
  end;
  if Length(tmpStr) > 0 then
    TextBlank := tmpStr[1];

  // Section 5 - user defined mask definitions
  // just use the rest of the string.  Note that it only copies
  // if i < lim - so there must be more than one character given.
  if i < Lim then
  begin
    UserMask := Copy(Source, i, MAXINT);
  end;
end;


function MakeMaskEnhString(MaskStr: string; SaveLit: boolean;
  EditBlank, TextBlank: char; UserMask: string): string;
begin
  // First the simple standard format (three sections)
  Result := MaskStr + eeMskFieldSeparator;
  if SaveLit then
    Result := Result + eeMskSave + eeMskFieldSeparator + EditBlank
  else
    Result := Result + eeMskNoSave + eeMskFieldSeparator + EditBlank;
  // Now the additional/optional sections
  if (TextBlank <> eeMskRealBlank) or (Length(UserMask) > 0) then
  begin
    Result := Result + eeMskFieldSeparator;
    if TextBlank <> eeMskRealBlank then
      Result := Result + TextBlank;
    if Length(UserMask) > 0 then
      Result := Result + eeMskFieldSeparator + UserMask;
  end;
end;





constructor TMaskEnh.Create;
begin
  inherited Create;
  FCtrlStr := '';
  FMaskStr := '';
  FTextStr := '';
  FLength := 0;
  FEditBlank := eeMskDefaultBlank;
  FTextBlank := eeMskRealBlank;
  FSave := true;
  FMode := eemNone;
  FPos := 0;
  FMaskList := nil;
  FUserMask := nil;
  FMaskListIdx := -1;
  FChoosingMask := false;
  FAssigningText := false;
  FValidateOnSetText := false;
  FMaxIntDigits := 0;
end;


destructor TMaskEnh.Destroy;
begin
  if FMaskList <> nil then
  begin
    FMaskList.Free;
    FMaskList := nil;
  end;
  if FUserMask <> nil then
  begin
    FUserMask.Free;
    FUserMask := nil;
  end;
  inherited Destroy;
end;


function TMaskEnh.GetMask: string;
begin
  Result := GetUserMask;
end;


procedure TMaskEnh.SetMask(AValue: string);
var
  SaveTxt: string;
begin
  if AValue <> GetMask then
  begin
    SaveMaskTextTo(SaveTxt);
    // If no mask mode just store the text for the moment
    if FMode = eemNone then
    begin
      FMaskStr := AValue;
    end
    // If restricted to standard masks...
    else
      if FMode <= eemMaskList then
      begin
        SetFromUserMask(eeMskStrStdChars, AValue);
      end
      // If numeric mode use the extended numeric set
      else
        if FMode = eemNumeric then
        begin
          SetFromUserMask(eeMskStrNumChars, AValue);
        end
        // Otherwise use full list of mask chars
        else
        begin
          SetFromUserMask(eeMskStrEnhChars, AValue);
        end;
    RestoreMaskTextFrom(SaveTxt, false);
  end;
end;

// Some bit-setting utility functions...

function SetLowerCase(Val: char): char;
begin
  Result := char(((integer(Val) or eecaLowerCase) and
    ($FFFFFFFF xor (eecaUpperCase or eecaProperCase))));
end;

function SetUpperCase(Val: char): char;
begin
  Result := char(((integer(Val) or eecaUpperCase) and
    ($FFFFFFFF xor (eecaLowerCase or eecaProperCase))));
end;

function SetProperCase(Val: char): char;
begin
  Result := char(((integer(Val) or eecaProperCase) and
    ($FFFFFFFF xor (eecaUpperCase or eecaLowerCase))));
end;

function SetNormalCase(Val: char): char;
begin
  Result := char((integer(Val) and
    ($FFFFFFFF xor (eecaProperCase or eecaUpperCase or eecaLowerCase))));
end;

function SetRightJust(Val: char): char;
begin
  Result := char(integer(Val) or eecaRightJust);
end;

function SetLeftJust(Val: char): char;
begin
  Result := char(integer(Val) and ($FFFFFFFF xor eecaRightJust));
end;


// Some bit-testing utility functions...

function IsLiteral(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaLiteral) <> 0);
end;

function IsIntlLiteral(Val: char): boolean;
begin
  Result := ((integer(Val) and (eecaIntlLiteral or eecaLiteral)) =
    (eecaIntlLiteral or eecaLiteral));
end;

function IsPrefix(Val: char): boolean;
begin
  Result := ((integer(Val) and (eecaPrefix or eecaLiteral)) =
    (eecaPrefix or eecaLiteral));
end;

function IsPostfix(Val: char): boolean;
begin
  Result := ((integer(Val) and (eecaPostfix or eecaLiteral)) =
    (eecaPostfix or eecaLiteral));
end;

function IsNumericSign(Val: char): boolean;
begin
  Result := ((integer(Val) and (eecaNumericSign or eecaLiteral)) =
    (eecaNumericSign or eecaLiteral));
end;

function IsMask(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaMask) <> 0);
end;

function IsLowerCase(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaLowerCase) <> 0);
end;

function IsUpperCase(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaUpperCase) <> 0);
end;

function IsProperCase(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaProperCase) <> 0);
end;

function IsNormalCase(Val: char): boolean;
begin
  Result := ((integer(Val) and (eecaLowerCase or eecaUpperCase or
    eecaProperCase)) = 0);
end;

function IsMandatory(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaMandatory) <> 0);
end;

function IsOptional(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaMandatory) = 0);
end;

function IsRightJust(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaRightJust) <> 0);
end;

function IsLeftJust(Val: char): boolean;
begin
  Result := ((integer(Val) and eecaRightJust) = 0);
end;

// We make an exception here for numeric mode - we do not
// automatically pass over the decimal point literal - it must
// be provided explicitly to prevent inappropriate assignments.

function IsNumericDecimal(Mode: TMaskEnhMode; Ctrl, Msk: char): boolean;
begin
  Result := ((Mode = eemNumeric) and IsIntlLiteral(Ctrl) and (Msk =
    eeMskDecimal));
end;



function TMaskEnh.GetUserMask: string;
var
  i: integer;
  MskCtrl, LitCtrl: char;
  tmpStr: string;
begin
  Result := '';
  MskCtrl := char(eecaMask);
  LitCtrl := char(eecaLiteral);
  for i := 1 to FLength do
  begin
    // If current mask control does not match justification of current position
    // then we need to insert a "reverse" directive...
    if (IsRightJust(FCtrlStr[i]) and IsLeftJust(LitCtrl)) or
      (IsLeftJust(FCtrlStr[i]) and IsRightJust(LitCtrl)) then
    begin
       // If current position is RightJustified then turn it off to match mask
      if IsRightJust(LitCtrl) then
      begin
        LitCtrl := SetLeftJust(LitCtrl);
        MskCtrl := SetLeftJust(MskCtrl);
      end
      // otherwise turn it on to match mask
      else
      begin
        LitCtrl := SetRightJust(LitCtrl);
        MskCtrl := SetRightJust(MskCtrl);
      end;
      Result := Result + eeMskReverse;
    end;
    // If current mask is for a literal...
    if IsLiteral(FCtrlStr[i]) then
    begin
      // Special case of prefix literal
      if IsPrefix(FCtrlStr[i]) then
        Result := Result + eeMskPrefixLiteral
      // Special case of postfix literal
      else
        if IsPostfix(FCtrlStr[i]) then
          Result := Result + eeMskPostfixLiteral
        else
          if (not IsNumericSign(FCtrlStr[i])) and
          (not IsIntlLiteral(FCtrlStr[i])) then
            Result := Result + eeMskLiteral;
      // Special case of numeric sign
      if IsNumericSign(FCtrlStr[i]) then
      begin
        if FMaskStr[i] <> eeMskNumSign then
          Result := Result + eeMskNumSignOpt
        else
          Result := Result + FMaskStr[i];
      end
      else
      begin
        // Otherwise just follow with actual mask character
        Result := Result + FMaskStr[i];
      end;
    end
    else
    begin
      if IsMask(FCtrlStr[i]) then
      begin
        if IsUppercase(FCtrlStr[i]) then
        begin
          if not IsUpperCase(MskCtrl) then
          begin
            Result := Result + eeMskUpperCase;
            MskCtrl := SetUpperCase(MskCtrl);
          end;
        end
        else
          if IsLowerCase(FCtrlStr[i]) then
          begin
            if not IsLowerCase(MskCtrl) then
            begin
              Result := Result + eeMskLowerCase;
              MskCtrl := SetLowerCase(MskCtrl);
            end;
          end
          else
            if IsProperCase(FCtrlStr[i]) then
            begin
              if not IsProperCase(MskCtrl) then
              begin
                Result := Result + eeMskProperCase;
                MskCtrl := SetProperCase(MskCtrl);
              end;
            end
            else
              if not IsNormalCase(MskCtrl) then
              begin
                Result := Result + eeMskLowerCase;
                Result := Result + eeMskUpperCase;
                MskCtrl := SetNormalCase(MskCtrl);
              end;
        if IsOptional(FCtrlStr[i]) then
        begin
          if FMaskStr[i] = eeMskNumeric then
            Result := Result + eeMskNumericOpt
          else
            Result := Result + AnsiLowerCase(FMaskStr[i]);
        end
        else
          Result := Result + FMaskStr[i];
      end;
    end;
  end;
  if Length(Result) > 0 then
  begin
    tmpStr := '';
    if FUserMask <> nil then
    begin
      for i := 1 to Length(FUserMask.Strings[0]) do
        tmpStr := tmpStr + FUserMask.Strings[0][i] + FUserMask.Strings[i];
    end;
    Result := MakeMaskEnhString(Result, FSave, FEditBlank, FTextBlank, tmpStr);
  end;
end;


procedure TMaskEnh.SetFromUserMask(ValidMaskChars, MaskSrc: string);
var
  i, Lim, UsrPos: integer;
  MskCtrl, LitCtrl, UsrChar: char;
  HadDecimal: boolean;
  Sect1, Sect5, tmpStr: string;

  procedure AddMaskChar(Msk, Ctrl: char);
  begin
    FCtrlStr := FCtrlStr + Ctrl;
    FMaskStr := FMaskStr + Msk;
    if (Mode = eemNumeric) and (not HadDecimal) then
    begin
      if (Msk = eeMskNumeric) or (Msk = eeMskNumericOpt) or
        (Msk = eeMskNumSymOpt) then
        Inc(FMaxIntDigits);
    end;
    if IsLiteral(Ctrl) then
    begin
      if IsIntlLiteral(Ctrl) then
      begin
        case Msk of
          eeMskTimeSeparator: Msk := TimeSeparator;
          eeMskDateSeparator: Msk := DateSeparator;
          eeMskDecimal:
            begin
              Msk := DecimalSeparator;
              HadDecimal := true;
            end;
          eeMskThousands: Msk := ThousandSeparator;
        end;
      end;
      FTextStr := FTextStr + Msk;
    end
    else
      FTextStr := FTextStr + eeMskEmptyChar;
    Inc(FLength);
  end;

begin
  FCtrlStr := '';
  FMaskStr := '';
  FTextStr := ''; // assume input has been saved already (see SetMask)
  FMaxIntDigits := 0;
  HadDecimal := false;
  FLength := 0;
  MskCtrl := char(eecaMask);
  LitCtrl := char(eecaLiteral);

  // Split the given string into its various sections...
  SplitMaskEnhString(MaskSrc, Sect1, FSave, FEditBlank, FTextBlank, Sect5);

  // Now parse the user defined mask definitions (if any)
  // into a more efficient structure.  Expected to be in the
  // form 'X[a-zA-Z]Y[0-9]'.  That is, one character followed
  // by the set of acceptable characters in square brackets
  // (for form acceptable to the Delphi MatchesMask function).
  // This structure can be repeated as often as required.
  if Length(Sect5) > 0 then
  begin
    if FUserMask = nil then
      FUserMask := TStringList.Create
    else
      FUserMask.Clear;
    // Add the character definition string.
    // Note that future processing will rely upon at least one string existing
    // in the FUserMask list if it is not nil.
    FUserMask.Add('');
    i := 1;
    while i <= Length(Sect5) do
    begin
      // Add the matching character - uppercase=mandatory, lowercase=optional
      // so store only uppercase version for the moment.
      FUserMask.Strings[0] := FUserMask.Strings[0] + AnsiUpperCase(Sect5[i]);
      Inc(i);
      tmpStr := '';
      // Each set is interpretted as all characters until the closing ']'
      // Note that MatchesMask does not support any escape/control character
      // so it will not permit the ']' character to exist within the set of
      // acceptable characters.  Certainly not perfect but it will do
      // for the moment.
      while i <= Length(Sect5) do
      begin
        tmpStr := tmpStr + Sect5[i];
        if Sect5[i] = ']' then
        begin
          Inc(i);
          break;
        end;
        Inc(i);
      end;
      // Now add the character set definition.  Note that future processing
      // will rely upon a stringlist entry for every character added to
      // the first string in the set.
      if tmpStr = '' then
        tmpStr := '[]';
      FUserMask.Add(tmpStr);
    end;
  end
  else
    if FUserMask <> nil then
    begin
      FUserMask.Free;
      FUserMask := nil;
    end;


  // Finally assemble the mask control string
  i := 1;
  Lim := Length(Sect1);
  while i <= Lim do
  begin
    if Pos(Sect1[i], ValidMaskChars) > 0 then
    begin
      case Sect1[i] of
        eeMskLiteral:
          begin
            if i < Lim then Inc(i);
            AddMaskChar(Sect1[i], LitCtrl);
          end;
        eeMskPrefixLiteral:
          begin
            if i < Lim then Inc(i);
            case Sect1[i] of
              eeMskTimeSeparator, eeMskDateSeparator,
                eeMskDecimal, eeMskThousands:
                AddMaskChar(Sect1[i], char(eecaPrefix or eecaIntlLiteral or
                  integer(LitCtrl)));
              eeMskLiteral:
                begin
                  if i < Lim then Inc(i);
                  AddMaskChar(Sect1[i], char(eecaPrefix or integer(LitCtrl)));
                end;
            else
              AddMaskChar(Sect1[i], char(eecaPrefix or integer(LitCtrl)));
            end;
          end;
        eeMskPostfixLiteral:
          begin
            if i < Lim then Inc(i);
            case Sect1[i] of
              eeMskTimeSeparator, eeMskDateSeparator,
                eeMskDecimal, eeMskThousands:
                AddMaskChar(Sect1[i], char(eecaPostfix or eecaIntlLiteral or
                  integer(LitCtrl)));
              eeMskLiteral:
                begin
                  if i < Lim then Inc(i);
                  AddMaskChar(Sect1[i], char(eecaPostfix or integer(LitCtrl)));
                end;
            else
              AddMaskChar(Sect1[i], char(eecaPostfix or integer(LitCtrl)));
            end;
          end;
        eeMskDefaultBlank: AddMaskChar(eeMskRealBlank, LitCtrl);
        eeMskReverse:
          begin
            if IsRightJust(LitCtrl) then
            begin
              LitCtrl := SetLeftJust(LitCtrl);
              MskCtrl := SetLeftJust(MskCtrl);
            end
            else
            begin
              LitCtrl := SetRightJust(LitCtrl);
              MskCtrl := SetRightJust(MskCtrl);
            end;
          end;
        eeMskUpperCase:
          begin
            MskCtrl := SetUpperCase(MskCtrl);
            if i < Lim then
            begin
                // If both upper and lower specified together turn off all
              if Sect1[i + 1] = eeMskLowerCase then
              begin
                MskCtrl := SetNormalCase(MskCtrl);
                Inc(i);
              end;
            end;
          end;
        eeMskLowerCase:
          begin
            MskCtrl := SetLowerCase(MskCtrl);
            if i < Lim then
            begin
              // If both upper and lower specified together turn off all
              if Sect1[i + 1] = eeMskUpperCase then
              begin
                MskCtrl := SetNormalCase(MskCtrl);
                Inc(i);
              end;
            end;
          end;
        eeMskProperCase:
          begin
            MskCtrl := SetProperCase(MskCtrl);
          end;
        eeMskAlpha: AddMaskChar(eeMskAlpha, char(integer(MskCtrl) or
          eecaMandatory));
        eeMskAlphaOpt: AddMaskChar(eeMskAlpha, MskCtrl);
        eeMskAlphaNum: AddMaskChar(eeMskAlphaNum, char(integer(MskCtrl) or
          eecaMandatory));
        eeMskAlphaNumOpt: AddMaskChar(eeMskAlphaNum, MskCtrl);
        eeMskAscii: AddMaskChar(eeMskAscii, char(integer(MskCtrl) or
          eecaMandatory));
        eeMskAsciiOpt: AddMaskChar(eeMskAscii, MskCtrl);
        eeMskNumeric: AddMaskChar(eeMskNumeric, char(integer(MskCtrl) or
          eecaMandatory));
        eeMskNumericOpt: AddMaskChar(eeMskNumeric, MskCtrl);
        eeMskNumSymOpt: AddMaskChar(eeMskNumSymOpt, MskCtrl);
        eeMskTimeSeparator, eeMskDateSeparator, eeMskDecimal,
          eeMskThousands: AddMaskChar(Sect1[i], char(integer(LitCtrl) or
            eecaIntlLiteral));
        eeMskNumSign: AddMaskChar(Sect1[i], char(integer(LitCtrl) or
          eecaNumericSign));
        eeMskNumSignOpt: AddMaskChar(eeMskRealBlank, char(integer(LitCtrl) or
          eecaNumericSign));
      end;                              // case
    end
    else
    begin
      case Sect1[i] of
        eeMskTimeSeparator, eeMskDateSeparator,
          eeMskDecimal, eeMskThousands:
          AddMaskChar(Sect1[i], char(eecaIntlLiteral or integer(LitCtrl)));
      else
        // If we got to here then no standard mask characters fit so
        // we check for user defined characters.
        if FUserMask = nil then
          UsrPos := 0
        else
          UsrPos := Pos(AnsiUpperCase(Sect1[i]), FUserMask.Strings[0]);
        // If matching a user defined mask character...
        if UsrPos > 0 then
        begin
          UsrChar := FUserMask.Strings[0][UsrPos];
          if Sect1[i] = UsrChar then    // match mandatory (uppercase) version?
            AddMaskChar(UsrChar, char(integer(MskCtrl) or eecaMandatory))
          else
            AddMaskChar(UsrChar, MskCtrl); // optional
        end
        else
        begin
          AddMaskChar(Sect1[i], LitCtrl);
        end
      end;
    end;
    Inc(i);
  end;

end;


function TMaskEnh.GetMaskList: TStrings;
begin
  if FMaskList = nil then
    Result := MaskEnh_EmptyList
  else
    Result := FMaskList;
end;


procedure TMaskEnh.SetMaskList(AValue: TStrings);
begin
  // If setting list to a non-empty list...
  if (AValue <> nil) and (AValue.Count > 0) then
  begin
    if FMaskList = nil then FMaskList := TStringList.Create;
    FMaskList.Clear;
    FMaskList.Assign(AValue);
    // Set the default edit mask...
    FMaskListIdx := 0;
    if (FMaskList.Count > 0) and
      ((FMode = eemMaskList) or (FMode = eemMaskEnhList)) then
      Mask := FMaskList.Strings[0];
  end
  // if setting list to an empty list...
  else
  begin
    FMaskList.Free;
    FMaskList := nil;
    FMaskListIdx := -1;
    Mask := '';
  end;
end;


function TMaskEnh.GetMode: TMaskEnhMode;
begin
  Result := FMode;
end;


procedure TMaskEnh.SetMode(AValue: TMaskEnhMode);
var
  HoldMask: string;
begin
  if AValue <> FMode then
  begin
    if (FMode = eemNone) and (Length(FMaskStr) > 0) then
      HoldMask := FMaskStr
    else
      HoldMask := Mask;
    Mask := '';
    FMode := AValue;
    if (FMode = eemMaskList) or (FMode = eemMaskEnhList) then
    begin
      if (FMaskList <> nil) and (FMaskList.Count > 0) then
        Mask := FMaskList.Strings[0]
    end
    else
      Mask := HoldMask;                 // May need reinterpretation
  end;
end;


function TMaskEnh.GetText: string;
begin
  Result := GetMaskedText(FSave, true, false);
end;


procedure TMaskEnh.SetText(AValue: string);
begin
  SetMaskedText(AValue, FSave);
end;


function TMaskEnh.GetMaskedText(IncLiterals, IncNulls, UseMaskBlank: boolean):
  string;
var
  RJStr: string;
  i, Lim, j, k, l: integer;
  HavePrev, HaveNext, FillToLeft: boolean;
  DispCh, HoldCh: char;
begin
  Result := '';
  Lim := FLength;
  HavePrev := false;
  i := 1;
  while i <= Lim do
  begin
    j := i;
    FillToLeft := false;
    // If right justified then work to right of right-justified section
    while IsRightJust(FCtrlStr[i]) do
    begin
      FillToLeft := true;
      if not IsLiteral(FCtrlStr[i]) and
        (FTextStr[i] <> eeMskEmptyChar) then
        HavePrev := true;
      Inc(i);
      if i > Lim then break;
    end;
    // And if right justified we fill back to the left...
    if FillToLeft then
    begin
      RJStr := '';
      HaveNext := false;
      k := i - 1;
      while k >= j do
      begin
        DispCh := FTextStr[k];
        HoldCh := eeMskEmptyChar;
        // Test for literals (dont test IncLiterals just yet)
        if IsLiteral(FCtrlStr[k]) then
        begin
          // If Postfix literal we simply test if any non-literals have been
          // entered previously, if so we include the literal character...
          if IsPostfix(FCtrlStr[k]) then
          begin
            // see if any entries to the left
            l := k - 1;
            while IsRightJust(FCtrlStr[l]) do
            begin
              if not IsLiteral(FCtrlStr[l]) and
                (FTextStr[l] <> eeMskEmptyChar) then
              begin
                HavePrev := true;
                HoldCh := DispCh;
                break;
              end;
              Dec(l);
              if l <= 0 then break;
            end;
          end
          // If Prefix literal then we need to see if any non-literals exist
          // in the remainder of the mask text, if so then include literal.
          else
            if IsPrefix(FCtrlStr[k]) then
            begin
              if not HaveNext then
              begin
                for l := (k + 1) to Lim do
                begin
                  if IsMask(FCtrlStr[l]) then
                  begin
                    if FTextStr[l] <> eeMskEmptyChar then
                    begin
                      HaveNext := true;
                      break;
                    end;
                  end;
                end;
              end;
              if HaveNext then
                HoldCh := DispCh;
            end
            // If numeric sign then show sign always if mask is set to
            // mandatory, otherwise only show if text is equal to the
            // optional sign.
            else
              if IsNumericSign(FCtrlStr[k]) then
              begin
                if FMaskStr[k] = eeMskNumSign then
                  HoldCh := DispCh
                else
                  if FTextStr[k] = eeMskNumSignOpt then
                    HoldCh := DispCh;
                // We ALWAYS include the numeric sign despite having
                // treated it as a literal, so I set it here and clear
                // HoldCh to prevent further treatment...
                if HoldCh = DispCh then
                begin
                  RJStr := HoldCh + RJStr;
                  HoldCh := eeMskEmptyChar;
                end;
              end
              else
                HoldCh := DispCh;
          // If we found a literal to include
          if HoldCh = DispCh then
          begin
            // If we want all literals just include them...
            if IncLiterals then
            begin
              RJStr := DispCh + RJStr;
            end
            // otherwise we only include literals necessary to the mode
            else
            begin
              if IsNumericDecimal( FMode, FCtrlStr[k], FMaskStr[k] ) then
                RJStr := DispCh + RJStr;
            end;
          end;
        end
        // Now test the mask inclusion stuff...
        else
          if IsMask(FCtrlStr[k]) then
          begin
            // if not provided...
            if FTextStr[k] = eeMskEmptyChar then
            begin
              // If we have been asked for blank placeholders...
              if IncNulls or UseMaskBlank then
              begin
                // For standard mask processing we always supply
                if FMode <= eemMaskList then
                begin
                  if UseMaskBlank then
                    RJStr := FEditBlank + RJStr
                  else
                    RJStr := FTextBlank + RJStr;
                end
                // For enhanced mask processing we only supply
                // mandatory items
                else
                begin
                  if IsMandatory(FCtrlStr[k]) then
                  begin
                    if UseMaskBlank then
                      RJStr := FEditBlank + RJStr
                    else
                      RJStr := FTextBlank + RJStr;
                  end;
                end;
              end;
            end
            else
            begin
              RJStr := DispCh + RJStr;
              HavePrev := true;
            end;
          end;
        Dec(k);
      end;
      Result := Result + RJStr;
    end
    // If not right justified then fill to right...
    else
    begin
      DispCh := FTextStr[i];
      HoldCh := eeMskEmptyChar;
      // Test for literal inclusion...
      if IsLiteral(FCtrlStr[i]) then
      begin
        // If Postfix literal we simply test if any non-literals have been
        // entered previously, if so we include the literal character...
        if IsPostfix(FCtrlStr[i]) then
        begin
          if HavePrev then
            HoldCh := DispCh;
        end
        // If Prefix literal then we need to see if any non-literals exist
        // in the remainder of the mask text, if so then include literal.
        else
          if IsPrefix(FCtrlStr[i]) then
          begin
            for j := i to Lim do
            begin
              if IsMask(FCtrlStr[j]) then
              begin
                if FTextStr[j] <> eeMskEmptyChar then
                begin
                  HoldCh := DispCh;
                  break;
                end;
              end;
            end;
          end
          // If numeric sign then show sign always if mask is set to
          // mandatory, otherwise only show if text is equal to the
          // optional sign.
          else
            if IsNumericSign(FCtrlStr[i]) then
            begin
              if FMaskStr[i] = eeMskNumSign then
                HoldCh := DispCh
              else
                if FTextStr[i] = eeMskNumSignOpt then
                  HoldCh := DispCh;
              // We ALWAYS include the numeric sign despite having
              // treated it as a literal, so I set it here and clear
              // HoldCh to prevent further treatment...
              if HoldCh = DispCh then
              begin
                Result := Result + HoldCh;
                HoldCh := eeMskEmptyChar;
              end;
            end
            else
              HoldCh := DispCh;
        // If we found a literal to include
        if HoldCh = DispCh then
        begin
          // If we want all literals just include them...
          if IncLiterals then
            Result := Result + HoldCh
          // otherwise we only include literals necessary to the mode
          else
          begin
            // Any literals immediately following a right justified
            // section are required!
            if (i > 1) and (IsRightJust(FCtrlStr[i - 1])) then
            begin
              Result := Result + HoldCh;
            end
            else
              if IsNumericDecimal( FMode, FCtrlStr[i], FMaskStr[i] ) then
                RJStr := DispCh + RJStr;
          end;
        end;
      end
      // Now test the mask inclusion stuff...
      else
        if IsMask(FCtrlStr[i]) then
        begin
          // if not provided...
          if FTextStr[i] = eeMskEmptyChar then
          begin
            // If we have been asked for blank placeholders...
            if IncNulls or UseMaskBlank then
            begin
              // For standard mask processing we always supply
              if FMode <= eemMaskList then
              begin
                if UseMaskBlank then
                  Result := Result + FEditBlank
                else
                  Result := Result + FTextBlank;
              end
              // For enhanced mask processing we only supply
              // mandatory items
              else
              begin
                if IsMandatory(FCtrlStr[i]) then
                begin
                  if UseMaskBlank then
                    Result := Result + FEditBlank
                  else
                    Result := Result + FTextBlank;
                end;
              end;
            end;
          end
          else
          begin
            Result := Result + DispCh;
            HavePrev := true;
          end;
        end;
      Inc(i);
    end;
  end;
end;


function TMaskEnh.GetIsEmpty: boolean;
var
  i: integer;
begin
  Result := true;
  for i := 1 to FLength do
  begin
    if IsMask(FCtrlStr[i]) then
    begin
      if FTextStr[i] <> eeMskEmptyChar then
      begin
        Result := false;
        break;
      end;
    end;
  end;
end;


procedure TMaskEnh.SetMaskedText(AValue: string; WithLiterals: boolean);
var
  i, Len: integer;
  SavedTxt: string;
  InvInput: boolean;
begin
  SaveMaskTextTo(SavedTxt);
  Len := Length(AValue);
  i := 1;
  InvInput := false;
  FAssigningText := true;
  try
    Clear;
    if (Len > 0) and (Mode = eemNumeric) then
    begin
      AValue := ExponentFilter(AValue, FMaxIntDigits);
      Len := Length(AValue);
    end;
    if Len > 0 then
    begin
      InitMaskInput(WithLiterals);      // prepare for mask input
      while (i <= Len) and (i <= FLength) do
      begin
        if (not InsertMaskChar(AValue[i], WithLiterals)) and
          (not WithLiterals) then
        begin
          // Only break if invalid and not with literals.  Literals may
          // cause "breaks" in various parts of a mask (depending upon the
          // justification) so we just ignore the errors.
          InvInput := true;
          // 12-Aug-1999 : I no longer break, but I do set the InvInput
          // value so that a beep will be issued to warn the user than
          // some input was ingored.
          //break;
        end;
        Inc(i);
      end;
      // What was inserted may be OK, but perhaps we didnt fill the string
      // note that we REJECT invalid text in this case - except clause will
      // reset the original value...
      if ValidateOnSetText then
        ValidateEdit;
    end;
    FAssigningText := false;
  except
    // If we get an exception then restore original value
    RestoreMaskTextFrom(SavedTxt, true);
    FAssigningText := false;
    raise;
  end;
  // If we got to here the input text is acceptable as a final value but
  // here are some potential problems...
  if InvInput then
  begin
    // If we truncated the input due to mask mismatch
    if ValidateOnSetText then
      DoValidateError(E_MASKENH_INVALIDTRUNC);
  end
  else
  begin
    // Or we ran out of mask with which to fill from the input string
    if (i <= Len) and ValidateOnSetText then // If we truncated the input
      DoValidateError(E_MASKENH_TRUNCATED);
  end;
end;


procedure TMaskEnh.SaveMaskTextTo(var ToStr: string);
var
  i: integer;
begin
  i := 1;
  ToStr := '';
  while i <= FLength do
  begin
    if IsMask(FCtrlStr[i]) then
      ToStr := ToStr + FTextStr[i]
    else
      if IsNumericSign(FCtrlStr[i]) then
      begin
        if (FMaskStr[i] = eeMskNumSign) then
          ToStr := ToStr + FTextStr[i]
        else
        begin
          if FTextStr[i] <> eeMskNumSignOpt then
            ToStr := ToStr + eeMskNumSign
          else
            ToStr := ToStr + FTextStr[i];
        end;
      end;
    Inc(i);
  end;
end;


procedure TMaskEnh.RestoreMaskTextFrom(var FromStr: string; ListChk: boolean);
var
  i, j, Skip: integer;
  SignCh: char;
begin
  if FMode = eemNumeric then
  begin
    SignCh := eeMskNumSign;
    Skip := -1;
    for i := 1 to Length(FromStr) do
    begin
      if (FromStr[i] = eeMskNumSign) or (FromStr[i] = eeMskNumSignOpt)
        then
      begin
        SignCh := FromStr[i];
        Skip := i;
        break;
      end;
    end;
    i := 1;
    j := 1;
    if j = Skip then
      Inc(j);
    while i <= FLength do
    begin
      if IsMask(FCtrlStr[i]) then
      begin
        if j > Length(FromStr) then
          FTextStr[i] := eeMskEmptyChar
        else
        begin
          FTextStr[i] := FromStr[j];
        end;
        Inc(j);
        if j = Skip then
          Inc(j);
      end;
      Inc(i);
    end;
    SetNumericSign(SignCh);
  end
  else
  begin
    i := 1;
    j := 1;
    while i <= FLength do
    begin
      if IsMask(FCtrlStr[i]) then
      begin
        if j > Length(FromStr) then
          FTextStr[i] := eeMskEmptyChar
        else
          FTextStr[i] := FromStr[j];
        Inc(j);
      end;
      Inc(i);
    end;
    // we may have to re-choose the edit mask according to the updated
    // contents - and then to another restore to be sure of getting it
    // all (since not all the masks in the list need be the same length)
    if ListChk and (not FChoosingMask) then
    begin
      FChoosingMask := true;
      try
        if ((FMode = eemMaskList) or (FMode = eemMaskEnhList)) and
          (FMaskList <> nil) and (FMaskList.Count > 0) then
        begin
          ChooseMask;
          RestoreMaskTextFrom(FromStr, false);
        end;
      finally
        FChoosingMask := false;
      end;
    end;
  end;
end;


procedure TMaskEnh.InitMaskInput(WithLiterals: boolean);
begin
  FPos := 0;
  AdjustMaskPos(WithLiterals);
end;


procedure TMaskEnh.AdjustMaskPos(WithLiterals: boolean);
begin
  // if no literal to be input then move to first non-literal position
  if not WithLiterals then
  begin
    while FPos < FLength do
    begin
      if IsNumericDecimal(FMode, FCtrlStr[FPos + 1],
        FMaskStr[FPos + 1]) then
        break;
      if (FPos > 0) and IsRightJust(FCtrlStr[FPos]) then
        break;
      if not IsLiteral(FCtrlStr[FPos + 1]) then
        break;
      Inc(FPos);
    end;
  end;
  // If new section right justified then move to right of the right-justified
  // section (while only empty positions are found).
  while (FPos < FLength) and
    IsRightJust(FCtrlStr[FPos + 1]) and
    ((IsMask(FCtrlStr[FPos + 1]) and (FTextStr[FPos + 1] = eeMskEmptyChar))
    or IsLiteral(FCtrlStr[FPos + 1])) do
  begin
    if IsNumericDecimal(FMode, FCtrlStr[FPos + 1],
      FMaskStr[FPos + 1]) then
      break;
    Inc(FPos);
  end;
  // If we are in enhanced mode then left-justified operates in the same way
  // except we have to keep checking back for a right-just change over...
  if FMode > eemMaskList then
  begin
    while (FPos > 0) and
      IsLeftJust(FCtrlStr[FPos]) and
      ((IsMask(FCtrlStr[FPos]) and (FTextStr[FPos] = eeMskEmptyChar))
      or IsLiteral(FCtrlStr[FPos])) do
    begin
      if IsNumericDecimal(FMode, FCtrlStr[FPos],
        FMaskStr[FPos]) then
        break;
      Dec(FPos);
      if (FPos > 0) and IsRightJust(FCtrlStr[FPos]) then
      begin
        while (FPos < FLength) and IsLiteral(FCtrlStr[FPos + 1]) do
          Inc(FPos);
        break;
      end;
    end;
    // if no literal to be input then move to first non-literal position
    // (since we may have backed up over some leading literals)...
    if not WithLiterals then
    begin
      while FPos < FLength do
      begin
        if IsNumericDecimal(FMode, FCtrlStr[FPos + 1],
          FMaskStr[FPos + 1]) then
          break;
        if (FPos > 0) and IsRightJust(FCtrlStr[FPos]) then
          break;
        if not IsLiteral(FCtrlStr[FPos + 1]) then
          break;
        Inc(FPos);
      end;
    end;
  end;
end;


function TMaskEnh.SetNumericSign(var AValue: char): boolean;
var
  i: integer;
  HoldCh: char;
begin
  Result := false;
  for i := 1 to FLength do
  begin
    if IsNumericSign(FCtrlStr[i]) then
    begin
      if FMaskStr[i] = eeMskNumSign then
        HoldCh := AValue
      else
      begin
        if AValue = eeMskNumSign then
          HoldCh := eeMskRealBlank
        else
          HoldCh := AValue;
      end;
      if HoldCh <> FTextStr[i] then
      begin
        FTextStr[i] := HoldCh;
      end;
      Result := true;
      break;
    end;
  end;
end;

// FPos corresponds to character position in mask control.
// Insert is done AFTER FPos specified (or if you rather, between the
// specified position and the "next" character).
// So FPos represents the character to the left of the insert position.
// eg. FPos = 0 (not a valid MaskPos) insert is done between 0 and 1
//     FPos = 1 insert is done between 1 and 2
//     FPos = last-character insert is done after last character

function TMaskEnh.InsertMaskChar(var AValue: char; WithLiterals: boolean):
  boolean;
var
  InsPos, EmptyPos, i: integer;
  AllowOnce: boolean;

  function CanInsertToLeft: boolean;
  var
    ii: integer;
  begin
    InsPos := -1;                       // initialise to invalid
    EmptyPos := -1;
    // If there is a right justified position to the left
    Result := (FPos > 0) and IsRightJust(FCtrlStr[FPos]);
    if Result then
    begin
      ii := FPos;
      Result := false;
      // Search the right justified section for un-entered data spaces
      while (ii > 0) and IsRightJust(FCtrlStr[ii]) do
      begin
        // Save position of first available input space
        if (InsPos < 0) and IsMask(FCtrlStr[ii]) then
          InsPos := ii;
        if FTextStr[ii] = eeMskEmptyChar then
        begin
          EmptyPos := ii;
          Result := true;
          break;
        end;
        if IsNumericDecimal(FMode, FCtrlStr[ii], FMaskStr[ii]) then
          break;
        Dec(ii);
      end;
      // If there was not room at the inn, adjust the mask position
      if AllowOnce and (FPos < FLength) and
        ((ii <= 0) or ((ii > 0) and IsLeftJust(FCtrlStr[ii])))
        and not IsNumericDecimal(FMode, FCtrlStr[FPos + 1], FMaskStr[FPos + 1])
        then
      begin
        Inc(FPos);
        AdjustMaskPos(WithLiterals);
        AllowOnce := false;
        Result := CanInsertToLeft;
      end;
    end;
  end;

  function CanInsertToRight: boolean;
  var
    ii: integer;
  begin
    InsPos := -1;
    EmptyPos := -1;
    // If there is a left justified position to the right
    Result := (FPos < FLength) and IsLeftJust(FCtrlStr[FPos + 1]);
    if Result then
    begin
      ii := FPos + 1;
      Result := false;
      // Search the left justified section for un-entered data spaces
      while (ii <= FLength) and IsLeftJust(FCtrlStr[ii]) do
      begin
        // Save position of first available input space
        if (InsPos < 0) and IsMask(FCtrlStr[ii]) then
          InsPos := ii;
        if FTextStr[ii] = eeMskEmptyChar then
        begin
          EmptyPos := ii;
          Result := true;
          break;
        end;
        if IsNumericDecimal(FMode, FCtrlStr[ii], FMaskStr[ii]) then
          break;
        Inc(ii);
      end;
    end;
  end;

  function MakeRoomToLeft: boolean;
  var
    ii, jj: integer;
  begin
    Result := true;
    // First verify that we are allowed to move each character...
    ii := EmptyPos;
    jj := ii + 1;
    while jj <= InsPos do
    begin
      if IsMask(FCtrlStr[jj]) then
      begin
        if not ValidateCharInPos(FTextStr[jj], jj, false) then
        begin
          Result := false;
          break;
        end;
        ii := jj;
        jj := ii + 1;
      end
      else
        Inc(jj);
    end;
    // If ok to move all characters then actually do the move...
    ii := EmptyPos;
    jj := ii + 1;
    if Result then
      while jj <= InsPos do
      begin
        if IsMask(FCtrlStr[jj]) then
        begin
          FTextStr[ii] := FTextStr[jj];
          ii := jj;
          jj := ii + 1;
        end
        else
          Inc(jj);
      end;
  end;

  function MakeRoomToRight: boolean;
  var
    ii, jj: integer;
  begin
    Result := true;
    // First verify that we are allowed to move each character...
    ii := EmptyPos;
    jj := ii - 1;
    while jj >= InsPos do
    begin
      if IsMask(FCtrlStr[jj]) then
      begin
        if not ValidateCharInPos(FTextStr[jj], jj, false) then
        begin
          Result := false;
          break;
        end;
        ii := jj;
        jj := ii - 1;
      end
      else
        Dec(jj);
    end;
    // If ok to move all characters then actually do the move...
    ii := EmptyPos;
    jj := ii - 1;
    if Result then
      while jj >= InsPos do
      begin
        if IsMask(FCtrlStr[jj]) then
        begin
          FTextStr[ii] := FTextStr[jj];
          ii := jj;
          jj := ii - 1;
        end
        else
          Dec(jj);
      end;
  end;

begin
  Result := false;
  AllowOnce := true;
  // Convert cr/lf as required...
  if (AValue = #13) or (AValue = #10) then
    AValue := ' ';
  // First check for special case of numeric sign...
  if (FMode = eemNumeric) and
    ((AValue = eeMskNumSign) or (AValue = eeMskNumSignOpt)) then
    Result := SetNumericSign(AValue);
  // If the next position is for a literal then...
  if (not Result) and (FPos < FLength) and
    IsLiteral(FCtrlStr[FPos + 1]) then
  begin
    // If literals are included in the text just test for a match...
    if WithLiterals and (AValue = FTextStr[FPos + 1]) then
    begin
      Inc(FPos);
      Result := true;
    end;
    // If we didnt get a match above we need to check for the special cases
    // where "literal" characters are required despite the mask specification
    // (change from Right-to-Left justification).
    if (not Result) and (not WithLiterals) then
    begin
      if (FPos > 0) and (AValue = FTextStr[FPos + 1]) and (
        (IsRightJust(FCtrlStr[FPos]) and IsLeftJust(FCtrlStr[FPos + 1])) or
        (IsLeftJust(FCtrlStr[FPos]) and IsRightJust(FCtrlStr[FPos + 1]))
        ) then
      begin
        Inc(FPos);
        Result := true;
      end;
    end;
  end;
  if not Result then
  begin
    if CanInsertToLeft then
    begin
      // Just because we CAN insert to the left (right justified text) does
      // not mean that this is what we want to do.  It is possible that
      // there is a right-justified-input-empty-space to the right then
      // this means we are inserting into the middle of a right-justifed
      // section so just overwrite the blank rather than insert.
      if FPos < FLength then
      begin
        if IsRightJust(FCtrlStr[FPos + 1]) and
          IsMask(FCtrlStr[FPos + 1]) and
          (FTextStr[FPos + 1] = eeMskEmptyChar) then
        begin
          if ValidateCharInPos(AValue, FPos + 1, false) then
          begin
            Inc(FPos);
            FTextStr[FPos] := AValue;
            Result := true;
          end;
        end;
      end;
      if not Result then
      begin
        if ValidateCharInPos(AValue, InsPos, false) then
        begin
          if MakeRoomToLeft then
          begin
            FTextStr[InsPos] := AValue;
            Result := true;
          end;
        end;
      end;
    end;
    // If not right justified, OR if right-justified entry was not accepted
    // we try to to a left-justified entry.  This open approach is to try
    // and make the control more flexible in its acceptance of input while
    // still operating within the confines of the mask.
    if (not Result) and CanInsertToRight then
    begin
      // If enhanced edit then left-justify can be terminated by the
      // specific input of a literal character...
      if (FMode > eemMaskList) then
      begin
        i := FPos + 1;
        while (i < FLength) do
        begin
          if IsRightJust(FCtrlStr[i]) then
          begin
            if IsLiteral(FCtrlStr[i]) and
              (FTextStr[i] = AValue) then
            begin
              FPos := i + 1;
              Result := true;
            end;
            break;
          end;
          if IsMask(FCtrlStr[i]) and (FTextStr[i] <> eeMskEmptyChar) then
            break;
          Inc(i);
        end;
      end;
      if (not Result) and ValidateCharInPos(AValue, InsPos, false) then
      begin
        if MakeRoomToRight then
        begin
          FTextStr[InsPos] := AValue;
          FPos := InsPos;
          Result := true;
        end;
      end;
    end;
  end;
  if Result then
  begin
    AdjustMaskPos(WithLiterals);
    if not FChoosingMask then
    begin
      FChoosingMask := true;
      try
        if ((FMode = eemMaskList) or (FMode = eemMaskEnhList)) and
          (FMaskList <> nil) and (FMaskList.Count > 0) then
          ChooseMask;
      finally
        FChoosingMask := false;
      end;
    end;
  end
  else
  begin
    if not WithLiterals then
    begin
      if FPos >= FLength then
        DoValidateError(E_MASKENH_CHAR_TRUNCATED)
      else
        DoValidateError(E_MASKENH_CHAR_INVALID);
    end;
  end;
end;


function TMaskEnh.ReplaceMaskChar(var AValue: char; WithLiterals, IncPos:
  boolean): boolean;

  procedure SlideExisting(DestPos: integer);
  var
    ii, jj: integer;
  begin
    if IsRightJust(FCtrlStr[DestPos]) then
    begin
      if AValue = eeMskEmptyChar then
        IncPos := true
      else
        IncPos := false;
      ii := DestPos;
      jj := ii - 1;
      while (jj > 0) and IsRightJust(FCtrlStr[jj]) do
      begin
        if IsMask(FCtrlStr[jj]) then
        begin
          FTextStr[ii] := FTextStr[jj];
          FTextStr[jj] := eeMskEmptyChar;
          ii := jj;
          jj := ii - 1;
        end
        else
          Dec(jj);
      end;
    end
    // we only do dynamic left-justified slides in enh mode...
    else
      if FMode > eemMaskList then
      begin
        IncPos := false;                // Force IncPos false if sliding data
        ii := DestPos;
        jj := ii + 1;
        while (jj <= FLength) and IsLeftJust(FCtrlStr[jj]) do
        begin
          if IsMask(FCtrlStr[jj]) then
          begin
            FTextStr[ii] := FTextStr[jj];
            FTextStr[jj] := eeMskEmptyChar;
            ii := jj;
            jj := ii + 1;
          end
          else
            Inc(jj);
        end;
      end;
  end;

begin
  Result := false;
  // Convert cr/lf as required...
  if (AValue = #13) or (AValue = #10) then
    AValue := ' ';
  // First check for special case of numeric sign...
  if (FMode = eemNumeric) and
    ((AValue = eeMskNumSign) or (AValue = eeMskNumSignOpt)) then
    Result := SetNumericSign(AValue);
  // If the next position is for a literal then...
  if (not Result) and (FPos < FLength) and
    IsLiteral(FCtrlStr[FPos + 1]) then
  begin
    // If literals are included in the text just test for a match...
    if WithLiterals and (AValue = FTextStr[FPos + 1]) then
    begin
      Result := true;
      if IncPos then
        Inc(FPos);
    end;
    // If we didnt get a match above we need to check for the special cases
    // where "literal" characters are required despite the mask specification
    // (change from Right-to-Left justification).
    if (not Result) and (not WithLiterals) then
    begin
      if (FPos > 0) and IsRightJust(FCtrlStr[FPos]) and
        (FPos < FLength) and
        (AValue = FTextStr[FPos + 1]) and
        IsLeftJust(FCtrlStr[FPos + 1]) then
      begin
        Result := true;
        if IncPos then
          Inc(FPos);
      end;
    end;
  end;
  if (not Result) and (FPos < FLength) and
    IsMask(FCtrlStr[FPos + 1]) then
  begin
    if ValidateCharInPos(AValue, FPos + 1, false) then
    begin
      FTextStr[FPos + 1] := AValue;
      if AValue = eeMskEmptyChar then
      begin
        SlideExisting(FPos + 1);
      end;
      Result := true;
      if IncPos then
        Inc(FPos);
    end;
  end;
  if Result then
  begin
    AdjustMaskPos(WithLiterals);
    if not FChoosingMask then
    begin
      FChoosingMask := true;
      try
        if ((FMode = eemMaskList) or (FMode = eemMaskEnhList)) and
          (FMaskList <> nil) and (FMaskList.Count > 0) then
          ChooseMask;
      finally
        FChoosingMask := false;
      end;
    end;
  end
  else
  begin
    if not WithLiterals then
    begin
      if FPos >= FLength then
        DoValidateError(E_MASKENH_CHAR_TRUNCATED)
      else
        DoValidateError(E_MASKENH_CHAR_INVALID);
    end;
  end;
end;


function TMaskEnh.GetMaskEx: string;
begin
  Result := MaskEnhModeToString(FMode);
  if FMode <> eemNone then
  begin
    if (FMode = eemMaskList) or (FMode = eemMaskEnhList) then
    begin
      if assigned(MaskList) then
        Result := Result + MaskList.CommaText;
    end
    else
      Result := Result + Mask;
  end;
end;

// Allows setting of mode and mask from a single string...

procedure TMaskEnh.SetMaskEx(MaskStr: string);
var
  TestStr: string;
  tmpList: TStringList;
  HoldMode: TMaskEnhMode;

begin
  if Length(MaskStr) >= 6 then
  begin
    HoldMode := FMode;
    tmpList := nil;
    try
      if StringToMaskEnhMode(Copy(MaskStr, 1, 6), FMode) then
      begin
        if FMode = eemNone then
        begin
          Mask := '';
        end
        else
          if (FMode = eemMaskList) or (FMode = eemMaskEnhList) then
          begin
            TestStr := Copy(MaskStr, 7, Length(MaskStr) - 6);
            tmpList := TStringList.Create;
            try
              tmpList.CommaText := TestStr;
              MaskList := tmpList;
            finally
              tmpList.Free;
              tmpList := nil;
            end;
          end
          else
          begin
            Mask := Copy(MaskStr, 7, Length(MaskStr) - 6);
          end;
      end
      else
      begin
        if HoldMode = eemNone then
          FMode := eemMask;
        Mask := MaskStr;
      end;
    except
      FMode := HoldMode;
      if tmpList <> nil then
        tmpList.Free;
      raise;
    end;
  end
  else
  begin
    if Length(MaskStr) <= 0 then
    begin
      Mask := '';
      FMode := eemNone;
    end
    else
    begin
      if FMode = eemNone then
        FMode := eemMask;
      Mask := MaskStr;
    end;
  end;
end;


procedure TMaskEnh.Clear;
var
  i: integer;

  function GetMaskChar(CPos: integer): char;
  begin
    if IsMask(FCtrlStr[CPos]) then
      Result := eeMskEmptyChar
    else
    begin
      Result := FMaskStr[i];
      if IsIntlLiteral(FCtrlStr[CPos]) then
      begin
        case Result of
          eeMskTimeSeparator: Result := TimeSeparator;
          eeMskDateSeparator: Result := DateSeparator;
          eeMskDecimal: Result := DecimalSeparator;
          eeMskThousands: Result := ThousandSeparator;
        end;
      end;
    end;
  end;
begin
  for i := 1 to FLength do
  begin
    FTextStr[i] := GetMaskChar(i);
  end;
  FPos := 0;
  if not FChoosingMask then
  begin
    FChoosingMask := true;
    try
      if ((FMode = eemMaskList) or (FMode = eemMaskEnhList)) and
        (FMaskList <> nil) and (FMaskList.Count > 0) then
        ChooseMask;
    finally
      FChoosingMask := false;
    end;
  end;
end;


function TMaskEnh.ValidateEdit: boolean;
var
  i: integer;
begin
  Result := true;
  for i := 1 to FLength do
  begin
    if IsMask(FCtrlStr[i]) then
    begin
      Result := ValidateCharInPos(FTextStr[i], i, true);
      if not Result then
        break;
    end;
  end;
  if not Result then
  begin
    DoValidateError(E_MASKENH_INVALID);
  end;
end;


{$IFDEF __COMPILE_WITH_IBO__}
function TMaskEnh.IsMasked: boolean;
begin
  Result := ((Mode <> eemNone) and (MaskLen > 0));
end;

function TMaskEnh.EmptyText: string;
begin
  Clear;
  Result := Text;
end;

function TMaskEnh.FormatText(AVal: string): string;
begin
  Text := AVal;
  Result := GetMaskedText( true, true, false );
end;
{$ENDIF}


procedure TMaskEnh.DoValidateError(ErrCode: integer);
begin
  if assigned(FOnError) then
    FOnError(Self, ErrCode);
end;


function TMaskEnh.ValidateCharInPos(var AValue: char; MskPos: integer; ChkOpt:
  boolean): boolean;
var
  HoldCh: char;
  UsrPos: integer;
begin
  // If given a null/empty character then
  if AValue = eeMskEmptyChar then
  begin
    // If we are checking for optional/mandatory character placements
    // and we need a mandatory then the test fails
    if ChkOpt and IsMandatory(FCtrlStr[MskPos]) then
    begin
      // If Numeric has been setup with '0' for real blank then we
      // will permit validation of mandatory values left "null" because
      // they will still show up properly in a returned string.
      if (FMode = eemNumeric) and (FTextBlank = '0') then
        Result := true
      else
        Result := false;
    end
    // otherwise always return true
    else
      Result := true
  end
  else
  begin
    Result := false;
    case FMaskStr[MskPos] of
      eeMskAlpha: Result := IsCharAlpha(AValue); // in US = A-Z,a-z
      eeMskAlphaNum: Result := IsCharAlphaNumeric(AValue); // in US = A-Z,a-z,0-9
      eeMskAscii: Result := true;
      eeMskNumeric: Result := (AValue in ['0'..'9']); // 0-9, no plus or minus
      eeMskNumSymOpt: Result := (AValue in [' ', '-', '+', '0'..'9']);
    else
      // Is there any user characters
      if FUserMask <> nil then
      begin
        UsrPos := Pos(FMaskStr[MskPos], FUserMask.Strings[0]);
        Result := MatchesMask(AValue, FUserMask.Strings[UsrPos]);
      end;
    end;
    if Result then
    begin
      if MskPos > 1 then
        HoldCh := FTextStr[MskPos - 1]
      else
        HoldCh := eeMskEmptyChar;
      if IsLowerCase(FCtrlStr[MskPos]) then
        AValue := ConvCharCase(HoldCh, AValue, 2) //Ord(ecLowerCase)
      else
        if IsUpperCase(FCtrlStr[MskPos]) then
          AValue := ConvCharCase(HoldCh, AValue, 1) //Ord(ecUpperCase)
        else
          if IsProperCase(FCtrlStr[MskPos]) then
            AValue := ConvCharCase(HoldCh, AValue, 3); //Ord(ecProperCase)
    end;
  end;
end;

// Get enhanced mask positioning
// if DispPos = true then return count of displayed characters
//                      from the given Limit of mask characters
// if DispPos = false then return mask position of Limit of displayed chars

function TMaskEnh.GetEnhMaskPos(Limit: integer; DispPos: boolean): integer;
var
  i, Lim, j: integer;
  HavePrev: boolean;
  DispCh, HoldCh: char;
begin
  Result := 0;
  Lim := FLength;
  HavePrev := false;
  i := 1;
  while (i <= Lim) and (
    (DispPos and (i <= Limit)) or
    ((not DispPos) and (Result < Limit))
    ) do
  begin
    DispCh := FTextStr[i];
    HoldCh := eeMskEmptyChar;
    // Test for literal inclusion...
    if IsLiteral(FCtrlStr[i]) then
    begin
      // If Postfix literal we simply test if any non-literals have been
      // entered previously, if so we include the literal character...
      if IsPostfix(FCtrlStr[i]) then
      begin
        if HavePrev then
          HoldCh := DispCh;
      end
      // If Prefix literal then we need to see if any non-literals exist
      // in the remainder of the mask text, if so then include literal.
      else
        if IsPrefix(FCtrlStr[i]) then
        begin
          for j := i to Lim do
          begin
            if IsMask(FCtrlStr[j]) then
            begin
              if FTextStr[j] <> eeMskEmptyChar then
              begin
                HoldCh := DispCh;
                break;
              end;
            end;
          end;
        end
        // If numeric sign then show sign always if mask is set to
        // mandatory, otherwise only show if text is equal to the
        // optional sign.
        else
          if IsNumericSign(FCtrlStr[i]) then
          begin
            if FMaskStr[i] = eeMskNumSign then
              HoldCh := DispCh
            else
              if FTextStr[i] = eeMskNumSignOpt then
                HoldCh := DispCh;
            // We ALWAYS include the numeric sign despite having
            // treated it as a literal, so I set it here and clear
            // HoldCh to prevent further treatment...
            if HoldCh = DispCh then
            begin
              Inc(Result);
              HoldCh := eeMskEmptyChar;
            end;
          end
          else
            HoldCh := DispCh;
      // If we found a literal to include
      if HoldCh = DispCh then
      begin
        // If we want all literals just include them...
        Inc(Result);
      end;
    end
    // Now test the mask inclusion stuff...
    else
      if IsMask(FCtrlStr[i]) then
      begin
        // if not provided...
        if FTextStr[i] = eeMskEmptyChar then
        begin
          // For standard mask processing we always supply
          if FMode <= eemMaskList then
          begin
            Inc(Result);
          end
          // For enhanced mask processing we only supply
          // mandatory items
          else
          begin
            if IsMandatory(FCtrlStr[i]) then
            begin
              Inc(Result);
            end;
          end;
        end
        else
        begin
          Inc(Result);
          HavePrev := true;
        end;
      end;
    Inc(i);
  end;
  if not DispPos then
    Result := i - 1;
end;


procedure TMaskEnh.ChooseMask;
var
  i, j: integer;
  tmpMatch, tmpEntry, tmpMask: string;
  FoundMask: integer;
begin
  FoundMask := -1;
  tmpMask := Mask;
  tmpMatch := '';
  for i := 1 to (MaskList.Count - 1) do
  begin
    tmpMatch := MaskList.Names[i];
    tmpEntry := GetMaskedText(false, false, false);  //Copy( GetMaskedText(false,false,false), 1, Length(tmpMatch) );
    //if CompareText( tmpMatch, tmpEntry ) = 0 then begin
    if MatchesMask(tmpEntry, tmpMatch) then
    begin
      tmpMask := Copy(MaskList.Strings[i], Length(tmpMatch) + 2,
        Length(MaskList.Strings[i]));
      FoundMask := i;
      break;
    end;
  end;
  if FoundMask < 0 then
  begin
    FoundMask := 0;
    tmpMask := MaskList.Strings[0];
  end;
  // If we actually found a different mask then...
  if FMaskListIdx <> FoundMask then
  begin
    // Before changing the mask, find out how many non-literal characters into
    // the mask we are - will will use this to try and match up positions after
    // the mask has changed...
    j := 0;
    for i := 1 to FPos do
    begin
      if IsMask(FCtrlStr[i]) then
        Inc(j);
    end;
    Mask := tmpMask;
    FMaskListIdx := FoundMask;
    // Having chosen the editmask we want to try and
    // setup the mask positioning to suite the new mask...
    if j > FLength then
      FPos := FLength
    else
    begin
      FPos := 1;
      while (j > 0) and (FPos <= FLength) do
      begin
        if IsMask(FCtrlStr[FPos]) then
          Dec(j);
        Inc(FPos);
      end;
      Dec(FPos);
      AdjustMaskPos(false);
    end;
  end;
end;


procedure TMaskEnh.MovePosition(Key: Word; Shift: TShiftState);
var
  tmpPos: integer;
begin
  case Key of
    VK_LEFT:
      begin
        if FPos > 0 then
        begin
          if FMode > eemMaskList then
          begin
            // If we are way over to the right of the enhanced mask
            // we might never get back to the actually displayed
            // characters.  So I work out what the right-most display
            // posiition is and make sure we are not beyond it.
            tmpPos := GetEnhMaskPos(FPos, true);
            tmpPos := GetEnhMaskPos(tmpPos, false);
            if (FPos > tmpPos) and (tmpPos > 0) then
              FPos := tmpPos;
          end;
          Dec(FPos);
        end;
      end;
    VK_RIGHT:
      begin
        if FPos < FLength then
        begin
          if FMode > eemMaskList then
          begin
            tmpPos := GetEnhMaskPos(FPos, true);
            tmpPos := GetEnhMaskPos(tmpPos + 1, false);
            if (FPos < tmpPos) and (tmpPos <= FLength) then
              FPos := tmpPos
            else
              Inc(FPos);
          end
          else
            Inc(FPos);
        end;
      end;
    VK_HOME:
      begin
        FPos := 0;
      end;
    VK_END:
      begin
        FPos := FLength;
      end;
  end;
end;


function TMaskEnh.MoveToPosition(NewPos: integer; Reselect: boolean): boolean;
begin
  Result := false;
  // For standard mask processing there is ALWAYS a particular
  // character present in the display string relative to the
  // masktext string...
  if FMode <= eemMaskList then
  begin
    FPos := NewPos;
    if Reselect then
    begin
      if not (((FPos < FLength) and
        IsRightJust(FCtrlStr[FPos + 1])) or
        ((FPos > 0) and IsRightJust(FCtrlStr[FPos]))) then
      begin
        Result := true;
      end;
    end;
  end
  // Otherwise the synchronisation between mask and caret is rather
  // more complicated - because the display text and the mask text
  // do not directly correlate.  No automatic selection is done in
  // enhanced mode because we are always in insert mode.
  else
  begin
    FPos := GetEnhMaskPos(NewPos, false);
  end;
end;


function TMaskEnh.GetDisplayPosition(var StartPos: integer; var StopPos:
  integer): boolean;
begin
  Result := false;
  // For standard mask processing there is ALWAYS a particular
  // character present in the display string relative to the
  // masktext string...
  if FMode <= eemMaskList then
  begin
    StartPos := FPos;
    StopPos := StartPos + 1;
    if StopPos > FLength then
    begin
      StartPos := FLength;
      StopPos := FLength;
    end;
    if ((FPos < FLength) and IsRightJust(FCtrlStr[FPos + 1])) or
      ((FPos > 0) and IsRightJust(FCtrlStr[FPos])) then
    begin
      Result := true;
    end;
  end
  // Otherwise the synchronisation between mask and caret is rather
  // more complicated...
  else
  begin
    StartPos := GetEnhMaskPos(FPos, true);
    StopPos := StartPos; // enh masks dont select a char
    Result := true;
  end;
end;


function TMaskEnh.GetIsLeftJust(TestPos: integer): boolean;
begin
  Result := IsLeftJust(FCtrlStr[TestPos])
end;


function TMaskEnh.GetIsRightJust(TestPos: integer): boolean;
begin
  Result := IsRightJust(FCtrlStr[TestPos])
end;


function TMaskEnh.DeleteBackSpace(WithUpd: boolean): boolean;
var
  HoldPos1: integer;
  IncPos: boolean;
  EmptyCh: char;
begin
  EmptyCh := eeMskEmptyChar;
  Result := false;
  if FPos > 0 then
  begin
    HoldPos1 := FPos;                   // May want to return to this position
    Dec(FPos);
    while (FPos > 0) and IsLiteral(FCtrlStr[FPos + 1]) do
      Dec(FPos);
    if (FPos > 0) and IsRightJust(FCtrlStr[FPos]) then
      IncPos := true
    else
      IncPos := false;
    if ReplaceMaskChar(EmptyCh, false, IncPos) then
    begin
      if WithUpd then
        Result := true;
    end
    else
      FPos := HoldPos1;                 // Return to prior position on fail
  end;
end;


function TMaskEnh.DeleteForward(WithUpd: boolean): boolean;
var
  EmptyCh: char;
begin
  Result := false;
  EmptyCh := eeMskEmptyChar;
  AdjustMaskPos(false);
  if FPos < FLength then
  begin
    if ReplaceMaskChar(EmptyCh, false, true) then
    begin
      if WithUpd then
        Result := true;
    end;
  end;
end;


function TMaskEnh.DeleteRange(WithUpd: boolean; Len: integer): boolean;
var
  HoldPos, TmpPos, MaskChars: integer;
begin
  Result := false;
  HoldPos := FPos;
  TmpPos := FPos;
  MaskChars := 0;
  while (TmpPos < FLength) and (Len > 0) do
  begin
    if IsMask(FCtrlStr[TmpPos + 1]) then
      Inc(MaskChars);
    Inc(TmpPos);
    Dec(Len);
  end;
  while (MaskChars > 0) do
  begin
    if (DeleteForward(WithUpd)) then
      Result := true;
    Dec(MaskChars);
  end;
  if not(WithUpd and Result) then
    FPos := HoldPos;
end;


{$IFDEF __COMPILE_WITH_IBO__}

var PrevCreateMaskProcessor: TIB_CreateMaskProcessor;

// If compiling for integration with IBO, we define a CreateMaskProcessor
// procedure for assignment to the function variable provided.
function CreateMaskEnhProcessor(const MaskStr: string): IIB_MaskProcessor;
var
  TmpStr: string;
begin
  TmpStr := Copy(MaskStr,1,6);
  if (TmpStr = '::NO::') or
    (TmpStr = '::SM::') or
    (TmpStr = '::SL::') or
    (TmpStr = '::EM::') or
    (TmpStr = '::EL::') or
    (TmpStr = '::NM::') then
  begin
    Result := TMaskEnh.Create;
    Result.SetMask( MaskStr );
  end
  else
    if assigned(PrevCreateMaskProcessor) then
      Result := PrevCreateMaskProcessor(MaskStr)
    else
      Result := nil;
end;
{$ENDIF}


initialization;
  MaskEnh_EmptyList := TStringList.Create;
  {$IFDEF __COMPILE_WITH_IBO__}
  // save any existing mask processor
  PrevCreateMaskProcessor := IB_CreateMaskProcessor;
  // and assign our own
  IB_CreateMaskProcessor := CreateMaskEnhProcessor;
  {$ENDIF}

finalization;
  MaskEnh_EmptyList.Free;
  MaskEnh_EmptyList := nil;
  {$IFDEF __COMPILE_WITH_IBO__}
  // Restore the previous mask processor function
  IB_CreateMaskProcessor := PrevCreateMaskProcessor;
  PrevCreateMaskProcessor := nil;
  {$ENDIF}

end.