{ Telesis_MemoEnh.pas                             SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Memo control derivation from TCustomEditEnh.
<br>
<br>
<strong>History</strong><br>
<br>
12-Apr-2011 - Use Length (not SizeOf) to make code a little more portable.
<br>
14-Feb-2008 - The old code, copied from TMemo, altered the CreateWindowHandle
so that the text was restored separately - but used the Caption property.  This
seems to be a bug as the text is lost during some recreate calls.  By changing
over to restore from WindowText (a property accessing the specially saved text
buffer, saved from just before the handle was destroyed) we seem to have solved
a long standing problem with text not saving in the Dream designer.<br>
<br>
22-Jan-2008 - [internal].<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_MemoEnh;

{$INCLUDE Telesis.inc}

interface

uses Classes,
  Messages,
  Types,
  Controls,
  StdCtrls,
  Telesis_EditEnh;


type

{: @summary Base class for Enhanced Memo Controls.
<br>
@desc Written as a derivation of CustomEditEnh it has all the features of
the base control, including embedded button, Ctl3D highlighting etc.   The
mask processing capabilities have not been published because their use in a
memo control would be unpredictable (the mask does not directly support
newline processing).<br>
<br>
This control supports automatic scrollbars, but possibily its biggest advantage
is that it shares the common ancestor to most other Enh controls allowing
code to be more generic.}
  TCustomMemoEnh = class(TCustomEditEnh)
  private
    FLines: TStrings;
    FScrollBars: TScrollStyle;
    FWantReturns: Boolean;
    FWantTabs: Boolean;
    FScrollChanging: boolean;
    FScrollBarsAuto: boolean;

    procedure SetScrollBarsAuto(AValue: boolean);
    // From TCustomMemo...
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMNCDestroy(var Message: TWMNCDestroy); message WM_NCDESTROY;

  protected
    function GetButton3KeyEx: TShortCut; override;
    procedure DoButton1; override;
    procedure DoButton2; override;
    procedure DoButton3; override;

    // My own...
    procedure Change; override;
    function GetLineHeight: integer; virtual;
    procedure CheckAutoScrollBars;

    // From TCustomMemo...
    function GetCaretPos: TPoint; virtual;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWindowHandle(const Params: TCreateParams); override;
    procedure DestroyWindowHandle; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Loaded; override;
    procedure SetLines(Value: TStrings);
    procedure SetScrollBars(Value: TScrollStyle);
    function GetWordWrap: boolean;
    procedure SetWordWrap(Value: Boolean);

    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars
      default ssNone;
{: @summary WantReturns indicates whether return characters (carriage return
and line feed) should be passed on to the control.
<br>
@desc Defaults to true.  It can be useful on dialogs to set this to false so
that the return character can be used to press the default button.}
    property WantReturns: Boolean read FWantReturns write FWantReturns
      default True;
{: @summary WantTabs indicates whether Tab characters should be passed on
to the control.
<br>
@desc Defaults to false.  Set true if you want the control to allow
input of tab characters for formatting.}
    property WantTabs: Boolean read FWantTabs write FWantTabs default False;
{: @summary Whether you want the control to automatically word wrap lines
that are too long to fit across the display area of the control.
<br>
@desc When true there will be no horizontal scroll bar.}
    property WordWrap: Boolean read GetWordWrap write SetWordWrap default True;
    property Options default
      [eeoAllowEmptyExit, eeoBeepOnTruncateErr, eeoBtnGlyphStretch,
      eeoBtnProportional, eeoExceptions, eeoWordWrap, eeoInputDecimalTrans];

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    // From TCustomMemo...
{: @summary Current position of the Caret (in pixels). }
    property CaretPos: TPoint read GetCaretPos;
{: @summary Access to the memo contents as TStrings.}
    property Lines: TStrings read FLines write SetLines;
{: @summary The height of individual rows in the control.}
    property LineHeight: integer read GetLineHeight;

    // My own...
{: @summary Automatically show/hide scrollbars.
<br>
@desc When true the vertical scroll bar is automatically shown or hidden
according to the number of lines in the control.}
    property ScrollBarsAuto: boolean read FScrollBarsAuto write SetScrollBarsAuto
      default true;

  end;                                  // TCustomMemoEnh

{: @summary The published variation of TCustomMemoEnh.}
  TMemoEnh = class(TCustomMemoEnh)
    {$DEFINE PUBLISH_CustomMemoEnh}
    {$INCLUDE Telesis_Publish.INC}
  end;                                  // TMemoEnh


implementation


uses
  Windows,
  Menus,
  SysUtils,
  Consts,
  Forms;

type
  // Essentially copied from StdCtrls...
  TMemoEnhStrings = class(TStrings)
  private
    Memo: TCustomMemoEnh;
  protected
    procedure DefineProperties(Filer: TFiler); override;
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetTextStr: string; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure SetTextStr(const Value: string); override;
    procedure SetUpdateState(Updating: Boolean); override;
  public
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
  end;


procedure TMemoEnhStrings.DefineProperties(Filer: TFiler);
begin
  if (Memo.WindowText <> nil) then
    Memo.HandleNeeded; // We need it to access the strings.
  inherited DefineProperties(Filer);
end;

function TMemoEnhStrings.GetCount: Integer;
begin
  Result := 0;
  if Memo.HandleAllocated or (Memo.WindowText <> nil) then
  begin
    Result := SendMessage(Memo.Handle, EM_GETLINECOUNT, 0, 0);
    if SendMessage(Memo.Handle, EM_LINELENGTH, SendMessage(Memo.Handle,
      EM_LINEINDEX, Result - 1, 0), 0) = 0 then Dec(Result);
  end;
end;

function TMemoEnhStrings.Get(Index: Integer): string;
var
  Text: array[0..4095] of Char;
begin
  Word((@Text)^) := Length(Text);
  SetString(Result, Text, SendMessage(Memo.Handle, EM_GETLINE, Index,
    Longint(@Text)));
end;

procedure TMemoEnhStrings.Put(Index: Integer; const S: string);
var
  SelStart: Integer;
begin
  SelStart := SendMessage(Memo.Handle, EM_LINEINDEX, Index, 0);
  if SelStart >= 0 then
  begin
    SendMessage(Memo.Handle, EM_SETSEL, SelStart, SelStart +
      SendMessage(Memo.Handle, EM_LINELENGTH, SelStart, 0));
    SendMessage(Memo.Handle, EM_REPLACESEL, 0, Longint(PChar(S)));
  end;
end;

procedure TMemoEnhStrings.Insert(Index: Integer; const S: string);
var
  SelStart, LineLen: Integer;
  Line: string;
begin
  if Index >= 0 then
  begin
    SelStart := SendMessage(Memo.Handle, EM_LINEINDEX, Index, 0);
    if SelStart >= 0 then
      Line := S + #13#10
    else
    begin
      SelStart := SendMessage(Memo.Handle, EM_LINEINDEX, Index - 1, 0);
      if SelStart < 0 then Exit;
      LineLen := SendMessage(Memo.Handle, EM_LINELENGTH, SelStart, 0);
      if LineLen = 0 then Exit;
      Inc(SelStart, LineLen);
      Line := #13#10 + s;
    end;
    SendMessage(Memo.Handle, EM_SETSEL, SelStart, SelStart);
    SendMessage(Memo.Handle, EM_REPLACESEL, 0, Longint(PChar(Line)));
  end;
end;

procedure TMemoEnhStrings.Delete(Index: Integer);
const
  Empty: PChar = '';
var
  SelStart, SelEnd: Integer;
begin
  SelStart := SendMessage(Memo.Handle, EM_LINEINDEX, Index, 0);
  if SelStart >= 0 then
  begin
    SelEnd := SendMessage(Memo.Handle, EM_LINEINDEX, Index + 1, 0);
    if SelEnd < 0 then
      SelEnd := SelStart +
        SendMessage(Memo.Handle, EM_LINELENGTH, SelStart, 0);
    SendMessage(Memo.Handle, EM_SETSEL, SelStart, SelEnd);
    SendMessage(Memo.Handle, EM_REPLACESEL, 0, Longint(Empty));
  end;
end;

procedure TMemoEnhStrings.Clear;
begin
  Memo.Clear;
end;

procedure TMemoEnhStrings.SetUpdateState(Updating: Boolean);
begin
  if Memo.HandleAllocated then
  begin
    SendMessage(Memo.Handle, WM_SETREDRAW, Ord(not Updating), 0);
    if not Updating then
    begin // WM_SETREDRAW causes visibility side effects in memo controls
      Memo.Perform(CM_SHOWINGCHANGED, 0, 0);  // This reasserts the visibility we want
      Memo.Refresh;
    end;
  end;
end;

function TMemoEnhStrings.GetTextStr: string;
begin
  Result := Memo.Text;
end;

procedure TMemoEnhStrings.SetTextStr(const Value: string);
var
  NewText: string;
begin
  NewText := AdjustLineBreaks(Value);
  if (Length(NewText) <> Memo.GetTextLen) or (NewText <> Memo.Text) then
  begin
    if SendMessage(Memo.Handle, WM_SETTEXT, 0, Longint(NewText)) = 0 then
      raise EInvalidOperation.Create(SInvalidMemoSize);
    Memo.Perform(CM_TEXTCHANGED, 0, 0);
  end;
end;



constructor TCustomMemoEnh.Create(AOwner: TComponent);
begin
  FScrollChanging := false;
  inherited Create(AOwner);
  FCRLFProtection := false;
  FScrollBarsAuto := true;
  Width := 185;
  Height := 89;
  AutoSize := False;
  Options := Options + [eeoWordWrap];
  FWantReturns := True;
  FLines := TMemoEnhStrings.Create;
  TMemoEnhStrings(FLines).Memo := Self;
end;

destructor TCustomMemoEnh.Destroy;
begin
  FLines.Free;
  inherited Destroy;
end;

procedure TCustomMemoEnh.CreateParams(var Params: TCreateParams);
const
  ScrollBar: array[TScrollStyle] of DWORD = (0, WS_HSCROLL, WS_VSCROLL,
    WS_HSCROLL or WS_VSCROLL);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or ScrollBar[FScrollBars];
  end;
end;

procedure TCustomMemoEnh.CreateWindowHandle(const Params: TCreateParams);
var
  P: TCreateParams;
begin
  if SysLocale.FarEast and (Win32Platform <> VER_PLATFORM_WIN32_NT) and
    ((Params.Style and ES_READONLY) <> 0) then
  begin
    P := Params;
    P.Style := P.Style and (not ES_READONLY);
    with Params do
      WindowHandle := CreateWindowEx(ExStyle, WinClassName, '', Style,
        X, Y, Width, Height, WndParent, 0, WindowClass.hInstance, Param);
    if WindowHandle <> 0 then
    begin
      if Assigned(WindowText) then
        SendMessage(WindowHandle, WM_SETTEXT, 0, Longint(WindowText));
      SendMessage(WindowHandle, EM_SETREADONLY, Ord(True), 0);
    end;
  end
  else
  begin
    with Params do
      WindowHandle := CreateWindowEx(ExStyle, WinClassName, '', Style,
        X, Y, Width, Height, WndParent, 0, WindowClass.hInstance, Param);
    if (WindowHandle <> 0) and Assigned(WindowText) then
      SendMessage(WindowHandle, WM_SETTEXT, 0, Longint(WindowText));
  end;
end;

procedure TCustomMemoEnh.DestroyWindowHandle;
begin
  inherited DestroyWindowHandle;
end;

function TCustomMemoEnh.GetCaretPos: TPoint;
begin
  Windows.GetCaretPos(Result);
end;

function TCustomMemoEnh.GetLineHeight: integer;
begin
  // This is a fudge until I can work out a more reliable method...
  Result := Canvas.TextHeight('Wg') + 2;
end;

procedure TCustomMemoEnh.Loaded;
begin
  inherited Loaded;
  Modified := False;
end;

procedure TCustomMemoEnh.SetLines(Value: TStrings);
begin
  FLines.Assign(Value);
end;

procedure TCustomMemoEnh.SetScrollBars(Value: TScrollStyle);
begin
  if FScrollBars <> Value then
  begin
    FScrollBars := Value;
    RecreateWnd;
  end;
end;

function TCustomMemoEnh.GetWordWrap: Boolean;
begin
  Result := (eeoWordWrap in Options);
end;

procedure TCustomMemoEnh.SetWordWrap(Value: Boolean);
begin
  if (Value and not (eeoWordWrap in Options)) then
  begin
    Options := Options + [eeoWordWrap];
  end
  else
    if (not Value and (eeoWordWrap in Options)) then
    begin
      Options := Options - [eeoWordWrap];
    end;
end;

procedure TCustomMemoEnh.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if FWantTabs then
    Message.Result := Message.Result or DLGC_WANTTAB
  else
    Message.Result := Message.Result and not DLGC_WANTTAB;
  if FWantReturns then
    Message.Result := Message.Result or DLGC_WANTALLKEYS
  else
    Message.Result := Message.Result and not DLGC_WANTALLKEYS;
end;

procedure TCustomMemoEnh.WMNCDestroy(var Message: TWMNCDestroy);
begin
  inherited;
end;

procedure TCustomMemoEnh.KeyDown(var Key: Word; Shift: TShiftState);
// Make these #$@#$ controls scroll all the way to the top and bottom!
// It drives me batty when normal memo controls wont do that with page-up and
// page-down.
//
// NOTE: Because the EM_CHARFROMPOS message returns the line index and
// character position information in a single 32bit integer (as hi and lo
// words) this means that there are some restrictions to the enhancements.
//
// * When the total bytes > 64k the selection capability may be a little
// inconsistent at the far end of the text - nothing serious - its is just
// when you change the select direction it does not work consistently with
// normal operation/expectation.
//
// * When the total number of lines > 64k then it may not work at all, but
// then this is really pushing a control that was only ever designed to
// operate within 64k bytes anyway.  I've been using this with text
// containing over 15000 lines (nearly 600k bytes) without any drama.
//
// IMPORTANT:  If this starts behaving erratically on page down it will
// quite likely be due to the GetLineHeight function (servicing the LineHeight
// property).  This function is really just a kludge and may not work well
// with all text sizes.  It may also be due to extended margins being defined
// (or possibly the control button aligned to top/bottom) since I am using
// the client height without taking into account the actual text area - I
// should probably fix this sometime. (I just need to make the GetTextRect
// function protected instead of private in the CustomEditEnh unit, so that
// I can use it from here.)
//
// Changes did not work under Delphi 5 - I am not sure why, so they have been
// completely removed!  Note that Ctrl+Home and Ctrl+End can be used to
// jump to the start/end of the text (holding shift at the same time can
// be used to select all the text).
//
//var
    //Start, Stop, LineNo: integer;
    //i, j, LCnt, CPos, NewCPos: integer;
    //CP: TPoint;
begin
  inherited KeyDown(Key, Shift);
  {if Key = VK_PRIOR then begin
      SendMessage(Handle, EM_GETSEL, Integer(@Start), Integer(@Stop));
      j := SendMessage(Handle, EM_LINEFROMCHAR, Start, 0 );
      i := j - (ClientHeight div LineHeight);
      if i < 0 then
          i := 0;
      i := SendMessage(Handle, EM_LINEINDEX, i, 0 );
      if ssShift in Shift then begin
          SendMessage(Handle, EM_SETSEL, Stop, i);
          end
      else begin
          SendMessage(Handle, EM_SETSEL, i, i);
          end;
      SendMessage(Handle, EM_SCROLLCARET, 0, 0 );
      Key := 0;
      end
  else if Key = VK_NEXT then begin
      SendMessage(Handle, EM_GETSEL, Integer(@Start), Integer(@Stop));
      j := SendMessage(Handle, EM_LINEFROMCHAR, Start, 0 );
      LCnt := SendMessage(Handle, EM_GETLINECOUNT, 0, 0 );
      i := j + (ClientHeight div LineHeight);
      if i > LCnt then
          i := LCnt - 1;
      i := SendMessage(Handle, EM_LINEINDEX, i, 0 );
      if ssShift in Shift then begin
          SendMessage(Handle, EM_SETSEL, Stop, i);
          end
      else begin
          SendMessage(Handle, EM_SETSEL, i, i);
          end;
      SendMessage(Handle, EM_SCROLLCARET, 0, 0 );
      Key := 0;
      end;
      //i := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
      //if Key = VK_PRIOR then begin
      //    j := -(ClientHeight div LineHeight);
      //    SendMessage(Handle, EM_LINESCROLL, 0, j );
      //    end
      //else begin
      //    j := (ClientHeight div LineHeight);
      //    SendMessage(Handle, EM_LINESCROLL, 0, j );
      //    end;
      //Key := 0; // stop standard processing
      //end;
  }
  {
  if Key = VK_PRIOR then begin
      i := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
      // if no change in visible line then must be at the top...
      if i = 0 then begin
          CaretP := CaretPos;
          CPos := SendMessage(Handle, EM_CHARFROMPOS, 0, MAKELPARAM( CaretP.X, CaretP.Y ));
          LineNo := LongRec( CPos ).Hi;
          CPos := LongRec( CPos ).Lo;
          // if we are not already on the first line...
          if LineNo > 0 then begin
              SendMessage(Handle, EM_GETSEL, Integer(@Start), Integer(@Stop));
              if CPos = Start then begin
                  NewCPos := Start - SendMessage( Handle, EM_LINEINDEX, LineNo, 0 );
                  end
              else begin
                  NewCPos := Stop - SendMessage( Handle, EM_LINEINDEX, LineNo, 0 );
                  Stop := Start;
                  end;
              if not(ssShift in Shift) then
                  Stop := NewCPos;
              SendMessage(Handle, EM_SETSEL, Stop, NewCPos);
              end;
          end;
      end
  else if Key = VK_NEXT then begin
      j := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
      inherited KeyDown( Key, Shift );
      i := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
      if i = j then begin
          CaretP := CaretPos;
          CPos := SendMessage(Handle, EM_CHARFROMPOS, 0, MAKELPARAM( CaretP.X, CaretP.Y ) );
          LineNo := LongRec( CPos ).Hi;
          CPos := LongRec( CPos ).Lo;
          i := SendMessage(Handle, EM_GETLINECOUNT, 0, 0) - 1;
          SendMessage(Handle, EM_GETSEL, Integer(@Start), Integer(@Stop));
          if CPos = Start then begin
              NewCPos := Start - SendMessage( Handle, EM_LINEINDEX, LineNo, 0 );
              NewCPos := NewCPos + SendMessage( Handle, EM_LINEINDEX, i, 0 );
              end
          else begin
              NewCPos := Stop - SendMessage( Handle, EM_LINEINDEX, LineNo, 0 );
              NewCPos := NewCPos + SendMessage( Handle, EM_LINEINDEX, i, 0 );
              Stop := Start;
              end;
          if not(ssShift in Shift) then
              Stop := NewCPos;
          SendMessage(Handle, EM_SETSEL, Stop, NewCPos);
          end;
      end
  else begin
      inherited KeyDown( Key, Shift );
      end;};
end;

procedure TCustomMemoEnh.KeyPress(var Key: Char);
var
  HoldCh: char;
begin
  HoldCh := Key;
  inherited KeyPress(Key);
  // CustomEditEnh removes all returns, so we reset the key here as
  // required.  The NulledReturn property tells me that CustomEditEnh
  // caused Key to be set to 0 and NOT some event handler.
  if ((HoldCh = Char(VK_RETURN)) and WantReturns and NulledReturn) or
    ((HoldCh = Char(VK_TAB)) and WantTabs and NulledReturn) or
    ((HoldCh = char(EE_VK_CTRLENTER)) and NulledReturn) then
  begin
    // A return char is the start of a modification...
    if EditBeginModify then
    begin
      Key := HoldCh;
      NulledReturn := false;
    end;
  end;
end;

procedure TCustomMemoEnh.DoButton1;
begin
    //[internal]
    inherited DoButton1;
end;

procedure TCustomMemoEnh.DoButton2;
begin
    //[internal]
    inherited DoButton2;
end;

procedure TCustomMemoEnh.DoButton3;
begin
    //[internal]
    inherited DoButton3;
end;

function TCustomMemoEnh.GetButton3KeyEx: TShortCut;
var
  tmp: TShortCut;
begin
  Result := inherited GetButton3KeyEx;
  //[internal]
end;

procedure TCustomMemoEnh.Change;
begin
  inherited;
  if ScrollBarsAuto then
    CheckAutoScrollBars;
end;

procedure TCustomMemoEnh.SetScrollBarsAuto(AValue: boolean);
begin
  if AValue <> FScrollBarsAuto then
  begin
    FScrollBarsAuto := AValue;
    CheckAutoScrollBars;
  end;
end;

procedure TCustomMemoEnh.CheckAutoScrollBars;
var
  i, h: integer;
  VSB, HSB: boolean;
begin
  if (not FScrollChanging) and HandleAllocated and
    not (csLoading in ComponentState) then
  begin
    VSB := false;
    HSB := false;
    FScrollChanging := true;
    try
      Canvas.Lock;
      try
        Canvas.Handle := GetWindowDC(Handle);
        try
          // First check whether vertical is required...
          if SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0) > 0 then
          begin
            VSB := true;
          end
          else
          begin
            h := LineHeight * Lines.Count;
            if h > ClientHeight then
              VSB := true;
          end;
          // If word-wrap is on then we will only ever need the vertical
          if not WordWrap then
          begin
            for i := 0 to (Lines.Count - 1) do
            begin
              h := Canvas.TextWidth(Lines.Strings[i]);
              if h > ClientWidth then
              begin
                HSB := true;
                break;
              end;
            end;
          end;
        finally
          ReleaseDC(Handle, Canvas.Handle);
          Canvas.Handle := 0;
        end;
      finally
        Canvas.Unlock;
      end;
      SendMessage(Handle, EM_GETSEL, Integer(@i), Integer(@h));
      if VSB and HSB then
        ScrollBars := ssBoth
      else
        if VSB then
          ScrollBars := ssVertical
        else
          if HSB then
            ScrollBars := ssHorizontal
          else
            ScrollBars := ssNone;
      if Focused then
      begin
        SendMessage(Handle, EM_SETSEL, i, h);
        SendMessage(Handle, EM_SCROLLCARET, i, h);
      end;
    finally
      FScrollChanging := false;
    end;
  end;
end;


end.