{ Telesis_SpinEditEnh.pas                         SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary SpinEdit derivation of TCustomEditEnh.
<br>
@desc Delphi 3 does not support streaming of variant datatypes, so in order to
provide Delphi3 support it was necessary to take over the streaming of variants
for all versions (to make it possible to share DFMs between versions).  The
variant properties (Increment, Large_Increment, Minimum, Maximum and Value)
will now be stored in the DFM under the name 'IncrementEx', 'ValueEx' etc.<br>
<br>
Old versions of DFMs should still load and initialise correctly (in Delphi4+).
Delphi will still loads properties that appear in the resource regardless of
the stored attribute.  However when the DFMs are saved it will be using the
new storage names.  Note: If an old property value has been saved with a
value between -127 and -1, it will be read as an byte value resulting in a
property value between 129 and 255.  Not sure why this is, but there does
not seem to be much I can do about it - hopefully there wont be too many
instances of these.<br>
<br>
I provided virtual "IsXxxxStored" procedures that can be overridden in
derivations if you need to change the default values of these variant
properties.<br>
<br>
<br>
<strong>History</strong><br>
<br>
12-Feb-2008 - MouseWheel support.<br>
<br>
28-Dec-2001 - Updated validation processing to match changes in EditEnh.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_SpinEditEnh;

{$INCLUDE Telesis.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  {$IFDEF __VCL60_OR_LATER__}
  Variants,
  {$ENDIF}
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  Telesis_MaskEnh,
  Telesis_Utils,
  Telesis_EditEnh;


type

  TCustomSpinEditEnh = class(TCustomEditEnh)
  private
    FIncrement: variant;
    FMaximum: variant;
    FMinimum: variant;
    FWrapValue: boolean;
    FLarge_Increment: variant;

  protected
    function GetButtonAlignDefault: TAlign; override;
    function GetButtonSizeDefault: integer; override;
    function GetButtonStyleDefault: TEditEnhButtonStyle; override;
    function GetButtonAutoRepeatDefault: boolean; override;
    function GetButton1KeyDefault: TShortCut; override;
    function GetButton2KeyDefault: TShortCut; override;

    procedure ReadIncrement(Reader:TReader);
    procedure WriteIncrement(Writer:TWriter);
    procedure ReadLargeIncrement(Reader:TReader);
    procedure WriteLargeIncrement(Writer:TWriter);
    procedure ReadMaximum(Reader:TReader);
    procedure WriteMaximum(Writer:TWriter);
    procedure ReadMinimum(Reader:TReader);
    procedure WriteMinimum(Writer:TWriter);
    procedure ReadValue(Reader:TReader);
    procedure WriteValue(Writer:TWriter);
    procedure DefineProperties(Filer: TFiler); override;
    function GetValue: variant; virtual;
    procedure SetValue(NewVal: variant); virtual;
    function IsValueStored: boolean; virtual;
    procedure SetIncrement(AVal: variant);
    function IsIncrementStored: boolean; virtual;
    procedure SetLarge_Increment(AVal: variant);
    function IsLarge_IncrementStored: boolean; virtual;
    procedure SetMinimum(AVal: variant);
    function IsMinimumStored: boolean; virtual;
    procedure SetMaximum(AVal: variant);
    function IsMaximumStored: boolean; virtual;
    procedure DoButton1; override;
    procedure DoButton2; override;
    procedure DoLargeUp;
    procedure DoLargeDown;
    //function DoTextChange: boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure DoValidateEdit(var ErrCode: integer); override;

{: @summary Contents of the control returned as a variant of type integer
or double.
<br>
@desc The control is setup to support spinediting both integer and floating
point values.  As such it parses the text contents of the control and forces
the result into a variant containing either an integer or a double.<br>
<br>
The control uses a special ExtractNumber function which is able to extract
a numeric value from the middle of a string of other characters.  This allows
the spin edit to operate with string values in which only part of the string
is the number - the number itself is altered, the characters outside the
number are left alone.}
    property Value: variant read GetValue write SetValue stored false;
{: @summary The amount to increment the control by with each button click.
<br>
@desc Clicking Button1 (the Up button) will add this amount to the Value.<br>
<br>
Clicking Button2 (the Down button) will subtract this amount to the value.}
    property Increment: variant read FIncrement write SetIncrement stored false;
{: @summary The amount to increment the control by when using the
VK_PRIOR (page up) and VK_NEXT (page down) keys.
<br>
@desc The keystrokes are hardcoded into the control and cannot be changed.
However you can disable this feature by setting Large_Increment to 0.  This
will let the keystrokes through for standard processing.}
    property Large_Increment: variant read FLarge_Increment
                                      write SetLarge_Increment
                                      stored false;
{: @summary Maximum value permitted in the control.
<br>
@desc The control is setup to prevent the Value property from exceeding this
value.  If a larger value is input it will be truncated back to this Maximum.}
    property Maximum: variant read FMaximum write SetMaximum
                                            stored false;
{: @summary Minimum value permitted in the control.
<br>
@desc The control is setup to prevent the Value property from being less than
this value.  If a smaller value is input it will be lifted back to this
Minimum.}
    property Minimum: variant read FMinimum write SetMinimum
                                            stored false;
{: @summary Whether the control should wrap its value when spinning.
<br>
@desc By default this is false, meaning that when the user clicks the spin
buttons (or uses the shortcut equivalent) then the value will stop changing
when it reaches the defined minimum and maximum values.<br>
<br>
If you set this property to true then when the you reach the maximum and click
the up button it will wrap back to the minimum value.  If you reach the
minimum and click the down button it will wrap around to the maximum value.}
    property WrapValue: boolean read FWrapValue write FWrapValue default false;
{: @summary Numeric values align to the Right by default.}
    property Alignment default taRightJustify;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  end;                                  // TCustomSpinEditEnh


{: @summary A derivation of TCustomEditEnh setup with 'spin buttons' to
allow the user to increment and decrement the numeric portion of the
control contents.
<br>
@desc Note that earlier versions of this control implemented a default
mask, but since true defaults for string properties cannot be achieved
I have removed the default mask - which makes the control more flexible.
However I still recommend that you install a simple mask such as
<code>
    !-999999990;0;0;0
</code><br>
with and EditMaskMode of eemNumeric (for integer values).<br>
<br>
The Value, Increment, LargeIncrement, Maximum and MinimumValues are now
defined as Variant (previously Integer).  But the values are forced to either
Integer or Double by internal processing.<br>
<br>
The Text and DisplayText properties have been made public not published since
the control is expected to be used with numeric values.  However the control
is capable of incrementing and decrementing numeric values embedded within
other text of the control. }
  TSpinEditEnh = class(TCustomSpinEditEnh)
    {$DEFINE PUBLISH_CustomSpinEditEnh}
    {$DEFINE PUBLISH_MaskEnh}
    {$INCLUDE Telesis_Publish.INC}
  end;

const
  DefaultMask_SpinEditEnh: string = '::NM::!-999999990;0;0;0';



implementation

uses Telesis_Locale,
  Telesis_Constants;

const
  SpinEdit_Default_Increment = 1;
  SpinEdit_Default_LargeIncrement = 100;
  SpinEdit_Default_Maximum = 999999999;
  SpinEdit_Default_Minimum = -999999999;
  SpinEdit_Default_Value = 0;



constructor TCustomSpinEditEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Alignment := taRightJustify;
  FIncrement := SpinEdit_Default_Increment;
  FLarge_Increment := SpinEdit_Default_LargeIncrement;
  FMaximum := SpinEdit_Default_Maximum;
  FMinimum := SpinEdit_Default_Minimum;
  SetValue( SpinEdit_Default_Value );
  // Prevent the caption from being assigned to the control
  ControlStyle := ControlStyle - [csSetCaption];
  //SetEditMaskEx(DefaultMask_SpinEditEnh);
end;


destructor TCustomSpinEditEnh.Destroy;
begin
  inherited Destroy;
end;


type
  TMyReader = class( TReader );

// Copied from Delphi5 Classes.pas subproc ReadVariantProp and
// modified for use with SpinEdit.  I am only worried about being
// able to retrieve numeric and string values.
function ReadVariantProperty( Reader: TReader ): Variant;
var
  ValType: TValueType;
begin
with TMyReader( Reader ) do
begin
  ValType := NextValue;
  case ValType of
    vaNil, vaNull:
      begin
        if ReadValue = vaNil then
          VarClear(Result) else
          Result := NULL;
      end;
    vaInt8, vaInt16, vaInt32: begin
      TVarData(Result).VType := varInteger;
      TVarData(Result).VInteger := ReadInteger;
      end;
    {$IFDEF __VCL40_OR_LATER__}
    vaSingle, vaExtended: begin
      TVarData(Result).VType := varDouble;
      TVarData(Result).VDouble := ReadFloat;
      end;
    {$ELSE}
    vaExtended: begin
      TVarData(Result).VType := varDouble;
      TVarData(Result).VDouble := ReadFloat;
      end;
    {$ENDIF}
    vaString, vaLString: begin
      TVarData(Result).VType := varString;
      Result := ReadString;
      end;
    {$IFDEF __VCL40_OR_LATER__}
    vaWString: begin
      TVarData(Result).VType := varOleStr;
      Result := ReadWideString;
      end;
    {$ENDIF}
  else
    raise Exception.Create( ENHERR_UNSUPPORTED_VARIANT );
  end;
end;
end;

// Simply to make the write functions easier.
procedure WriteVariantProperty( Writer: TWriter; Value: Variant );
var
  VType: Integer;
begin
  with Writer do begin
      VType := VarType(Value);
      case VType and varTypeMask of
      {$IFDEF __VCL40_OR_LATER__}
        varOleStr: WriteWideString(Value);
      {$ENDIF}
        varString: WriteString(Value);
        varByte, varSmallInt, varInteger: WriteInteger(Value);
      {$IFDEF __VCL40_OR_LATER__}
        varSingle: WriteSingle(Value);
      {$ENDIF}
        varDouble: WriteFloat(Value);
      end;
  end;
end;

function TCustomSpinEditEnh.GetButtonAlignDefault: TAlign;
begin
  Result := alRight;
end;

function TCustomSpinEditEnh.GetButtonSizeDefault: integer;
begin
  Result := 17;
end;

function TCustomSpinEditEnh.GetButtonStyleDefault: TEditEnhButtonStyle;
begin
  Result := eebUpDownArrow;
end;

function TCustomSpinEditEnh.GetButtonAutoRepeatDefault: boolean;
begin
  Result := true;
end;

function TCustomSpinEditEnh.GetButton1KeyDefault: TShortCut;
begin
  Result := TextToShortCut('Up');
end;

function TCustomSpinEditEnh.GetButton2KeyDefault: TShortCut;
begin
  Result := TextToShortCut('Down');
end;

procedure TCustomSpinEditEnh.ReadIncrement(Reader:TReader);
begin
  SetIncrement( ReadVariantProperty(Reader) );
end;

procedure TCustomSpinEditEnh.WriteIncrement(Writer:TWriter);
begin
  WriteVariantProperty( Writer, Increment );
end;

procedure TCustomSpinEditEnh.ReadLargeIncrement(Reader:TReader);
begin
  SetLarge_Increment( ReadVariantProperty(Reader) );
end;

procedure TCustomSpinEditEnh.WriteLargeIncrement(Writer:TWriter);
begin
  WriteVariantProperty( Writer, Large_Increment);
end;

procedure TCustomSpinEditEnh.ReadMaximum(Reader:TReader);
begin
  DisableChangeEvent;
  try
  SetMaximum( ReadVariantProperty(Reader) );
  finally
    // do NOT fire the event as a result of loading the value
    EnableChangeEvent( false );
  end;
end;

procedure TCustomSpinEditEnh.WriteMaximum(Writer:TWriter);
begin
  WriteVariantProperty( Writer, Maximum);
end;

procedure TCustomSpinEditEnh.ReadMinimum(Reader:TReader);
begin
  DisableChangeEvent;
  try
    SetMinimum( ReadVariantProperty(Reader) );
  finally
    // do NOT fire the event as a result of loading the value
    EnableChangeEvent( false );
  end;
end;

procedure TCustomSpinEditEnh.WriteMinimum(Writer:TWriter);
begin
  WriteVariantProperty( Writer, Minimum);
end;

procedure TCustomSpinEditEnh.ReadValue(Reader:TReader);
begin
  DisableChangeEvent;
  try
    SetValue( ReadVariantProperty(Reader) );
  finally
    // do NOT fire the event as a result of loading the value
    EnableChangeEvent( false );
  end;
end;

procedure TCustomSpinEditEnh.WriteValue(Writer:TWriter);
begin
  WriteVariantProperty( Writer, Value);
end;

procedure TCustomSpinEditEnh.DefineProperties(Filer: TFiler);
begin
  inherited; // allow base classes to define properties
  Filer.DefineProperty('IncrementEx', ReadIncrement, WriteIncrement, IsIncrementStored);
  Filer.DefineProperty('Large_IncrementEx', ReadLargeIncrement, WriteLargeIncrement, IsLarge_IncrementStored);
  Filer.DefineProperty('MaximumEx', ReadMaximum, WriteMaximum, IsMaximumStored);
  Filer.DefineProperty('MinimumEx', ReadMinimum, WriteMinimum, IsMinimumStored);
  Filer.DefineProperty('ValueEx', ReadValue, WriteValue, IsValueStored);
end;

//---------------------------------------------------------------------------

function TCustomSpinEditEnh.GetValue: variant;
begin
  Result := ExtractNumber( Text );
end;
//---------------------------------------------------------------------------

procedure TCustomSpinEditEnh.SetValue(NewVal: variant);
var
  ns, ne: integer;
  tmpVar, tmpOld: variant;
  tmpStr: string;
begin
  tmpVar := ExtractNumber( VariantToStr(NewVal) );
  if tmpVar > Maximum then
  begin
    if WrapValue then
      tmpVar := Minimum
    else
      tmpVar := Maximum;
  end;
  if tmpVar < Minimum then
  begin
    if WrapValue then
      tmpVar := Maximum
    else
      tmpVar := Minimum;
  end;
  tmpOld := ExtractNumberEx( Text, ns, ne );
  if tmpOld <> tmpVar then
  begin
    tmpStr := VariantToStr(tmpVar); // force back to string
    if ns > 1 then begin
      tmpStr := Copy( Text, 1, ns - 1 ) + tmpStr;              // copy prefix
      if ne <= Length(Text) then
        tmpStr := tmpStr + Copy( Text, ne, Length(Text) ); // copy postfix
    end;
    Text := tmpStr;
  end;
end;
//---------------------------------------------------------------------------
function TCustomSpinEditEnh.IsValueStored: boolean;
begin
  Result := (Value <> SpinEdit_Default_Value);
end;
//---------------------------------------------------------------------------
// We force the given value into either integer or float
procedure TCustomSpinEditEnh.SetIncrement(AVal: variant);
begin
  FIncrement := ExtractNumber( VariantToStr(AVal) );
end;
//---------------------------------------------------------------------------
function TCustomSpinEditEnh.IsIncrementStored: boolean;
begin
  Result := (Increment <> SpinEdit_Default_Increment);
end;
//---------------------------------------------------------------------------
// We force the given value into either integer or float
procedure TCustomSpinEditEnh.SetLarge_Increment(AVal: variant);
begin
  FLarge_Increment := ExtractNumber( VariantToStr(AVal) );
end;
//---------------------------------------------------------------------------
function TCustomSpinEditEnh.IsLarge_IncrementStored: boolean;
begin
  Result := (Large_Increment <> SpinEdit_Default_LargeIncrement);
end;
//---------------------------------------------------------------------------
// We force the given value into either integer or float
procedure TCustomSpinEditEnh.SetMinimum(AVal: variant);
var
  tmpVar: variant;
begin
  tmpVar := ExtractNumber( VariantToStr(AVal) );
  if tmpVar > FMaximum then
    tmpVar := FMaximum;
  if tmpVar <> FMinimum then
  begin
    FMinimum := tmpVar;
    SetValue(GetValue);
  end;
end;
//---------------------------------------------------------------------------
function TCustomSpinEditEnh.IsMinimumStored: boolean;
begin
  Result := (Minimum <> SpinEdit_Default_Minimum);
end;
//---------------------------------------------------------------------------
// We force the given value into either integer or float
procedure TCustomSpinEditEnh.SetMaximum(AVal: variant);
var
  tmpVar: variant;
begin
  tmpVar := ExtractNumber( VariantToStr(AVal) );
  if tmpVar < FMinimum then
    tmpVar := FMinimum;
  if tmpVar <> FMaximum then
  begin
    FMaximum := tmpVar;
    SetValue(GetValue);
  end;
end;
//---------------------------------------------------------------------------
function TCustomSpinEditEnh.IsMaximumStored: boolean;
begin
  Result := (Maximum <> SpinEdit_Default_Maximum);
end;
//---------------------------------------------------------------------------

procedure TCustomSpinEditEnh.DoButton1;
begin
  if EditCanModify then
  begin
    EditBeginModify;
    SetValue(Value + FIncrement);
  end;
  inherited;
end;
//---------------------------------------------------------------------------

procedure TCustomSpinEditEnh.DoButton2;
begin
  if EditCanModify then
  begin
    EditBeginModify;
    SetValue(Value - FIncrement);
  end;
  inherited;
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//function TCustomSpinEditEnh.DoTextChange: boolean;
//begin
//SetValue( GetValue );
//Result := inherited DoTextChange;
//end;

procedure TCustomSpinEditEnh.DoValidateEdit(var ErrCode: integer);
var
  tmpVar: variant;
begin
  if FInternalUpd = 0 then
  begin
    inherited DoValidateEdit(ErrCode);
    if ErrCode = E_EDITENH_NOERROR then
    begin
      tmpVar := ExtractNumber( Text );
      if tmpVar > Maximum then
        SetValue( Maximum )
      else
        if tmpVar < Minimum then
          SetValue( Minimum );
    end;
  end;
end;

procedure TCustomSpinEditEnh.KeyDown(var Key: Word; Shift: TShiftState);
begin
  // Up/Down are configured in TCustomEditEnh according to the keys
  // defined for Button1 and Button2.
  if FLarge_Increment <> 0 then
  begin
    if Key = VK_PRIOR then
    begin
      DoLargeUp;
      Key := 0;
    end
    else
      if Key = VK_NEXT then
      begin
        DoLargeDown;
        Key := 0;
      end;
  end;
  inherited;
end;

procedure TCustomSpinEditEnh.DoLargeDown;
begin
  if EditCanModify then
  begin
    EditBeginModify;
    SetValue(Value - FLarge_Increment);
  end;
end;

procedure TCustomSpinEditEnh.DoLargeUp;
begin
  if EditCanModify then
  begin
    EditBeginModify;
    SetValue(Value + FLarge_Increment);
  end;
end;

end.