{ Telesis_Utils.pas                               SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Various generic utilities.
<br>
@desc Note that this contains items previously found in
Telesis_EditUtil/CustomEditUtil as well as some others.<br>
<br>
<br>
<strong>History</strong><br>
<br>
22-Jan-2007 - The old IsMatch and matche code is both not-reliable and now
also out of date.  Use the Delphi6 MatchesMask method instead.<br>
<br>
21-Apr-2006 - ExtractDateTime changed so that input ADateTime parameter
is only changed if returning a valid (no error) value.<br>
<br>
06-Feb-2005 - Split ExtractDateTime into two parts.  The old ExtractDateTime
does what it always did but now calls ExtractDateTimeParts to do it.  This
gives us the function ExtractDateTimeParts that lets us get the various
bits and pieces of a date/time directly - allowing us to see what was set
and what was not.<br>
<br>
26-Feb-2004 - Check that XP Themes are enabled (thanks Dany Marmur).<br>
<br>
09-Oct-2003 - Support for XP Themes.<br>
<br>
30-Jan-2002 - Correction to detect invalid year in ExtractDateTime.<br>
<br>
24-Dec-2001 - Addition of Locale strings.  Addition of ExtractDateTime
and DateTimeToLayoutString functions.<br>
<br>
13-Aug-2001 - Replaced SameText function with CompareText to support
versions prior to VCL50.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_Utils;

{$INCLUDE Telesis.INC}

interface

uses SysUtils,
  Windows,
  Messages,
  Classes,
  Graphics,
  Controls,
  Forms,
  ImgList,
  Telesis_Constants;

type
{: @summary Type used to describe three states.
@desc Used with the LockKeyState function the type defines three states,
as an extension to the standard boolean true/false type.}
TBoolStateEnh = (bsOn, bsOff, bsNoChange);


{: @summary Specifies what type of button to display.
<br>
@desc Enumeration of the various types of buttons supported by EditEnh and
related edit controls.<br>
<br>
<br>
<br>
When only one button is displayed it is always "Button1".<br>
<br>
When two buttons are displayed,<br>
<br>
* Up-Down Buttons - top button is "Button1", bottom button is "Button2"<br>
<br>
* Left-Right Buttons - Left button is "Button1", Right button is "Button2"<br>
<br>
@enum eebNone
No button
<br>
@enum eebCustom
User Paint everything
<br>
@enum eebBlank
Single button, blank
<br>
@enum eebGlyph
Single button, use Glyph
<br>
@enum eebDownArrow
Single Down arrow button
<br>
@enum eebUpArrow
Single Up arrow button
<br>
@enum eebLeftArrow
Single Left arrow button
<br>
@enum eebRightArrow
Single Right arrow button
<br>
@enum eebUpDownBlank
Horizontal split, blank
<br>
@enum eebUpDownGlyph
Horizontal split, use Glyph
<br>
@enum eebUpDownArrow
Horizontal split, Up and Down arrow buttons
<br>
@enum eebLeftRightBlank
Vertical split, blank
<br>
@enum eebLeftRightGlyph
Vertical split, use Glyph
<br>
@enum eebLeftRightArrow
Vertical split, Left and Right arrow buttons
<br>
@enum eebEllipsis
Draw ellipsis (like a grid cell button gets)
<br>
@enum eebReturnKey
Draw a Return Key symbol
<br>
@enum eebLeftRightEllipsis
Vertical split, down-arrow on left, ellipsis on right
<br>
@enum eebCheckMark
Draw a Checkmark symbol
<br>
@enum eebQuery
Draw a Question symbol
}
TEditEnhButtonStyle = (
    eebNone,
    eebCustom,
    eebBlank,              // Start Single Buttons
    eebGlyph,
    eebDownArrow,
    eebUpArrow,
    eebLeftArrow,
    eebRightArrow,
    eebEllipsis,
    eebReturnKey,
    eebCheckMark,
    eebQuery,              // End Single Buttons
    eebUpDownBlank,        // Start Horz Split Buttons
    eebUpDownGlyph,
    eebUpDownArrow,        // End Horz Split Buttons
    eebLeftRightBlank,     // Start Vert Split Buttons
    eebLeftRightGlyph,
    eebLeftRightArrow,
    eebLeftRightEllipsis   // End Horz Split Buttons
    );


{: @summary Set of available Vector images.
<br>
@desc Vector images are predefine images that drawn using standard API lineto,
moveto functions.  As such they can be scaled larger or smaller without the
problems that can occur with regular bitmap glyphs.<br>
<br>
This set of images are defined for use with the drawing routines in the
Telesis_Utils unit.}
TVectorGlyphEnh = (
    vgNone,
    vgUpArrow,
    vgDownArrow,
    vgLeftArrow,
    vgRightArrow,
    vgEllipsis,
    vgReturnKey,
    vgCheckMark,
    vgQuery,
    vgWindowsClose,
    vgWindowsMaximise,
    vgWindowsMinimise,
    vgWindowsRestore,
    vgWindowsHelp,  // Win95 only?
    vgWindowsMenuCheck,
    vgWindowsMenuBullet,
    vgWindowsCheck,
    vgWindowsCheckChecked,
    vgWindowsCheckGrayed,
    vgWindowsRadio,
    vgWindowsRadioChecked,
    vgWindowsSizeGrip );


{: @summary Style of button border.
<br>
@desc This style information is used by the drawing routines in Telesis_Utils. }
TButtonBorderEnh = (
    bbCtl3D,
    bbShallow,
    bbFlat );


{: @summary EditEnh button states.
<br>
@desc The edit control supports three states relating to the use of the
button and related button/control keys - these are mostly for internal
checking but I make the state available for public reading in case it
is of interest.<br>
<br>

@enum eesNoButtonDown
No button down
<br>
@enum eesButton1Down
Key1, Top, Left or Only Button is Down
<br>
@enum eesButton2Down
Key2, Bottom or Right button is Down
<br>
@enum eesButton3Down
Key3 is down
<br>
}
  TEditEnhButtonState = (
    eesNoButtonDown,
    eesButton1Down,
    eesButton2Down,
    eesButton3Down
    );


{$IFNDEF __VCL50_OR_LATER__}
{: @summary Type to describe the image index in an imagelist.
<br>
@desc Only defined for VCL40 and before.  The type is already defined by
VCL50 and later.}
    TImageIndex = integer;
{$ENDIF}

{: @summary Structure for passing information to drawing functions.
<br>
@desc The various drawing functions in Telesis_Utils use this structure to
allow complex drawing instructions to be passed between functions without
needing many parameters etc.<br>
<br>
Its only a record so it is best to initialise all elements as desired.<br>
<br>
TestState is used tell the various functions what state in Pressed or Hover
should be tested.  eg:  For one of these utility functions to see Pressed
as true it tests "Pressed = TestState".  This is used, for example, by the
DrawEditEnhButton function to tell the other functions which button they are
currently painting.  For most other uses simply assign the Pressed and
TestState to eesButton1Down when you want the drawing as Pressed, otherwise
assign Pressed to eesNoButtonDown and TestState to eesButton1Down.  And similar
logic should be applied for the Hover attribute. }
TDrawInfoEnh = record
    VectorGlyph: TVectorGlyphEnh;
    Canvas: TCanvas;
    VectorColor: TColor;
    VectorBrush: TBrushStyle;
    BkColor: TColor;
    Rect: TRect;
    Border: TButtonBorderEnh;
    Proportional: boolean;
    Enabled: boolean;
    Stretch: boolean;
    Pressed: TEditEnhButtonState;
    Hover: TEditEnhButtonState;
    TestState: TEditEnhButtonState;
    ImageList: TCustomImageList;
    ImageIndex: TImageIndex;
    ImageDisabled: TImageIndex;
    ImagePressed: TImageIndex;
    end;


{: @summary Used to discover the next control in a TWinControl.
<br>
@desc I dont know why, but SelectNext and FindNextControl are *protected*
TWinControl functions which means I cannot just call either of these
functions to jump to the next control.  So what I have done is create
my own "FindNextControl" equivalent function which uses is essentially just
a copy of the code from TWinControl.  I have separated this from the edit
component and declared the function so that I can reuse it later if I want.}
function GetNextControl(ParControl, CurControl: TWinControl;
  GoForward, CheckTabStop, CheckParent: Boolean): TWinControl;


{: @summary Draw a button as described by TDrawInfoEnh.
<br>
@desc Procedure used to draw the inplace buttons on TCustomEditEnh and
derived controls.  Separated out so that it can also be used in other
circumstances - such as my own variations of TSpeedButton and TBitBtn.}
procedure DrawEditEnhButton( var DrawInfo: TDrawInfoEnh;
    BtnStyle: TEditEnhButtonStyle );

{: @summary Draw an image as described by TDrawInfoEnh.
<br>
@desc The advantage of this function and the draw functions used by this, is
that it reduces the number of bitmaps required in the system.  If you
use one of the TVectorGlyph images they are vector based and do not need
any bitmap or imagelist at all - AND they can be sized to any extent.}
procedure DrawStandardImage( var DrawInfo: TDrawInfoEnh );


{: @summary Draw a 3D border as described by TDrawInfoEnh.}
procedure Draw3DBorder( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a 3D border on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure Draw3DBorder( Cvs: TCanvas; var Rect: TRect; Shallow, Down: Boolean ); overload;

{: @summary Draw a polygon as described by TDrawInfoEnh.
<br>
@desc Draw the polygon defined by the give points and drawinfo.  The Width
and Height parameters control the sizing of a polygon within the given
rectangle, making it possible to draw multiple stretched polygons into
the same area and still maintain relative positioning. }
procedure DrawPolygon( var DrawInfo: TDrawInfoEnh; Points: array of TPoint;
    Width, Height: integer ); overload;
{: @summary Draw a Ploygon on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawPolygon( Cvs: TCanvas; Points: array of TPoint; var Rect: TRect;
    Proport, Stretch, Enabled: Boolean; Width, Height: integer ); overload;

{: @summary Draw a down arrow as described by TDrawInfoEnh.
<br>
@desc A down arrow similar to the symbol that appears on ComboBox buttons.}
procedure DrawDownArrow( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a down arrow on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawDownArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean ); overload;

{: @summary Draw an up arrow as described by TDrawInfoEnh.}
procedure DrawUpArrow( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw an up arrow on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawUpArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean ); overload;

{: @summary Draw a left arrow as described by TDrawInfoEnh.}
procedure DrawLeftArrow( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a left arrow on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawLeftArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean ); overload;

{: @summary Draw a right arrow as described by TDrawInfoEnh.}
procedure DrawRightArrow( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a right arrow on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawRightArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean ); overload;

{: @summary Draw a glyph as described by TDrawInfoEnh.}
procedure DrawGlyph( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a glyph on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawGlyph( Cvs: TCanvas; var Rect: TRect; Proport, Stretch: Boolean; ImgList: TImageList; ImgIdx: integer ); overload;

{: @summary Draw a bitmap as described by TDrawInfoEnh.
@desc Similar to DrawGlyph.  Note however this this procedure does NOT account
for Enabled status.  (But does manage stretch and proportional).  So it if you
want to draw a disabled image I suggest that you write your bitmap using the
Windows API DrawState function or some other method to make bitmap appear as
required in its disabled state before passing to this procedure.  }
procedure DrawBitmap( var DrawInfo: TDrawInfoEnh; Bitmap: TBitmap );

{: @summary Draw an Ellipsis (...) as described by TDrawInfoEnh.}
procedure DrawEllipsis( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw an ellipsis (...) on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawEllipsis( Cvs: TCanvas; var Rect: TRect; Stretch, Enabled: Boolean ); overload;

{: @summary Draw a return key (enter key) symbol as described by TDrawInfoEnh.}
procedure DrawReturnKey( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a return key symbol on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawReturnKey( Cvs: TCanvas; var Rect: TRect; Proport, Stretch, Enabled: Boolean ); overload;

{: @summary Draw a check mark (tick) as described by TDrawInfoEnh.}
procedure DrawCheckMark( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a check mark (tick) on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawCheckMark( Cvs: TCanvas; var Rect: TRect; Proport, Stretch, Enabled: Boolean ); overload;

{: @summary Draw a query symbol (question mark) as described by TDrawInfoEnh.}
procedure DrawQuery( var DrawInfo: TDrawInfoEnh ); overload;
{: @summary Draw a query symbol (question mark) on canvas.
@deprecated Please use TDrawInfoEnh function instead.}
procedure DrawQuery( Cvs: TCanvas; var Rect: TRect; Proport, Stretch, Enabled: Boolean ); overload;


{: @summary Calculate appropriate text color from supplied background color.
<br>
@desc Attempts to determine the most appropriate text color from the given
background color (BkColor).  This is done by a simple test against the
available system color definitions (if BkColor=clWindow return
clWindowText etc).
<br>
<br>If no match is found in the standard system colors (those colors that
have a *Text definition, then a weighting calculation will attempt to
return either clWhite or clBlack according to the brightness of the color.}
function TextColorFromColor( BkColor: TColor ): TColor;

{: @summary Returns an integer indicating the relative brightness of a color.
<br>
@desc Used internally by TextColorFromColor to return an integer indicating
the relative "brightness" of a color, I decided to make this function
more generally available in case it has other uses.  It will return a
value between 0 and ColorBrightness(clWhite). }
function ColorBrightness( BkColor: TColor ): integer;

{: @summary Check for exponent definition in string.
<br>
@desc This function is used to check for exponent definitions in a string
(which is assumed to be a valid numeric string according to the usual
VCL/Pascal ffGeneric number formatting rules) and if supplied the value will
be reformatted to remove the exponent.
<br>
<br>Exponents are checked by simply looking for 'E' or 'e' in the string.
If neither of these characters are found then the parameter value is simply
returned as it is.  If either of these characters are found then the value
if parsed into an Extended value using StrToFloat.  An EConvertError may
result if the string is not acceptable - see VCL help for StrToFloat.
<br>
<br>The reformatting is done by parsing the extended value into
the FormatFloat function with a format string of
"#################0.0#################".  The parameter MaxIntDigit tells
the function the maximum number of acceptable digits prior to the decimal
point, this function will raise an exception if more digits that this
are required by the AValue parameter given.
<br>
<br>This is called by TMaskEnh.SetMaskedText when MaskMode = eemNumeric to
ensure that exponents (which are not supported by numeric masks) are removed
from from the string prior to processing. }
function ExponentFilter( AValue: string; MaxIntDigit: integer ): string;

{: @summary Character case conversion supporting proper case.
@desc This is the function used by TMaskEnh, TCustomEditEnh and derived
controls to check/convert input characters to particular case.  PrevChar is the
character preceeding the NewChar in the destination string.  ChChase specifies
the case to which you want NewChar converted. }
function ConvCharCase( PrevChar, NewChar: char; ChCase: integer ): char;



{: @summary Convert given string into a name acceptable to Delphi.
<br>
@desc Leading numerics are dropped, all special characters converted
to '_'.<br>
<br>
A blank parameter will be returned as 'x'.}
function MakeDelphiName(AName: string): string;

{: @summary Center the given TControl inside its parent.
<br>
@desc If the control is too large to fit inside the parent then the TopLeft
parameter is used to determine if the control should still be centred or
placed against the top left corner of the parent.}
procedure CenterInParent(Control: TControl; TopLeft: boolean = true);

{: @summary Center the given TControl inside the given rectangle.
<br>
@desc If the control is too large to fit inside the rectangle then the TopLeft
parameter is used to determine if the control should still be centred or
placed against the top left corner.}
procedure CenterInRect(Control: TControl; const Rect: TRect;
  TopLeft: boolean = true);

{: @summary Activate the form and focus first control.
<br>
@desc Attempts to do a "smart" activation of a form.  It checks whether the
form is visible and can be focused, if so it attempts to set focus to the first
available TWinControl on the form.}
function ActivateForm(AForm: TCustomForm): boolean;

{: @summary Determine if given Control is a child window.
<br>
@desc Uses the Windows API to determine if the wincontrol has the child window
attribute set.  This avoids having to rely on VCL "parent" property which may
not always be accurate.}
function IsChildWindow(Control: TWinControl): boolean;

{: @summary Convert a component to its string representation.}
function ComponentToString(Component: TComponent): string;

{: @summary Initialise AInstance from a string representation of a component.}
function StringToComponent(Value: string; AInstance: TComponent): TComponent;

{: @summary Read a stream to initialise a component.}
function StreamToComponent(Stream: TStream; AInstance: TComponent): TComponent;

{: @summary Save a string to file.
<br>
@desc Just a convenient way to automatically save a string value, including
the ability to show a save dialog (if Prompt parameter is true).}
function SaveStringToFile(AValue: string; var FileName: string;
  Prompt: boolean): boolean;

{: @summary Read a string from file.
<br>
@desc Just a convenient way to automatically read a string value from file,
including the ability to show an open dialog (if Prompt parameter is true). }
function ReadStringFromFile(var FileName: string; Prompt: boolean): string;

{: @summary Find a component in another component.
<br>
@desc Much like the TComponent.FindComponent function but looks through all
children, grandchildren etc of the AOwner parameter.}
function FindComponentIn(AOwner: TComponent; AName: string): TComponent;

{: @summary Reads the current keyboard state information.
<br>
@desc Useful when not inside a keyboard handler where such information
has been supplied.}
function GetShiftKeyStates: TShiftState;

{: @summary Read or Alter lock key states.
<br>
@desc "Lock Keys" include VK_CAPITAL, VK_NUMLOCK, VK_SCROLL.  This function
allows you to turn the key on/off or simply read the current value by
supplying the State parameter as bsNoChange. }
function LockKeyState(Key: integer; State: TBoolStateEnh): boolean;

{: @summary Read the current work area of the desktop.
<br>
@desc Excludes the taskbar etc. }
function GetScreenWorkArea: TRect;

{: @summary Invalidate the control AND all of is child controls.}
procedure InvalidateAll(Control: TWinControl);

{: @summary Update the control AND all of is child controls.}
procedure UpdateAll(Control: TWinControl);

{: @summary Repaint the control AND all of is child controls.}
procedure RepaintAll(Control: TWinControl);

{: @summary Retrieve the value of the named environment variable.}
function GetEnvironmentValue(VarName: string): string;

{: @summary Get the network name of the current workstation.}
function GetWorkstationName: string;

{: @summary Get the Microsoft Terminal Server session name.
<br>
@desc Returns 'CONSOLE' if not found (or not a Terminal Server) or if
the session is running at the terminal server console.}
function GetSessionName: string;

{: @summary Return interval value as string.
@desc The supplied Interval parameter is expected to be the difference
between two TDateTime values.  This function will return a string
interpretation of this value in the form:
<br>
d h:m:s.p<br>
<br>
where 'd' = the number of days<br>
'h' = the number of hours<br>
'm' = the number of minutes<br>
's' = the number of seconds<br>
'p' = points of seconds.<br>
<br>
the '.p' (points of seconds) part is only included if the IncPts parameter
is true.
}
function TimeIntervalString( Interval: double; IncPts: boolean = true ): string;



{: @summary A simplified version of ExtractNumberEx.
<br>
@desc You can use this to avoid needing to provide the Var parameters that
ExtractNumberEx uses to return the positions within the string.  (Note that
this is simply a routine that declares the required var parameters internally
and calls ExtractNumberEx.) }
function ExtractNumber( StrVal: string ): variant;

{: @summary Extract the numeric part of a string, ignoring non-numeric parts.
<br>
@desc A possibly useful routine to extract the numeric portion of a string,
returning the number value as a variant and setting the two var parameters
as the start and end positions if the number in the string.  That is;
NumStart is set to the first numeric character, NumEnd is set to the position
immediately after the last numeric character. }
function ExtractNumberEx( StrVal: string; var NumStart, NumEnd: integer ): variant;

{: @summary A "safe" way to convert Variant to String.
<br>
@desc Use this in preference to direct casting or the VarToStr VCL function.
<br><br>
This procedure simply detects the problem types and forces the conversion
using the traditional VCL to string routines to produce a consistent result.
(That is, a result using the Delphi global variables, not the operating system
definitions which may be different.)
<br><br>
Notes:
<br><br>
With standard VCL processing a Variant is converted to string via OLE
functions - which use Windows regional settings and not Delphi globals
(such as DecimalSeparator).  This can lead to problems when the operating
system regional settings do not match the application global settings.
While this should be rare it does sometimes happen during development of
products for multiple regions.
<br><br>
Also the problem can happen in the reverse.  If you assign a numeric value
as a string to a variant (such as assigning from the contents of an edit
control) then the variant will continue to store it as a string.  However
if you then try to assign the variant to an integer, float or date data-type
then the OLE conversion will attempt to convert using operating system
regional settings.  Take the following for example...
<br>
<br>    TmpVariant := '10000.12'; // 1. TmpVariant becomes a string data type
<br>    TmpFloat := TmpVariant;   // 2. Successful conversion to integer!
<br>    TmpFloat := StrToFloat( TmpVariant ); // 3. May fail
<br><br>
Line 3 may fail, or give an invalid value, if the operating system settings
are different to the global variables ThousandSeparator and DecimalSeparator.
<br><br>
It is probably relevant to note at this point that the OLE will convert string
representations of numbers that contain thousand separator characters. (Again
using the operating system definition for the thousand separator.)  Whereas
the StrToFloat and StrToInt functions will raise an exception with such
strings.  You can avoid this difficulty by using the ExtractNumber and
ExtractNumberEx functions which extract numeric values from strings and
return a variant of type integer or double.
<br><br>
There is no way to easily provide a generic function to manage the assignment
of values to variant (other than what is already provided).  The key to
minimising problems is to make sure you assign the value using the most
appropriate datatype.  For example, dont assign a string to a variant and
expect to be able to assign that variant back to a non-string datatype.
<br><br>
For example the following is not advisable...
<br>    TmpVariant := '10000.12';
<br>    TmpFloat := TmpVariant;
<br><br>
Instead use something like...
<br>    TmpVariant := StrToFloat('10000.12');
<br>    TmpFloat := TmpVariant;
<br><br>
This ensures that the variant is carrying data in the most appropriate form.
You can then use this VariantToString function if you need to retrieve the
value as a string reliably.
}
function VariantToStr( VarVal: variant ): string;


{: @summary Convert DateTime formatting string to ExtractDateTime Layout
parameter string
<br>
@desc This function can be used to convert a DateTime formatting string (see
VCL online help for FormatDateTime) into a string compatible with the
ExtractDateTime function.<br>
<br> Refer to the ExtractDateTime function for details of the Layout string.}
function DateTimeFormatToLayout( AFormat: string ): string;
{: @summary Extract the various possible parts of date/time information.
<br>
@desc Used by ExtractDateTime to build date/time info from freeform input.
Provided as a separate function to allow the the specific information to
be isolated if desired.<br>
<br> Refer to the ExtractDateTime function for details.}
function ExtractDateTimeParts(ALayout, AText: string;
    var RDayDate, RMnth, RYear, RHour, RMinute, RSecond, RPoints: word): integer;

{: @summary Extract a TDateTime from the given AText string, based on the
given ALayout parameter string.
<br>
@desc ALayout is a string of characters provided to give the order of the
date/time elements contained within AText.  The characters used by ALayout
mostly correspond to those used by the VCL FormatDateTime function, but
ALayout should contain only these characters (no formatting or delimiter
characters) and there should only be one instance of each character in
ALayout.<br>
<br>
The function does not care what delimiters are used in AText, or even if
there are no delimiters, provided there is a way to identify specific
separate values from the input string.  Examples showing how this works
given later.<br>
<br>
ALayout accepts the following characters in any order (the order determines
the order expected within the AText parameter)...<br>
d or D = day date (numeric)<br>
m or M = month (numeric or string)<br>
y or Y = year (numeric)<br>
h or H = hour (12 or 24 hour)<br>
n or N = minute (numeric)<br>
s or S = second (numeric)<br>
z or Z = points of second (numeric)<br>
a or A = am/pm indicator (if specified input is expected in 12hr format)<br>
x or X = day name.  Day name is always accepted in the first position<br>
but to accept in any other position use 'x'<br>
<br>
Use lower case values from the above characters indicate that the value
is optional (may not appear in the string).  This is not to say that it
will skip to the next element in the layout, but only that the AText input
need only provide enough information for the initial (uppercase) inputs.
If there is insufficient inputs in AText to fill all upper case layout
characters from the above set, then an error will be returned.  If there
is sufficient information to fill all upper case instances, but not lower
case instances, then the lower cases instances not provided will be
defaulted as follows...<br>
d = todays date<br>
m = current month<br>
y = current year<br>
h = 0<br>
n = 0<br>
s = 0<br>
z = 0<br>
a = am (accepts hours in 12 or 24 hour format, if am/pm not given
then presumes 24hour time format)<br>
x = (not defined)<br>
<br>
You can prefix the layout string with either an 'E' or 'e' character.
If this prefixing character is not provided then processing will only check
against current locale strings to match month and day names etc.
If you prefix the layout with 'E' then only english strings will be matched.
If you prefix the layout with 'e' then the process will try to match the
current locale first, and if not found will then attempt to match with
english strings.<br>
<br>
Function also accepts the following special values for AText (and ignores
the ALayout value if found)...<br>
today               (todays date, time = 0 or as read)<br>
tomorrow            (tomorrows date, time = 0 or as read)<br>
yesterday           (yesterdays date, time = 0 or as read)<br>
now                 (current date and time)<br>
none                (set zero)<br>
<br>
these special values can be abbreviated to the first three characters
(as described in ExtractDT_ShortStrings constant).<br>
<br>
The special keyword values must be the first word of AText.<br>
<br>
You may enter the name of a day of the week as the first word in the string.
If you provide no other value after the day name then the function will
return the date next 'day name'.  If you enter the day name and a valid
date, then the function will validate that the day name and the date
correspond.<br>
<br>
You can automatically generate a valid ALayout by calling the function
DateTimeFormatToLayout, which will compress the given datetime format
string into a layout value applicable to ExtractDateTime.  eg:<br>
ErrorCode := ExtractDateTime(<br>
    DateTimeFormatToLayout( 'dd/mm/yyyy hh:nn:ss' ),<br>
    Edit1.Text,<br>
    ADateTimeVar );<br>
<br>
Like the VCL FormatDateTime function, the DateTimeFormatToLayout function
will accept input time formats as "hh:nn" or "hh:mm" - so that when "m"
immediately follows "h" it will be interpretted as minutes and not months.<br>
<br>
After all required information is extracted from the string, and length
of string remaining is ignored.<br>
<br>
Return code should be E_EDITENH_NOERROR if the function was able to extract
a valid datetime value from AText.  If a problem occurs then one of the
following integer error codes are returned to indicate the probable
problem...<br>
  E_DATEENH_VALUE_NOT_FOUND (required entry value not found)<br>
  E_DATEENH_INVALID_LAYOUT  (invalid or blank Layout string)<br>
  E_DATEENH_INPUT_TOO_SHORT (required input values not completed)<br>
  E_DATEENH_NO_AM_OR_PM (no am/pm input and 'A' specified)<br>
  E_DATEENH_INVALID_AM_TIME (time invalid for AM entry)<br>
  E_DATEENH_INVALID_PM_TIME (time invalid for PM entry)<br>
  E_DATEENH_INVALID_DATE (invalid date input, eg wrong # of days)<br>
  E_DATEENH_MISMATCHED_DAYNAME (input of day name did not match date)<br>
  E_DATEENH_INVALID_DAY_DATE (the day date number was invalid)<br>
  E_DATEENH_INVALID_MONTH (the month number or str was invalid)<br>
  E_DATEENH_INVALID_YEAR (the year number was invalid)<br>
  E_DATEENH_INVALID_HOUR (the hour number was invalid)<br>
  E_DATEENH_INVALID_MINUTE (the minute number was invalid)<br>
  E_DATEENH_INVALID_SECOND (the second number was invalid)<br>
  E_DATEENH_INVALID_POINTS (the points of seconds number was invalid)<br>
  E_DATEENH_INVALID_AMPM (the am/pm string was invalid)<br>
<br>
These error codes can be translated into text (for exception processing) using
the GetEditEnhErrorMsg function.<br>
<br>
<br>
Examples<br>
<br>
To support input of dates in American format use a layout of...<br>
'mdy'<br>
<br>
To support input of dates in European format use a layout of...<br>
'dmy'<br>
}
function ExtractDateTime(ALayout, AText: string; var ADateTime: TDateTime): integer;


var
{: @summary Have the Enhanced utility function use XP themes when drawing.
<br>
@desc Defaults to true so that XP themes will be used if available.  Option
is ignored if XP themes are not available.}
  EnhUtils_UseXPThemes: boolean;

  function UseXPThemes: boolean;

implementation

uses Dialogs, Telesis_Locale
  {$IFDEF SUPPORT_XP_THEMES}
  , Themes
  {$ENDIF};


function GetNextControl(ParControl, CurControl: TWinControl;
  GoForward, CheckTabStop, CheckParent: Boolean): TWinControl;
var
  I, StartIndex: Integer;
  List: TList;
begin
Result := nil;
List := TList.Create;
try
    ParControl.GetTabOrderList(List);
    if List.Count > 0 then begin
        StartIndex := List.IndexOf(CurControl);
        if StartIndex = -1 then
            if GoForward then
                StartIndex := List.Count - 1
            else StartIndex := 0;
        I := StartIndex;
        repeat
            if GoForward then begin
                Inc(I);
                if I = List.Count then
                    I := 0;
                end
            else begin
                if I = 0 then
                    I := List.Count;
                    Dec(I);
                    end;
            CurControl := List[I];
            if CurControl.CanFocus and
                (not CheckTabStop or CurControl.TabStop) and
                (not CheckParent or (CurControl.Parent = ParControl)) then
                Result := CurControl;
            until (Result <> nil) or (I = StartIndex);
        end;
finally
    List.Destroy;
    end;
end;


procedure DrawEditEnhButton( var DrawInfo: TDrawInfoEnh;
    BtnStyle: TEditEnhButtonStyle );
var
    TmpRect: TRect;
begin
// Ignore custom and none button styles
if BtnStyle <= eebCustom then begin
    Exit;
    end
// Single Button drawing...
else if BtnStyle <= eebQuery then begin
    with DrawInfo do begin
        case BtnStyle of
            eebBlank, eebGlyph:   VectorGlyph := vgNone;
            eebDownArrow:         VectorGlyph := vgDownArrow;
            eebUpArrow:           VectorGlyph := vgUpArrow;
            eebLeftArrow:         VectorGlyph := vgLeftArrow;
            eebRightArrow:        VectorGlyph := vgRightArrow;
            eebEllipsis:          VectorGlyph := vgEllipsis;
            eebReturnKey:         VectorGlyph := vgReturnKey;
            eebCheckMark:         VectorGlyph := vgCheckMark;
            eebQuery:             VectorGlyph := vgQuery;
        else                      VectorGlyph := vgNone;
            end;
        TestState := eesButton1Down;
        end;
    Draw3DBorder( DrawInfo );
    DrawStandardImage( DrawInfo );
    end
// Horizontal Split (up/down) buttons
else if BtnStyle <= eebUpDownArrow then begin
    with DrawInfo do begin
        // Split the given rectangle horizontally
        TmpRect.Top := Rect.Top + ((Rect.Bottom - Rect.Top) div 2);
        TmpRect.Left := Rect.Left;
        TmpRect.Right := Rect.Right;
        TmpRect.Bottom := Rect.Bottom;
        Rect.Bottom := TmpRect.Top;
        // Keep the two rects the same height!
        if (TmpRect.Bottom - TmpRect.Top) > (Rect.Bottom - Rect.Top) then
            TmpRect.Top := TmpRect.Top + 1;
        // Assign the appropriate style for the first button
        case BtnStyle of
            eebUpDownArrow:       VectorGlyph := vgUpArrow;
        else                      VectorGlyph := vgNone;
            end;
        TestState := eesButton1Down;
        end;
    Draw3DBorder( DrawInfo );
    DrawStandardImage( DrawInfo );
    with DrawInfo do begin
        // Now prepare for the second button
        Rect := TmpRect;
        // Assign the appropriate style for the first button
        case BtnStyle of
            eebUpDownArrow:       VectorGlyph := vgDownArrow;
        else                      VectorGlyph := vgNone;
            end;
        TestState := eesButton2Down;
        // Adjust the image index as +1 for the second button
        if (BtnStyle = eebUpDownGlyph) and assigned(ImageList) then begin
            if (ImageIndex >=0) and (ImageIndex < (ImageList.Count - 1)) then
                Inc(ImageIndex);
            if (ImagePressed >=0) and (ImagePressed < (ImageList.Count - 1)) then
                Inc(ImagePressed);
            if (ImageDisabled >=0) and (ImageDisabled < (ImageList.Count - 1)) then
                Inc(ImageDisabled);
            end;
        end;
    Draw3DBorder( DrawInfo );
    DrawStandardImage( DrawInfo );
    end
// Vertical Split (left/right) buttons
else if BtnStyle <= eebLeftRightEllipsis then begin
    with DrawInfo do begin
        // Split the given rectangle horizontally
        TmpRect.Left := Rect.Left + ((Rect.Right - Rect.Left) div 2);
        TmpRect.Top := Rect.Top;
        TmpRect.Right := Rect.Right;
        TmpRect.Bottom := Rect.Bottom;
        Rect.Right := TmpRect.Left;
        // Keep the two rects the same Width!
        if (TmpRect.Right - TmpRect.Left) > (Rect.Right - Rect.Left) then
            TmpRect.Left := TmpRect.Left + 1;
        // Assign the appropriate style for the first button
        case BtnStyle of
            eebLeftRightArrow:    VectorGlyph := vgLeftArrow;
            eebLeftRightEllipsis: VectorGlyph := vgDownArrow;
        else                      VectorGlyph := vgNone;
            end;
        TestState := eesButton1Down;
        end;
    Draw3DBorder( DrawInfo );
    DrawStandardImage( DrawInfo );
    with DrawInfo do begin
        // Now prepare for the second button
        Rect := TmpRect;
        // Assign the appropriate style for the first button
        case BtnStyle of
            eebLeftRightArrow:    VectorGlyph := vgRightArrow;
            eebLeftRightEllipsis: VectorGlyph := vgEllipsis;
        else                      VectorGlyph := vgNone;
            end;
        TestState := eesButton2Down;
        // Adjust the image index as +1 for the second button
        if (BtnStyle = eebLeftRightGlyph) and assigned(ImageList) then begin
            if (ImageIndex >=0) and (ImageIndex < (ImageList.Count - 1)) then
                Inc(ImageIndex);
            if (ImagePressed >=0) and (ImagePressed < (ImageList.Count - 1)) then
                Inc(ImagePressed);
            if (ImageDisabled >=0) and (ImageDisabled < (ImageList.Count - 1)) then
                Inc(ImageDisabled);
            end;
        end;
    Draw3DBorder( DrawInfo );
    DrawStandardImage( DrawInfo );
    end;
end;


procedure DrawStandardImage( var DrawInfo: TDrawInfoEnh );
var
  WType, WFlags, C1: integer;
  TmpColor: TColor;
  {$IFDEF SUPPORT_XP_THEMES}
  Details : TThemedElementDetails;
  Offset: integer;
  {$ENDIF}
begin
  {$IFDEF SUPPORT_XP_THEMES}
  if UseXPThemes then
  begin
    Offset := 0;
    with DrawInfo do
    begin
      if not Enabled then
        Offset := 3
      else if Pressed = TestState then
        Offset := 2
      else if Hover = TestState then
        Offset := 1;
      case VectorGlyph of
        vgUpArrow:
          begin
            // ScrollBar buttons cant shrink much before the arrow almost
            // disappears.  So to make spin buttons look better we use the
            // TThemedSpin theme - but we inflate the rect, because for some
            // reason they are normally painted smaller than the specified rect.
            if (Rect.Bottom - Rect.Top) < 13 then
            begin
              Details := ThemeServices.GetElementDetails(
                TThemedSpin(Ord(tsUpNormal) + Offset));
              InflateRect(Rect, 1, 1);
            end
            else
              Details := ThemeServices.GetElementDetails(
                TThemedScrollBar(Ord(tsArrowBtnUpNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgDownArrow:
          begin
            if (Rect.Bottom - Rect.Top) < 13 then
            begin
              Details := ThemeServices.GetElementDetails(
                TThemedSpin(Ord(tsDownNormal) + Offset));
              InflateRect(Rect, 1, 1);
            end
            else
              Details := ThemeServices.GetElementDetails(
                TThemedScrollBar(Ord(tsArrowBtnDownNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgLeftArrow:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedScrollBar(Ord(tsArrowBtnLeftNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgRightArrow:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedScrollBar(Ord(tsArrowBtnRightNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgEllipsis:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedScrollBar(Ord(tsThumbBtnHorzNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
            DrawEllipsis( DrawInfo );
          end;
        vgReturnKey:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedScrollBar(Ord(tsThumbBtnHorzNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
            DrawReturnKey( DrawInfo );
          end;
        vgCheckMark:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedScrollBar(Ord(tsThumbBtnHorzNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
            DrawCheckMark( DrawInfo );
          end;
        vgQuery:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedScrollBar(Ord(tsThumbBtnHorzNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
            DrawQuery( DrawInfo );
          end;
        vgWindowsClose:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedWindow(Ord(twCloseButtonNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsMaximise:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedWindow(Ord(twMaxButtonNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsMinimise:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedWindow(Ord(twMinButtonNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsRestore:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedWindow(Ord(twRestoreButtonNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsHelp:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedWindow(Ord(twHelpButtonNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsMenuCheck:
          begin
            Details := ThemeServices.GetElementDetails(tmMenuItemSelected);
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsMenuBullet:
          begin
            Details := ThemeServices.GetElementDetails(tmMenuItemDemoted);
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsCheck:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedButton(Ord(tbCheckBoxUncheckedNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsCheckChecked:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedButton(Ord(tbCheckBoxCheckedNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsCheckGrayed:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedButton(Ord(tbCheckBoxMixedNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsRadio:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedButton(Ord(tbRadioButtonUncheckedNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsRadioChecked:
          begin
            Details := ThemeServices.GetElementDetails(
              TThemedButton(Ord(tbRadioButtonCheckedNormal) + Offset));
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
        vgWindowsSizeGrip:
          begin
            Details := ThemeServices.GetElementDetails(tsGripperPane);
            ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
          end;
      else // including vgNone,          ;
        Details := ThemeServices.GetElementDetails(
          TThemedScrollBar(Ord(tsThumbBtnHorzNormal) + Offset));
        ThemeServices.DrawElement(Canvas.Handle, Details, Rect);
      end;
    end;
  end
  else
  {$ENDIF}
  begin
    if DrawInfo.Pressed = DrawInfo.TestState then
    begin // slightly offset image
      Inc(DrawInfo.Rect.Top);
      Inc(DrawInfo.Rect.Left);
    end;
    // If no VectorGlyph specified check for and draw the appropriate
    // imagelist item
    if DrawInfo.VectorGlyph = vgNone then
    begin
      if DrawInfo.ImageList <> nil then
        DrawGlyph( DrawInfo );
    end
    // Otherwise select and draw the necessary vector glyphs...
    else with DrawInfo do
    begin
      if VectorGlyph < vgWindowsClose then
      begin
        case VectorGlyph of
          vgUpArrow:    DrawUpArrow( DrawInfo );
          vgDownArrow:  DrawDownArrow( DrawInfo );
          vgLeftArrow:  DrawLeftArrow( DrawInfo );
          vgRightArrow: DrawRightArrow( DrawInfo );
          vgEllipsis:   DrawEllipsis( DrawInfo );
          vgReturnKey:  DrawReturnKey( DrawInfo );
          vgCheckMark:  DrawCheckMark( DrawInfo );
          vgQuery:      DrawQuery( DrawInfo );
        end;
      end
      else
      begin
        case VectorGlyph of
            vgWindowsClose: begin
                WType := DFC_CAPTION;
                WFlags := DFCS_CAPTIONCLOSE;
                end;
            vgWindowsMaximise: begin
                WType := DFC_CAPTION;
                WFlags := DFCS_CAPTIONMAX;
                end;
            vgWindowsMinimise: begin
                WType := DFC_CAPTION;
                WFlags := DFCS_CAPTIONMIN;
                end;
            vgWindowsRestore: begin
                WType := DFC_CAPTION;
                WFlags := DFCS_CAPTIONRESTORE;
                end;
            vgWindowsHelp: begin
                WType := DFC_CAPTION;
                WFlags := DFCS_CAPTIONHELP;  // Win95 Only?
                end;
            vgWindowsMenuCheck: begin
                WType := DFC_MENU;
                WFlags := DFCS_MENUCHECK;
                end;
            vgWindowsMenuBullet: begin
                WType := DFC_MENU;
                WFlags := DFCS_MENUBULLET;
                end;
            vgWindowsCheck: begin
                WType := DFC_BUTTON;
                WFlags := DFCS_BUTTONCHECK;
                end;
            vgWindowsCheckChecked: begin
                WType := DFC_BUTTON;
                WFlags := DFCS_BUTTONCHECK or DFCS_CHECKED;
                end;
            vgWindowsCheckGrayed: begin
                WType := DFC_BUTTON;
                WFlags := DFCS_BUTTON3STATE or DFCS_CHECKED;
                end;
            vgWindowsRadio: begin
                WType := DFC_BUTTON;
                WFlags := DFCS_BUTTONRADIO;
                end;
            vgWindowsRadioChecked: begin
                WType := DFC_BUTTON;
                WFlags := DFCS_BUTTONRADIO or DFCS_CHECKED;
                end;
            vgWindowsSizeGrip: begin
                WType := DFC_SCROLL;
                WFlags := DFCS_SCROLLSIZEGRIP;
                end;
        else WType := DFC_CAPTION;
             WFlags := DFCS_CAPTIONCLOSE;
            end;
        //if Border = bbFlat then
        // Always flat - draw out own borders using Draw3DBorder or whatever
            WFlags := WFlags or DFCS_FLAT;
        if not Enabled then
            WFlags := WFlags or DFCS_INACTIVE;
        if Pressed = TestState then
            WFlags := WFlags or DFCS_PUSHED;
        DrawFrameControl( Canvas.Handle, Rect, WType, WFlags );
        // if Caption button - redraw the rectangle in the background
        // color to clear the rectangle drawn by Windows...
        if WType = DFC_CAPTION then begin
            //Canvas.FloodFill( Rect.Left, Rect.Top, BkColor, fsSurface );
            TmpColor := Canvas.Brush.Color;
            Canvas.Brush.Color := BkColor;
            Canvas.FrameRect( Rect );
            Canvas.Brush.Color := TmpColor;
            end;
        // If we are given a VectorColor different to BtnText
        // (Which appears to be the color used for the drawings)
        // then fill the given images with the requested color.
        // Note that this is not really recommended (pixel access if
        // far from efficient) and not totally comprehensive, since
        // I simply do a single scan down the vertical center looking
        // for BtnText color to fill with VectorColor.  It seems to
        // work OK for all access the ScrollSizeGrip.
        C1 := ColorToRGB( clBtnText );
        if ColorToRGB( VectorColor ) <> C1 then begin
            TmpColor := Canvas.Brush.Color;
            Canvas.Brush.Color := VectorColor;
            // Reusing WType and WFlags as iterator controls
            WType := Rect.Left + ((Rect.Right - Rect.Left) div 2);
            for WFlags := Rect.Top to Pred(Rect.Bottom) do begin
                if ColorToRGB(Canvas.Pixels[WType,WFlags]) = C1 then
                    Canvas.FloodFill( WType, WFlags, C1, fsSurface );
                end;
            Canvas.Brush.Color := TmpColor;
            end;
        end;
      end;
    end;
end;


// Draws a 3D border within the give rectangle, and reduces the rectangle
// accordingly...
procedure Draw3DBorder( var DrawInfo: TDrawInfoEnh );
var
    TmpBrushColor, TmpPenColor: TColor;
begin
if DrawInfo.Border = bbFlat then
  Exit;
with DrawInfo, DrawInfo.Canvas do begin
    TmpBrushColor := Brush.Color;
    TmpPenColor := Pen.Color;
    Pen.Width := 1;
    Rect.Right := Rect.Right - 1;
    Rect.Bottom := Rect.Bottom - 1;
    if Border = bbShallow then begin
        if Pressed = TestState then begin
            Pen.Color := clBtnShadow;
            MoveTo( Rect.Left, Rect.Bottom );
            LineTo( Rect.Left, Rect.Top );
            LineTo( Rect.Right, Rect.Top );
            Pen.Color := clBtnHighLight;
            MoveTo( Rect.Right, Rect.Top );
            LineTo( Rect.Right, Rect.Bottom );
            LineTo( Rect.Left - 1, Rect.Bottom );
            end
        else begin
            Pen.Color := clBtnHighlight;
            MoveTo( Rect.Left, Rect.Bottom );
            LineTo( Rect.Left, Rect.Top );
            LineTo( Rect.Right, Rect.Top );
            Pen.Color := clBtnShadow;
            MoveTo( Rect.Right, Rect.Top );
            LineTo( Rect.Right, Rect.Bottom );
            LineTo( Rect.Left - 1, Rect.Bottom );
            end;
        end
    else if Border = bbCtl3D then begin
        if Pressed = TestState then begin
            Pen.Color := clBlack;
            MoveTo( Rect.Left, Rect.Bottom );
            LineTo( Rect.Left, Rect.Top );
            LineTo( Rect.Right, Rect.Top );
            Pen.Color := clBtnHighLight;
            MoveTo( Rect.Right, Rect.Top );
            LineTo( Rect.Right, Rect.Bottom );
            LineTo( Rect.Left - 1, Rect.Bottom );
            end
        else begin
            //Pen.Color := clBtnHighlight; this is what most "buttons" do
            Pen.Color := clBtnFace;  // this is what combo-box buttons do
            MoveTo( Rect.Left, Rect.Bottom );
            LineTo( Rect.Left, Rect.Top );
            LineTo( Rect.Right, Rect.Top );
            Pen.Color := clBlack;
            MoveTo( Rect.Right, Rect.Top );
            LineTo( Rect.Right, Rect.Bottom );
            LineTo( Rect.Left - 1, Rect.Bottom );
            end;
        InflateRect( Rect, -1, -1 );
        if Pressed = TestState then begin
            Pen.Color := clBtnShadow;
            MoveTo( Rect.Left, Rect.Bottom );
            LineTo( Rect.Left, Rect.Top );
            LineTo( Rect.Right, Rect.Top );
            Pen.Color := clBtnFace;
            MoveTo( Rect.Right, Rect.Top );
            LineTo( Rect.Right, Rect.Bottom );
            LineTo( Rect.Left - 1, Rect.Bottom );
            end
        else begin
            //Pen.Color := clBtnFace;  this is what most buttons do
            Pen.Color := clBtnHighlight; // this is what combo-box buttons do
            MoveTo( Rect.Left, Rect.Bottom );
            LineTo( Rect.Left, Rect.Top );
            LineTo( Rect.Right, Rect.Top );
            Pen.Color := clBtnShadow;
            MoveTo( Rect.Right, Rect.Top );
            LineTo( Rect.Right, Rect.Bottom );
            LineTo( Rect.Left - 1, Rect.Bottom );
            end;
        end;
    Rect.Left := Rect.Left + 1;
    Rect.Top := Rect.Top + 1;
    // Then make sure the rect is still valid....
    if Rect.Right <= Rect.Left then
        Rect.Right := Rect.Left + 1;
    if Rect.Bottom <= Rect.Top then
        Rect.Bottom := Rect.Top + 1;
    Brush.Color := TmpBrushColor;
    Pen.Color := TmpPenColor;
    end; // with Cvs
end;

// Provided for backward compatibility
procedure Draw3DBorder( Cvs: TCanvas; var Rect: TRect; Shallow, Down: Boolean );
var
    DrawInfo: TDrawInfoEnh;
begin
DrawInfo.Rect := Rect;
with DrawInfo do begin
    VectorGlyph := vgNone;
    Canvas := Cvs;
    VectorColor := Cvs.Brush.Color;
    BkColor := Cvs.Brush.Color;
    VectorBrush := Cvs.Brush.Style;
    // Rect := already done
    if Shallow then
        Border := bbShallow
    else Border := bbCtl3D;
    Proportional := true;
    Enabled := true;
    Stretch := true;
    if Down then
      Pressed := eesButton1Down
    else
      Pressed := eesNoButtonDown;
    Hover := eesNoButtonDown;
    TestState := eesButton1Down;
    ImageList := nil;
    ImageIndex := -1;
    ImageDisabled := -1;
    ImagePressed := -1;
    end;
Draw3DBorder( DrawInfo );
end;



// Utility function for the arrow drawing functions...
procedure GrowthLimiter( var Rect: TRect; WLim, HLim: integer );
var
    w, h: integer;
begin
w := Rect.Right - Rect.Left;
h := Rect.Bottom - Rect.Top;
if (w > WLim) and (h > HLim) then begin
    // '4' is a magic number used to limit how fast the rectangle can grow.
    w := w div 4;
    h := h div 4;
    InflateRect( Rect, -w, -h );
    end;
end;


procedure DrawPolygon( var DrawInfo: TDrawInfoEnh; Points: array of TPoint;
    Width, Height: integer );
var
    i, j, xAdj, yAdj, ReqHeight, p: integer;
    //RKWidth, RKHeight: integer;
    TmpBrushColor, TmpPenColor: TColor;
    TmpBrushStyle: TBrushStyle;
begin
with DrawInfo, DrawInfo.Canvas do begin
    //RKWidth := 0;
    //for i := Low(Points) to High(Points) do
    //    if Points[i].x >= RKWidth then RKWidth := Points[i].x + 1;
    //RKHeight := 0;
    //for i := Low(Points) to High(Points) do
    //    if Points[i].y >= RKHeight then RKHeight := Points[i].y + 1;
    i := (Rect.Right - Rect.Left);
    j := (Rect.Bottom - Rect.Top);
    if Stretch and Proportional then begin
        // Height limited
        ReqHeight := ((Height * i) div Width);
        if j >= ReqHeight then begin           // Higher than required
            yAdj := ((j - ReqHeight) div 2);// + 1;
            // Convert the points according available width
            for p := Low(Points) to High(Points) do begin
                Points[p].x := ((Points[p].x * i) div Width) + Rect.Left;
                Points[p].y := ((Points[p].y * i) div Width) + Rect.Top + yAdj;
                end;
            end
        else begin                          // Wider than required
            xAdj := ((i - ((Width * j) div Height)) div 2);// + 1;
            // Convert the points according available height
            for p := Low(Points) to High(Points) do begin
                Points[p].x := ((Points[p].x * j) div Height) + Rect.Left + xAdj;
                Points[p].y := ((Points[p].y * j) div Height) + Rect.Top;
                end;
            end;
        end
    else if Stretch then begin
        // Convert the points according actual size of the rectangle
        for p := Low(Points) to High(Points) do begin
            Points[p].x := ((Points[p].x * i) div Width) + Rect.Left;
            Points[p].y := ((Points[p].y * j) div Height) + Rect.Top;
            end;
        end
    else begin
        // otherwise just center...
        xAdj := (i + 1 - Width) div 2;
        yAdj := (j + 1 - Height) div 2;
        // Convert the points according available height
        for p := Low(Points) to High(Points) do begin
            Points[p].x := Points[p].x + Rect.Left + xAdj;
            Points[p].y := Points[p].y + Rect.Top + yAdj;
            end;
        end;
    TmpBrushColor := Brush.Color;
    TmpPenColor := Pen.Color;
    TmpBrushStyle := Brush.Style;
    if Enabled then begin
        Brush.Color := VectorColor;
        Pen.Color := VectorColor;
        Brush.Style := VectorBrush;
        Polygon( Points );
        end
    else begin
        Brush.Style := VectorBrush;
        for p := Low(Points) to High(Points) do begin
            Points[p].x := Points[p].x + 1;
            Points[p].y := Points[p].y + 1;
            end;
        Pen.Color := clBtnHighlight;
        Brush.Color := clBtnHighlight;
        Polygon( Points );
        for p := Low(Points) to High(Points) do begin
            Points[p].x := Points[p].x - 1;
            Points[p].y := Points[p].y - 1;
            end;
        Pen.Color := clBtnShadow;
        Brush.Color := clBtnShadow;
        Polygon( Points );
        end;
    Brush.Color := TmpBrushColor;
    Brush.Style := TmpBrushStyle;
    Pen.Color := TmpPenColor;
    end;
end;

// Backwards compatibility
procedure DrawPolygon( Cvs: TCanvas; Points: array of TPoint; var Rect: TRect;
    Proport, Stretch, Enabled: Boolean; Width, Height: integer );
var
    DrawInfo: TDrawInfoEnh;
begin
DrawInfo.Rect := Rect;
DrawInfo.Enabled := Enabled;
DrawInfo.Stretch := Stretch;
with DrawInfo do begin
    VectorGlyph := vgNone;
    Canvas := Cvs;
    VectorColor := Cvs.Brush.Color;
    BkColor := Cvs.Brush.Color;
    VectorBrush := Cvs.Brush.Style;
    // Rect := already done
    Border := bbFlat;
    Proportional := Proport;
    Pressed := eesNoButtonDown;
    Hover := eesNoButtonDown;
    TestState := eesButton1Down;
    ImageList := nil;
    ImageIndex := -1;
    ImageDisabled := -1;
    ImagePressed := -1;
    end;
DrawPolygon( DrawInfo, Points, Width, Height );
end;



// Down Arrow always takes an "odd" number of pixels in width
// and tends to fall to the right side of the rectangle if necessary
procedure DrawDownArrow( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 8, 4 );
Points[0].x := 0;  Points[0].y := 0;
Points[1].x := 8;  Points[1].y := 0;
Points[2].x := 4;  Points[2].y := 4;
// Good to keep away from bottom
DrawInfo.Rect.Bottom := DrawInfo.Rect.Bottom - 1;
// Arrows draw better with odd width rectangles...
if ((DrawInfo.Rect.Right - DrawInfo.Rect.Left) mod 2) = 0 then
    DrawInfo.Rect.Right := DrawInfo.Rect.Right - 1;
DrawPolygon( DrawInfo, Points, 9, 5 );
end;

// Backwards compatibility
procedure DrawDownArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( Rect, 8, 4 );
Points[0].x := 0;  Points[0].y := 0;
Points[1].x := 8;  Points[1].y := 0;
Points[2].x := 4;  Points[2].y := 4;
Rect.Bottom := Rect.Bottom - 1; // Good to keep away from bottom
// Arrows draw better with odd width rectangles...
if ((Rect.Right - Rect.Left) mod 2) = 0 then
    Rect.Right := Rect.Right - 1;
DrawPolygon( Cvs, Points, Rect, Proport, true, Enabled, 9, 5 );
end;



procedure DrawUpArrow( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 8, 4 );
Points[0].x := 0;  Points[0].y := 4;
Points[1].x := 8;  Points[1].y := 4;
Points[2].x := 4;  Points[2].y := 0;
// Good to keep away from bottom
DrawInfo.Rect.Bottom := DrawInfo.Rect.Bottom - 1;
// Arrows draw better with odd width rectangles...
if ((DrawInfo.Rect.Right - DrawInfo.Rect.Left) mod 2) = 0 then
    DrawInfo.Rect.Right := DrawInfo.Rect.Right - 1;
DrawPolygon( DrawInfo, Points, 9, 5 );
end;

// Backwards compatibility
procedure DrawUpArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( Rect, 8, 4 );
Points[0].x := 0;  Points[0].y := 4;
Points[1].x := 8;  Points[1].y := 4;
Points[2].x := 4;  Points[2].y := 0;
Rect.Bottom := Rect.Bottom - 1; // Good to keep away from bottom
// Arrows draw better with odd width rectangles...
if ((Rect.Right - Rect.Left) mod 2) = 0 then
    Rect.Right := Rect.Right - 1;
DrawPolygon( Cvs, Points, Rect, Proport, true, Enabled, 9, 5 );
end;



procedure DrawLeftArrow( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 4, 8 );
Points[0].x := 0;  Points[0].y := 4;
Points[1].x := 4;  Points[1].y := 0;
Points[2].x := 4;  Points[2].y := 8;
// Good to keep away from right
DrawInfo.Rect.Right := DrawInfo.Rect.Right - 1;
DrawPolygon( DrawInfo, Points, 5, 9 );
end;

// Backwards compatibility
procedure DrawLeftArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( Rect, 4, 8 );
Points[0].x := 0;  Points[0].y := 4;
Points[1].x := 4;  Points[1].y := 0;
Points[2].x := 4;  Points[2].y := 8;
Rect.Right := Rect.Right - 1; // Good to keep away from right
DrawPolygon( Cvs, Points, Rect, Proport, true, Enabled, 5, 9 );
end;



procedure DrawRightArrow( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 4, 8 );
Points[0].x := 0;  Points[0].y := 0;
Points[1].x := 4;  Points[1].y := 4;
Points[2].x := 0;  Points[2].y := 8;
// Good to keep away from right
DrawInfo.Rect.Right := DrawInfo.Rect.Right - 1;
DrawPolygon( DrawInfo, Points, 5, 9 );
end;

// Backwards compatibility
procedure DrawRightArrow( Cvs: TCanvas; var Rect: TRect; Proport, Enabled: Boolean );
var
    Points: array [0..2] of TPOINT;
begin
GrowthLimiter( Rect, 4, 8 );
Points[0].x := 0;  Points[0].y := 0;
Points[1].x := 4;  Points[1].y := 4;
Points[2].x := 0;  Points[2].y := 8;
// Good to keep away from right
Rect.Right := Rect.Right - 1;
DrawPolygon( Cvs, Points, Rect, Proport, true, Enabled, 5, 9 );
end;


procedure DrawGlyph( var DrawInfo: TDrawInfoEnh );
var
    Idx: integer;
    Bitmap: TBitmap;
begin
with DrawInfo, DrawInfo.Canvas do begin
    Idx := ImageIndex;
    if not Enabled then begin
        if ImageDisabled >= 0 then
            Idx := ImageDisabled;
        end
    else if Pressed = TestState then begin
        if ImagePressed >= 0 then
            Idx := ImagePressed;
        end;
    // Can only proceed with valid image list and index
    if assigned(ImageList) and (Idx >= 0) then begin
        Bitmap := TBitmap.Create;
        try
            {$IFDEF __VCL30_OR_LATER__}
            Bitmap.Transparent := true;
            {$ENDIF}
            Bitmap.Width := ImageList.Width;
            Bitmap.Height := ImageList.Height;
            Bitmap.Canvas.Brush.Style := bsSolid;
            //Bitmap.Canvas.Brush.Color := Pixels[Rect.Left,Rect.Top];
            // I need a background color - the "transparent" color.
            // Originally tried picking it up from the Canvas but this was
            // very unreliable.  At least by picking a specific consistent
            // color it is something that can be documented and the behaviour
            // will be consistent.  Put simply, try to avoid clOlive in your
            // Glyphs as this will be treated as transparent.
            //Bitmap.Canvas.Brush.Color := clOlive;
            // Added BkColor to DrawInfo so now we have something to work with
            Bitmap.Canvas.Brush.Color := BkColor;
            Bitmap.Canvas.FillRect( Bounds(0,0, ImageList.Width, ImageList.Height) );
            ImageList.Draw( Bitmap.Canvas, 0, 0, Idx, (Enabled or (Idx <> ImageIndex)) );
            DrawBitmap( DrawInfo, Bitmap );
        finally
            Bitmap.Free;
            end;
        end;
    end;
end;

// Backwards compatibility
procedure DrawGlyph( Cvs: TCanvas; var Rect: TRect; Proport, Stretch: Boolean;
    ImgList: TImageList; ImgIdx: integer );
var
    DrawInfo: TDrawInfoEnh;
begin
DrawInfo.Rect := Rect;
DrawInfo.Stretch := Stretch;
with DrawInfo do begin
    VectorGlyph := vgNone;
    Canvas := Cvs;
    VectorColor := Cvs.Brush.Color;
    BkColor := Cvs.Brush.Color;
    VectorBrush := Cvs.Brush.Style;
    // Rect := already done
    Border := bbFlat;
    Proportional := Proport;
    Enabled := true;
    Pressed := eesNoButtonDown;
    Hover := eesNoButtonDown;
    TestState := eesButton1Down;
    ImageList := ImgList;
    ImageIndex := ImgIdx;
    ImageDisabled := -1;
    ImagePressed := -1;
    end;
DrawGlyph( DrawInfo );
end;



procedure DrawBitmap( var DrawInfo: TDrawInfoEnh; Bitmap: TBitmap );
var
    X, Y, Adj: integer;
begin
with DrawInfo, DrawInfo.Canvas do begin
    // if not stretched it is a simple draw
    if not Stretch then begin
        X := Rect.Left + (((Rect.Right - Rect.Left) - ImageList.Width) div 2);
        Y := Rect.Top + (((Rect.Bottom - Rect.Top) - ImageList.Height) div 2);
        Draw( X, Y, Bitmap )
        end
    else begin
        if Proportional then begin
            X := Rect.Right - Rect.Left;
            Y := Rect.Bottom - Rect.Top;
            if X >= Y  then begin
                Adj := (X - Y);
                Rect.Left := Rect.Left + (Adj div 2);
                Rect.Right := Rect.Right - ((Adj+1) div 2);
                end
            else begin
                Adj := (Y - X);
                Rect.Top := Rect.Top + (Adj div 2);
                Rect.Bottom := Rect.Bottom - ((Adj+1) div 2);
                end;
            end;
        StretchDraw( Rect, Bitmap );
        end;
    end;
end;



procedure DrawEllipsis( var DrawInfo: TDrawInfoEnh );
var
    i, j, c, h, w: integer;
    TmpBrushColor: TColor;
    TmpBrushStyle: TBrushStyle;
begin
with DrawInfo, DrawInfo.Canvas do begin
    Rect.Right := Rect.Right - 1; // Pull in just a little from the right
    i := (Rect.Right - Rect.Left);
    j := (Rect.Bottom - Rect.Top);
    w := 2;
    c := Rect.Left + (i div 2);
    if Stretch then begin
        w := i div 6;
        if w = 0 then
            w := 1;
        end;
    h := Rect.Top + (3*(j div 4)); // draw 3/4 of way down button
    // Make sure we stay within confines of the button...
    if (h + w) >= Rect.Bottom then begin
        if( (w + 4) > j ) then
            w := j - 4;
        h := Rect.Bottom - (w + 2);
        end;
    TmpBrushColor := Brush.Color;
    TmpBrushStyle := Brush.Style;
    Brush.Color := VectorColor;
    Brush.Style := VectorBrush;
    if Enabled then begin
        PatBlt(Handle, c, h, w, w, PATCOPY);
        PatBlt(Handle, c - (w * 2), h, w, w, PATCOPY);
        PatBlt(Handle, c + (w * 2), h, w, w, PATCOPY);
        end
    else begin
        Inc( c );
        Inc( h );
        Brush.Color := clBtnHighlight;
        PatBlt(Handle, c, h, w, w, PATCOPY);
        PatBlt(Handle, c - (w * 2), h, w, w, PATCOPY);
        PatBlt(Handle, c + (w * 2), h, w, w, PATCOPY);
        Dec( c );
        Dec( h );
        Brush.Color := clBtnShadow;
        PatBlt(Handle, c, h, w, w, PATCOPY);
        PatBlt(Handle, c - (w * 2), h, w, w, PATCOPY);
        PatBlt(Handle, c + (w * 2), h, w, w, PATCOPY);
        end;
    Brush.Color := TmpBrushColor;
    Brush.Style := TmpBrushStyle;
    end;
end;

// Backwards compatibility
procedure DrawEllipsis( Cvs: TCanvas; var Rect: TRect; Stretch, Enabled: Boolean );
var
    DrawInfo: TDrawInfoEnh;
begin
DrawInfo.Rect := Rect;
DrawInfo.Enabled := Enabled;
DrawInfo.Stretch := Stretch;
with DrawInfo do begin
    VectorGlyph := vgNone;
    Canvas := Cvs;
    VectorColor := Cvs.Brush.Color;
    BkColor := Cvs.Brush.Color;
    VectorBrush := Cvs.Brush.Style;
    // Rect := already done
    Border := bbFlat;
    Proportional := true;
    Pressed := eesNoButtonDown;
    Hover := eesNoButtonDown;
    TestState := eesButton1Down;
    ImageList := nil;
    ImageIndex := -1;
    ImageDisabled := -1;
    ImagePressed := -1;
    end;
DrawEllipsis( DrawInfo );
end;



procedure DrawReturnKey( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..9] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 7, 9 );
// Initialise polygon with drawing that I want
Points[0].x := 1;  Points[0].y := 4;
Points[1].x := 3;  Points[1].y := 2;
Points[2].x := 3;  Points[2].y := 4;
Points[3].x := 6;  Points[3].y := 4;
Points[4].x := 6;  Points[4].y := 1;
Points[5].x := 7;  Points[5].y := 1;
Points[6].x := 7;  Points[6].y := 5;
Points[7].x := 3;  Points[7].y := 5;
Points[8].x := 3;  Points[8].y := 7;
Points[9].x := 1;  Points[9].y := 5;
DrawPolygon( DrawInfo, Points, 8, 8 );
end;

// Backwards compatibility
procedure DrawReturnKey( Cvs: TCanvas; var Rect: TRect; Proport, Stretch, Enabled: Boolean );
var
    Points: array [0..9] of TPOINT;
begin
GrowthLimiter( Rect, 7, 9 );
// Initialise polygon with drawing that I want
Points[0].x := 1;  Points[0].y := 4;
Points[1].x := 3;  Points[1].y := 2;
Points[2].x := 3;  Points[2].y := 4;
Points[3].x := 6;  Points[3].y := 4;
Points[4].x := 6;  Points[4].y := 1;
Points[5].x := 7;  Points[5].y := 1;
Points[6].x := 7;  Points[6].y := 5;
Points[7].x := 3;  Points[7].y := 5;
Points[8].x := 3;  Points[8].y := 7;
Points[9].x := 1;  Points[9].y := 5;
DrawPolygon( Cvs, Points, Rect, Proport, Stretch, Enabled, 8, 8 );
end;



procedure DrawCheckMark( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..5] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 8, 8 );
// Initialise polygon with drawing that I want
Points[0].x := 8;  Points[0].y := 1;
Points[1].x := 8;  Points[1].y := 3;
Points[2].x := 3;  Points[2].y := 8;
Points[3].x := 1;  Points[3].y := 6;
Points[4].x := 1;  Points[4].y := 4;
Points[5].x := 3;  Points[5].y := 6;
DrawPolygon( DrawInfo, Points, 9, 9 );
end;

// Backwards compatibility
procedure DrawCheckMark( Cvs: TCanvas; var Rect: TRect; Proport, Stretch, Enabled: Boolean );
var
    Points: array [0..5] of TPOINT;
begin
GrowthLimiter( Rect, 8, 8 );
// Initialise polygon with drawing that I want
Points[0].x := 8;  Points[0].y := 1;
Points[1].x := 8;  Points[1].y := 3;
Points[2].x := 3;  Points[2].y := 8;
Points[3].x := 1;  Points[3].y := 6;
Points[4].x := 1;  Points[4].y := 4;
Points[5].x := 3;  Points[5].y := 6;
DrawPolygon( Cvs, Points, Rect, Proport, Stretch, Enabled, 9, 9 );
end;



procedure DrawQuery( var DrawInfo: TDrawInfoEnh );
var
    Points: array [0..15] of TPOINT;
    Points2: array [0..3] of TPOINT;
begin
GrowthLimiter( DrawInfo.Rect, 5, 9 );
// If given a small rectangle draw a small query
if (DrawInfo.Rect.Bottom - DrawInfo.Rect.Top) < 18 then begin
    InflateRect( DrawInfo.Rect, 0, 1 ); // Just bit more room is nice
    Points[0].x := 0;  Points[0].y := 3;
    Points[1].x := 0;  Points[1].y := 2;    // |
    Points[2].x := 2;  Points[2].y := 0;    //  /
    Points[3].x := 3;  Points[3].y := 0;    //   -
    Points[4].x := 5;  Points[4].y := 2;    //     \
    Points[5].x := 5;  Points[5].y := 3;    //      |
    Points[6].x := 3;  Points[6].y := 5;    //     /
    Points[7].x := 3;  Points[7].y := 6;    //    |
    Points[8].x := 2;  Points[8].y := 6;    //   -
    Points[9].x := 2;  Points[9].y := 5;    //  |
    Points[10].x := 4; Points[10].y := 3;   //   /
    Points[11].x := 4; Points[11].y := 2;   //     |
    Points[12].x := 3; Points[12].y := 1;   //    \
    Points[13].x := 2; Points[13].y := 1;   //   -
    Points[14].x := 1; Points[14].y := 2;   //  /
    Points[15].x := 1; Points[15].y := 3;   // |
    DrawPolygon( DrawInfo, Points, 6, 10 );
    Points2[0].x := 2;  Points2[0].y := 8;
    Points2[1].x := 3;  Points2[1].y := 8;
    Points2[2].x := 3;  Points2[2].y := 9;
    Points2[3].x := 2;  Points2[3].y := 9;
    DrawPolygon( DrawInfo, Points2, 6, 10 );
    end
// otherwise draw a bigger one...
else begin
    Points[0].x := 0;  Points[0].y := 6;
    Points[1].x := 0;  Points[1].y := 4;     // |
    Points[2].x := 4;  Points[2].y := 0;     //  /
    Points[3].x := 6;  Points[3].y := 0;     //   -
    Points[4].x := 10;  Points[4].y := 4;    //     \
    Points[5].x := 10;  Points[5].y := 6;    //      |
    Points[6].x := 6;  Points[6].y := 10;    //     /
    Points[7].x := 6;  Points[7].y := 12;    //    |
    Points[8].x := 4;  Points[8].y := 12;    //   -
    Points[9].x := 4;  Points[9].y := 10;    //  |
    Points[10].x := 8; Points[10].y := 6;    //   /
    Points[11].x := 8; Points[11].y := 4;    //     |
    Points[12].x := 6; Points[12].y := 2;    //    \
    Points[13].x := 4; Points[13].y := 2;    //   -
    Points[14].x := 2; Points[14].y := 4;    //  /
    Points[15].x := 2; Points[15].y := 6;    // |
    DrawPolygon( DrawInfo, Points, 11, 18 );
    Points2[0].x := 4;  Points2[0].y := 15;
    Points2[1].x := 6;  Points2[1].y := 15;
    Points2[2].x := 6;  Points2[2].y := 17;
    Points2[3].x := 4;  Points2[3].y := 17;
    DrawPolygon( DrawInfo, Points2, 11, 18 );
    end;
end;

// Backwards compatibility
procedure DrawQuery( Cvs: TCanvas; var Rect: TRect; Proport, Stretch, Enabled: Boolean );
var
    DrawInfo: TDrawInfoEnh;
begin
DrawInfo.Rect := Rect;
DrawInfo.Enabled := Enabled;
DrawInfo.Stretch := Stretch;
with DrawInfo do begin
    VectorGlyph := vgNone;
    Canvas := Cvs;
    VectorColor := Cvs.Brush.Color;
    BkColor := Cvs.Brush.Color;
    VectorBrush := Cvs.Brush.Style;
    // Rect := already done
    Border := bbFlat;
    Proportional := Proport;
    Pressed := eesNoButtonDown;
    Hover := eesNoButtonDown;
    TestState := eesButton1Down;
    ImageList := nil;
    ImageIndex := -1;
    ImageDisabled := -1;
    ImagePressed := -1;
    end;
DrawQuery( DrawInfo );
end;



function TextColorFromColor( BkColor: TColor ): TColor;
begin
case BkColor of
    clWindow:            Result := clWindowText;
    clMenu:              Result := clMenuText;
    clBtnFace:           Result := clBtnText;
    clActiveCaption:     Result := clCaptionText;
    clInactiveCaption:   Result := clInactiveCaptionText;
    clHighlight:         Result := clHighlightText;
    clInfoBk:            Result := clInfoText;
else
    // Color BRG=E0C000 is just an abitrary color to switch between
    // white text and black text which seems to work OK.
    if ColorBrightness(BkColor) >= ColorBrightness(TColor($E0C000)) then
        Result := clBlack
    else Result := clWhite;
    end;
end;

// The color weight system used in this function was determined by
// experimentation with formula that appeared to give the desired result.
// ie. It is probably not perfect.
// The resulting formula goes like this (^=power)...
//     (Blue^2) + ((Green^2)*4) + ((Red^2)*2) = "Brightness"
// (brightness being a term used here to indicate whether a color is
// bright enough to warrant having black text)
// The constant EditUtil_BrightnessMax defines the maximum possible
// value of the brightness that may be calculated with this formula.
function ColorBrightness( BkColor: TColor ): integer;
var
   b,g,r: integer;
begin
b := (integer(BkColor) and $FF0000) shr 16;
g := (integer(BkColor) and $00FF00) shr 8;
r := (integer(BkColor) and $0000FF);
b := b * b * b * 8;
g := (g * g * g) * 32;
r := (r * r * r) * 9;
Result := b + g + r;
end;


function ExponentFilter( AValue: string; MaxIntDigit: integer ): string;
var
   tmpExt, Lim: Extended;
begin
if (Pos( 'E', AValue ) > 0) or (Pos( 'e', AValue ) > 0) then begin
    tmpExt := StrToFloat( AValue );
    Lim := 1.0;
    while MaxIntDigit > 0 do begin
        Lim := Lim * 10;
        Dec(MaxIntDigit);
        end;
    if tmpExt >= Lim then begin
        raise Exception.Create(ENHERR_NUMBER_TOO_LARGE);;
        end
    else begin
        Result := FormatFloat( '#################0.0#################', tmpExt );
        end;
    end
else Result := AValue;
end;


function ConvCharCase( PrevChar, NewChar: char; ChCase: integer ): char;
begin
case ChCase of
    1{ecUpperCase}: Result := char(CharUpper(PChar(NewChar)));
    2{ecLowerCase}: Result := char(CharLower(PChar(NewChar)));
    3{ecProperCase}: begin
        if (PrevChar = #0) or
            (Pos( PrevChar, ' -,.;/()!@#$%^&*+="') <> 0) then
            Result := char(CharUpper(PChar(NewChar)))
        else Result := NewChar;
        end;
else
    Result := NewChar;
    end;
end;




function MakeDelphiName(AName: string): string;
var
  i: integer;
begin
  Result := '';
  for i := 1 to Length(AName) do
  begin
    if AName[i] in [' ','@','#','$'] then
      Result := Result + '_'
    else
      if AName[i] in ['A'..'Z', 'a'..'z', '_'] then
        Result := Result + AName[i]
      else
        if (i > 1) and (AName[i] in ['0'..'9']) then
          Result := Result + AName[i];
  end;
  if AName = '' then
    Result := 'x';
end;

procedure CenterInParent(Control: TControl; TopLeft: boolean);
var
  l, t: integer;
begin
  if (Control.Parent <> nil) and (Control.Parent.HandleAllocated) then
  begin
    l := (Control.Parent.ClientWidth - Control.Width) div 2;
    t := (Control.Parent.ClientHeight - Control.Height) div 2;
    if (l < 0) and TopLeft then
      l := 0;
    if (t < 0) and TopLeft then
      t := 0;
    Control.SetBounds(l, t, Control.Width, Control.Height);
  end;
end;

procedure CenterInRect(Control: TControl; const Rect: TRect; TopLeft: boolean);
var
  l, t: integer;
begin
  l := Rect.Left + (((Rect.Right - Rect.Left) - Control.Width) div 2);
  t := Rect.Top + (((Rect.Bottom - Rect.Top) - Control.Height) div 2);
  if (l < Rect.Left) and TopLeft then
    l := Rect.Left;
  if (t < Rect.Top) and TopLeft then
    t := Rect.Top;
  Control.SetBounds(l, t, Control.Width, Control.Height);
end;

function ActivateForm(AForm: TCustomForm): boolean;
var
  TmpC: TWinControl;
begin
  Result := false;
  if AForm.CanFocus and AForm.Visible and AForm.Enabled then
  begin
    if (AForm.ActiveControl <> nil) and AForm.ActiveControl.CanFocus then
    begin
      AForm.ActiveControl.SetFocus;
      Result := true;
    end
    else
    begin
      TmpC := GetNextControl(AForm, nil, true, true, false);
      if (TmpC <> nil) and TmpC.CanFocus then
        TmpC.SetFocus
      else
        AForm.SetFocus;
      Result := true;
    end;
  end;
end;

function IsChildWindow(Control: TWinControl): boolean;
begin
  Result := ((Windows.GetWindowLong(Control.Handle, GWL_STYLE) and WS_CHILD) <>
    0);
end;

function ComponentToString(Component: TComponent): string;
var
  BinStream: TMemoryStream;
  StrStream: TStringStream;
  s: string;
begin
  BinStream := TMemoryStream.Create;
  try
    StrStream := TStringStream.Create(s);
    try
      BinStream.WriteComponent(Component);
      BinStream.Seek(0, soFromBeginning);
      ObjectBinaryToText(BinStream, StrStream);
      StrStream.Seek(0, soFromBeginning);
      Result := StrStream.DataString;
    finally
      StrStream.Free;
    end;
  finally
    BinStream.Free
  end;
end;

function StringToComponent(Value: string; AInstance: TComponent): TComponent;
var
  StrStream: TStringStream;
  BinStream: TMemoryStream;
begin
  StrStream := TStringStream.Create(Value);
  try
    BinStream := TMemoryStream.Create;
    try
      ObjectTextToBinary(StrStream, BinStream);
      BinStream.Seek(0, soFromBeginning);
      Result := BinStream.ReadComponent(AInstance);
    finally
      BinStream.Free;
    end;
  finally
    StrStream.Free;
  end;
end;

function StreamToComponent(Stream: TStream; AInstance: TComponent): TComponent;
var
  BinStream: TMemoryStream;

  function IsTextStream(S: TStream): boolean;
  var
    buffer: array[1..length('inherited')] of byte;
    savepos: integer;

    function CheckIdent(const Ident: string): boolean;
    begin
      Result := CompareMem(@buffer, pointer(Ident), length(Ident));
    end;
  begin
    savepos := S.Position;
    try
      S.ReadBuffer(buffer, sizeof(buffer));
      result := CheckIdent('object') or CheckIdent('inherited') or
        CheckIdent('inline');
    finally
      S.Position := savepos;
    end;
  end;

  function SmartSkipResHeader(S: TStream; OnlyCheck: boolean): boolean;
  var
    ReadCount: DWORD;
    Header: array[0..79] of Char;
    OldPos: Integer;
  begin
    with S do
    begin
      FillChar(Header, SizeOf(Header), 0);
      OldPos := Position;
      ReadCount := Read(Header, SizeOf(Header) - 1);
      Result := (Byte((@Header[0])^) = $FF) and (Word((@Header[1])^) = 10);
      if Result and (not OnlyCheck) then
        Seek(StrLen(Header + 3) + 10 - ReadCount, 1)
      else
        Seek(OldPos, soFromBeginning);
    end;
  end;

begin
  if IsTextStream(Stream) then
  begin
    BinStream := TMemoryStream.Create;
    try
      ObjectTextToBinary(Stream, BinStream);
      BinStream.Seek(0, soFromBeginning);
      Result := BinStream.ReadComponent(AInstance);
    finally
      BinStream.Free;
    end;
  end
  else
  begin
    SmartSkipResHeader(Stream, false);
    Result := Stream.ReadComponent(AInstance);
  end;
end;


function SaveStringToFile(AValue: string; var FileName: string; Prompt:
  boolean): boolean;
var
  Strs: TStringList;
  Continue: boolean;
  Dlg: TSaveDialog;
begin
  Result := false;
  Continue := true;
  // If no file specified, prompt for one...
  if Prompt then
  begin
    Dlg := TSaveDialog.Create(nil);
    try
      if Dlg.Execute then
        FileName := Dlg.FileName
      else
        Continue := false;
    finally
      Dlg.Free;
    end;
  end;
  if Continue and (FileName <> '') then
  begin
    Strs := TStringList.Create;
    try
      Strs.Text := AValue;
      Strs.SaveToFile(FileName);
    finally
      Strs.Free;
    end;
  end;
end;

function ReadStringFromFile(var FileName: string; Prompt: boolean): string;
var
  Strs: TStringList;
  Continue: boolean;
  Dlg: TOpenDialog;
begin
  Result := '';
  Continue := true;
  // If no file specified, prompt for one...
  if Prompt then
  begin
    Dlg := TOpenDialog.Create(nil);
    try
      if Dlg.Execute then
        FileName := Dlg.FileName
      else
        Continue := false;
    finally
      Dlg.Free;
    end;
  end;
  if Continue and (FileName <> '') then
  begin
    Strs := TStringList.Create;
    try
      Strs.LoadFromFile(FileName);
      Result := Strs.Text;
    finally
      Strs.Free;
    end;
  end;
end;

function FindComponentIn(AOwner: TComponent; AName: string): TComponent;
var
  i: integer;
begin
  Result := nil;
  with AOwner do
  begin
    for i := 0 to ComponentCount - 1 do
    begin
      if CompareText(Components[i].Name, AName) = 0  then
        Result := Components[i]
      else
        Result := FindComponentIn(Components[i], AName);
      if Result <> nil then
        break;
    end;
  end;
end;

function GetShiftKeyStates: TShiftState;
// Short function to retrieve keyboard shift state
// (essentially a copy of similar defined in Control.PAS, but this one
// also uses GetKeyState for the Alt-key (VK_MENU) since this information
// does not come with key messages.
begin
  Result := [];
  if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
  if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
  if GetKeyState(VK_MENU) < 0 then Include(Result, ssAlt);
end;

function GetScreenWorkArea: TRect;
// Simple function to gather the screen workarea coordinates
begin
  // Uf GetWorkArea fails just use screen rectangle
  if not SystemParametersInfo(SPI_GETWORKAREA, 0, @Result, 0) then
  begin
    Result.Left := 0;
    Result.Top := 0;
    Result.Right := Screen.Width;
    Result.Bottom := Screen.Height;
  end;
end;

function LockKeyState(Key: integer; State: TBoolStateEnh): boolean;
// Simple functions to switch lock keys on/off
// returns true if on when complete
// ie. VK_CAPITAL, VK_NUMLOCK, VK_SCROLL
begin
  if State = bsOn then
  begin
    // If not on already turn on...
    if (Windows.GetKeyState(Key) and 1) = 0 then
    begin
      Windows.keybd_event(Key, 0, 0, 0); // key down
      Windows.keybd_event(Key, 0, KEYEVENTF_KEYUP, 0); // key up again
    end;
  end
  else
    if State = bsOff then
    begin
      // If on turn off...
      if (Windows.GetKeyState(Key) and 1) > 0 then
      begin
        Windows.keybd_event(Key, 0, 0, 0); // key down
        Windows.keybd_event(Key, 0, KEYEVENTF_KEYUP, 0); // key up again
      end;
    end;
  Result := (Windows.GetKeyState(Key) and 1) > 0;
end;

procedure InvalidateAll(Control: TWinControl);
var
  i: integer;
begin
  with Control do
  begin
    for i := 0 to ControlCount - 1 do
    begin
      if Controls[i] is TWinControl then
        InvalidateAll(TWinControl(Controls[i]));
      Controls[i].Invalidate;
    end;
  end;
end;

procedure UpdateAll(Control: TWinControl);
var
  i: integer;
begin
  with Control do
  begin
    for i := 0 to ControlCount - 1 do
    begin
      if Controls[i] is TWinControl then
        UpdateAll(TWinControl(Controls[i]));
      Controls[i].Update;
    end;
  end;
end;

procedure RepaintAll(Control: TWinControl);
var
  i: integer;
begin
  with Control do
  begin
    for i := 0 to ControlCount - 1 do
    begin
      if Controls[i] is TWinControl then
        RepaintAll(TWinControl(Controls[i]));
      Controls[i].Repaint;
    end;
  end;
end;

function GetEnvironmentValue(VarName: string): string;
var
  eLen: DWORD;
begin
  eLen := 0;
  try
    SetLength(Result, 1024);
    eLen := GetEnvironmentVariable(PChar(VarName), PChar(Result), 1024);
  finally
    SetLength(Result, eLen);
  end;
end;

function GetWorkstationName: string;
var
  eLen: DWORD;
begin
  eLen := 256;
  SetLength(Result, eLen);
  if not GetComputerName(PChar(Result), eLen) then
    eLen := 0;
  SetLength(Result, eLen);
end;

function GetSessionName: string;
begin
  Result := GetEnvironmentValue(MTS_SessionEnvironmentVar);
  if Result = '' then
    Result := GetEnvironmentValue(MTS_SessionEnvironmentVar2);
  if Result = '' then
    Result := SessionDefaultName;
end;

function TimeIntervalString( Interval: double; IncPts: boolean ): string;
begin
  Result := '';
  if Interval >= 0.0 then
  begin
    if Interval >= 1.0 then
      Result := FloatToStr(Int(Interval)) + '+';
    Interval := 24 * Frac(Interval);
    Result := Result + FloatToStr(Int(Interval)) + ':';
    Interval := 60 * Frac(Interval);
    Result := Result + FloatToStr(Int(Interval)) + ':';
    Interval := 60 * Frac(Interval);
    if IncPts then  // include points of seconds?
      Result := Result + FormatFloat('#0.###',Interval)
    else
      Result := Result + FloatToStr(Int(Interval));
  end;
end;




function ExtractNumber( StrVal: string ): variant;
var
  StartPos, EndPos: integer;
begin
Result := ExtractNumberEx( StrVal, StartPos, EndPos );
end;


function ExtractNumberEx( StrVal: string; var NumStart, NumEnd: integer ): variant;
type
  TValidNumericChars = set of char;
var
  i: integer;
  tmpStr: string;
  tmpChr: char;
  NumberChars, ProvisionalChars: TValidNumericChars;
  IsFloat: boolean;
  IntCount: integer;
begin
  i := 1;
  IntCount := 0;
  tmpStr := '';
  NumberChars := ['-','+','0'..'9'];
  ProvisionalChars := ['e','E',DecimalSeparator];
  IsFloat := false;
  // Initialise return variables to 0
  NumStart := 0;
  NumEnd := 0;

  // Find first instance of a numeric character.
  // Sign only available as first character in number (or with exp).
  while i <= Length( StrVal ) do begin
    tmpChr := StrVal[i];
    if (tmpChr in NumberChars) or
       ( (tmpChr in ProvisionalChars) and (i < Length(StrVal)) and
         (StrVal[i+1] in NumberChars) )
      then begin
      tmpStr := tmpChr;
      NumStart := i;
      NumEnd := i;
      NumberChars := NumberChars - ['-','+']; // remove sign as acceptable
      // Thousand separate not permitted as first character
      ProvisionalChars := ProvisionalChars + [ThousandSeparator];
      if tmpChr = DecimalSeparator then begin
          IsFloat := true;
          end
      else begin
          if tmpChr in NumberChars then
              Inc( IntCount );
          end;
      Inc(i);
      break;
    end;
    Inc(i);
  end;

  // While we have valid numeric characters (possibly including exponent)
  // extract those characters into tmpStr
  while i <= Length( StrVal ) do begin
    tmpChr := StrVal[i];
    // Simple - character is a number...
    if (tmpChr in NumberChars) then begin
      tmpStr := tmpStr + tmpChr;
      NumberChars := NumberChars - ['-','+']; // remove sign as acceptable
      if (not IsFloat) and (tmpChr in NumberChars) then
          Inc( IntCount );
    // More complicated, not a number but may format/syntax character
    end else if (tmpChr in ProvisionalChars) and
      (i < Length(StrVal)) and (StrVal[i+1] in NumberChars) then begin
      if tmpChr = ThousandSeparator then begin
         ; // do nothing - ignore thousand separators
      end else begin
        tmpStr := tmpStr + tmpChr;
        if tmpChr in ['e','E'] then begin
          IsFloat := true;
          ProvisionalChars := ProvisionalChars - ['e','E'];
          NumberChars := NumberChars + ['-','+']; // make sign acceptable
        end else if tmpChr = DecimalSeparator then begin
          IsFloat := true;
          ProvisionalChars := ProvisionalChars - [DecimalSeparator];
        end;
      end;
    end else begin
      NumEnd := i;
      break;
    end;
    Inc( i );
    NumEnd := i;
  end;
  if Length(tmpStr) <= 0 then begin
    Result := 0;
  end else begin
    if IsFloat then begin
      Result := StrToFloat( tmpStr );
    end else begin
      // If reliably within 32bit integer range...
      if IntCount < 9 then begin
        Result := StrToInt( tmpStr );
      // otherwise try to store as float value...
      end else begin
        Result := StrToFloat( tmpStr );
      end;
    end;
  end;
end;

function VariantToStr( VarVal: variant ): string;
begin
case TVarData(VarVal).VType of
    varNull:     Result := '';
    varSmallint: Result := IntToStr( VarVal );
    varInteger:  Result := IntToStr( VarVal );
    varSingle:   Result := FloatToStr( VarVal );
    varDouble:   Result := FloatToStr( VarVal );
    varCurrency: Result := CurrToStr( VarVal );
    varDate:     Result := DateTimeToStr( VarVal );
    else
        Result := VarVal;
    end;
end;



function DateTimeFormatToLayout( AFormat: string ): string;
var
  i: integer;
  tmpCh: char;
  LastWasHour, InsideQuotes, ProcessChar: boolean;
begin
  Result := '';
  LastWasHour := false;
  InsideQuotes := false;
  AFormat := AnsiUpperCase(AFormat); // simplifies later code
  i := 1;
  while i <= Length(AFormat) do
  begin
    tmpCh := ' ';
    case AFormat[i] of
      'C': tmpCh := 'c';
      'D': tmpCh := 'd';
      'M': tmpCh := 'm';
      'Y': tmpCh := 'y';
      'H': tmpCh := 'h';
      'N': tmpCh := 'n';
      'S': tmpCh := 's';
      'Z': tmpCh := 'z';
      'A': tmpCh := 'a';
      'T': tmpCh := 't';
      '''', '"': InsideQuotes := not InsideQuotes;
    end;
    if not InsideQuotes then
    begin
      ProcessChar := true;
      // Special parsing to avoid day names (which give no useful information)
      // and to accept some special format requests supported by FormatDateTime.
      case tmpCh of
        'c': // autoformat to locale specified strings...
          begin
            if (Length(AFormat) - i) >= 1 then
              AFormat := Copy(AFormat,1,i) + ShortDateFormat + ' ' +
                ShortTimeFormat + Copy(AFormat, i + 1, MAXINT )
            else
              AFormat := Copy(AFormat,1,i) + ShortDateFormat + ' ' +
                ShortTimeFormat;
            AFormat := AnsiUpperCase(AFormat);
            ProcessChar := false;
          end;
        'd':
          begin
            // Insert long date format on request
            if ((Length(AFormat) - i) >= 5) and
              (Copy(AFormat,i,6) = 'DDDDDD') then
            begin
              if (Length(AFormat) - i) > 5 then
                AFormat := Copy(AFormat,1,i) + LongDateFormat +
                  Copy(AFormat, i + 6, MAXINT )
              else
                AFormat := Copy(AFormat,1,i) + LongDateFormat;
              AFormat := AnsiUpperCase(AFormat);
              ProcessChar := false;
            end
            else
              // insert short date format on request
              if ((Length(AFormat) - i) >= 4) and
                (Copy(AFormat,i,5) = 'DDDDD') then
              begin
                if (Length(AFormat) - i) > 4 then
                  AFormat := Copy(AFormat,1,i) + ShortDateFormat +
                    Copy(AFormat, i + 5, MAXINT )
                else
                  AFormat := Copy(AFormat,1,i) + ShortDateFormat;
                AFormat := AnsiUpperCase(AFormat);
                ProcessChar := false;
              end
              else
                // translate long day name in formatting
                if ((Length(AFormat) - i) >= 3) and
                  (Copy(AFormat,i,4) = 'DDDD') then
                begin
                  tmpCh := 'x';
                  Inc(i,3); // fourth one happens below
                end
                else
                  // translate short day name in formatting
                  if ((Length(AFormat) - i) >= 2) and
                    (Copy(AFormat,i,3) = 'DDD') then
                  begin
                    tmpCh := 'x';
                    Inc(i,2); // third one happens below
                  end;
          end;
        'a':
          begin
            if ((Length(AFormat) - i) >= 2) and
              (Copy(AFormat,i,3) = 'A/P') then
            begin
              Inc(i,2); // third one happens below
            end
            else
              if ((Length(AFormat) - i) >= 3) and
                (Copy(AFormat,i,4) = 'AMPM') then
              begin
                Inc(i,3); // fourth one happens below
              end
              else
                if ((Length(AFormat) - i) >= 4) and
                  (Copy(AFormat,i,5) = 'AM/PM') then
                begin
                  Inc(i,4); // fifth one happens below
                end;
          end;
        't':
          begin
            if ((Length(AFormat) - i) >= 1) and
              (Copy(AFormat,i,2) = 'TT') then
            begin
              if (Length(AFormat) - i) > 1 then
                AFormat := Copy(AFormat,1,i) + LongTimeFormat +
                  Copy(AFormat, i + 2, MAXINT )
              else
                AFormat := Copy(AFormat,1,i) + LongTimeFormat;
              AFormat := AnsiUpperCase(AFormat);
              ProcessChar := false;
            end
            else
            begin
              if (Length(AFormat) - i) > 0 then
                AFormat := Copy(AFormat,1,i) + ShortTimeFormat +
                  Copy(AFormat, i + 1, MAXINT )
              else
                AFormat := Copy(AFormat,1,i) + ShortTimeFormat;
              AFormat := AnsiUpperCase(AFormat);
              // only required inc happens below
              ProcessChar := false;
            end;
          end;
      end;
      if ProcessChar then
      begin
        if tmpCh = 'h' then
          LastWasHour := true
        else
        begin
          if (tmpCh = 'm') and LastWasHour then
            tmpCh := 'n';
          if LastWasHour and not(tmpCh in [' ','n']) then
            LastWasHour := false;
        end;
        if (tmpCh <> ' ') and (Pos(tmpCh, Result) < 1) then
          Result := Result + tmpCh;
      end;
    end;
    Inc( i );
  end;
end;

function ExtractDateTimeParts(ALayout, AText: string;
  var RDayDate, RMnth, RYear, RHour, RMinute, RSecond, RPoints: word): integer;
type
  TCharType = (ctNumeric, ctDelimiter, ctCharacter);
const
  EnglishShortDays: array[1..7] of string =
  ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
  EnglishLongDays: array[1..7] of string =
  ('Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
    'Saturday');
  EnglishShortMonths: array[1..12] of string =
  ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
  EnglishLongMonths: array[1..12] of string =
  ('January', 'February', 'March', 'April', 'May', 'June',
    'July', 'August', 'September', 'October', 'November', 'December');
  EnglishShortAMPM: array[1..2] of string = ('a', 'p');
  EnglishLongAMPM: array[1..2] of string = ('am', 'pm');
var
  i, LayoutPos, FirstPos: integer;
  CurCharType, LastCharType: TCharType;
  CurChar, CurStr: string;
  DayNum, DayDate, Mnth, Year, Hour, Minute, Second, Points: word;
  TmpDayDate, TmpMnth, TmpYear: word;
  Accept12Hr, Only12Hr, IsAM, IsPM, CheckEnglish, EnglishOnly: boolean;
  tmpDT: TDateTime;

  // Generate default ALayout value from global variables

  function FindMatchInStringArrays(AValue: string;
    ShortStrings: array of string; LongStrings: array of string): integer;
  var
    ii: integer;
  begin
    Result := -1;
    AValue := AnsiUpperCase(AValue);
    for ii := Low(ShortStrings) to High(ShortStrings) do
    begin
      if (Length(AValue) >= Length(ShortStrings[ii])) and
        (Length(AValue) <= Length(LongStrings[ii])) then
      begin
        if (AValue = AnsiUpperCase(ShortStrings[ii])) or
          (AValue = AnsiUpperCase(Copy(LongStrings[ii], 1, Length(AValue))))
            then
        begin
          // All input arrays are 1 based, but open array params always base
          // on zero, so we need to adjust accordingly
          Result := ii + 1;
          break;
        end;
      end;
    end;
  end;

  function ReadDayNum(var ValueStr: string; AType: TCharType): smallint;
  var
    tmpVal: integer;
  begin
    Result := E_DATEENH_VALUE_NOT_FOUND;
    if AType = ctCharacter then
    begin
      tmpVal := -1;
      // first check locale specific strings...
      if not EnglishOnly then
        tmpVal := FindMatchInStringArrays(ValueStr, ShortDayNames,
          LongDayNames);
      if (tmpVal < 0) and (EnglishOnly or CheckEnglish) then
        tmpVal := FindMatchInStringArrays(ValueStr, EnglishShortDays,
          EnglishLongDays);
      if tmpVal >= 0 then
      begin
        DayNum := tmpVal;
        ValueStr := '';                 // clear to indicate all was used.
        Result := E_EDITENH_NOERROR;
      end;
    end;
  end;

  function ReadSpecialWord(var ValueStr: string; AType: TCharType): smallint;
  var
    tmpVal: integer;
  begin
    Result := E_DATEENH_VALUE_NOT_FOUND;
    if AType = ctCharacter then
    begin
      tmpVal := FindMatchInStringArrays(ValueStr, ExtractDT_ShortStrings,
        ExtractDT_LongStrings);
      case tmpVal of
        1:                              // Today
          begin
            tmpDT := Trunc(SysUtils.Now);
            ValueStr := '';             // clear to indicate all was used.
            DecodeDate(tmpDT, Year, Mnth, DayDate);
            Result := E_EDITENH_NOERROR;
          end;
        2:                              // Tomorrow
          begin
            tmpDT := Trunc(SysUtils.Now) + 1;
            ValueStr := '';             // clear to indicate all was used.
            DecodeDate(tmpDT, Year, Mnth, DayDate);
            Result := E_EDITENH_NOERROR;
          end;
        3:                            // Yesterday
          begin
            tmpDT := Trunc(SysUtils.Now) - 1;
            ValueStr := '';             // clear to indicate all was used.
            DecodeDate(tmpDT, Year, Mnth, DayDate);
            Result := E_EDITENH_NOERROR;
          end;
        4:                            // Now
          begin
            tmpDT := SysUtils.Now;
            ValueStr := '';             // clear to indicate all was used.
            DecodeDate(tmpDT, Year, Mnth, DayDate);
            DecodeTime(tmpDT, Hour, Minute, Second, Points);
            Result := E_EDITENH_NOERROR;
          end;
        5:                            // None
          begin
            tmpDT := 0;
            ValueStr := '';             // clear to indicate all was used.
            Result := E_EDITENH_NOERROR;
          end;
      end;
    end;
  end;

  function ReadDayDate(var ValueStr: string; AType: TCharType): smallint;
  begin
    if AType <> ctNumeric then
      Result := E_DATEENH_INVALID_DAY_DATE
    else
    begin
      Result := E_EDITENH_NOERROR;
      DayDate := StrToInt( Copy(ValueStr,1,2) );
      if (DayDate > 31) or (DayDate <= 0) then
        Result := E_DATEENH_INVALID_DAY_DATE;
      if Length(ValueStr) > 2 then
        ValueStr := Copy(ValueStr, 3, Length(ValueStr) )
      else
        ValueStr := '';
    end;
  end;

  function ReadMonth(var ValueStr: string; AType: TCharType): smallint;
  var
    tmpVal: integer;
  begin
    Result := E_EDITENH_NOERROR;
    if AType = ctCharacter then
    begin
      tmpVal := -1;
      // first check locale specific strings...
      if not EnglishOnly then
        tmpVal := FindMatchInStringArrays(ValueStr, ShortMonthNames,
          LongMonthNames);
      if (tmpVal < 0) and (EnglishOnly or CheckEnglish) then
        tmpVal := FindMatchInStringArrays(ValueStr, EnglishShortMonths,
          EnglishLongMonths);
      if tmpVal >= 0 then
      begin
        Mnth := tmpVal;
        ValueStr := '';                 // clear to indicate all was used.
      end;
    end
    else  // AType = ctNumeric
    begin
      Mnth := StrToInt( Copy(ValueStr,1,2) );
      if Length(ValueStr) > 2 then
        ValueStr := Copy(ValueStr, 3, Length(ValueStr) )
      else
        ValueStr := '';
    end;
    if (Mnth > 12) or (Mnth <= 0) then
      Result := E_DATEENH_INVALID_MONTH;
  end;

  function ReadYear(var ValueStr: string; AType: TCharType): smallint;
  var
    CurrYear, tmpMnth, tmpDay, CenturyBase: word;
  begin
    if AType <> ctNumeric then
      Result := E_DATEENH_INVALID_YEAR
    else
    begin
      Result := E_EDITENH_NOERROR;
      // If a year prefixes another number it must provide all four digits!
      Year := StrToInt( Copy(ValueStr,1,4) );
      // adjust two digit years to four digits.
      if Year < 100 then
      begin
        DecodeDate( SysUtils.Now, CurrYear, tmpMnth, tmpDay );
        CenturyBase := CurrYear - TwoDigitYearCenturyWindow;
        Inc(Year, ((CenturyBase div 100) * 100));
        if (TwoDigitYearCenturyWindow > 0) and (Year < CenturyBase) then
          Inc(Year, 100);
      end;
      // I hope no one will have a problem with a 9999 limit on the year :-)
      if (Year > 9999) or (Year <= 0) then
        Result := E_DATEENH_INVALID_YEAR;
      if Length(ValueStr) > 4 then
        ValueStr := Copy(ValueStr, 5, Length(ValueStr) )
      else
        ValueStr := '';
    end;
  end;

  function ReadHours(var ValueStr: string; AType: TCharType): smallint;
  begin
    if AType <> ctNumeric then
      Result := E_DATEENH_INVALID_HOUR
    else
    begin
      Result := E_EDITENH_NOERROR;
      Hour := StrToInt( Copy(ValueStr,1,2) );
      if (Hour > 23) then
        Result := E_DATEENH_INVALID_HOUR;
      if Length(ValueStr) > 2 then
        ValueStr := Copy(ValueStr, 3, Length(ValueStr) )
      else
        ValueStr := '';
    end;
  end;

  function ReadMinutes(var ValueStr: string; AType: TCharType): smallint;
  begin
    if AType <> ctNumeric then
      Result := E_DATEENH_INVALID_MINUTE
    else
    begin
      Result := E_EDITENH_NOERROR;
      Minute := StrToInt( Copy(ValueStr,1,2) );
      if (Minute > 59) then
        Result := E_DATEENH_INVALID_MINUTE;
      if Length(ValueStr) > 2 then
        ValueStr := Copy(ValueStr, 3, Length(ValueStr) )
      else
        ValueStr := '';
    end;
  end;

  function ReadSeconds(var ValueStr: string; AType: TCharType): smallint;
  begin
    if AType <> ctNumeric then
      Result := E_DATEENH_INVALID_SECOND
    else
    begin
      Result := E_EDITENH_NOERROR;
      Second := StrToInt( Copy(ValueStr,1,2) );
      if (Second > 59) then
        Result := E_DATEENH_INVALID_SECOND;
      if Length(ValueStr) > 2 then
        ValueStr := Copy(ValueStr, 3, Length(ValueStr) )
      else
        ValueStr := '';
    end;
  end;

  function ReadPoints(var ValueStr: string; AType: TCharType): smallint;
  begin
    if AType <> ctNumeric then
      Result := E_DATEENH_INVALID_POINTS
    else
    begin
      Result := E_EDITENH_NOERROR;
      if Length(ValueStr) = 1 then // .n  will need * 100
      begin
        Points := StrToInt( Copy(ValueStr,1,1) ) * 100;
        ValueStr := '';
      end
      else
        if Length(ValueStr) = 2 then // .nn  will need * 10
        begin
          Points := StrToInt( Copy(ValueStr,1,2) ) * 10;
          ValueStr := '';
        end
        else
          if Length(ValueStr) = 3 then
          begin
            Points := StrToInt( Copy(ValueStr,1,3) );
            ValueStr := '';
          end
          else
            Result := E_DATEENH_INVALID_POINTS;
    end;
  end;

  function ReadAMPM(var ValueStr: string; AType: TCharType): smallint;
  var
    tmpVal: integer;
  begin
    Result := E_DATEENH_VALUE_NOT_FOUND;
    if AType = ctCharacter then
    begin
      // first check locale specific strings...
      if not EnglishOnly then
      begin
        if AnsiUpperCase(ValueStr) = AnsiUpperCase(TimeAMString) then
        begin
          IsAM := true;
          ValueStr := '';
          Result := E_EDITENH_NOERROR;
        end
        else
          if AnsiUpperCase(ValueStr) = AnsiUpperCase(TimePMString) then
          begin
            IsPM := true;
            ValueStr := '';
            Result := E_EDITENH_NOERROR;
          end;
      end;
      if (Result = E_DATEENH_VALUE_NOT_FOUND) and (EnglishOnly or CheckEnglish) then
      begin
        tmpVal := FindMatchInStringArrays(ValueStr, EnglishShortAMPM,
          EnglishLongAMPM);
        if tmpVal = 1 then
        begin
          IsAM := true;
          ValueStr := '';
          Result := E_EDITENH_NOERROR;
        end
        else
          if tmpVal = 2 then
          begin
            IsPM := true;
            ValueStr := '';
            Result := E_EDITENH_NOERROR;
          end;
      end;
    end;
    if Result = E_DATEENH_VALUE_NOT_FOUND then
      Result := E_DATEENH_INVALID_AMPM;
  end;

  function ExtractValue(ValueStr: string; AType: TCharType): smallint;
  begin
    Result := E_DATEENH_VALUE_NOT_FOUND;
    // we accept a day name in the first position - even though it is
    // not explicitly used to set the date unless it is the only entry
    if LayoutPos = FirstPos then
    begin
      // only do first position matching of day name if not explicitly
      // going to look for one below (so that LayoutPos will be incremented
      // if necessary).
      if not(ALayout[LayoutPos] in ['x','X']) then
        Result := ReadDayNum(ValueStr, AType);
      // we accept special words to set the datetime to particular offsets
      if Result = E_DATEENH_VALUE_NOT_FOUND then
        Result := ReadSpecialWord(ValueStr, AType);
    end;
    if Result = E_DATEENH_VALUE_NOT_FOUND then
    begin
      case ALayout[LayoutPos] of
        'd', 'D': Result := ReadDayDate(ValueStr, AType);
        'm', 'M': Result := ReadMonth(ValueStr, AType);
        'y', 'Y': Result := ReadYear(ValueStr, AType);
        'h', 'H': Result := ReadHours(ValueStr, AType);
        'n', 'N': Result := ReadMinutes(ValueStr, AType);
        's', 'S': Result := ReadSeconds(ValueStr, AType);
        'z', 'Z': Result := ReadPoints(ValueStr, AType);
        'a', 'A': Result := ReadAMPM(ValueStr, AType);
        'x', 'X': Result := ReadDayNum(ValueStr, AType);
      else
        Result := E_DATEENH_INVALID_LAYOUT;
      end;
      Inc(LayoutPos);
      // This recursive call supports concatenated formats like...
      //  20010101  (YMD layout string)
      if (Length(ValueStr) > 0) and (LayoutPos <= Length(ALayout)) and
        (Result = E_EDITENH_NOERROR) then
        Result := ExtractValue(ValueStr, AType);
    end;
  end;

begin
  Result := E_EDITENH_NOERROR;
  LayoutPos := 1;
  CurStr := '';
  LastCharType := ctCharacter;
  DayNum := 0;
  Year := 0;
  Mnth := 0;
  DayDate := 0;
  Hour := MaxWord;
  Minute := MaxWord;
  Second := MaxWord;
  Points := MaxWord;
  tmpDT := 0;
  Accept12Hr := false;
  Only12Hr := false;
  IsAM := false;
  IsPM := false;
  CheckEnglish := false;
  EnglishOnly := false;
  if Trim(ALayout) = '' then
    ALayout := DateTimeFormatToLayout(ShortDateFormat + ' ' + ShortTimeFormat);
  // Check for ALayout problems...
  if Length(ALayout) < 1 then
    Result := E_DATEENH_INVALID_LAYOUT;
  for i := 1 to Length(ALayout) do
  begin
    // Make sure Layout string contains only valid characters
    if not (ALayout[i] in ['E', 'D', 'M', 'Y', 'H', 'N', 'S', 'Z', 'A', 'X',
      'e', 'd', 'm', 'y', 'h', 'n', 's', 'z', 'z', 'a', 'x']) then
    begin
      Result := E_DATEENH_INVALID_LAYOUT;
      break;
    end;
  end;
  if Result <> E_EDITENH_NOERROR then
    Exit;
  // Check whether english strings are accepted/needed
  if ALayout[1] = 'E' then
  begin
    CheckEnglish := true;
    EnglishOnly := true;
    Inc(LayoutPos);
  end
  else
    if ALayout[1] = 'e' then
    begin
      CheckEnglish := true;
      EnglishOnly := false;
      Inc(LayoutPos);
    end;
  FirstPos := LayoutPos;                // track first useful position
  // Check whether am/pm strings are accepted/needed
  if Pos('A', ALayout) > 0 then
  begin
    Accept12Hr := true;
    Only12Hr := true;
  end
  else
    if Pos('a', ALayout) > 0 then
    begin
      Accept12Hr := true;
      Only12Hr := false;
    end;
  // Now the main loop that extracts sections of AText for parsing...
  for i := 1 to Length(AText) do
  begin
    CurChar := AText[i];
    if AnsiPos(CurChar, ExtractDT_Numbers) > 0 then
      CurCharType := ctNumeric
    else
    begin
      if AnsiPos(CurChar, ExtractDT_Delimiters) > 0 then
        CurCharType := ctDelimiter
      else
        CurCharType := ctCharacter;
    end;
    // If reading a contiguous value then continue collecting the characters
    if CurCharType = LastCharType then
      CurStr := CurStr + CurChar
    else
    begin
      // if changed from one character type to another then we can process
      // the string acquired so far and see what we can find.  If last string
      // processing was all delimiter characters then they are ignored.
      if (LastCharType <> ctDelimiter) and (CurStr <> '') then
      begin
        Result := ExtractValue(CurStr, LastCharType);
        if (Result <> E_EDITENH_NOERROR) or (LayoutPos > Length(ALayout))
          then
        begin
          CurStr := '';                 // reset to avoid problems below
          break;
        end;
      end;
      CurStr := CurChar;
    end;
    LastCharType := CurCharType;
  end;
  // After loop finished make sure we processing anything left in CurStr
  if (Result = E_EDITENH_NOERROR) and (LayoutPos <= Length(ALayout)) and
    (CurStr <> '') and (LastCharType <> ctDelimiter) then
    Result := ExtractValue(CurStr, LastCharType);
  // Make sure all required parts of the layout have been filled
  if Result = E_EDITENH_NOERROR then
  begin
    for i := LayoutPos to Length(ALayout) do
    begin
      if ALayout[i] in ['E', 'D', 'M', 'Y', 'H', 'N', 'S', 'Z', 'A'] then
      begin
        Result := E_DATEENH_INPUT_TOO_SHORT;
        break;
      end;
    end;
  end;
  // Make sure the time input is acceptable
  if (Result = E_EDITENH_NOERROR) and Accept12Hr then
  begin
    if Only12Hr and not (IsAM or IsPM) then
      Result := E_DATEENH_NO_AM_OR_PM
    else
    begin
      if IsPM then
      begin
        // convert to 24hour time
        if Hour < 12 then
          Hour := Hour + 12;
        if Hour > 23 then
          Result := E_DATEENH_INVALID_PM_TIME;
      end
      else
        if IsAM then
        begin
          if (Hour < 1) or (Hour > 12) then
            Result := E_DATEENH_INVALID_AM_TIME
          else // convert 12 to 0 for actual processing
            if Hour = 12 then
              Hour := Hour - 12;
        end;
    end;
  end;
  if Result = E_EDITENH_NOERROR then
  begin
    if (DayNum <> 0) then
    begin
      // If given a day of week but no date values get the next
      // date from today that matches that day of week.
      if (Year = 0) and (Mnth = 0) and (DayDate = 0) then
      begin
        tmpDT := Trunc(Sysutils.Now);
        while DayOfWeek(tmpDT) <> DayNum do
          tmpDT := tmpDT + 1;
        DecodeDate(tmpDT, Year, Mnth, DayDate);
      end
      else
      begin
        // If given day of week and any part of the date we fill any
        // missing values from current date
        if (Year = 0) or (Mnth = 0) or (DayDate = 0) then
        begin
          DecodeDate(Sysutils.Now, TmpYear, TmpMnth, TmpDayDate);
          if Year = 0 then
            Year := TmpYear;
          if Mnth = 0 then
            Mnth := TmpMnth;
          if DayDate = 0 then
            DayDate := TmpDayDate;
        end;
        // Then check that everything is consistent
        try
          tmpDT := EncodeDate(Year, Mnth, DayDate);
        except
          Result := E_DATEENH_INVALID_DATE;
        end;
        // if a DayNum was set then we can use it to double-check the input
        if DayOfWeek(tmpDT) <> DayNum then
          Result := E_DATEENH_MISMATCHED_DAYNAME;
      end;
    end;
    if (Year > 0) and (Year <= 9999) then
      RYear := Year;
    if (Mnth > 0) and (Mnth <= 12) then
      RMnth := Mnth;
    if (DayDate > 0) and (DayDate <= 31) then
      RDayDate := DayDate;
    if Hour <= 23 then
      RHour := Hour;
    if Minute <= 59 then
      RMinute := Minute;
    if Second <= 59 then
      RSecond := Second;
    if Points <= 999 then
      RPoints := Points;
  end;
end;

function ExtractDateTime(ALayout, AText: string; var ADateTime: TDateTime):
  integer;
var
  DayDate, Mnth, Year, Hour, Minute, Second, Points: word;
  TmpDayDate, TmpMnth, TmpYear: word;
  tmpDT: TDateTime;
begin
  Year := 0;
  Mnth := 0;
  DayDate := 0;
  Hour := 0;
  Minute := 0;
  Second := 0;
  Points := 0;
  Result := ExtractDateTimeParts(ALayout, AText, DayDate, Mnth, Year,
    Hour, Minute, Second, Points);
  if Result = E_EDITENH_NOERROR then
  begin
    // Fill in missing date values from current date
    if (Year = 0) or (Mnth = 0) or (DayDate = 0) then
    begin
      DecodeDate(Sysutils.Now, TmpYear, TmpMnth, TmpDayDate);
      if Year = 0 then
        Year := TmpYear;
      if Mnth = 0 then
        Mnth := TmpMnth;
      if DayDate = 0 then
        DayDate := TmpDayDate;
    end;
    // Then check that everything is consistent
    TmpDT := 0;
    try
      tmpDT := EncodeDate(Year, Mnth, DayDate);
    except
      Result := E_DATEENH_INVALID_DATE;
    end;
    if Result = E_EDITENH_NOERROR then
    begin
      // We can always add the time, because 0 time is the default above
      ADateTime := TmpDT + EncodeTime(Hour, Minute, Second, Points);
    end;
  end;
end;



function UseXPThemes: boolean;
begin
  Result := false;
  if EnhUtils_UseXPThemes then
  begin
    {$IFDEF SUPPORT_XP_THEMES}
    Result := ThemeServices.ThemesAvailable and ThemeServices.ThemesEnabled;
    {$ELSE}
    Result := false;
    {$ENDIF}
  end;
end;


initialization;
  EnhUtils_UseXPThemes := true;


end.
